* BENCH1.PRG ** Sieve of Erotosthenes, using numeric array or string
*
* Release $Date: 2004/10/18 18:42:31 $

// Notes:
// This program runs on DOS using Clipper 5.x
// or on any UNIX using FlagShip 4.3 or later
// For DOS programs running Fox or dBASE, use BENCH1FX.PRG
//
// Compile on DOS:  CLIPPER  bench1 -L  ; RTLINK|BLINKER bench1 ; bench1
// Compile on UNIX: FlagShip bench1.prg -nl -nL -nd -obench1 ; bench1


#ifndef FlagShip
#define LOCAL_LONG LOCAL
#else

#Cinline
#define ARRSIZE  1001
#define SIZEC    1000
static int cflags [ARRSIZE];
#endCinline
#endif

#define ITERATIONS 1000
#define SIZE       1000

LOCAL_LONG iters := ITERATIONS
LOCAL_LONG count := 0
LOCAL_LONG prime := 0
LOCAL_LONG i, j, k

LOCAL flags [SIZE], sflags
LOCAL duration, start

CLEAR

#ifndef FlagShip
   ? 'Clipper 5.x (Locals)'
#else
   ? 'FlagShip (Locals + Typed)'
#endif
?? " Sieve of Erotosthenes,", ltrim(str(ITERATIONS)), "iterations"
?
#ifdef FlagShip5
	set font to "courier",10
#endif

? "1. Test for numeric calculations"
************************************
start = SECONDS()

FOR k := 1 TO ITERATIONS
   count := 0
   FOR i := 1 TO SIZE
      flags[i] = 1
   NEXT
   FOR i := 1 TO SIZE
      IF flags[i] != 0
         prime = 2 * i + 3
         j     = i + prime
         WHILE j <= SIZE
            flags[j] = 0
            j += prime
         ENDDO
         count++
      ENDIF
   NEXT
NEXT

duration = SECONDS() - start
? '   Number of primes were '
?? LTRIM(STR(count)) + ".  "
?? LTRIM(STR(iters)), 'iterations took '
   dispout(transform(duration, '999.99'), "W+", "b+")
?? ' seconds'

? "2. Test for string handling"
*********************************
sflags := replicate(CHR(1), SIZE)

start = SECONDS()

FOR k := 1 TO ITERATIONS
   count := 0
   FOR i := 1 TO SIZE
      IF STRPEEK (@sflags,i) != 2
         prime = 2 * i + 3
         j     = i + prime
         WHILE j <= SIZE
            STRPOKE (@sflags, j, 2)
            j += prime
         ENDDO
         count++
      ENDIF
   NEXT
NEXT

duration = SECONDS() - start
? '   Number of primes were '
?? LTRIM(STR(count)) + ".  "
?? LTRIM(STR(iters)), 'iterations took '
   dispout(transform(duration, '999.99'), "W+", "b+")
?? ' seconds'

#ifdef FlagShip
? "3. Test for C-inline usage"
*****************************
start = SECONDS()

FOR k := 1 TO ITERATIONS
   count := 0
#Cinline
   for (i = 0; i <= SIZEC; i++)
      cflags[i] = 1;

   for (i = 0; i <= SIZEC; i++) {
      if (cflags[i]) {
         prime = 2 * i + 3;
         j     = i + prime;
         while (j <= SIZEC) {
            cflags[j] = 0;
            j += prime;
         }
         count++;
      }
   }
#endCinline
NEXT

duration = SECONDS() - start
? '   Number of primes were '
?? LTRIM(STR(count)) + ".  "
?? LTRIM(STR(iters)), 'iterations took '
   dispout(transform(duration, '999.99'), "W+", "B+")
?? ' seconds'
#endif

wait

?
? "Guideline values for identical HW :"
?
set color to "W+"
? "P3/700 Win2K vs. Linux :"
? "                                          Clipper 5.3  FlagShip 5.0   "
DISPOUT ("faster", "GR+", "r+")
set color to "W"
? "1. numer.calculation,   " + str(SIZE,4) + " interations     8.74 sec    1.91 sec  = "
DISPOUT ("  4.6 x", "GR+", "r+")
? "2. string manipulation, " + str(SIZE,4) + " interations    97.53 sec    3.24 sec  = "
DISPOUT (" 30.1 x", "GR+", "r+")
? "3. num.calc, C-inline,  " + str(SIZE,4) + " interations     8.74 sec    0.07 sec  = "
DISPOUT ("124.8 x", "GR+", "r+")
?

wait

#ifndef FlagShip
   FUNCTION strpeek (flags, pos)
   RETURN ASC( SUBSTR(flags, pos, 1))

   FUNCTION strpoke (flags, pos, val)
   flags := STUFF (flags, pos, 1, CHR(val))
   RETURN flags
#endif

** eof **

