/* For FlagShip 6
 * Copyright (c) 2008 by multisoft Datentechnik, Munich, Germany 
 *
 * You may freely modify this source according to your needs and/or
 * re-use any parts of it in your source code.
 * 
 * Release $Date: 2008/01/14 08:21:54 $
 *--------------------------------------------------------------------------
 * This application demonstrates check for cause of IndexCheck() error 
 * (or run-time-error RTE 331)
 *
 * Copy the source to your working directory, and
 *
 * Compile Linux  : FlagShip checkindex.prg -o checkindex
 * Run Linux      : ./checkindex dbfname indexname [index2 [index3 ...]]
 *
 * Compile Windows: FlagShip checkindex.prg
 * Run Windows    : checkindex dbfname indexname [index2 [index3 ...]]
 */

PARAMETERS par1, par2  // and more...
local   ii, iCount, cName, cName2, iErrCnt, cDbf := "", aIdx := {}
local   cMsg, lPrint := .F.
private bStdErr, iRteErr := 0

set font "courier",10

? "Check for index corruption"
?

/* get/check parameters
 */
iCount := Pcount()      // number of passed parameters
if iCount < 2
   DisplayHelp()
   wait
   quit
endif
for ii := 1 to iCount
   cName := alltrim(Param(ii))
   if lower(cName) == "-tofile"
      if !lPrint
         lPrint := .T.
         SET PRINTER TO checkindex.log
         SET PRINTER ON
         SET CONSOLE OFF
         ? "Check for index corruption by", execname(), "on", date(), time()
         ?
         SET CONSOLE ON
      endif
      loop
   endif
   cName := strtran(cname, ",", "")
   cName := strtran(cname, '"', "")
   cName := strtran(cname, "'", "")
   if empty(cName)
      loop
   endif
   cMsg := ""
   if !file(cName)
      cName2 := if(empty(cDbf), ".dbf", ".idx")
      if cName2 $ cName
         cMsg := "*** sorry, '" + cName + "' not available"
      else
         cName2 := cName + cName2
         if !file(cName2)
            cMsg := "*** sorry, neither '" + cName + "' nor '" + cName2 + "' available"
         endif
         cName := cName2
      endif
   endif
   if !empty(cMsg)                  // display error, file n/a
      DisplayHelp()
      ? ; ? "Parameter: " 
      for ii := 1 to iCount
         ?? Param(ii) + " "
      next
      ? ; ? cMsg ; ?
      wait
      quit
   endif
   if empty(cDbf)
      cDbf := cName
   else
      aadd(aIdx, cName)
   endif
next

? "Parameter: '" + cDbf + "'"
aeval(aIdx, {|x| qqout(" '" + x + "'")})
?
if empty(cDbf) .or. len(aIdx) < 1   // not enough parameters
   DisplayHelp()
   wait
   quit
endif

/* open database
 */
USE (cDbf) EXCLUSIVE       
if !used()
   DisplayHelp()
   ? ; ? "*** sorry, cannot open " + cDbf + " in exclusive mode, try later"
   wait
   quit
endif

/* assign indices
 */
for ii := 1 to len(aIdx)   
   /* Note: wrong index may cause RTE 307 (undefined variable), 
    * we therefore use here special error handler (below)
    * to display additional data and to print it to log file
    */
   bStdErr := ErrorBlock( {|obj| myError307(obj, aIdx[ii], lPrint) } )    
   dbSetIndex(aIdx[ii])  // SET INDEX TO ...
   ErrorBlock(bStdErr)   // restore default error handler
   if neterr()
      ? ; ? "*** sorry, cannot open index " + aIdx[ii]
      wait
      quit
   endif
   cName := IndexDbf()
#ifdef FS_WIN32
   if upper(cName) != upper(Dbf()) 
#else
   if cName != Dbf()
#endif
      ? ; ? "*** Index " + aIdx[ii] + " was created from database " + cName
      wait "Continue (n/y) ? " 
      if !(chr(lastkey()) $ "yY")
         quit
      endif
   endif               
next   
SET ORDER TO 1

/* avoid displaying RTE 331 by redirecting errors 
 * to own error handler myError331() = below
 */
bStdErr := ErrorBlock( {|obj| myError331(obj) } )    

/* check all given indices unconditionally
 */   
iErrCnt := myIndexCheck(.T.)     // or (.F.) for conditional check

if (iErrCnt + iRteErr) == 0
   ? "No problems detected :-)"
elseif iErrCnt == 0 .and. iRteErr > 0  // errors only from RTE307
   SET ORDER TO 1
   ? "     database: " + cDbf
   ? "        index: " + aIdx[1]
   ? "     indexkey: " + IndexKey()
   ? "   indexCheck: " + ltrim(IndexCheck(1)) 
   ? " .dbf records: " + ltrim(Reccount())
   ? " .idx records: " + ltrim(DbObject():OrderKeyCount(1))
   for ii := 2 to len(aIdx)
      SET ORDER TO (ii)
      ?
      ? "        index: " + aIdx[ii]
      ? "     indexkey: " + IndexKey()
      ? "   indexCheck: " + ltrim(IndexCheck(ii)) 
      ? " .idx records: " + ltrim(DbObject():OrderKeyCount(ii))
   next
   REINDEX                       // ok, dbf is exclusively open
   ?
   ? "*** " + ltrim(iRteErr) + " error(s) fixed by Reindex"
endif
?
ErrorBlock(bStdErr)              // restore std. errors
USE
if lPrint 
   SET PRINTER TO
   SET PRINTER OFF
endif
wait TIMEOUT 10 "Done ..." 
quit

/*********************************************************************
 * General function to check current index (indices) for corruption 
 * and repair this/these index(es)
 * 
 * Parameter, optional: if .T., the IndexCheck() value is ignored 
 *    and the whole index is checked. Suggested when RTE 331 is
 *    reported but IndexCheck() returns 0. Default is .F. which
 *    checks the index only when IndexCheck() returns value > 0
 * Note: when you get RTE 331 during this check, press "Ignore" 
 *    and continue.
 */
FUNCTION myIndexCheck(lMustCheck) 
#include "rddsys.fh"
local idx, cDbf, cIdx, cDbfKey, cIdxKey, lFirstErr, iError := 0
local lNeedsUnlock := .F.

if !used() .or. NetErr()       
   ? "could not open dbf or index"
   return -1                     
endif
if !IsDbExcl() .and. !IsDbFlock()
   while !flock()    // avoid changing by others during the check
      SleepMs(50)
   enddo
   lNeedsUnlock := .T.
endif
for idx := 1 to IndexCount()     // for all open indices
   if IndexCheck(idx) >= 1 .or. lMustCheck
      cDbf := Dbf(,.T.)
      cIdx := IndexNames()[idx]
      SET ORDER TO (idx)         // for oRdd:OrderInfo(...)             
      if IndexCheck(idx) == 2 .or. ;
           (Reccount() != DbObject():OrderKeyCount(idx) ;
            .and. empty(DbObject():OrderInfo(DBOI_CONDITION)) ;
            .and. ! DbObject():OrderInfo(DBOI_ISCOND) )
         // record count in database vs. index differs, but this 
         // may be ok when created with INDEX..ON..FOR.. condition 
         ? "Error: database: " + cDbf
         ? "       index   : " + cIdx
         ? "       indexkey: " + IndexKey()
         ? "     indexCheck: " + ltrim(IndexCheck(idx)) 
         ? "   .dbf records: " + ltrim(Reccount())
         ? "   .idx records: " + ltrim(DbObject():OrderKeyCount(idx))
         iError++               
      else
         // go thru dbf and check index keys, display errors
         lFirstErr := .T.
         GO TOP
         while !eof()
            cDbfKey := &(IndexKey())
            cIdxKey := DbObject():OrderInfo(DBOI_KEYVAL)
            if !(cDbfKey == cIdxKey)
               if lFirstErr
                  ? "Error: database: " + cDbf
                  ? "       index   : " + cIdx
                  ? "       indexkey: " + IndexKey()
                  ? "     indexCheck: " + ltrim(IndexCheck(idx)) 
                  lFirstErr := .F. 
               endif
               ? " - record", str(Recno(),6), "key = '" + cDbfKey + "'"
               ? "   index         key = '" + cIdxKey + "'"
               iError++               
            endif
            SKIP
         enddo
      endif
      SET ORDER TO 1
   endif
next
if lNeedsUnlock
   UNLOCK
endif
if iError > 0 
   if IsDbExcl()  
      REINDEX  
   else
      local lSave := Set(_SET_GOTOP, .F.)
      DbAutoReopen(.T.)  // in ../system/autolock.prg: re-open exclusive
      REINDEX  
      Set(_SET_GOTOP, lSave)
      DbAutoReopen(.F.)  // re-open shared
   endif
   ? "*** " + ltrim(iError) + " error(s) fixed by Reindex"
   wait
endif
return iError

/*********************************************************************
 * helpers for main
 */
FUNCTION DisplayHelp()
   ? "Syntax: " + execname() + " [-tofile] dbfname indexname [index2 [index3 ...]]"
   ? 
   ? "   dbfname  : name of the database,  optional with path"
   ? "   indexname: name of the 1st index, optional with path"
   ? "   index2   : name of the 2nd index, optional with path, etc."
   ? '              if the path contain space(s), use "/my path/filename"'
   ? "   -tofile  : echo the output to file named checkindex.log"
   ? 
   ? " At least dbfname and indexname is required."
   ? " Best to assign all used indices for this database."
return

FUNCTION myError331(oErr)
MEMVAR bStdErr, iRteErr
if oErr:GenCode == 331     // ignore here RTE 331 (index file corrupted)
   iRteErr++               // count RTE331 errors
   return .T.              // and continue
endif
return Eval(bStdErr, oErr) // but display other errors by std. err handler

FUNCTION myError307(oErr, cName, lPrint)
MEMVAR bStdErr, iRteErr
#ifdef FS_WIN32
# define CRLF chr(13,10)
#else
# define CRLF chr(10)
#endif
if oErr:GenCode == 307     // RTE 307 (undefined variable) in DbSetIndex()
   oErr:Description += "; ;Cause: index " + cName + ;
          ";probably not for this database;Exit now to avoid index corruption" 
   if lPrint
      SET CONSOLE OFF
      ? "*** RTE 307 on DbSetIndex(): " + ;
        strtran(oErr:Description, ";", CRLF + "*** ") // print err msg
      SET CONSOLE ON
   endif
endif
return Eval(bStdErr, oErr) // display error by std. err handler

// eof
