// This example demonstrates creating of database with indices
// and it access/browsing by using DbEdit() Function.
// Release $Date: 2005/11/21 13:38:25 $
//
// Compile: FlagShip database1.prg -o database1       # compile for hybrid mode
// Execute: database1         (or ./database1)        # GUI mode (auto detect)
//      or: database1 -io=t   (or ./database1 -io=t)  # force Terminal mode
//      or: newfswin ./database1                      # Unix/Linux Terminal 
//
// -> see also: database2.prg, tbrowse_db.prg, tbrowse_ar.prg
// ->    fsman: FSC, CMD, FUN, LNG.9, REL

#include "fspreset.fh"              // see <FlagShip_dir>/include directory
#include "box.fh"         

if !file("mydata.dbf")              // database available?
   aDbfStruct := {{"Id",      "N",  8, 0}, ;
                  {"Name",    "C", 25, 0}, ;
                  {"Address", "C", 50, 0}, ;
                  {"City",    "C", 30, 0}, ;
                  {"ZIP",     "C",  8, 0}, ;
                  {"Date",    "D",  8, 0}, ;
                  {"Turnover","N", 10, 2} }
   ? "creating database"
   DbCreate("mydata", aDbfStruct)   // create new database
# ifndef FS_WIN32
   run ("chmod 666 mydata.dbf")     // Linux: set rw permission for all
# endif
endif
SET OPENERROR OFF                   // don't display RTE on open failure
USE mydata SHARED                   // open database in non-exclusive mode
if !used()                          // could open?
   alert("cannot open 'mydata.dbf', check permission")
   quit
endif
if !file("mydata1.idx") .or. ;      // index available?
   !file("mydata2.idx")
   ? "creating indices"
   INDEX ON id TO mydata1           // n/a, create indices
   INDEX ON upper(left(name,10)) TO mydata2 
# ifndef FS_WIN32
   run ("chmod 666 mydata1.idx mydata2.idx") // Linux: set rw permission for all
# endif
endif
SET INDEX TO mydata1, mydata2       // assign indices
if neterr()
   alert("could not open 'mydata?.idx', check permission")
   quit
endif
GO TOP                              // go to first logical record
if eof()                            // database empty?
   ? "filling database"
   for ii := 1 to 20                // yes, fill with some records
      APPEND BLANK
      REPLACE id with ii, Name with "Name " + ltrim(ii), ;
              Address with "Address " + ltrim(ii), ;
              City with "City " + ltrim(ii), ;
              Zip with ltrim(ii + 1000)
      REPLACE Date with Date(), Turnover with ii * 100 + ii / 100
   next
endif

GO TOP                              // go to first logical record
nLine := row() +1                   // next screen row
@ nLine,10 SAY "Browse database sorted on ID - exit with ESC or Enter"
@ nLine +1, 0, 17, 79 BOX B_SINGLE  // box around (in Terminal mode only)
DbEdit(nLine +2, 1, 16, 78)         // browse database

SET ORDER TO 2                      // select index 2 = Name
GO TOP                              // go to first logical record
@ nLine,0 clear                     // clear part of the screen
@ nLine,10 SAY "Browse database sorted on Name - exit with ESC or Enter"
iShowLines := min(15, maxrow() - nLine - 1, reccount() -5)
@ nLine +1, 0, nLine + iShowLines +1, maxcol() BOX B_SINGLE 
DbEdit(nLine +2, 1, nLine + iShowLines, maxcol() -2)

USE                                 // close database & indices
setpos(maxrow() -1,0)               // set sursor position
wait                                // press any key...
quit
