/* For FlagShip 6. 
 * Copyright (c) 2004..05 by multisoft GmbH, Munich, Germany
 * You may modify this source according to your needs and distribute the
 * object code freely with your executable. You may distribute also the 
 * modified source code when your application is distributed in source, 
 * but may not remove the copyright notice.
 * 
 * Release $Date: 2005/11/21 13:41:51 $
 */
// ---------------------------------------------------------------------
// This example demonstrates FlagShip's capability and several mechanisms
// how to handle and display image (bitmap) files and how to store them
// in database fields.
//
// Copy this file to your working directory and un-zip there also
// the file images.zip there. Then invoke:
// - on MS-Windows: FlagShip images.prg ; images
// - on Linux/Unix: FlagShip images.prg -o images ; ./images
// ---------------------------------------------------------------------
#include "box.fh"
#include "mouse.fh"

if AppIoMode() != "G"
   if alert("Sorry, graphic images are supported in GUI mode only. You;" + ;
            "however may continue, but will see only the plain text...", ;
            {"Abort", "Continue"} ) != 2
      quit
   endif
endif

cPath     := ""
lUseDbf   := file("images.dbf")
lHasFiles := .T.

/* Note: since MemoRead() does not use SET PATH, we need to
 * check current directory and the subdirectory ./images
 */
if !file(cPath + "anim1.bmp") .or. !file(cPath + "frust01.gif")
   cPath := "." + PATH_SLASH + "images" + PATH_SLASH
endif
if !file(cPath + "anim1.bmp") .or. !file(cPath + "frust01.gif")
   lHasFiles := .F.
endif

do case
case !lUseDbf .and. !lHasFiles
   alert("Sorry, could not locate images.;Unzip the file " + ;
         FlagShip_dir() + PATH_SLASH + "examples" + PATH_SLASH + ;
         "images.zip;into your working directory " + curdir())
   quit
case !lUseDbf .and. lHasFiles
   /* choice to use plain image files or import them to database
    */
   lUseDbf := import2dbf()
case lUseDbf .and. lHasFiles
   /* choice to use plain image files or the database
    */
   lUseDbf := choiceDataFrom()
otherwise
   /* has database but no image files, will use database
    */
endcase
cls

/* display the images
 */
anim_hello()
anim_frust()
jpeg_img()

wait
quit

/* --------------------------------------------------------------------
 * A small animation of smiley/hello
 * the image size w*h is 126*32 pixel
 */
function anim_hello()
local yPos, ii, iPos, key, acImg := {}
local aFiles := {"anim1.bmp", "anim2.bmp", "anim3.bmp", "anim4.bmp", ;
                 "anim5.bmp", "anim6.bmp", "anim7.bmp", "anim8.bmp" }

yPos := 20
if lUseDbf
   @ 24,0 say "Note: using images from database with "
   use images shared new
   locate for left(FILENAME,4) == "anim"
   ?? if(valtype(MEMODATA) == "M", "MEMO", "BLOB") + " fields"
   while !eof()
      if valtype(MEMODATA) == "M"   // MEMO field
         aadd(acImg, MemoDecode(MEMODATA) )
      else
         aadd(acImg, IMGDATA)       // BLOB field
      endif
      continue
      @ yPos,10 SAY IMAGE atail(acImg) PIXEL
      yPos += 37
   enddo
   use
   // you may add check for len(acImg) == len(aFiles)
else
   @ 24,0 say "Note: using plain image files"
   acImg := array(len(aFiles))
   for ii := 1 to len(aFiles)
      acImg[ii] := memoread(cPath + aFiles[ii])
      @ yPos,10 SAY IMAGE acImg[ii] PIXEL
      yPos += 37
   next
endif

@ 120,145,173,281 BOX guicolor "N/W+" PIXEL LINEWIDTH 1
@ 180,160 say "Press ESC to abort" PIXEL

iPos := 1
while (key := inkey(0.2)) != K_ESC
   if iPos > len(aFiles)
      iPos := 1
   endif

   /* Note: you may read/display the files directly from
    * disk, but using memory variable is more effective for
    * this repetitive purpose and avoids heavy load
    */
// @ 130,150 SAY IMAGE FILE (cPath + aFiles[iPos]) PIXEL
   @ 130,150 SAY IMAGE acImg[iPos] PIXEL

   iPos++
enddo
@ 180,160 clear PIXEL
return NIL

/* --------------------------------------------------------------------
 * Cartoon animation of frustrated programmer :-)
 * from http://camelotdesign.com/cberanim6.gif
 * the image size w*h is 400*320 pixels, so we scale them
 */
function anim_frust()
local ii, iPos, xPos, yPos, xSize, ySize, key
local lCursOn := set(_SET_GUICURSOR)
local aFiles := {"frust01.gif", "frust02.gif", "frust03.gif", "frust04.gif", ;
                 "frust05.gif", "frust06.gif", "frust07.gif", "frust08.gif", ;
                 "frust09.gif", "frust10.gif" }
local aSeqTim := {{1, 0.1}, {2, 0.1}, {3, 0.1}, ;
                  {1, 0.1}, {2, 0.1}, {3, 0.1}, ;
                  {1, 0.1}, {2, 0.1}, {3, 0.1}, ;
                  {1, 0.1}, {2, 0.1}, {3, 0.1}, ;
                  {1, 0.1}, {2, 0.1}, {3, 0.1}, ;
                  {1, 0.1}, {2, 0.1}, {3, 0.1}, ;
                  {1, 0.1}, {2, 0.1}, {3, 0.1}, ;
                  {4, 2.5}, {1, 0.1}, {2, 0.1}, ;
                  {3, 0.1}, {5, 2.5}, {6, 2.5}, ;
                  {7, 0.1}, {8, 0.1}, {9, 0.1}, ;
                  {8, 0.1}, {9, 0.1}, {7, 0.1}, ;
                  {8, 0.1}, {9, 0.1}, {10, 3,0} }

local acImg := array(len(aFiles))
// local cTitle := " Frustrated Programmer "
local cTitle := " Programming Stages "

/* read image data from file and display it miniature
 */
xSize := Col2Pixel(6)
ySize := int(xSize * (320 / 400))
xPos  := Col2pixel(67)
yPos  := 20

if lUseDbf
   @ 24,0 say "Note: using images from database with "
   use images shared new
   ?? if(valtype(MEMODATA) == "M", "MEMO", "BLOB") + " fields"
   locate for left(FILENAME,5) == "frust"
else
   @ 24,0 say "Note: using plain image files"
endif
for ii := 1 to len(aFiles)
   if lUseDbf
      if valtype(MEMODATA) == "M"   // MEMO field
         acImg[ii] := MemoDecode(MEMODATA)
      else
         acImg[ii] := IMGDATA       // BLOB field
      endif
   else
      acImg[ii] := memoread(cPath + aFiles[ii])
   endif
   @ yPos,xPos,,xPos+xSize SAY IMAGE acImg[ii] PIXEL SCALE
   yPos += ySize + max(10, Col2Pixel(0.7))
   if ii == 5
      xPos += xSize + max(10, Row2pixel(0.5))
      yPos := 20
   endif
   if lUseDbf
      continue
   endif
next
use

/* simulates WAIT behavior
 */
SET GUICURSOR ON
SET GUICURSOR TO CURSOR_HAND
@ 22,44 say "Press ESC to abort"
SET GUICURSOR OFF

/* display title
 */
oFont := Font{"Helvetica"}
oFont:Bold := .T.
@ 1,44 SAY cTitle GUICOLOR "G/W+" FONT oFont

/* Do the animation according to image sequence (1..36) and it
 * delay (in fraction of seconds) in above array. Note that we
 * use only 10 images for the whole sequence of 36 frames.
 * Note also that the scaling used here is computing expensive,
 * so you may decrease the CPU load on slower computers by
 * providing 1:1 images.
 */
iPos := 1
while .T.
   @ 3,37,16,67 SAY IMAGE acImg[aSeqTim[iPos,1]] SCALE FRAME BOX_RAISED

// @ 23,44 say "Seq="+ltrim(iPos)+" img="+ ltrim(aSeqTim[iPos,1])+" delay="+ltrim(aSeqTim[iPos,2])
   key := inkey(aSeqTim[iPos,2])   // wait, variable delay
// @ 23,44 clear to 23,68
   if key == K_ESC
      exit
   endif
   if ++iPos > len(aSeqTim)
      iPos := 1
   endif
enddo
set(_SET_GUICURSOR, lCursOn)
setpos(23,0)
return NIL

/* --------------------------------------------------------------------
 * Display jpeg image from file
 * the image size w*h is  pixel
 */
function jpeg_img()
local cFile := cPath + "space.jpg"
local cImg := ""

cls
for ii := 2 to 24
   @ ii,0 say strzero(ii,2) guicolor "B/W"
next
@ 0,0 say "JPEG image scalled to fit to rows 2..23 and columns 5..79 "

if lUseDbf
   use images shared new
   locate for left(FILENAME,5) == "space"
   if valtype(MEMODATA) == "M"   // MEMO field
      cImg := MemoDecode(MEMODATA)
   else
      cImg := IMGDATA            // BLOB field
   endif

   @ 2,5,23,79 say image cImg
else
   @ 2,5,23,79 say image file (cFile)
endif
wait

cls
@ 0,0 say "JPEG image 1:1 "
if lUseDbf
   ?? "using database records"
   @ 2,0 say image cImg
else
   ?? "using image file"
   @ 2,0 say image file (cFile)
endif
// wait
return

/* --------------------------------------------------------------------
 * a small dialog to import plain image files into database
 */
function import2dbf()
local ii, jj, nCol, aDir, cImg, oRadio[3], nRadio := 1
local aFiles := {"anim*.bmp", "frust*.gif", "space.jpg"}
aTxt := {"use image files", ;
         "import into database using BLOB fields", ;
         "import into database using MEMO fields" }
@ 5,0 say "Do you wish to import image files into" ; nCol := col()
@ 6,0 say "database or use plain image files ?"
oRadio[1] := RadioButton{4.5, nCol +4, aTxt[1]}
oRadio[2] := RadioButton{5.5, nCol +4, aTxt[2]}
oRadio[3] := RadioButton{6.5, nCol +4, aTxt[3]}
@ 4, nCol +3, 7.5, nCol + Strlen2col(aTxt[2]) +8  GET nRadio RADIOGROUP oRadio
read

if nRadio != 2 .and. nRadio != 3
   return .F.
endif

if nRadio == 2
   // use .dbf and .dbv
   dbcreate("images", {{"FILENAME", "C", 20,0},{"IMGDATA", "VBZ", 10,0}} )
else
   // use .dbf and .dbt
   dbcreate("images", {{"FILENAME", "C", 20,0},{"MEMODATA", "M", 10,0}} )
endif
use images
if !used()
   alert("sorry, could not create database;" + ;
         curdir()+PATH_SLASH+"images.db?;check access rights")
   return .F.
endif
for ii := 1 to len(aFiles)
   aDir := directory(cPath + aFiles[ii])
   for jj := 1 to len(aDir)
      cImg := memoread(cPath + aDir[jj,1])
      if len(cImg) == 0
         alert("File " + cPath + aDir[jj,1] + " is empty or not accessible")
         quit
      endif
      append blank
      REPLACE filename WITH aDir[jj,1]
      if nRadio == 2
         REPLACE IMGDATA with cImg
      else
         REPLACE MEMODATA with MemoEncode(cImg)
         if len(MemoEncode(cImg)) > 64000
            alert("Warning, data does not fit in MEMO field")
         endif
      endif
   next
next
use
return .T.

/* --------------------------------------------------------------------
 * a small dialog for choice to use plain image files or database
 */
function choiceDataFrom()
local oRadio[2], nRadio := 1
local cTxt2 := "use database 'images.dbf' with BLOB fields"

use images new shared
if valtype(MEMODATA) == "M"   // MEMO field
   cTxt2 := strtran(cTxt2, "BLOB", "MEMO")
endif
use

oRadio[1] := RadioButton{4, 10, "use plain image files"}
oRadio[2] := RadioButton{5, 10, cTxt2}
@ 3, 9, 6.3, 9 + Strlen2col(cTxt2) +5  GET nRadio RADIOGROUP oRadio ;
    CAPTION "Do you wish to"
read
cls
return nRadio == 2


/* eof */

