/* For FlagShip 5 and 6
 * Copyright (c) 2005 by multisoft GmbH, Munich, Germany
 *
 * Test for keyboard mappings and keyboard flags (LEDs)
 *
 * You may freely modify this source according to your needs and/or
 * re-use any parts of it in your source code.
 *
 * Release $Date: 2005/11/23 10:05:40 $
 */
//--------------------------------------------------------------------
// compile: FlagShip kyboard.prg -o keyboard
// execute: keyboard
// or term: keyboard -io=t
//
// or: Clipper keyboard ; blinker fi keyboard

// If you wish to protocol it, enable following #define:
// #define PROTOCOL
// or compile : FlagShip -DPROTOCOL kyboard.prg -o keyboard
// and execute: keyboard 2>protocol.log
//          or: keyboard -io=t 2>protocol.log
//
// You may also display/protocol the raw-codes (scan codes) when setting
// environment variable FSPRINTKEYRAW, i.e.
//   on Unix/Linux: export FSPRINTKEYRAW=1
//   on MS-Windows: SET FSPRINTKEYRAW=1
//--------------------------------------------------------------------

cls

#ifdef FlagShip
  /* -- FlagShip specific part --
   */
  set font "courier", 10                     // apply for GUI mode
  if AppIoMode() == "T"
    /*
     * executing in Terminal i/o mode
     */
#   ifdef FS_WIN32
//   Run ("MODE CON CP select=858")          // optional: set code page
//   FS_SET("inmap", "FSkeymap.cp858")       // optional: assign keyboard mapping
#   endif
     private termOutFile, termOutName, termInFile, termInName
     fs_set("term", @termOutFile, @termOutName, @termInFile, @termInName)
     ? "Input mapping file :", termInFile
     ? "Input mapping entry:", termInName
#   ifdef PROTOCOL
     ??## "Input mapping file :", termInFile
     ??## "Input mapping entry:", termInName
#   endif
  else
    /*
     * executing in GUI mode
     */
#    include "font.fh"
//   FS_SET("guikey","FSguikey.def")          //optional: default keyboard
//   m->oApplic:Font:CharSet(FONT_ISO8859_2)  // optional: set default font to Latin-2
//   FS_SET("guikey","x:\your_path\FSguikey.cp1250_pl")  //optional: polish 'programmer' keyboard
     cMapFile := FS_SET("guikey")
     if empty(cMapFile)
        cMapFile := "none set, using default, corresponds to " + ;
                    FlagShip_dir() + "\terminfo\FSguikey.def"
     endif
     ? "Input mapping file :", cMapFile
#   ifdef PROTOCOL
     ??## "Input mapping file :", cMapFile
#   endif
  endif // GUI or Terminal i/o mode
#else
  /* -- non-FlagShip = backward Clipper compatibility --
   */
# include "inkey.ch"
# command ??## [ <list,...> ] => outerr ( <list> ) ; outerr (chr(10))
#endif

#ifdef PROTOCOL
 ??## "-- inkey() test"
#endif

/* input loop until Escape
 */
?
key := 0
while key != K_ESC
   ? "Press any key or ESC : "
   key := inkey(0)                           // read a key
   ?? padr(str(key,4) + " = " + inkey2str(key), 40) + ;
      GetInkeyFlags()                        // display it
# ifdef PROTOCOL
   ??## padr("--> " + str(key,4) + " = " + inkey2str(key),50) + ;
      GetInkeyFlags()                        // protocol it
# endif
enddo
?

#ifdef PROTOCOL
 ??## "-- inkey() test done"
#endif

wait
quit

#ifdef FlagShip
// -------------------------------------------------------------------
FUNCTION GetInkeyFlags()
local cTxt := ""
if InkeyStatus(.T.) > 0
  cTxt += if(InkeyStatus(STATUS_INSERT),     "Insert-ON ",    "")
  cTxt += if(InkeyStatus(STATUS_CAPSLOCK),   "CapsLock-ON ",  "")
  cTxt += if(InkeyStatus(STATUS_NUMLOCK),    "NumLock-ON ",   "")
  cTxt += if(InkeyStatus(STATUS_SCROLL),     "ScrollLock-ON ","")
  cTxt += if(InkeyStatus(STATUS_SHIFT),      "Shift ",        "")
  cTxt += if(InkeyStatus(STATUS_CTRL),       "Ctrl ",         "")
  cTxt += if(InkeyStatus(STATUS_ALT),        "Alt ",          "")
  cTxt += if(InkeyStatus(STATUS_LWINDOW),    "LeftWin ",      "")
  cTxt += if(InkeyStatus(STATUS_RWINDOW),    "RightWin ",     "")
  cTxt += if(InkeyStatus(STATUS_MENU),       "Menu ",         "")
  cTxt += if(InkeyStatus(STATUS_META),       "Meta ",         "")
endif
// if BinAnd(InkeyStatus(), STATUS_ANYMOUSE) > 0
//   cTxt += if(InkeyStatus(STATUS_MOUSEDOUBLE),"Double-",    "")
//   cTxt += if(InkeyStatus(STATUS_LEFTDOWN),   "MouseLeft",  "")
//   cTxt += if(InkeyStatus(STATUS_MIDDOWN),    "MouseMid",   "")
//   cTxt += if(InkeyStatus(STATUS_RIGHTDOWN),  "MouseRight", "")
//   cTxt += if(InkeyStatus(STATUS_WHEEL),      "MouseWheel", "")
//   cTxt += if(InkeyStatus(STATUS_MOUSESHIFT), "+Shift",     "")
//   cTxt += if(InkeyStatus(STATUS_MOUSECTRL),  "+Ctrl",      "")
//   cTxt += if(InkeyStatus(STATUS_MOUSEALT),   "+Alt",       "")
// endif
if !empty(cTxt)
   cTxt := " Flags: " + cTxt
endif
return trim(cTxt)
#else
// -------------------------------------------------------------------
// Clipper compatibility to FlagShip
function Inkey2str(key)                // Clipper: Inkey2str()   n/a
return ""
function GetInkeyFlags()               // Clipper: InkeyStatus() n/a
return ""
#endif

