/* For FlagShip 5 and 6
 * Copyright (c) 2002..05 by multisoft GmbH, Munich, Germany
 *
 * You may freely modify this source according to your needs and/or
 * re-use any parts of it in your source code.
 *
 * Release $Date: 2005/11/21 13:46:15 $
 *
 *--------------------------------------------------------------------------
 * This application demonstrates different printing options
 *
 * Copy the source to your working directory, and invoke
 * - on Linux/Unix: FlagShip printer.prg -o printer ; ./printer [-io=t|g]
 * - on MS-Windows: FlagShip printer.prg ; printer [-io=t|g]
 *--------------------------------------------------------------------------
 */
#include "inkey.fh"

// comment-out defines for parts which should not be tested
#define DO_PART1
#define DO_PART2
#define DO_PART3
#define DO_PART4
#define DO_PART5
#define DO_PART6
#define DO_PART7

local fileOut, oMenuItem, key, iLdpi, iLdpiMM
local BOLD_ON, BOLD_OFF, UNDERL_ON, UNDERL_OFF, ITALIC_ON, ITALIC_OFF as char
memvar _oPrinter           // instantiated in initio.prg

if AppIoMode() == "G"      // apply/resolved for GUI mode only
   BOLD_ON    := "<b>"
   BOLD_OFF   := "</b>"
   UNDERL_ON  := "<u>"
   UNDERL_OFF := "</u>"
   ITALIC_ON  := "<i>"
   ITALIC_OFF := "</i>"
endif

set century on             // for formatted header
set date german            // for formatted header

aPrint := FS_SET("prset", { chr(13)+chr(10), chr(12)+chr(13) } )

fileOut := execname() + ".prn"

inkey()                    // clean input buffer

// assign <Ctrl-PrtScr> key to printScreen output
set key K_CTRL_PRTSCR to myHelper()

// Create ouput spooler file with pre-defined name, e.g. 'printer.prn'
// via standard ? and SET PRINTER statements
//
? "Creating output to '" + fileOut + "'"
set printer to (fileOut)
createOutput()                      // create output to 'printer.prn'
set printer to

// select printer, either forced setup selection:
_oPrinter:Setup()
// or by user's click on menu item:
* wait noecho "<html>You may " + BOLD_ON + "set printer" + BOLD_OFF + ;
*     " now via Menu " + ITALIC_ON + "File -> Printer Setup" +  ITALIC_OFF + ;
*     " or via " + BOLD_ON + "Alt+R" + BOLD_OFF + " or print to default printer..."

// 1: print the spooler file via COPY to LPT1:
//    Apply for MS-Windows only.
//    In Linux, use RUN ("lpr -P<printerName> " + fileOut)
#ifdef DO_PART1
#ifdef FS_WIN32
?  "Printing file '" + fileOut + "' unformatted directly to LPT1: via copy"
copy file (fileOut) to lpt1:        // or prn: etc.
if doserror() == 0
   delete file (fileOut)
else
   ? "could not execute 'copy " + fileOut + " lpt1:'"
endif
wait
#endif   /* FS_WIN32 */
#endif   /* DO_PART1 */

// create spooler/output file with default name, e.g. 'printer.12345'
createOutput()

// get/set some printer characteristics
* _oPrinter:MarginTop     := 10    // margins are in LPI = Lines Per Inch
* _oPrinter:MarginBottom  := 40
* _oPrinter:MarginLeft    := 20
* _oPrinter:MarginRight   := 15

// get the logical (i.e. driver's, not physical) printer
// resolution, here in Lines_per_mm = Lines_per_inch / 25.4
local nLpmm := m->_oPrinter:driverHldpi / 25.4

// set tabulators at position 20, 40, 60, 80 and 100 mm
m->_oPrinter:TabStop := {20*nLpmm, 40*nLpmm, 60*nLpmm, 80*nLpmm, 100*nLpmm} // tabs in mm
* m->_oPrinter:TabStop := {50,100,200,250,350} // or set tabs natively in lpi

// set fixed font = Courier,8
m->_oPrinter:Font := Font{"Courier", 8}

// 2: printing raw
//    You may print the file programatically,
//    either via RUN ("lpr -PmyPrinter " + fileOut)
//    or via the method Exec() of Printer class
#ifdef DO_PART2
?
? "Printing output file " + _oPrinter:InputFileName + ", raw"
_oPrinter:Exec()
wait
#endif

// 3: Creating ouput spooler file to given name and
//    print this file formatted
//    This differs from raw printing (2) for GUI where
//    additional headers, margins, tabs etc can be specified
//
set print to (fileOut)
// createOutput(.T.)             // create output with HTML formatting
createOutput()                   // create output plain
set printer to
#ifdef DO_PART3
?
? "Printing output file " + fileOut + ", formatted, Courier 8"
m->_oPrinter:ExecFormatted(fileOut, Font{"Courier", 8})
wait
#endif

// 4: now the same but with proportional fonts
//
#ifdef DO_PART4
? "Printing output file " + fileOut + ", formatted, Helvetica/10"
m->_oPrinter:ExecFormatted(fileOut, Font{"Helvetica",10} )
wait
#endif

// 5. Re-defining Menu item "Print" (GUI only)
//
#ifdef DO_PART5
if AppIoMode() == "G"
   oMenuItem := m->oTopBar:FindMenuProperty( {|obj| obj:Id == 50002 } )
   if valtype(oMenuItem) == "O"
      oFont := Font{"Helvetica",12}
      oMenuItem:Data    := {|obj, menuID| m->_oPrinter:ExecFormatted(m->_oPrinter:InputFileName, oFont)}
      oMenuItem:Caption := "&Print formatted"
      ?
      ? "<html>Redefining the Menu item " + BOLD_ON + "'File -> Print'" + BOLD_OFF + ;
        " from raw to formatted printing (Helv/12)"
      wait "<html>(you may now print the output via menu click, then) " + ;
         BOLD_ON + ITALIC_ON + "press any key..." + ITALIC_OFF + BOLD_OFF
   endif
endif
wait
#endif

// 6: PrintScreen
//
#ifdef DO_PART6
?
? "Printing current screen image, adapted to fit in " + m->_oPrinter:PageSize
m->_oPrinter:ExecPrintScreen(.T.)
wait
#endif

// 7: Dependence on the used editor
//
#ifdef DO_PART7
cls
// set console off
// set device to printer
set printer on
set printer to lpt1

? "We will demonstrate here the output dependence on the used source editor."
?
SET FONT "Courier"
? "This paragraph was created with an editor using DOS = PC-8 = ASCII = OEM"
? "character set as coming e.g. from Clipper (in UltraEdit32: view->OEMchar)"
?
? "Umlauts and some accents: -- A O U s E --"
cBytePC8 := "A"+chr(142)+chr(132)+chr(32)+ "O"+chr(153)+chr(148)+ ;
            " U"+chr(154)+chr(129)+chr(32)+ "s"+chr(225) + ;
            " E"+chr(144, 130, 138, 136)
? "       dito as bytes/PC8: -- " + cBytePC8 + " --"
? "These lines will be printed correctly, same as in Clipper, when the"
? "printer uses ASCII character set"
?
? "However, the screen output in GUI mode uses different character set,"
? "and in Terminal i/o mode it depends on the used "
#ifdef FS_WIN32
?? "CodePage."
#else
?? "TERM or FSTERM environment."
#endif
? "The best way here is to use SET GUITRANSL ASCII ON to convert this"
? "ASCII/PC8 input charset automatically to ANSI/ISO screen output, or"
? "insert the line"
? ' #include "fspreset.fh"'
? "at the begin of your application (see the .../include/fspreset.fh file)."
? "You may additionally use SET GUITRANSL TEXT ON, ...BOX ON and ...LINES ON"
? "to draw PC-8 semi-graphic characters/boxes on the screen as well."
?
SET GUITRANSL ASCII ON
? "Umlauts and some accents: -- A O U s E --"
? "       dito as bytes/PC8: -- " + cBytePC8 + " --"
SET GUITRANSL ASCII OFF // reset defaults
? replicate("-", 75)
EJECT
wait
cls

? "This paragraph was created with Windows or X11 editor using ISO = ANSI"
? "character set (in UltraEdit32: view->disable OEM charset, in Nedit: Unix)"
? "Umlauts and some accents: -- A O U s E --"
cByteISO := "A"+chr(196)+chr(228)+chr(32)+ "O"+chr(214)+chr(246)+ ;
            " U"+chr(220)+chr(252)+chr(32)+ "s"+chr(223) + ;
            " E"+chr(201, 233, 232, 234)
? "       dito as bytes/ISO: -- " + cByteISO + " --"
? "You will see it correctly on screen in GUI mode, but not in textual mode"
? "with DOS/PC8 character set, and also the ASCII printer will output garbage."
?
? "But FlagShip supports character set conversion also here :-)"
? "You have different choices (see details in fsman section LNG.5.4):"
?
? "a) SET SOURCE ISO (or compile the source with -iso switch)"
SET SOURCE ISO // or ANSI
? "   Umlauts and some accents: -- A O U s E --"
? "          dito as bytes/ISO: -- " + cByteISO + " --"
? "          dito as bytes/PC8: -- " + cBytePC8 + " --"
SET SOURCE ASCII // reset defaults
?
? "b) explicitly using the Ansi2oem() or Oem2ansi() translation:"
? ansi2oem("   Umlauts and some accents: -- A O U s E --")
? ansi2oem("    ansi2oem() of bytes/ISO: -- " + cByteISO + " --")
? ansi2oem("    ansi2oem() of bytes/PC8: -- " + cBytePC8 + " --")
// ? oem2ansi("    oem2ansi() of bytes/PC8: -- " + cBytePC8 + " --")
// ? oem2ansi("    oem2ansi() of bytes/ISO: -- " + cByteISO + " --")
? replicate("-", 75)
? "so decide the best and most comfortable way for you depending on"
? "the used editor, environment and probably available source code..."
EJECT

set device to screen
set printer off
set printer to
set console on
#endif

wait "Done..."

/* ------------------------------------------------------------------
 * create the printer output, with or w/o font attributes
 */
function createOutput(lWithAttrib)
   local ii, iNum := 5, TAB := chr(9)
   local BOLD_ON, BOLD_OFF as char
   memvar _oPrinter                  // instantiated in initio.prg

   SET HTML ON                      // enable HTML/RichText tags per default

   if valtype(lWithAttrib) == "L" .and. lWithAttrib .and. AppIoMode() == "G"
      BOLD_ON  := "<b>"
      BOLD_OFF := "</b>"
   endif

   set console off
   set print on
   for ii := 1 to 70 // 100
      if ii % 10 == 0
         ?? repli("XXXX5XXXX0",iNum)
      else
         ?? "line", strzero(ii,4)
      endif
      iNum++
      ?
   next

   eject
   // or:
   // ? chr(12)   // FF == eject new page
   if !empty(_oPrinter:TabStop)
      iLdpi   := _oPrinter:DriverHldpi    // logic.pixels per inch
      iLdpiMM := iLdpi / 25.4             // logic.pixels per mm
      ?? BOLD_ON + "Tabulators" + BOLD_OFF + ;
         " set on logical dpi position:"
      for ii := 1 to len(_oPrinter:TabStop)
         ? TAB + ltrim(_oPrinter:TabStop[ii]) + TAB + ;
           "= at " + ltrim(round(_oPrinter:TabStop[ii] / iLdpiMM,1)) + " mm" + ;
           TAB + "= at " + ltrim(round(_oPrinter:TabStop[ii] / iLdpi,2)) + " inch"
      next
      ?
   endif
   for ii := 1 to 10
      ? "aa" + TAB + "bb" + TAB + "cc" + TAB + "dd" + TAB + "ee" + TAB + "ff" + TAB + "gg"
   next
   ?
   ? BOLD_ON + "Some global printer settings modifiable by FS_SET('prset'):" + BOLD_OFF
   local aPrintData := FS_SET("prset")
   ? "new line"      + TAB + TAB + str2chr(aPrintData[1])
   ? "new page"      + TAB + TAB + str2chr(aPrintData[2])
   ? "print head column"   + TAB + str2chr(aPrintData[3])
   ? "print head leftmost" + TAB + str2chr(aPrintData[4])
   ? "print head one left" + TAB + str2chr(aPrintData[5])
   ? "print head one right" + TAB + str2chr(aPrintData[6])
   ? "print seq margin left" + TAB + str2chr(aPrintData[7])
   ?
   ? BOLD_ON + "Some Printer Characteristics modifiable by _oPrinter object:" + BOLD_OFF
   ? ":pageSize"     + TAB + TAB + _oPrinter:pagesize
   ? ":headerLeft"   + TAB + TAB + "'" + _oPrinter:headerLeft + "'"
   ? ":headerMid"    + TAB + TAB + "'" + _oPrinter:headerMid + "'"
   ? ":headerRight"  + TAB + TAB + "'" + _oPrinter:headerRight + "'"
   ? ":inputFileName" + TAB + TAB + _oPrinter:inputFileName
   ? ":marginTop"    + TAB + TAB + ltrim(_oPrinter:marginTop)
   ? ":marginBottom" + TAB + TAB + ltrim(_oPrinter:marginBottom)
   ? ":marginLeft"   + TAB + TAB + ltrim(_oPrinter:marginLeft)
   ? ":marginRight"  + TAB + TAB + ltrim(_oPrinter:marginRight)
   ? ":printExecutable" + TAB +  _oPrinter:printExecutable
   ? ":driver"       + TAB + TAB + TAB + if(empty(_oPrinter:driver),"(not set, using default)",_oPrinter:driver)
   ? ":driverHldpi"  + TAB + TAB + ltrim(_oPrinter:DriverHldpi) + ;
     " ldpi (logical dots per inch) = " + ;
     ltrim(round(_oPrinter:DriverHldpi / 25.4,2)) + " dots/pixel per mm"
   ? ":driverVldpi"  + TAB + TAB + ltrim(_oPrinter:DriverVldpi) + ;
     " ldpi (logical dots per inch) = " + ;
     ltrim(round(_oPrinter:DriverVldpi / 25.4,2)) + " dots/pixel per mm"
   ?
   ? BOLD_ON + "All m->_oPrinter object properties:" + BOLD_OFF
   SET HTML OFF
   local aOut := _displObj(_oPrinter, "_oPrinter")
   aeval(aOut, {|x| Qout(x)} )

   eject
   set printer off
   set console on
   return

static function MyHelper
   _oPrinter:ExecPrintScreen(.T.)
   return

static function str2chr(cStr)
   local cRet := "chr("
   local ii
   if valtype(cStr) != "C" .or. len(cStr) == 0
      return "not set"
   endif
   for ii := 1 to len(cStr)
      cRet += if(ii > 1, ",", "") + ltrim(strpeek(cStr, ii))
   next
   cRet += ")"
   return cRet

* eof *

