/* For FlagShip 5
 * Copyright (c) 2002..05 by multisoft GmbH, Munich, Germany
 *
 * You may freely modify this source according to your needs and/or
 * re-use any parts of it in your source code.
 *
 * Release $Date: 2005/11/21 13:47:07 $
 *
 *--------------------------------------------------------------------------
 * This application demonstrates the use of SET ANSI and Ansi2oem() function
 *
 * Copy the source to your working directory, and
 * Compile: FlagShip setansi.prg -na -Mmain -o setansi
 * Run:     ./setansi [-io=t|g] [on|off]
 *--------------------------------------------------------------------------
 */
#include "set.fh"

#define U_UML_PC8 chr(129)
#define U_UML_ISO chr(252)

FUNCTION main(cmdParam1)    // optional command-line parameter: ON or OFF

local xKey, xVal, u_uml, saveAnsi, alreadySet := .F.
local BOLD_ON := "", BOLD_OFF := "", NSPACE := " ", NSPACE3 := "   "

SET HTMLTEXT ON   // enables RichText tags in text

? "This application demonstrates the use of SET ANSI and Ansi2oem() function to" 
? "access data containing national characters > chr(127), e.g. umlauts & accents." 
?
? "Info: You are now using " 
if AppIoMode()== "G"
   BOLD_ON  := "<b>"
   BOLD_OFF := "</b>"
   NSPACE   := "&nbsp;"
   NSPACE3  := replicate(NSPACE, 3)
   ?? "GUI i/o mode"
   if empty(cmdParam1) .or. lower(cmdParam1) != "off"
      // Info: you may set another sorting table now, e.g.
      //    FS_SET("loadlang",2,"FSsortab.ger")
      //    FS_SET("setlang",2)
      // or disable the currently used sorting table via
      //    FS_SET("setlang",0)
      // but you should also remove the old indices ('rm setansi*.idx') since 
      // the used sorting table may influence the upper(...) and lower(...)
      // translation of the index key and sorting order.
      //
      ?? ", we therefore use '" + BOLD_ON + "SET ANSI ON" + BOLD_OFF + "' to translate "
      ? "DOS database chars > 127 automatically to ISO/ANSI.  See additional details"
      ? "later in this example."
      SET ANSI ON
      alreadySet := .T.
   endif
elseif AppIoMode()== "T"
   ?? "Terminal i/o with environment TERM=" + getenv("TERM")
else
   ?? "Basic i/o mode"
endif

if !empty(cmdParam1) .and. lower(cmdParam1) == "on" .and. !alreadySet
   ?
   ? 'Info: parameter "on" passed. Enabling ' + BOLD_ON + "SET ANSI ON" + BOLD_OFF + "' to translate "
   ? "DOS database chars > 127 automatically to ISO/ANSI charset. See additional"
   ? "details later."
endif
?
if FS_SET("setlang") == 0
   ? 'Info: FS_SET("loadlang"/"setlang") is ' + BOLD_ON + "disabled" + BOLD_OFF + ', natural sort order is used'
   ? "without support of upper/lowercase conversion of chr(128..255)" 
   ? "See additional description later."
else
   ? 'Info: FS_SET("loadlang"/"setlang") is ' + BOLD_ON + "enabled" + BOLD_OFF + ', probably from setsortab.prg'
   ? "which provides upper/lowercase conversion of chr(128..255)" 
   ? "See additional description later."
endif
?
? "1. First, a small DOS compatible database 'setansi.dbf' containing"
? "    few records with umlauts (Mueller) is created and indexed by"
? "   - INDEX ON field1        TO setansi1.idx"
? "   - INDEX ON upper(field1) TO setansi2.idx"
? "   - INDEX ON lower(field1) TO setansi3.idx"

if !file("setansi.dbf")
   u_uml := if(Set(_SET_ANSI), U_UML_ISO, U_UML_PC8)
   
   dbcreate("setansi",{{"field1","c",30,0},{"field2","n",3,0}} )
   USE setansi
   append blank
   replace field1 with "mueller" , field2 with 1
   append blank
   replace field1 with "Maier" , field2 with 2
   append blank
   replace field1 with "M" + u_uml + "ller =PC8/DOS Mueller/umlaut", field2 with 3
   append blank
   replace field1 with "your entries", field2 with 4
   use
   delete file setansi1.idx
endif 

USE setansi
if !file("setansi1.idx")
   INDEX ON field1        TO setansi1.idx
   INDEX ON upper(field1) TO setansi2.idx
   INDEX ON lower(field1) TO setansi3.idx
endif
SET INDEX TO setansi1, setansi2, setansi3
?
wait
if lastkey() == 27
   quit
endif

cls
?? BOLD_ON + "2. The current data of '" + dbf() + "' are:" + BOLD_OFF
Displaydata()
?
wait

cls
?? BOLD_ON + "3. Perform SEEK(..) with and w/o SET ANSI" + BOLD_OFF + " (abort by ESC)"
u_uml := if(Set(_SET_ANSI), U_UML_ISO, U_UML_PC8)
goto 4
xVal  := field1
xKey  := "M" + u_uml + "ller" + space(10) // =ANSI or PC8 Mueller/umlaut
while .T.
   @ 2,0 SAY "opt: enter value for record#4 " GET xVal
   @ 3,0 SAY "enter search key value or ESC " GET xKey
   read
   if lastkey() == 27
      exit
   endif
   replace field1 with xVal
   saveAnsi := Set(_SET_ANSI, .T.)  // SET ANSI ON
   DisplaySeek(5, trim(xKey), "M" + U_UML_ISO + "ller", "M" + U_UML_PC8 + "ller")
   SET ANSI OFF
   DisplaySeek(5, trim(xKey), "M" + U_UML_ISO + "ller", "M" + U_UML_PC8 + "ller")
   Set(_SET_ANSI,saveAnsi)          // restore SET ANSI
   goto 4
enddo

cls
?? BOLD_ON + "4. The final database content is:" + BOLD_OFF
Displaydata()
?
wait

cls
SET FONT TO "Helvetica", 12
?? BOLD_ON + "Conclusion" + BOLD_OFF
?
? "1. If you are planning to use only " + BOLD_ON + "plain ASCII" + BOLD_OFF + " character set <= chr(127)," 
? "   or don't need to keep the DOS compatiblility of your databases,"
? NSPACE3+"best not to use SET ANSI at all (" + BOLD_ON + "SET ANSI OFF" + BOLD_OFF + " is the default),"
? "   and simply ignore the rest of this conclusion (press ESC) ..."
?  
? "2. When the database contains extended character set " + BOLD_ON + "above chr(127)" + BOLD_OFF + ", "
? NSPACE3+"e.g. with German umlauts, French accents etc. _" + BOLD_ON + "and" + BOLD_OFF + "_ the application is"
? "   running in GUI mode (or Terminal i/o in X11 without TERM translation),"
? NSPACE3+"best to " + BOLD_ON + "SET ANSI ON" + BOLD_OFF + " to automatically translate the ISO/ANSI character"
? "   set (used for display in this mode) to/from the PC8/OEM/ASCII charset"
? "   (used as character set in a DOS compatible database)."
? 
? "3. The " + BOLD_ON + "index data" + BOLD_OFF + " always correspond to the database, i.e. if the database"
? "   contain a-umlaut in PC8/OEM/ANSI char set = chr(132), the corresponding"
? "   key in the index also contain chr(132) -- or chr(142) resp. if UPPER(key)"
? "   was specified as the index key."
?
wait
if lastkey() == 27
   quit
endif
cls
? "4. For the " + BOLD_ON + "upper()" + BOLD_OFF + " and " + BOLD_ON + ;
      "lower()" + BOLD_OFF + " translation of extended character set > 127,"
? "   FlagShip use freely re-definable sorting table (e.g. the file FSsortab.def)."
? "   On program start, the initio.prg module calls the user redefinable function"
? NSPACE3 + BOLD_ON + "_SetSortab()" + BOLD_OFF + " (source code is in setsortab.prg) which sets per default"
? "   the FSsortab.def as a current sorting table. It may be disabled or reset"
? '   in the application at any time later by FS_SET("setlang"/"loadlang").'
?
? "   The current sorting table (at the time of INDEX ON) is stored in the"
? NSPACE3 + BOLD_ON + "index header" + BOLD_OFF + " and cannot be changed later during the same index"
? "   operation, otherwise run-time error (RTE) occurs. This avoids possible"
? "   seek() problems when searching e.g. for upper(key) translated by"
? "   other table or w/o sorting table at all."
?
wait
cls
? '5. The SET ANSI ON automatic work fine only for strings which include'
? NSPACE3 + BOLD_ON + 'ANSI/ISO characters' + BOLD_OFF + ' (i.e. chr(228) for a-umlaut), not for strings with'
? '   PC8/OEM/ASCII characters (i.e. chr(132) for a-umlaut). '
?
? '   If you need to replace DOS formatted strings or constants (i.e. with'
? '   PC8 character set), you may '
? '   - temporarily unset the auto translation by SET ANSI OFF, e.g.'
? '       SET ANSI OFF'
? '       REPLACE field WITH cDosString'
? '       SET ANSI ON'
? '   - or transform the PC8/OEM/ASCII string to ANSI/ISO via oem2ansi(), e.g.'
? '       REPLACE field WITH oem2ansi(cDosString)'
? '   which stores the string in the database in DOS compatible format.'
?
wait
cls
?? '6. Generally: With enabled SET ANSI ON, the ' + BOLD_ON + 'auto translation' + BOLD_OFF + ' of '
? '   extended charset (PC8 -> ANSI or ANSI -> PC8) is done automatically'
? NSPACE3 + 'by FlagShip in the ' + BOLD_ON + 'DbfIdx RDD' + BOLD_OFF + ' driver as follows:'
? '   - on a field access, the statement "cVar := field" is treated same as'
? '     "cVar := Oem2ansi(field)" with disabled SET ANSI'
? '   - the character expression passed to SEEK (cVar) is handled same as'
? '     SEEK (Ansi2oem(cVar)) with disabled SET ANSI'
? '   - for a field assign, the statement "REPLACE field WITH cVar" is handled'
? '     same as "REPLACE field WITH Ansi2oem(cVar)" with disabled SET ANSI'
?
? 'Note: The SET ANSI ON is there mainly ' + BOLD_ON + 'for your convenience' + BOLD_OFF + '. For mixed'
? 'ANSI and PC8 strings in your application, best to handle it by yourself'
? 'with the Ansi2oem(..) and Oem2ansi(..) functions as described above.'
? 'You also ' + BOLD_ON + 'may fully ignore' + BOLD_OFF + ' the extended characters handling and save the'
? 'data in the database "' + BOLD_ON + 'as is' + BOLD_OFF + ;
  '". FlagShip for MS-Windows or other systems'
? 'will handle it correctly by the same way. But if you get undesirable output'
? 'of extended chars with DOS based applications (like Clipper) or in Terminal'
? 'i/o mode (which assumes DOS/PC8 character set), you will need to handle it.'
?
wait "Done, press any key ..."
quit

/*--------------------------------------------------------------------------
 * Helper: display current data,
 *   simulate the index key display
 */
FUNCTION DisplayData()
local ii, idx, lSetAnsi, xData

for idx := 1 to indexcount()
   if idx == 2
      xData := "upper(field1)"
   elseif idx == 3
      xData := "lower(field1)"
   else
      xData := "field1"
   endif

   set order to (idx)
   go top
   
   ?
   ? "----- order#" + ltrim(indexord()) + " = " + indexkey()   
   while !eof()
      ? "rec#" +ltrim(recno()) + "= '" + &(xData) + "'"
      /* 
       * to display the REAL (untranslated) values of the 
       * database, SET ANSI must be disabled
       */
      lSetAnsi := Set(_SET_ANSI, .F.)           // SET ANSI OFF
      iCol := if(AppIoMode() == "G", 23, 10 + len(field1))
      @ row(), iCol SAY " dbf = chr(" 
      for ii := 1 to 6 
         ?? ltrim(asc(substr(&(xData),ii,1))) + " "
      next
      ?? "...)"
      Set(_SET_ANSI, lSetAnsi)                  // SET ANSI ON
      if lSetAnsi
          ?? " - you see '" + left(&(xData),7) + "' as chr(" 
          for ii := 1 to 6
            ?? ltrim(asc(substr(&(xData),ii,1))) + " "
         next
         ?? "...)"
      endif
      skip
   enddo
next
return

/*--------------------------------------------------------------------------
 * Helper: seek entered value + "Mller" and display results
 */
FUNCTION DisplaySeek(iRow, cKey, cAnsi, cPC8)
local ii

@ iRow, 0 clear

? "SET ANSI is " + if(Set(_SET_ANSI), ;
   "ON  = seek automatically converts the value to PC8", ;
   'OFF = seek handles the value "as is"')
?
SET ORDER TO 1
? "----- order#" + ltrim(indexord()) + " = " + indexkey()  
SEEK (cKey)
? 'seek ("' + cKey + '")'
   @ row(), 30 SAy '= seek chr(' 
   for ii := 1 to 5 
      ?? ltrim(asc(substr(cKey,ii,1))) + " "
   next
   ?? "...)"
   @ row(), 60 say if(found(),"found, record#" + ltrim(recno()), "NOT FOUND")

? 'seek ansi2oem("' + cKey + '"))' 
SEEK (ansi2oem(cKey))
   @ row(), 30 SAy '= seek chr(' 
   for ii := 1 to 5 
      ?? ltrim(asc(substr(ansi2oem(cKey),ii,1))) + " "
   next
   ?? "...)"
   @ row(), 60 say if(found(),"found, record#" + ltrim(recno()), "NOT FOUND")

if !(cKey == cAnsi)
   SEEK (cAnsi)
   ? 'seek ("' + cAnsi + '") =ANSI' 
      @ row(), 60 say if(found(),"found, record#" + ltrim(recno()), "NOT FOUND")
   ? 'seek ansi2oem("' + cAnsi + '")) =ANSI' 
   SEEK (ansi2oem(cAnsi))
      @ row(), 60 say if(found(),"found, record#" + ltrim(recno()), "NOT FOUND")
endif
if !(cKey == cPc8)
   SEEK (cPc8)
   ? 'seek ("' + cPc8 + '")'
      @ row(), 30 SAy '(DOS const) = seek chr(' 
      for ii := 1 to 5 
         ?? ltrim(asc(substr(cPc8,ii,1))) + " "
      next
      ?? "...)"
      @ row(), 60 say if(found(),"found, record#" + ltrim(recno()), "NOT FOUND")
endif

****** index 2 = upper(...)
SET ORDER TO 2
? "----- order#" + ltrim(indexord()) + " = " + indexkey()  
? 'seek upper("' + cKey + '")' 
SEEK (upper(cKey))
   @ row(), 30 SAy '= seek chr(' 
   for ii := 1 to 5 
      ?? ltrim(asc(substr(upper(cKey),ii,1))) + " "
   next
   ?? "...)"
   @ row(), 60 say if(found(),"found, record#" + ltrim(recno()), "NOT FOUND")

? 'seek "upper(ansi2oem(' + cKey + '"))' 
SEEK (upper(ansi2oem(cKey)))
   @ row(), 30 SAy '= seek chr(' 
   for ii := 1 to 5 
      ?? ltrim(asc(substr(upper(ansi2oem(cKey)),ii,1))) + " "
   next
   ?? "...)"
   @ row(), 60 say if(found(),"found, record#" + ltrim(recno()), "NOT FOUND")

if !(cKey == cAnsi)
   ? 'seek upper("' + cAnsi + '") = ANSI' 
   SEEK (upper(cAnsi))
      @ row(), 60 say if(found(),"found, record#" + ltrim(recno()), "NOT FOUND")
   ? 'seek upper(ansi2oem("' + cAnsi + '"))' 
   SEEK (upper(ansi2oem(cAnsi)))
      @ row(), 60 say if(found(),"found, record#" + ltrim(recno()), "NOT FOUND")
endif

if !(cKey == cPc8)
   SEEK (upper(cPc8))
   ? 'seek upper("' + cPc8 + '")'
      @ row(), 30 SAy '(DOS const) = seek chr(' 
      for ii := 1 to 5 
         ?? ltrim(asc(substr(upper(cPc8),ii,1))) + " "
      next
      ?? "...)"
      @ row(), 60 say if(found(),"found, record#" + ltrim(recno()), "NOT FOUND")
endif

***** index 3 = lower(...)
SET ORDER TO 3
? "----- order#" + ltrim(indexord()) + " = " + indexkey()  
? 'seek lower("' + cKey + '")' 
SEEK (lower(cKey))
   @ row(), 30 SAy '= seek chr(' 
   for ii := 1 to 5 
      ?? ltrim(asc(substr(lower(cKey),ii,1))) + " "
   next
   ?? "...)"
   @ row(), 60 say if(found(),"found, record#" + ltrim(recno()), "NOT FOUND")

? 'seek lower(ansi2oem("' + cKey + '"))' 
SEEK (lower(ansi2oem(cKey)))
   @ row(), 30 SAy '= seek chr(' 
   for ii := 1 to 5 
      ?? ltrim(asc(substr(lower(ansi2oem(cKey)),ii,1))) + " "
   next
   ?? "...)"
   @ row(), 60 say if(found(),"found, record#" + ltrim(recno()), "NOT FOUND")

if !(cKey == cAnsi)
   ? 'seek lower("' + cAnsi + '") = ANSI' 
   SEEK (lower(cAnsi))
      @ row(), 60 say if(found(),"found, record#" + ltrim(recno()), "NOT FOUND")
   ? 'seek lower(ansi2oem("' + cAnsi + '"))' 
   SEEK (lower(ansi2oem(cAnsi)))
      @ row(), 60 say if(found(),"found, record#" + ltrim(recno()), "NOT FOUND")
endif

if !(cKey == cPc8)
   SEEK (lower(cPc8))
   ? 'seek lower("' + cPc8 + '")'
      @ row(), 30 SAy '(DOS const) = seek chr(' 
      for ii := 1 to 5 
         ?? ltrim(asc(substr(lower(cPc8),ii,1))) + " "
      next
      ?? "...)"
      @ row(), 60 say if(found(),"found, record#" + ltrim(recno()), "NOT FOUND")
endif

wait
@ iRow, 0 clear

return

* eof *

