// This example demonstrates the DbEdit() function with user defined
// handler supporting edit and auto refresh. Start the executable by 
// two or more users to see the effect of auto refresh.
//
// You may freely modify this source according to your needs and/or
// re-use any parts of it in your source code.
// 
// Release $Date: 2007/01/03 10:17:04 $
//
// Compile: FlagShip tdbedit.prg 
// Execute: ./a.out  (or tdbedit.exe)  in GUI mode
//          tdbedit -io=t              for Terminal i/o in MS-Windows
//          newfswin ./a.out -io=t     Terminal i/o on Unix

#define DEFAULT_AUTOREFRESH   // disable for own auto refresh

#include "dbedit.fh"
#include "box.fh"

if !file("tdbedit.dbf")
   ? "creating database tdbedit.dbf"
   dbcreate("tdbedit",{{"field1","c",10,0},{"field2","N",7,2}, ;
                       {"field3","d",8,0}, {"field4","c",10,0}} )
   use tdbedit
   for ii := 1 to 100
      append blank
      replace field1 with "recno " + ltrim(ii), field2 with ii, ;
              field3 with date(), field4 with time()
   next
   use
endif
use tdbedit new shared

set color to "W+/B,B/W,,,B/W"       // color apply for terminal i/o (or SET GUICOLOR ON)
clear
@ 1,1,21,51 BOX B_SINGLE            // box apply for terminal i/o (or SET GUITRANSL ...)

#ifdef DEFAULT_AUTOREFRESH
   dbedit(2,2, 20,50, ,"myUdf1")   // handler with default refresh by Tbrowse 
#else
   dbedit(2,2, 20,50, ,"myUdf2")   // handler with my own refresh 
#endif

setpos(21,0)
wait
quit

#ifdef DEFAULT_AUTOREFRESH
**************************************************************
* simple DbEdit() handler with edit and default auto-refresh 
*
#include "rddsys.fh"
function myUdf1(nMode, nCol, oTbr)
LOCAL  lk := LASTKEY()
LOCAL  getlist := {}, fieldvar
LOCAL  cSetcolor := SetColor()

//display curr. action, un-comment for debugging:
// displUdfData(nMode, nCol, lk, NIL, usersdbf())

DO CASE
CASE nMode == 5                        // at setup time:
   oTbr:AutoRefresh := 2               //    auto refresh every 2 seconds 
   return 1                            
CASE nMode == 0 .and. lk == 0          // idle or time out: 
   return 1                            //    nothing to do here
CASE nMode == 1 .or. nMode == 2        // BOF, EOF
   ?? CHR(7)
CASE nMode == DE_EXCEPT                // exception
   IF lk = 27                          //    ESC = Exit DBEDIT()
      RETURN 0
   ENDIF
   if lk != K_ENTER  
      lastkey(.T.)                     // clear lastkey
      return 1
   endif
   /* Enter = edit current cell
    */
   set color to "W+/B,R+/GR,,,B/W"     // color apply for terminal i/o
// fieldvar := &(field(nCol))
   fieldvar := FieldGet(nCol)
   @ ROW(), COL() GET fieldvar
   READ CLEAR
   IF UPDATED()
//    REPLACE &(field(nCol)) WITH fieldvar
      FieldPut(nCol, fieldvar)
      if nCol != 3
         FieldPut(3, date())
      endif
      if nCol != 4
         FieldPut(4, time())
      endif
   ENDIF
   SetColor(cSetcolor)                 // restore colors for Terminal i/o
   lastkey(.T.)                        // clear lastkey
   oTbr:refreshall()
ENDCASE
RETURN 1

#else /* not DEFAULT_AUTOREFRESH */
**************************************************************
* simple DbEdit() handler with edit and own auto-refresh 
*
#include "rddsys.fh"
function myUdf2(nMode, nCol, oTbr)
LOCAL  lk := LASTKEY()
LOCAL  getlist := {}, fieldvar, iUpdCount
LOCAL  cSetcolor := SetColor()
LOCAL  nUsers := usersdbf()            // number of users on current database
STATIC lastDbfUpdate := 0, lastUsers := 0

if nUsers > 1 .or. lastUsers > nUsers
   iUpdCount := DbObject():Info(DBI_INTGRCOUNT)  // internal dbf update counter
endif

//display curr. action, un-comment for debugging:
// displUdfData(nMode, nCol, lk, iUpdCount, nUsers)

DO CASE
CASE nMode == 5                        // at setup time:
   oTbr:TimeOut  := 2                  //    time-out every 2 seconds 
   lastDbfUpdate := 0                  //    init update counter
   lastUsers     := nUsers             //    init dbf users
   return 1                            
CASE nMode == 0 .and. lk == 0          // idle or time out: 
   if (nUsers > 1 .or. lastUsers > nUsers) .and. ;
      iUpdCount != lastDbfUpdate       //    if database changed by others,
      oTbr:refreshall()                //    refresh data display
      lastDbfUpdate := iUpdCount       //    save current dbf status
      lastUsers     := nUsers          //    consider last change of others
   endif     
CASE nMode == 1 .or. nMode == 2        // BOF, EOF
   ?? CHR(7)
CASE nMode == DE_EXCEPT                // exception
   IF lk = 27                          //    ESC = Exit DBEDIT()
      RETURN 0
   ENDIF
   if lk != K_ENTER  
      lastkey(.T.)                     // clear lastkey
      return 1
   endif
   /* Enter = edit current cell
    */
   set color to "W+/B,R+/GR,,,B/W"     // color apply for terminal i/o
// fieldvar := &(field(nCol))
   fieldvar := FieldGet(nCol)
   @ ROW(), COL() GET fieldvar
   READ CLEAR
   IF UPDATED()
//    REPLACE &(field(nCol)) WITH fieldvar
      FieldPut(nCol, fieldvar)
      if nCol != 3
         FieldPut(3, date())
      endif
      if nCol != 4
         FieldPut(4, time())
      endif
   ENDIF
   SetColor(cSetcolor)                 // restore colors for Terminal i/o
   lastkey(.T.)                        // clear lastkey
   oTbr:refreshall()
   lastDbfUpdate := DbObject():Info(DBI_INTGRCOUNT) 
ENDCASE
RETURN 1
#endif  /* DEFAULT_AUTOREFRESH */

**************************************************************
* helper: display current UDF data 
*  - in GUI mode on console window
*  - in terminal mode in last row
*
function displUdfData(nMode, nCol, lk, iUpdCount, user)
if AppIoMode() == "G"
   ??## user > 0, "users:" + ltrim(user) + " " + ;
      procstack(1) + " " + time() + " mode=" + ltrim(nMode) + ;
      " column=" + ltrim(nCol) + " LastKey=" + ltrim(lk) + ;
      if(iUpdCount != NIL, " dbfUpdateCounter=" + ltrim(iUpdCount), "")
else
   local row := row(), col := col()
   @ maxrow(),0 
   @ maxrow(),0 say "users:" + ltrim(user) + " " + ;
      time() + " mode=" + ltrim(nMode) + ; 
      " column=" + ltrim(nCol) + " LastKey=" + ltrim(lk) + ;
      if(iUpdCount != NIL, " dbfUpdateCounter=" + ltrim(iUpdCount), "")
   setpos(row,col)
endif
return NIL

*eof
