/* For FlagShip 5 and 6
 * Copyright (c) 2002..08 by multisoft GmbH, Munich, Germany
 *
 * You may freely modify this source according to your needs and/or
 * re-use any parts of it in your source code.
 *
 * Release $Date: 2008/01/14 08:22:58 $
 *
 *--------------------------------------------------------------------------
 * This application demonstrates used-defined menues with several
 * different, useful features. You may disable (comment-out) the
 * #define USE_PART* statements to run only the required part.
 *
 * Copy the source to your working directory, and
 * Compile Linux: FlagShip  [-io=g|a|t] topmenu.prg -o topmenu
 * Run          : ./topmenu [-io=t|g]
 *
 * MS-Windows   : FlagShip topmenu.prg [-io-a|g|t|b]
 * Run          : topmenu  -or-  topmenu -io=t
 */
#define USE_PART1
#define USE_PART2
#define DISPLAY_LONG

#include "boxclass.fh"

PUBLIC _menu_quit := {|obj, menuID| MyQuitQuest(obj, menuID) }  // optional: API to initiomenu.prg, below

#ifdef USE_PART1
/*
 * 1. Using the pre-defined menu (available in initiomenu.prg)
 *    The TopBar object is available in the 'oTopBar' global variable
 */
if AppIoMode() == "G"   // gui ?
   Part1()
else
   ? "The TopBar menu is pre-defined for GUI only. Skipping this test."
   wait
   cls
endif
#endif

#ifdef USE_PART2
/*
 * 2. Creating new, user defined menu
 */
Part2()
#endif

wait "Press any key to exit gracefully ..."
quit

/*--------------------------------------------------------------------------
 * 1. Using the pre-defined menu (available in initiomenu.prg)
 *    The TopBar object is available in the 'oTopBar' global variable
 */
Static FUNCTION Part1()
local oMenu, oPopup, cFont, nFont as usual
memvar oTopBar    // declared in .../sys/initiomenu.prg

@ 8,0 say "Click the 'Help' menu item now, you will see a pop-up window"
? "with four items: Context Help, About, Separator, About FlagShip."
? "Click on the currently available 'About' item."
?
? "We will modify the 'About' text, change the caption to 'Hello' and"
? "will add an additional item 'NewTop' now."
wait

@ 8,0 clear to 15,maxcol()

oMenu := oTopBar:FindMenuProperty( {|obj| obj:Caption == "&About"} )

if valtype(oMenu) != "O" .or. !IsObjClass(oMenu,"_MenuItem")
   alert("Sorry, there is no menu item 'About' available")
   return
endif
oMenu:Caption := "He&llo"
oMenu:Data := {|obj,pos| InfoBoxNew(NIL, "Hello ...", ;
      "Hello, this is a modified message of the; ;" + ;
      "'Help -> About' menu item, now redefined to;" + ;
      "'Help -> Hello'; ;The Menu-Id is " + ltrim(oMenu:Id) + ;
      if(oMenu:Id == 501,""," and will be changed now to 501")):show() }

setpos(8,0)
? "Check the 'Help' menu again, both the caption and text of the"
? "second menu item was redefined."
wait "The menu-id will be changed now (check and press any key)..."

oMenu:Id := 501
? "Menu item Id changed (check). "
?
wait "We will insert TopBar item 'NewTop' with one item 'NewTop Item' now...."

set message to 22
oMenu := MenuItem("NewTop &Item" ,{|o,n| DisplayStatus(o,n, oTopBar) },, ;
         "Status message of NewTop Item", 221)
oPopUp := PopUp()
oTopBar:InsItem(4, MenuItem("New&Top",oPopUp))
oPopUp:AddItem(oMenu)
oTopBar:Display()
wait "Watch/try the top bar anew, check the new item ..."


cls
? "The whole menu structure is (in short form):"
cFont := set(_SET_FONTNAME)
nFont := set(_SET_FONTSIZE)   // save current font
set font "courier" size 10

aeval(oTopbar:MenuStruct(), ;
      {|x| qout(space(x[1] *2), ltrim(x[2]), x[4])} )

set(_SET_FONTNAME, cFont)     // restore font
set(_SET_FONTSIZE, nFont)

#ifdef DISPLAY_LONG
? "The long form is printed to stderr, you may redefine it to a file via"
? " ./menu  2>myoutput "
DisplStruToErr(oTopbar)
#endif

?
wait "Press any key to finish Part1 ..."
cls
return


/*--------------------------------------------------------------------------
 * 2. Creating new, user defined menu
 */
Static FUNCTION Part2()
local oPopMenu1, oPopMenu2, oSub1, oSub2 as object
local lYes := .T. as logic
public oMenu
memvar oMenu

cls
setpos(8,0)
? "We will define own, simple menu structure now, stored in PUBLIC oMenu."
? "We will use only one common codeblock, although each item can have its own."
?
? "Hint: for a full featured menu, see source in .../system/initiomenu.prg"
?
wait "Press any key to create the new menu structure..."

m->oTopBar:Clear()            // delete default menu, see initiomenu.prg

oMenu := TopBar( 0, 0, 78)    // oMenu must be public

****** menu 1
oPopMenu1:=PopUp()                                                // node in main menu
oMenu:AddItem(MenuItem("Top&1", oPopMenu1, , , 100) )
oPopMenu1:AddItem(MenuItem ("Top1 &Item1", {|o,n| DisplayStatus(o,n,oMenu) }, ;
                         K_ALT_X, "Alt+X", 101) )
oPopMenu1:AddItem(MenuItem("_SEPARATOR_"))
oPopMenu1:AddItem(MenuItem ("Top1 I&tem2", {|o,n| DisplayStatus(o,n,oMenu) }, ;
                         K_CTRL_X, "Ctrl+X", 102) )
oPopMenu1:GetItem( oPopMenu1:GetLast() ):Enabled := .F.           // disable last item

****** menu 2
oPopMenu2  := PopUp()                                             // node in main menu
oMenu:AddItem(MenuItem("Top&2", oPopMenu2),,,200)                 // assign node to main menu
oPopMenu2:AddItem(MenuItem("Top2 Item1",{|o,n| DisplayStatus(o,n,oMenu) },,,210))
oPopMenu2:AddItem(MenuItem("Top2 Item2",{|o,n| SayHello(o,n) },,,220))
oPopMenu2:AddItem(MenuItem("Top2 Item3",{|o,n| DisplayStatus(o,n,oMenu) },,,230))
   oSub1 := PopUp()                                               // node in menu 2 = sub-menu
   oPopMenu2:AddItem(MenuItem("Top2 Item4 ...",oSub1,,,240))      // assign node to menu2
   oSub1:AddItem(MenuItem("Top2-4/item1",{|o,n| DisplayStatus(o,n,oMenu) },,,241))
   oSub1:AddItem(MenuItem("Top2-4/item2",{|o,n| SayHello(o,n) },,,242))
      oSub2 := PopUp()                                            // node in sub-menu = sub-sub-menu
      oSub1:AddItem(MenuItem("My sub-sub-menu",oSub2,,,310))      // assign node to sub-menu
      oSub2:AddItem(MenuItem("Text 1",{|o,n| SayHello(o,n) },,,311))
      oSub2:AddItem(MenuItem("Text 2",{|o,n| SayHello(o,n) },,,312))
   oSub1:AddItem(MenuItem("Top2-4/item3",{|o,n| DisplayStatus(o,n,oMenu) },,,243))
oPopMenu2:AddItem(MenuItem("Top2 Item5",{|o,n| DisplayStatus(o,n,oMenu) },,,250))

oMenu:Display()

wait

set color to "w/b,gr+/r,w+/b,w+/b,w+/bg"
cls
setpos(8,0)
? "You may activate the new menu items by mouse click now."
? "Hint: You also may apply features or checks from above par1() here"
wait
#ifdef DISPLAY_LONG
?
? "The long form is printed to stderr, you may redefine it to a file when invoking"
#ifdef FS_WIN32
? "  C:> TOPMENU 2>topmenu.txt "
#else
? "  ./" + execname() + "  2>topmenu.txt "
#endif
if AppIoMode() == "T"
   lYes := .F.
   @ row(), col() say " - do you want to print it (t/f) ?" get lYes
   read
endif
if lYes
   DisplStruToErr(oMenu)
endif
#endif
?
wait "Press any key to finish Part2"
cls
return

/*--------------------------------------------------------------------------
 * Helper: display status on mouse click
 *
 * The MenuItem codeblock receives 2 params: the pushed object self and
 * the menu ID#. The 3rd param was passed in the codeblock to this UDF
 * directly and is used here for the TopBar object.
 * If the codeblock (i.e. this UDF) return .F., the menu selection
 * via oTopBar:Exec() or MenuModal() is not terminated.
 */
Static Function DisplayStatus(obj, nId, oTop)
local cStack := "Menu", cMsg := "" as char
local aStack   as usual                // FindSelectedItem() may return NIL
local oBox     as object

aStack  := oTop:FindSelectedItem(.T.)  // creates 2-dim array with callstack
* oItem := oTop:FindSelectedItem()     // returns the pushed item object

aeval(aStack, {|x| cStack += " -> " + x[1] + " (" + ltrim(x[2]) + ")" } )
cStack := strtran(cStack, "&", "")

cMsg := "Currently clicked: Item# " + ltrim(nId)
cMsg += ";with displayed caption '" + strtran(obj:Caption,"&","") + "'"
cMsg += ";;The menu callstack is: ;" + cStack

oBox := InfoBoxNew(NIL, "Menu click status", cMsg)
oBox:Show()
return

/*--------------------------------------------------------------------------
 * another function triggered from menu
 */
Function SayHello(obj, nId)
alert("Hello from " + procname() + ";menu #" + ;
      ltrim(nId) + ";caption = " + obj:Caption)
return

/*--------------------------------------------------------------------------
 * Helper: display menu structure to stderr, starting at TopBar.
 * It may be used also for any PopUp object.
 */
Static Function DisplStruToErr(oTop)
local aStru as array
local ii    as intvar
local oMenu as object

aStru := oTop:MenuStruct()
?# "TopBar menu structure:"
for ii := 1 to len(aStru)
   oMenu := aStru[ii,3]
   ?# space(aStru[ii,1] *2), ltrim(aStru[ii,2])
   if IsObjClass(oMenu,"_MenuItem")
      ??# "",aStru[ii,4]
      if !oMenu:Enabled
         ??# " (Disabled)"
      endif
      if oMenu:IsSeparator()
         ??# " (Separator)"
      endif
      ??# " Id#",ltrim(oMenu:Id),"| class", IsObjClass(oMenu)[1]
      if valtype(oMenu:Data) == "O"
         ??# " | MenuItem:Data = class", IsObjClass(oMenu:Data)[1], "MenuItem:Data:Current=", ltrim(oMenu:Data:Current)
      endif
   else
      ??# " (" + aStru[ii,4] + ") Current=" + ltrim(oMenu:current),"| class", IsObjClass(oMenu)[1]
   endif
next
?#
return NIL

/*--------------------------------------------------------------------------
 * API to initiomenu.prg, see additional description there
 */
Function MyQuitQuest(obj, menuID)
local iPos
iPos := alert("Quit the application '" + execname() + "' now?", {"Yes","No"})
if iPos == 1
   quit       // <-- note!
endif
return .F.

/* eof */

