/* For FlagShip 5 and 6. 
 * Copyright (c) 2004..05 by multisoft GmbH, Munich, Germany
 * You may modify this source according to your needs and distribute the
 * object code freely with your executable. You may distribute also the 
 * modified source code when your application is distributed in source, 
 * but may not remove the copyright notice.
 * 
 * Release $Date: 2005/11/21 13:50:44 $
 */
// ---------------------------------------------------------------------
// This small example demonstrates FlagShip's capability and several
// mechanisms to achieve your internalization, independent on the used
// source-code editor. It displays Mnchen or Mnchen in dependence 
// on the used character set of your editor (DOS/OEM/ASCII-ANSI/ISO),
// the embedded ASCII/ISO translation and the compiler switch -iso. 
//
// Copy this file to your working directory and compile/run on
//   Unix/Linux: FlagShip umlauts.prg -o umlauts ; ./umlauts
//      -and-  : FlagShip umlauts.prg -iso -o umlauts ; ./umlauts
//               (for Terminal i/o, use newfswin|newfscons ./umlauts -io=t)
//   MS-Windows: FlagShip umlauts.prg ; umlauts
//      -and-  : FlagShip umlauts.prg -iso ; umlauts
//               (for Terminal i/o, invoke umlauts -io=t)
//
// and watch the difference. You may add your national characters in
// similar way below (using your preferred editor) and see the results 
// to decide your programming style or the used FlagShip translation.
//
// See also fsman section LNG.5.3 and LNG.5.4 for further details, or
// pc8lines.prg and fscheck.prg for more examples and SET ANSI ON/OFF
// for controlling the input/output of database storage corresponding 
// to your internalization.
//
// In the second stage, it demonstrates how to use umlauts and 
// internalization for databases to remain compatible to DOS/OEM
// ---------------------------------------------------------------------

set font "courier"  // you may comment-out this line to use default font

// You also may enable one of following statements for an automatic
// source conversion in dependence of the used editor:
* SET SOURCE ASCII
* SET SOURCE ISO

// or the general lower/upper conversion and common settings
// via fspreset.fh (available in <FlagShip_dir>/include):
* #include "fspreset.fh"

? '-- plain "as is" --'
mydisplay()
?
? "-- the same data with SET GUITRANSL ASCII ON --"
SET GUITRANSL ASCII ON
mydisplay()
wait
myDatabase()
wait
quit

/* -------------------------------------------------------------------
 * display umlauts (here Muenchen) in different character sets
 */
function mydisplay()
local x, y, z, rr, cc
x := "Mnchen in ISO/ANSI  -or- "
y := "Mnchen in DOS/OEM/ASCII format"
z := ""   // <-- insert your entry in the quotas...
? x, y, z
? "oem2ansi:", oem2ansi(x), oem2ansi(y), oem2ansi(z)
? "ansi2oem:", ansi2oem(x), ansi2oem(y), ansi2oem(z)

cc := 2
rr := row() +1
@ rr++, cc say "Mnchen   = ISO/ANSI"
@ rr++, cc say "M\374nchen   = ISO/ANSI direct per chr(252) = octal 374"
@ rr++, cc say "Mnchen   = DOS/OEM/ASCII"
@ rr++, cc say "M\201nchen   = DOS/OEM/ASCII direct per chr(129) = octal 201"
@ rr++, cc say "a u o s   = ISO/ANSI"
@ rr++, cc say "a u o s   = DOS/OEM/ASCII"
@ rr++, cc say "para    = ISO/ANSI and direct  \247  as chr(167)"
@ rr++, cc say "para    = DOS/OEM/ASCII and direct  \025  as chr(21)"
@ rr++, cc say "" // <-- insert your entry in the quotas...
return

/* -------------------------------------------------------------------
 * store and display umlauts in database, using PC-8 (ASCII,OEM)
 * character set to remain compatible to DOS databases and applications
 */
function myDatabase()
local lFirst, cData1, cData2
local lGUIAsci, lDbfAnsi
local cMsg := "Enter your data or ESC"

cData1 := "M" + chr(129) + "nchen in DOS/PC8/OEM/ASCII format" 
cData2 := "a\216\204 u o\231\224 s\341 in DOS/PC8/OEM/ASCII format" 

SET GUITRANSL ASCII OFF

if !file("umlauts.dbf")
   dbcreate("umlauts", {{"data", "C", 42, 0}} )

   // don't convert ISO/ANSI to PC-8/ASCII/OEM,
   // since the string is already available in PC-8 format
   SET ANSI OFF   

   USE umlauts
   append blank
   replace DATA with cData1
   append blank
   replace DATA with cData2
   append blank
   replace DATA with "your data"
   USE
endif

/* since GUI mode and ISO terminal uses ISO/ANSI 
 * character set, but we want to store it in ASCII (DOS/OEM)
 * format, we use here the automatic conversion
 */
if AppIoMode() == "G" .or. FS_SET("term") == "xterm"
  SET ANSI ON            // auto convert ISO/ANSI to PC-8/ASCII/OEM
endif

SET FONT "Helvetica", 10
cls
? "Using database 'umlauts'"

USE umlauts SHARED

lFirst := .T.
while .T.
   lGUIAsci := set(_SET_GUIASCII)
   lDbfAnsi := set(_SET_ANSI)
   @ 3,1 GET lGUIAsci CHECKBOX VALID (set(_SET_GUIASCII,lGUIAsci), .T.) 
   @ 3,4 say "SET GUITRANSL ASCII ON (translate PC-8 to ANSI for GUI display)" 
   @ 4,1 GET lDbfAnsi CHECKBOX VALID (set(_SET_ANSI,lDbfAnsi), .T.) 
   @ 4,4 say "SET ANSI ON (translate given ANSI chars to PC-8 from/into database)" 
   if lFirst
      CLEAR GETS
      lFirst := .F.
   else
      READ
   endif
   
   go top
   @ 6,1 say "Record 1: you should see Muenchen with u-umlaut"
   @ 7,1 say "not modifiable"
   @ 7,1 + Strlen2col(cMsg) +1 GET umlauts->data
   DisplayData(9,5, data) 

   skip
   @ 12,1 say "Record 2: you should see different umlauts"
   @ 13,1 say "not modifiable"
   @ 13,1 + Strlen2col(cMsg) +1 GET umlauts->data
   DisplayData(15,5, data) 
   CLEAR GETS

   skip
   DisplayData(21,5, data) 
   @ 18,1 say "Record 3: user modifiable"
   @ 19,1 say cMsg GET umlauts->data
   READ

   if lastkey() == K_ESC
      setpos(22,0)
      exit
   endif
   @ 21,0 clear 
   DisplayData(21,5, data) 
enddo
return

/* -------------------------------------------------------------------
 * helper, displays <cData> raw and using current SET ANSI translation
 */
function DisplayData(row, col, cData)
   local cc, lAnsi := set(_SET_ANSI)
   SET ANSI OFF
   @ row, col say "Raw Database " 
   cc := col()
   @ row, cc say "[" + trim(data) + "] = " + myChrOut(data,10) 
   set(_SET_ANSI, lAnsi)
   @ row+1,col say "And you see"
   @ row+1,cc  say "[" + trim(data) + "] = " + myChrOut(data,10) 
return NIL

/* -------------------------------------------------------------------
 * helper, converts <cData> in length <nSize> to chr(...) syntax
 */
function myChrOut(cData, nSize)
   local ii, cRet := "CHR("
   for ii := 1 to min(len(trim(cData)) +1, len(cData), nSize)
      cRet += if(ii > 1, ",", "") + ltrim(strpeek(cData,ii))
   next
   if nSize < len(cData)
      cRet += ",..."
   endif
   cRet += ")"
return cRet

/* eof */

