#ifdef COMMENT_ONLY
* For FlagShip 5 and 6 
* Copyright (c) 2007 by multisoft, Munich, Germany 
*
* You may modify this source according to your needs and distribute the
* object code freely with your executable. You may distribute also the 
* modified source code when your application is distributed in source, 
* but may not remove the copyright notice.
* 
* Release $Date: 2007/08/23 15:27:15 $
*
* Default Input handler for Achoice() with io=g and -io=t called via 
* ListBox:Show(). In terminal i/o mode, applicable only when  
* _aGlobSetting[GSET_T_L_ACHOICE_LBOX] is .T. (the default)
* You may replace this handler by your own, or modify it by 
* copying this source to your working directory, make changes and 
*
* Compile     : FlagShip -c -na -m -io=a achoicehand.prg
* Link Linux  : FlagShip [-io=t|g|a] myapp*.prg achoicehand.o
* Link Windows: FlagShip [-io=t|g|a] myapp*.prg achoicehand.obj
*
* Note: the 1:1 backward compatibility to FS4.48 for terminal mode is 
* also available via the Achoice4(...) instead of Achoice() function.
#endif

#include "boxclass.fh"
#include "inkey.fh"
#include "box.fh"
#include "button.fh"
#include "set.fh"
#include "achoice.fh"
#include "FSerrmsg.h"

/* --------------------------------------------------------------------------
 * Achoice() input handler, similar to ListBoxHandler() but w/o ComboBox
 * All parameters are mandatory here
 *
 * All parameters are passed from Achoice():
 * obj     = current ListBox object
 * item    = current item#
 * nDummy1 = not used here
 * nDummy2 = not used here
 * cUdfName= name of user defined function or "" or NIL
 */
FUNCTION AchoiceHandler(obj, item, nDummy1, nDummy2, cUdfName)
   local key, ii, iHeight, iStatus, retVal := 0    as intvar
   local lFound, isUdf, lReadOnly, lUdfCalled      as logic
   local uRet, iWait                               as usual
   local cAppType                                  as char
   local iOldPos, iSetKey, iCursor, iPos, keyOrig  as intvar
   local lUdfSearchHotkey, lEnterUdf               as logic
   local lAbortLeftRight, lShouldExit              as logic
   local lConsiderConfirm                          as logic
      
   memvar _aGlobSetting
   
   if valtype(obj) != "O" .or. !IsObjClass(obj, "_ListBox") 
      _Rt_Error(RTARGTYP, "AchoiceHandler(): wrong object class - " + ;
         if(valtype(obj) != "O", "not object", IsObjClass(obj)[1]) )
      return retVal 
   endif
   cAppType := AppIoMode() // or: m->oApplic:AppType == i/o mode: G,T,B
   
   isUdf := valtype(cUdfName) == "C" .and. !empty(cUdfName) .and. ;
            IsFunction(cUdfName) 

   lReadOnly := .T.
   for ii := 1 to obj:ItemCount
      if obj:ItemSelectable(ii)
         lReadOnly := .F.
         exit
      endif
   next

   /* consider global settings
    */
   lAbortLeftRight         := _aGlobSetting[GSET_L_ACHOICEABORTLR   ] // def = .F.
   lUdfSearchHotkey        := _aGlobSetting[GSET_L_ACHOICEUDFSEARCH ] // def = .F.
   obj:SelectBySingleClick := _aGlobSetting[GSET_L_ACHOICESINGLE    ] // def = .T.
   obj:SelectBySpace       := _aGlobSetting[GSET_L_ACHOICESPACESEL  ] // def = .T.
   lConsiderConfirm        := _aGlobSetting[GSET_L_ACHOICECONFIRM   ] // def = .T.
   
   obj:FilterKeys := .F.
   CLEAR TYPEAHEAD
   
*   if !lReadOnly
*      obj:SetFocus()
*   endif
*   obj:Display()
   
   iWait := 0

   // if you want to execute your UDF with AC_IDLE status 
   // before processing Achoice(), e.g. for special setup,
   // un-comment the next statement:
*  iWait := if(isUdf, 0.1, 0)    // first Inkey() wait - forces AC_IDLE in UDF

   while .T.
      iStatus     := 0
      iHeight     := obj:ItemsVisible
      lEnterUdf   := .F.                  // triggers entering UDF if present
      lUdfCalled  := .F.                  // avoid twice UDF invocation
      lShouldExit := .F.                  // handle SET CONFIRM
      
      if cAppType == "T"                  // Terminal i/o
         iCursor := SetCursor()           
         SetCursor(0)                     // cursor off
      endif

      /* wait for user key or mouse press, 
       * check for SET KEY or ON KEY or SET FUNCTION redirection. If so,
       * don't process standard keys. Redirected ESC exits Acoice()
       * only if LastKey() is still ESC, i.e. is not changed in UDF
       */
      key := Inkey(iWait, INKEY_ALL_BUT_MOVE)   
      if key != 0 
         iSetKey := _SetKeyEval(key, .T., .T.)  // user redirection?
         if iSetKey != 0                        // yes, was executed 
            if key == K_ESC                     // if ESC was redirected,
               if lastkey() != K_ESC            // exit only if lastkey() = K_ESC
                  loop
               endif
            else
               loop                             // other keys = new input
            endif
         endif
      endif
      keyOrig := key
         
      /* Mouse double-click simulates RETURN,
       * Mouse left click too, when obj:SelectBySingleClick is enabled
       * Also space may simulate RETURN if obj:SelectBySpace is enabled
       */
      if (key == K_SPACE .and. obj:SelectBySpace) .or. ;
         ( obj:HitTest(Mrow(.T.), Mcol(.T.), .T.) > 0 .and. ;
           ( key == K_LDBLCLK .or. ;       
            ( key == K_LBUTTONUP .and. obj:SelectBySingleClick) ) )
      
         if !obj:ItemSelectable(obj:Buffer)     // only selectable items
            loop
         endif
         key := K_RETURN
         keyboard chr(key)       // UDF will probably check lastkey()
         inkey()                 // so simulate Return key press
      endif
      
      /* if you want to evaluate also mouse events in your UDF,
       * disable the following if ... endif sequence
       */
      if key >= K_MOUSEMOVE .and. key <= K_ALT_RBUTTONUP
         loop                    // ignore other mouse actions
      endif

      iWait := 0                 
      
      // if you want to trigger your UDF frequently even without key press
      // with the AC_IDLE status, un-comment the next statement:
*     iWait := if(isUdf, 2, 0)   // 2nd and next Inkey() wait 
      
      do case
      case key == K_ESC .and. !isUdf            // ESC = exit
         retVal := 0  
         obj:KillFocus() 
         exit

      case key == K_RETURN .and. !isUdf         // RETURN = exit
         if !obj:ItemSelectable(obj:Buffer)     // only selectable items
            loop
         endif
         retVal := obj:Buffer   
         lShouldExit  := retVal > 0 .and. ;        // exit if found and
                         lConsiderConfirm  .and. ; // consider SET CONFIRM ON/OFF 
                         !set(_SET_CONFIRM)        // SET CONFIRM is OFF 
         if lShouldExit .or. keyOrig == K_RETURN
            obj:KillFocus() 
            exit
         endif
         lShouldExit := .F.  

      case key == K_UP                          // previous item
         iStatus := if(obj:CurrItemNo == 1, 1, iStatus)
         if lReadOnly
            obj:Select(obj:CurrItemNo -1)
         else
            obj:PrevItem()
         endif
         if isUdf .and. obj:HitTop
            iStatus := AC_HITTOP                // 1: notify UDF 
            uRet := &cUdfName(iStatus, obj:CurrItemNo, obj:CurrItemNo - obj:TopItem, obj)
            lUdfCalled := .T.
         endif

      case key == K_DOWN                        // next item
         iStatus := if(obj:CurrItemNo == obj:ItemCount, 2, iStatus)
         if lReadOnly
            obj:Select(obj:CurrItemNo +1)
         else
            obj:NextItem()
         endif
         if isUdf .and. obj:HitBottom
            iStatus := AC_HITBOTTOM             // 2: notify UDF 
            uRet := &cUdfName(iStatus, obj:CurrItemNo, obj:CurrItemNo - obj:TopItem, obj)
            lUdfCalled := .T.
         endif

      case (key == K_LEFT .or. key == K_RIGHT) .and. !isUdf  ; // abort
            .and. lAbortLeftRight                              // user setting
         retVal := 0  
         obj:KillFocus() 
         exit
         
//Enable for backward compatibility to VFS releases 5.1.4 and below which
//then triggers the UDF with status/mode=0. If disabled, mode=3 is used.
//    case key == K_LEFT .and. cAppType == "T"  // for Terminal mode only
//       obj:ColumnLeft := max(1, obj:ColumnLeft - int((obj:Right - obj:Left) / 2) )
//       obj:Display()
//    case key == K_RIGHT .and. cAppType == "T"  // for Terminal mode only
//       obj:ColumnLeft := obj:ColumnLeft + int((obj:Right - obj:Left) / 2)
//       obj:Display()

      case key == K_PGUP                        // page up
         iOldPos := obj:Buffer
         if valtype(obj:VScroll) $ "CO"
            obj:Scroll(HTSCROLLBLOCKDEC)
         else
            obj:TopItem := max(1, obj:TopItem - iHeight)
            obj:Select(obj:Buffer - iHeight) 
         endif
         if !lReadOnly .and. !obj:ItemSelectable(obj:CurrItemNo)  // if not selectable,
            obj:PrevItem()                      // skip to last selectable
         endif
         if isUdf .and. (obj:HitTop .or. iOldPos == obj:Buffer)
            iStatus := AC_HITTOP                // 1: notify UDF 
            uRet := &cUdfName(iStatus, obj:CurrItemNo, obj:CurrItemNo - obj:TopItem, obj)
            lUdfCalled := .T.
         endif

      case key == K_PGDN                        // page down
         iOldPos := obj:Buffer
         if valtype(obj:VScroll) $ "CO"
            obj:Scroll(HTSCROLLBLOCKINC)
         else
            obj:TopItem := min(obj:TopItem + iHeight, obj:ItemCount)
            obj:Select(obj:Buffer + iHeight) 
         endif
         if !lReadOnly .and. !obj:ItemSelectable(obj:CurrItemNo)  // if not selectable,
            obj:NextItem()                      // skip to last selectable
         endif
         if isUdf .and. (obj:HitBottom .or. iOldPos == obj:Buffer)
            iStatus := AC_HITBOTTOM             // 2: notify UDF 
            uRet := &cUdfName(iStatus, obj:CurrItemNo, obj:CurrItemNo - obj:TopItem, obj)
            lUdfCalled := .T.
         endif

      case key == K_CTRL_PGUP .or. ;
          (key == K_HOME .and. !isUdf)          // first item in list
          if(lReadOnly, obj:Select(1), obj:FirstItem())

      case key == K_CTRL_PGDN .or. ;
          (key == K_END  .and. !isUdf)          // last item in list
         obj:Select(obj:ItemCount)
         if !lReadOnly .and. !obj:ItemSelectable(obj:CurrItemNo) // if not selectable,
            obj:PrevItem()                      // skip to last selectable
         endif

      case key == K_CTRL_HOME                   // first in window
         obj:Select(obj:TopItem)
         while !lReadOnly .and. obj:CurrItemNo <= obj:ItemCount ;
            .and. !obj:ItemSelectable(obj:CurrItemNo) 
            obj:NextItem()
         enddo

      case key == K_CTRL_END                    // last in window
         obj:Select(obj:TopItem + obj:ItemsVisible -1)
         while !lReadOnly .and. obj:CurrItemNo > 1 ;
            .and. !obj:ItemSelectable(obj:CurrItemNo) 
            obj:PrevItem()
         enddo

//    case key == -71                           // GUI only
//       /* GUI: selected by mouse
//        * if SET CONFIRM is on, loop to Inkey() for ENTER or other key
//        * otherwise accept == simulate ENTER press
//        */
//       if !set(_SET_CONFIRM)
//          KEYBOARD chr(K_RETURN) 
//       endif

      case key >= K_MOUSEMOVE .and. ;
           key <=  K_ALT_RBUTTONUP
         lEnterUdf := .T.

      case key >= 32 .and. key <= 255 .and. ;     // check for key in list text
           (!isUdf .or. lUdfSearchHotkey)
         iPos   := obj:CurrItemNo
//       retVal := obj:FindText(chr(key), iPos +1, .F., , .T., .T.)      // ignore selectable
         retVal := obj:FindText(chr(key), ;
            if(iPos < obj:ItemCount,iPos+1,1), .F., , .T., .T., .T.)     // only selectable
         if (retVal == 0 .or. retVal == iPos) .and. chr(key) != " "
//          retVal := obj:FindText(chr(key), iPos +1, .F., , , .T.)      // ltrim(text) 
            retVal := obj:FindText(chr(key), ;
               if(iPos < obj:ItemCount,iPos+1,1), .F., , , .T., .T.)     // only selectable
         endif
         if retVal > 0
            obj:Select(retVal)
         else
            obj:Select(iPos)
         endif
         lEnterUdf    := .T.
         lShouldExit  := retVal > 0 .and. ;        // exit if found and
                         lConsiderConfirm  .and. ; // consider SET CONFIRM ON/OFF 
                         !set(_SET_CONFIRM)        // SET CONFIRM is OFF 
         
      case isUdf
         lEnterUdf := .T.

      endcase

      if isUdf .and. lEnterUdf .and. !lUdfCalled
         /* calls UDF, passing parameters
          *   par1) num: status 0..4
          *   par2) num: current item#
          *   par3) num: relat.position
          *   par4) obj: current ListBox object (new in FS5)
          */
         if key != 0 .and. lReadOnly
            iStatus := 4                 // AC_NOITEM = Display mode, no choice
         elseif key != 0 .and. iStatus == 0
            iStatus := 3                 // AC_EXCEPT = Exception key pressed
         else
            iStatus := 0                 // AC_IDLE   = Idle, no more key pressed
         endif   
         iPos := obj:CurrItemNo
         uRet := &cUdfName(iStatus, obj:CurrItemNo, obj:CurrItemNo - obj:TopItem, obj)
         lUdfCalled := .T.
      endif
     
      if lUdfCalled
         iPos := obj:CurrItemNo
         if valtype(uRet) == "N" 
            do case
            case uRet == AC_ABORT      // 0 = abort ACHOICE() return 0
               retVal := 0  
               obj:KillFocus() 
               exit
            case uRet == AC_SELECT     // 1 = leave ACHOICE() return element number
               retVal := obj:Buffer   
               obj:KillFocus() 
               exit
            case uRet == AC_CONT       // 2 = continue
               obj:Select(iPos)
               loop
            case uRet == AC_GOTO       // 3 = go to next item started with the same keypressed
               retVal := obj:FindText(chr(lastkey()), , .F.)
               obj:Select(if(retVal > 0, retVal, iPos))
               loop
            case uRet < 0              // goto abs(uRet)
               obj:Select(abs(uRet))
               loop
            endcase
         endif
         obj:Select(if(retVal > 0, retVal, iPos))
      endif
      
      if lShouldExit
         exit
      endif
   enddo
   if cAppType == "T"
      SetCursor(iCursor)            // cursor on
   endif
   return retVal

* eof *
