/****************************************************************************
 * rddcb4a.c     RDD API for the CB4FOX, CB4CDX, CB4NDX, CB4MDX and CB4NTX  *
 *               by using CodeBase 5.1 multiplatform version (not included  *
 *               in the FlagShip package, contact Sequiter Inc.)            *
 *                                                                          *
 *               Note: all CodeBase function & defines are marked by CB5    *
 *                                                                          *
 * (c) 1995..2004 multisoft Datentechnik GmbH, Munich. All rights reserved  *
 * Release $Date: 2004/10/27 12:52:09 $                                     *
 *--------------------------------------------------------------------------*
 * This file contains 1st part of the minimal/full driver:                  *
 *    CB4xxxNEW      FUNCTION object creator for different CB4... drivers   *
 *    oCb4:init()    METHOD   init class for one working area               *
 *    oCb4:close()   METHOD   close current working area                    *
 *    oCb4:used      ACCESS   is the database open ? (T/F)                  *
 *    oCb4:axit()    METHOD   estroy object                                 *
 * For additional methods, see rddcb4b.c to rddcb4?.c                       *
 ****************************************************************************/

/* Driver rddcb4fs.h #define  .o file  indexext()    CB #define  CB lib     */
/*                                                                          */
/* CB4CDX   CB4FS_DBFCDX   rddcb4a1.o  .cdx (C5,Fox) S4FOX      libcbfox.a  */
/* CB4FOX   CB4FS_DBFFOX   rddcb4a2.o  .idx (Fox)    S4FOX      libcbfox.a  */
/* CB4NDX   CB4FS_DBFNDX   rddcb4a3.o  .ndx (dB3)    S4NDX      libcbndx.a  */
/* CB4MDX   CB4FS_DBFMDX   rddcb4a4.o  .mdx (dB4)    S4MDX      libcbmdx.a  */
/* CB4NTX   CB4FS_DBFNTX   rddcb4a5.o  .ntx (C5,C87) S4CLIPPER  libcbntx.a  */

#include "rddcb4fs.h"                            /* RDD API for CodeBase    */
#include "FSextend.h"                            /* for malloc (_xalloc)    */

static CODE4 code_base;                          /* CB5 global structure    */
static FSvar *currObject = NULL;

extern FSint   fgsGetExclusive();                /* SET EXCLUSIVE ON/OFF ?  */
extern FSint   fgsGetDelete();                   /* SET DELETED   ON/OFF ?  */
extern FSint   fgsGetUpperFiles();               /* FS_SET("upper") ?       */
extern FSint   fgsGetLowerFiles();               /* FS_SET("lower") ?       */
extern FSchar *fgsPathFile();                    /* search for file         */
extern FSchar *fgsGetAbsPath();                  /* add SET DEFA path       */
extern FSvoid  fgsConvertFile();                 /* conv.acc.FS_SET(up/low) */
extern FSvoid  fgsAddDefPath();                  /* add SET DEFAULT path    */

extern FSint  CB4FSudf(callError)();             /* in rddcb4b.c            */
extern FSint  CB4FSudf(checkFile)();             /* in rddcb4a.c            */
extern FSint  CB4FSudf(getCurrAutolock)();       /* in rddcb4a.c            */

extern FSvar  *fsFnNAME(dbselectar)();           /* std fn DbSelectArea()   */
extern FSvar  *fsFnNAME(dbobject)();             /* std fn DbObject()       */
extern FSvar  *fsFnNAME(errornew)();             /* std fn ErrorNew()       */

extern FSvar  *CB4FS_CALL_METH(axit)();          /* cb4:axit()  rddcb4a.c   */
extern FSvar  *CB4FS_CALL_METH(close)();         /* cb4:close() rddcb4a.c   */
extern FSvar  *CB4FS_CALL_METH(init)();          /* cb4:init()  rddcb4a.c   */
#ifndef CB4FS_MINIMAL_RDD
 extern FSvar * CB4FS_CALL_METH(clearindex)();   /* cb4:clearindex()        */
#endif
static FSvar * rddCreateMethArray();             /* forward declaration     */

#define CB4_INITPARAM    "cb4:Init(cDbfName, [...])"
#define CB4_NOTCREATED   "cb4:Init() create object by CB4xxxNEW first"
#define CB4_STILOPEN     "cb4:Init() database still open, use cb4:Close() first"
#define CB4_OPENERROR    "cb4:Init() open error %d for file '%s'"
#define CB4_CANNOTCLOSE  "cb4:Close() cannot close database, CodeBase error %d"
#define CB4_WRONGARG     "%s;wrong type or missing argument # %d"
#define CB4_NAMETOOLONG  "%s;file name too long"
#define CB4_ALIASERROR   "%s;cannot create/change ALIAS"
#define CB4_FILENOTFOUND "%s;file %s not found"
#define CB4_CB4NEWPAR    "%sNEW (cDbfName,[lShareMode],[lReadOnly],[NIL,NIL,lNew]);wrong type (%c) or missing argument # %d"

FSchar  fileNameBuff[PATHMAX];                   /* global buffer           */
FSchar  charBuff500[501];                        /* global buffer           */

/*--------------------------------------------------------------------------*
 | oCb4:Init() METHOD, called from CB4xxxNew() creator function, available  |
 |             in the rddcb4.prg file                                       |
 |                                                                          |
 | Automatically called from DBUSEAREA() or USE ... VIA CB4xxx              |
 |   or from the object creator obj := CB4xxx {param} or the equivalent     |
 |   CB4xxxNew(param) function. Note: the CB4xxx:Init() should NOT be       |
 |   called directly from the user application.                             |
 |                                                                          |
 | See parameters description in FS manual OBJ.DBSERVERNEW()                |
 |                                                                          |
 | Parameters of CB4xxxNEW() [or from DBUSEAREA()] are passed as:           |
 |   argv[0] = object:self                            must                  |
 |   argv[1] = expC1 [expC3] = file name (and path)   must                  |
 |   argv[2] = expL2 [expL5] = shared                 opt, def = SET EXCLUS |
 |   argv[3] = expL3 [expL6] = read only              opt, def = r/w        |
 |   argv[4] = unused                                                       |
 |   argv[5] = unused                                                       |
 |   argv[6] = expL6 [expL1] = new                    opt, def = .T. !!     |
 |                                                                          |
 | Returns                                                                  |
 |   always the object itself, NIL if argv[0] is not passed.                |
 |   Subsequent check by USED() or oCbx:INIT is recommended.                |
 |                                                                          |
 | Description:                                                             |
 |   1. check parameters, set defaults if neccessary.                       |
 |   2. if a file is already opened in this area, close it first            |
 |   3. convert file name [& path] according to FS_SET(...)                 |
 |   4. search for .dbf file in current directory,                          |
 |      if not found, search in SET PATH                                    |
 |   5. open the specified file using the low-level function                |
 |      open the associated .dbt if any                                     |
 |      goto top (recno 1)                                                  |
 |   6. allocates FlagShip variable storage of corresponding fields         |
 |   7. allocates object-hiden data                                         |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (init)              /* e.g.  _bb_cb4cdx_init_me(argc, argv) */
{

   FSvar    *ovSelf;                             /* object                  */
   FSvar    **arrFSvars;                         /* array of ptr to FS vars */
   FSvar    *ptrFSvar;
   FSvar    *parv[1];                            /* temp.parameters passing */
   FSuchar  *str;
   FSchar   *cPtr, cType;
   FSint    err = 0, ok = 0, ii, jj;
   FSint    iShared   = 0;                       /* default: no shared      */
   FSint    iReadOnly = 0;                       /* default: no read only   */
   FSint    iOpenNew  = 1;                       /* default: new work area  */

   static int cb4init = 0;                       /* CodeBase initialized ?  */

   CODE4    *codeBasePtr = NULL;                 /* CB5 structure           */
   DATA4    *dataPtr = NULL;                     /* CB5 structure           */
   FIELD4   *fieldPtr;                           /* CB5 structure           */

   iShared = !fgsGetExclusive();                 /* SET EXCLUSIVE status    */

/*
 *  parameter checking, set defaults
 */
   if ((argc < 1) || !(IS_VAR_OBJ(argv[0])) ) {
      rt_error(RTARGCNT, CB4_INITPARAM);       /* RTE: wrong par, no object */
      return (NIL_VAR);
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("Init",ovSelf);

   SET_VAR_COPY(VAR_OBJ (ovSelf,CB4OBJ_ERROBJECT),    /* create new error   */
                fsFnNAME(errornew)(0,0) );            /*  object for RTE    */

/*
 *  Set default extentions
 */
   SET_VAR_CHR ( VAR_OBJ (ovSelf,CB4OBJ_DBF_EXT), CB4FSdbfExt);
   SET_VAR_CHR ( VAR_OBJ (ovSelf,CB4OBJ_DBT_EXT), CB4FSdbtExt);
   SET_VAR_CHR ( VAR_OBJ (ovSelf,CB4OBJ_IDX_EXT), CB4FSindexExt);

   if (argc > 7) argc = 7;
   switch (argc) {
      case 7: if (!IS_VAR_LOG(argv[6]) && !IS_VAR_NIL(argv[5])) err = 7;
              else if (! (IS_VAR_TRUE(argv[6])) ) iOpenNew = 0;
      case 4: if (!IS_VAR_LOG(argv[3]) && !IS_VAR_NIL(argv[3])) err = 4;
              else iReadOnly = IS_VAR_TRUE(argv[3]);
      case 3: if (!IS_VAR_LOG(argv[2]) && !IS_VAR_NIL(argv[2])) err = 3;
              else if(IS_VAR_LOG(argv[2])) iShared = IS_VAR_TRUE(argv[2]);
/*
 *  If the init routine receive a NIL for the file name, then it
 *  returns the object back to FlagShip, because someone want's
 *  to make an ::*Info() query for which a database is not necessary
 */
      case 2: if (IS_VAR_NIL(argv[1]))  goto short_out;
		      if (!IS_VAR_CHR(argv[1])) err = 2;
   }
   if (err) {
      cType = ((argc >= err) ? VAR_ISTYPE(argv[err-1]) : '?');
#ifdef TEST_ON
      fprintf (stderr,"\n\r** error %d in %s/%d, varType = %c",
        err, __FILE__, __LINE__, cType);
#endif
      sprintf (repdum, CB4_CB4NEWPAR, CB4FSrddName, cType, err);
      CB4FSudf(callError)(ovSelf, 1, RTARGTYP,
               CB4FSrddName, repdum, __FILE__, __LINE__);           /* RTE  */
      return CB4FS_DEBUG_END ("Init",ovSelf);
   }

/*
 *  Check if .dbf is still open. May happened on subsequent instantiation
 *  with "the same WA" w/o oCb4:close(), otherwise checked in DBUSEAREA().
 */
   if (!iOpenNew && cb4init && IS_VAR_TRUE(VAR_OBJ(ovSelf,CB4OBJ_USED))) {
      codeBasePtr = VAR_SPECIAL(VAR_OBJ(ovSelf,CB4OBJ_CODE4));
      if (codeBasePtr) {
         dataPtr = VAR_SPECIAL(VAR_OBJ(ovSelf,CB4OBJ_DATA4));
         if (dataPtr) {
            CB4FS_CALL_METH(close) (1, argv);    /* call oCb4:close(...)    */
         }
      }
   }
   SET_VAR_LOG(VAR_OBJ (ovSelf,CB4OBJ_USED), 0);      /* oCb4:USED = .F.    */

/*
 *  init CodeBase CODE4 structure, but only once
 *  and set default/global CB5 switches
 */
   if (!cb4init) {
     d4init (&code_base);                        /* CB5 fn: init structure  */
     cb4init++;
     code_base.auto_open        = 0;             /* don't open index autom  */
     code_base.go_error         = 0;             /* don't print error GOTO  */
     code_base.field_name_error = 0;             /* don't print errr fget() */
     code_base.read_lock        = 0;             /* don't auto-readlock     */
     code_base.safety           = 0;             /* SET SAFETY = OFF        */
     code_base.lock_attempts    = 0;             /* handled by AUTOxLOCK()  */
#ifdef CODEBASE6
     code_base.singleOpen       = 0;
	 code4unlockAutoSet(&code_base, LOCK4OFF);   /* see the README.CB4      */
#endif
   }

   code_base.exclusive = !iShared;               /* CB5 SET EXCLUSIVE on/off*/
   code_base.read_only =  iReadOnly;             /* CB5 USE ... READONLY    */

/*
 *  init the object instances
 */
   SET_VAR_SPECIAL( VAR_OBJ (ovSelf,CB4OBJ_CODE4),    /* CB5: ptr to        */
                    (CODE4 *) &code_base);            /*  CODE4 structure   */

   SET_VAR_CHR (VAR_OBJ (ovSelf,CB4OBJ_DRIVERNAME),   /* init driver name,  */
                CB4FSrddName);                        /*  compil.specific   */

/*
 *  Consider global/current .prg settings
 */
   if (fgsGetUpperFiles()) {                          /* FS_SET("upper") ?  */
      SET_VAR_UPPER ( VAR_OBJ (ovSelf,CB4OBJ_DBF_EXT));
      SET_VAR_UPPER ( VAR_OBJ (ovSelf,CB4OBJ_DBT_EXT));
      SET_VAR_UPPER ( VAR_OBJ (ovSelf,CB4OBJ_IDX_EXT));
   }

   jj = !iShared;                                        /* xxx1 =exclusive */
   jj = (iReadOnly * 10) + jj;                           /* xx1x =readonly  */
   SET_VAR_INT (VAR_OBJ (ovSelf,CB4OBJ_FILEOPENMODE), jj);
   SET_VAR_INT (VAR_OBJ (ovSelf,CB4OBJ_RELCOUNT), 0);

/*
 *  The object is initialized now.
 *  Add .dbf extention, if not given. The given name is ALLTRIMmed.
 *  Convert file name [& path] according to FS_SET(...)
 *  Check if file exists, consider the SET PATH
 */

   if (! CB4FSudf(checkFile)
         (VAR_CHR(argv[1]), fileNameBuff, &cPtr,
         VAR_CHR( VAR_OBJ(ovSelf,CB4OBJ_DBF_EXT)), CB4_INITPARAM, 1))
      return CB4FS_DEBUG_END ("Init",ovSelf);

   SET_VAR_CHR (VAR_OBJ(ovSelf,CB4OBJ_FILEINPNAME),  /* partial name        */
                fileNameBuff);                       /* store in ovSelf[7]  */
   SET_VAR_CHR (VAR_OBJ(ovSelf,CB4OBJ_FILEFULLNAME), /* full name incl.path */
                cPtr);                               /* store in ovSelf[8]  */
/*
 *  Open the file, set flags, Goto top
 */
   dataPtr = d4open(&code_base, cPtr);               /* CB5 open file       */
   if ((!dataPtr) || code_base.error_code) {
      sprintf (repdum, CB4_OPENERROR,
            code_base.error_code, cPtr);             /* RTE CB5 error code  */
      CB4FSudf(callError)(ovSelf, 2, IOOPEN, "INIT", repdum,
                          __FILE__, __LINE__);
      return CB4FS_DEBUG_END ("Init",ovSelf);
   }

   d4go (dataPtr, 1L);                               /* CB5 go to recno 1   */
   SET_VAR_SPECIAL( VAR_OBJ (ovSelf,CB4OBJ_DATA4),   /* store CB5 data ptr  */
               (void *) dataPtr);                    /*   into object       */

/*
 *  Set NEW work area, if required
 */
   if (iOpenNew) {
      VAR_NEW_ARGS (parv, 1);                         /* invoke the         */
      SET_VAR_INT  (parv[0], 0);                      /* standard function  */
      fsFnNAME(dbselectar)(1, parv);                  /* DBSELECTAREA(0)    */
   }

/*
 *  Announce ALIAS name to FlagShip for hybride processing
 *  Note: Specify file or alias name in param2, or leave it NULL
 *        to determine the name automatically by using ovSelf:info()
 *        The 3rd param must be char string[11].
 */
   ok = fgsAnnounceAlias(ovSelf, fileNameBuff,
         dataPtr->alias);                            /* 0=ok, 1=new alias   */
   if (ok < 0) {                                     /* -1=cannot crea alias*/
      sprintf (repdum, CB4_ALIASERROR, CB4_INITPARAM);
      CB4FSudf(callError)(ovSelf, 1, RTILLVAL, "INIT", repdum,
                          __FILE__, __LINE__);
      return CB4FS_DEBUG_END ("Init",ovSelf);
   }

/*
 *  Announce fields of .dbf record to FlagShip & create corresponding
 *  FS variables, allocate memory for FSvar pointers (will be freed in
 *  cb4:close() ), store FSvar ptrs into object
 */
   jj = d4num_fields(dataPtr);                       /* CB5 no of fields   */
   arrFSvars =
      (FSvar **) _xalloc (sizeof(FSvar *) * jj);     /* allocate memory    */

   for (ii = 1; ii <= jj; ii++) {                    /* field 1 ...n       */
      fieldPtr = d4field_j (dataPtr, ii);            /* CB5 field struct   */
      ptrFSvar = fgsAnnounceField (fieldPtr->name,   /*     field name     */
                    ii,                              /*     field pos      */
                    (FSchar) fieldPtr->type,         /*     field type     */
                    1,                               /*     late access    */
                    fieldPtr->len,                   /*     field length   */
                    fieldPtr->dec,                   /*     field decim    */
                    f4ptr (fieldPtr) );              /* CB5 ptr to value   */
      arrFSvars[ii-1] = ptrFSvar;
   }

   SET_VAR_SPECIAL( VAR_OBJ(ovSelf,CB4OBJ_FIELDVARS), /* store arr of ptrs  */
               (void *) arrFSvars);                   /*   into ovSelf[5]   */
   VAR_ISLEN (VAR_OBJ(ovSelf,CB4OBJ_FIELDVARS)) = jj; /* store count 1..    */

   SET_VAR_LOG(VAR_OBJ (ovSelf,CB4OBJ_USED), 1);      /* oCb4:USED = .T.    */

 short_out:
   
   return CB4FS_DEBUG_END ("Init",ovSelf);
}

/*--------------------------------------------------------------------------*
 | oCb4:used ACCESS    database in use ?                                    |
 |                                                                          |
 | Parameters : ovSelf  = object:self                                       |
 | Returns    : TRUE_VAR if .dbf is open, FALSE otherwise                   |
 | Description: checks if the data4 structure of CodeBase is NULL           |
 | FS invocat.: expL := oCb4:USED                                           |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_ACCESS (used)                /* e.g.  _bb_cb4cdx_used_ac(ovSelf) */
{
   CB4FS_DEBUG_BEGIN ("Used",ovSelf);
   return CB4FS_DEBUG_END ("Used",
      (CB4FS_GETDATA4PTR && IS_VAR_TRUE(VAR_OBJ(ovSelf,CB4OBJ_USED))) ?
      TRUE_VAR : FALSE_VAR );
}

/*--------------------------------------------------------------------------*
 | oCb4:Close() METHOD closes the open database, indicies                   |
 |                     frees the allocated memory for object                |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 | Returns    : TRUE_VAR or FALSE_VAR                                       |
 | Description: if indices are open for this WA, close them. Delete         |
 |              relations, if any set. Close the open .dbf & .dbt file      |
 |              of this WA.                                                 |
 | FS invocat.: retL := oCb4:CLOSE()                                        |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (close)            /* e.g.  _bb_cb4cdx_close_me(argc, argv) */
{
   FSvar  *ovSelf;
   DATA4  *dataPtr;                              /* CB5 structure           */
   FSvar  *fieldVar;
   FSvar  **arrFSvars;                           /* array of ptr to FSvar   */
   FSint  ii, jj;
   FSint  ok = 1, isOpen = 1;
   FSvar  *retVar = FALSE_VAR;

   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("Close",ovSelf);
   if (!ovSelf || ! IS_VAR_OBJ(ovSelf) ||
      IS_VAR_NIL(VAR_OBJ(ovSelf,CB4OBJ_CODE4)) )  /* avoid recursion        */
      return CB4FS_DEBUG_END ("Close",FALSE_VAR); /* object not init.       */

   if (! VAR_SPECIAL (VAR_OBJ(ovSelf,CB4OBJ_CODE4)) ||
       ! CB4FS_GETDATA4PTR ||
       ! IS_VAR_TRUE(VAR_OBJ(ovSelf,CB4OBJ_USED)) )
      return CB4FS_DEBUG_END ("Close",FALSE_VAR); /* object not init.        */

   dataPtr = CB4FS_GETDATA4PTR;
   fieldVar = VAR_OBJ(ovSelf,CB4OBJ_FIELDVARS);  /* object element created  */

#ifdef CODEBASE6
   if(!dataPtr || !dataPtr->dataFile) {
      isOpen = 0;
      goto cleanup;
   }
#else
   if(!dataPtr) {
      isOpen = 0;
      goto cleanup;
   }
#endif

#ifndef CB4FS_MINIMAL_RDD
   CB4FS_CALL_METH(clearindex)(1, argv);         /* call cb4:ClearIndex()   */
#endif

/*  Tell FlagShip that fields of the curr WA/Object are not valid any more
 *  and the corresponding FS field variables are also invalid.
 *  Works opposite of fgsAnnounceField() in cb4:init()
 *  NOTE: fgsRetireAlias() is the LAST valid operation on the object if
 *  the default DBCLOSEAREA() or CLOSE was used.
 */
   if (IS_VAR_SPECIAL( fieldVar) ) {            /*  in cb4:init() contains */
	   jj        = VAR_ISLEN  (fieldVar);       /* stored arr length 1...  */
	   for (ii = 1; ii <= jj; ii++)
		  fgsRetireField(ovSelf, d4field_j(dataPtr, ii)->name); /*CB5 fld   */
   }

   if (ii = d4close(dataPtr)) {                /* CB5 close files in WA   */
      sprintf (repdum, CB4_CANNOTCLOSE, ii);   /* I/O error               */
      CB4FSudf(callError)(ovSelf, 2, IOCLOS, "CLOSE", repdum,
                          __FILE__, __LINE__);
      goto cleanup;
   }
   retVar = TRUE_VAR;
   
cleanup:

/*  Free the allocated memory (if any) for all SPECIAL elements of the object.
 *  Note: All object elements containing usual variables or usual FlagShip
 *  arrays will be deleted/freed automatically with the object, when the
 *  object scope becomes invalid.
 */
   if (IS_VAR_SPECIAL( fieldVar) )               /*  in cb4:init() contains */
   {
      arrFSvars = VAR_SPECIAL(fieldVar);         /*  C array of ptr to vars */
      if (arrFSvars)                             /* free allocated memory   */
         _xfree (arrFSvars);                     /* (fn from FSextend.h)    */
   }

   SET_VAR_LOG (VAR_OBJ (ovSelf,CB4OBJ_USED),0); /* oCb4:USED = .F.         */
   VAR_SPECIAL (VAR_OBJ (ovSelf,CB4OBJ_CODE4)) = 0;
   
   fgsRetireAlias(ovSelf);                       /* clar link to working ar.*/

   return CB4FS_DEBUG_END ("Close (T)", retVar);
}

/*--------------------------------------------------------------------------*
 | oCb4:Axit() METHOD destroys the class, frees all memory allocated        |
 |                    for the object                                        |
 |                                                                          |
 | Invoked just before the object variable is deleted via FS garbage collect|
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 | Returns    : ovSelf                                                      |
 | FS invocat.: not designed for .prg invocation                            |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (axit)              /* e.g.  _bb_cb4cdx_axit_me(argc, argv) */
{
   FSvar    *ovSelf;                             /* object                  */
   DATA4    *dataPtr;                            /* CB5 structure           */

   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("Axit",ovSelf);

   if (!ovSelf || ! IS_VAR_OBJ(ovSelf) ||
       ! VAR_SPECIAL (VAR_OBJ(ovSelf,CB4OBJ_CODE4)) ||
       ! IS_VAR_TRUE(VAR_OBJ(ovSelf,CB4OBJ_USED)) )
      return CB4FS_DEBUG_END ("Axit",FALSE_VAR); /* object not init.        */

/*
 *  Close .dbf, .dbt and indicies if still open. Avoid recursion via
 *  cb4:close(), here checked via NIL_VAR in CB4OBJ_CODE4 elem.
 *  Destroys and frees all SPECIAL variables of object elements.
 */

   CB4FS_CALL_METH(close) (argc, argv);         /*  call oCb4:close(...)    */

   currObject = NULL;                           /* for test purposes        */

   return CB4FS_DEBUG_END ("Axit",ovSelf);
}

/*--------------------------------------------------------------------------*
 | Internal purpose: determine the current WA and Alias from the            |
 |    field corresponding FSvariable. Used mainly for error output purpos.  |
 |    Requires inialized object and announced fields.                       |
 *--------------------------------------------------------------------------*/
int CB4FSudf(getWAdata)(ovSelf, fieldPtr, fldPos, currWA, currAlias)
                                     /* e.g.  _bb_cb4cdx_getWAdata_fn(....) */
FSvar  *ovSelf;
FIELD4 *fieldPtr;                                /* CB5 structure           */
FSint  fldPos;
FSint  *currWA;                                  /* output: curr.WA number  */
char   **currAlias;                              /* output: curr.alias name */
{
   DATA4  *dataPtr;                              /* CB5 structure           */
   FSint  fieldPos = 0;
   FSvar  **arrFSvars;                           /* array of ptr to FS vars */

   dataPtr = fieldPtr->data;
   if (!fldPos)
      fieldPos = fldPos;
   if (!fieldPos)
      fieldPos = d4field_number (dataPtr, fieldPtr->name );     /* CB5   */
   if (fieldPos > 0) {
      arrFSvars = VAR_SPECIAL( VAR_OBJ(
                    ovSelf,CB4OBJ_FIELDVARS));   /* ptr to C arr of var ptr */
      *currWA    = VAR_ISFLDWA(arrFSvars[fieldPos -1]);
      *currAlias = dataPtr->alias;
      return 1;
   }
   return 0;
}

/*--------------------------------------------------------------------------*
 | Internal purpose: add file extension if not given,                       |
 |                   convert file name to upper/lower accord to fs_set(..), |
 |                   check if the file exist                                |
 |                                                                          |
 | in:   oldFileName = programmer's entry                                   |
 | out:  newFileName (must be allocated) = converted oldFileName, see below |
 | out:  fullFileName(ptr only) = found file name, see below                |
 | in:   defExt      = default extension (e.g. ".dbf")                      |
 | in:   funText     = parameter syntax                                     |
 | in:   howToFound  = 0: search for file, don't display not-foud error  (1)|
 |                     1: search for file,       display not-foud error  (1)|
 |                    -1: don't search for file, add ext & convert l/u   (2)|
 |                    -2: don't search, add SET DEFA path                (3)|
 |                                                                          |
 |       (1) add extention if not given, search for file. If the file name  |
 |           does not include path, look for the file in SET PATH direct.   |
 |           Return the qualif. file name (upper/lower) in newFileName      |
 |           and the full name incl. path in fullFileName.                  |
 |       (2) add extention if not given. Convert to upper/lower, return the |
 |           name in newFileName.                                           |
 |       (2) add extention if not given. Convert to upper/lower, return the |
 |           name in newFileName.                                           |
 |       (3) add extention if not given. Convert to upper/lower. If the     |
 |           file name does not include path, add the SET DEFAULT path      |
 |           and return the name in newFileName.                            |
 |                                                                          |
 | Return: 1 = ok, 0 = error                                                |
 *--------------------------------------------------------------------------*/
FSint CB4FSudf(checkFile)(oldFileName, newFileName, fullFileName,
                          defExt, funText, howToFind)
char  *oldFileName, *newFileName, **fullFileName, *defExt, *funText;
FSint howToFind;
{

	static FSchar tmpBuff[512];
/*
 *  Add .dbf extention, if not given. The given name is ALLTRIMmed.
 *  Convert file name [& path] according to FS_SET(...)
 *  Check if file exists
 */

   if (! fgsAddExtension(
            newFileName,                         /* new file name           */
            oldFileName,                         /* specified name          */
            defExt,                              /* default ext             */
            PATHMAX) ) {                         /* max length              */
      sprintf(repdum, CB4_NAMETOOLONG, funText);
      rt_error (RTARGVAL, repdum);               /* RTE: name too long      */
      return 0;
   }

   fgsConvertFile(newFileName, PATHMAX);         /* FS_SET(upper/lower)     */
   if (howToFind < 0) {
      if (howToFind == -2)
         fgsAddDefPath (newFileName, PATHMAX);   /* add SET DEFAULT path    */
      *fullFileName = newFileName;
      return 1;
   }

   *fullFileName = fgsPathFile(newFileName, tmpBuff); /* srch 4 file: local,*/
   if (! (*fullFileName)[0])                     /* SET DEFA, SET PATH      */
      fgsGetAbsPath (newFileName);               /* add SET DEFAULT path    */
   if (! (*fullFileName)[0] && (howToFind > 0)) {
      sprintf(repdum, CB4_FILENOTFOUND,
              funText, newFileName);
      io_error (IOOPEN, repdum);                 /* file not found          */
      return 0;
   }
   return 1;
}

/*--------------------------------------------------------------------------*
 | Internal purpose: returns the current object                             |
 |                   used mainly for testing purposes only                  |
 *--------------------------------------------------------------------------*/
FSvar * CB4FSudf(getCurrRDDobject)()
{
   FSvar *objVar;

   objVar = fsFnNAME(dbobject)(0,0);            /* return SELECT()ed object */
   if (objVar != currObject)
      currObject = objVar;
   return currObject;
}

/*--------------------------------------------------------------------------*
 | Internal purpose: sets the current object - is NOT the DBSELECT() fn     |
 |                   used mainly for testing purposes only                  |
 *--------------------------------------------------------------------------*/
FSvar * CB4FSudf(setCurrRDDobject)(ovSelf)
FSvar *ovSelf;
{
   currObject = ovSelf;
   return currObject;
}

/**** eof rddcb4a.c ****/

