/****************************************************************************
 * rddcb4b.c     RDD API for the CB4FOX, CB4CDX, CB4NDX, CB4MDX and CB4NTX  *
 *               by using CodeBase 5.1 multiplatform version (not included  *
 *               in the FlagShip package, contact Sequiter Inc.)            *
 *                                                                          *
 *               Note: all CodeBase function & defines are marked by CB5    *
 *                                                                          *
 * (c) 1995..2004 multisoft Datentechnik GmbH, Munich. All rights reserved  *
 * Release $Date: 2004/10/27 09:41:37 $                                     *
 *--------------------------------------------------------------------------*
 * This file contains second part of the minimal/full driver:               *
 *    oCb4:alias             ASSIGN store alias                             *
 *    oCb4:alias             ACCESS retrieve alias                          *
 *    oCb4:bof               ACCESS bof() status                            *
 *    oCb4:eof               ACCESS eof() status                            *
 *                                                                          *
 *    oCb4:error()           METHOD event handler on error                  *
 *    oCb4:fieldget()        METHOD retrieve field value                    *
 *    oCb4:fieldput()        METHOD replace field value                     *
 *    oCb4:goto()            METHOD goto specific record                    *
 *    oCb4:gobottom()        METHOD goto last  logical record               *
 *    oCb4:gotop()           METHOD goto first logical record               *
 *    oCb4:skip()            METHOD skip N records                          *
 *    oCb4:quickfieldget()   METHOD quick-retrieve field value              *
 *    oCb4:quickfieldput()   METHOD quick-replace field value               *
 *                                                                          *
 * For additional methods, see rddcb4a.c to rddcb4?.c                       *
 ****************************************************************************/

#include "rddcb4fs.h"                            /* RDD API for CodeBase    */

extern FSint  fgsAnnounceAlias();                /* pass Alias to FS        */
extern FSint  fgsIntVar();                       /* converts N,I,F to int   */
extern FSint  fgsGetDelete();                    /* SET DELETE ON/OFF       */
extern FSint  CB4FSudf(getWAdata)();             /* in rddcb4a.c            */
extern FSint  CB4FSudf(initErr)();               /* in rddcb4b.c            */
extern FSint  CB4FSudf(callError)();             /* in rddcb4b.c            */
extern FSint  CB4FSudf(CSkip)();                 /* in rddcb4b.c            */
extern FSint  CB4FSudf(moveChildRelation)();     /* in rddcb4r.c            */
extern FSvar *CB4FS_CALL_METH(error)();          /* in rddcb4b.c            */
extern FSvar *CB4FS_CALL_METH(gotop)();          /* in rddcb4b.c            */
extern FSint  CB4FSudf(myAutoRFlock)();          /* in rddcb4b.c            */
extern FSint  CB4FSudf(myAutoUnlock)();          /* in rddcb4b.c            */

extern FSvar  *fsFnNAME(errorblock)();           /* std fn ErrorBlock()     */
extern FSvar  *fsFnNAME(_deferror)();            /* std fn DefError()       */
extern FSvar  *fsFnNAME(eval)();                 /* std fn eval()           */
extern FSvar  *fsFnNAME(procfile)();             /* std fn procfile()       */
extern FSvar  *fsFnNAME(procname)();             /* std fn procname()       */
extern FSvar  *fsFnNAME(procline)();             /* std fn procline()       */

#ifndef CB4FS_MINIMAL_RDD
extern FSvar  *fsFnNAME(autorlock)();            /* std fn AutoRLock()      */
extern FSvar  *fsFnNAME(autoflock)();            /* std fn AutoFLock()      */
extern FSvar  *fsFnNAME(autounlock)();           /* std fn AutoUnLock()     */
#endif

static FSvar *myFieldGet (FSvar *, FIELD4 *, DATA4 *);  /* forward declaration     */
static FSvar *myFieldPut ();                     /* forward declaration     */
static FSint  myRlockCheck ();                   /* forward declaration     */

#define CB4_REPLNOOPEN  "Attempt to replace field in closed database"
#define CB4_REPLONEOF   "Attempt to replace field %s;being on EOF() in area/alias %d/%s"
#define CB4_REPLREADON  "Attempt to replace field %s;on READ ONLY file in area/alias %d/%s"
#define CB4_REPLNOLOCK  "Attempt to replace field %s;in area/alias %d/%s;record %d not locked"
#define CB4_ERRORPAR    "cb4:ERROR([oErrObj],[cFnName])"
#define CB4_FILDGETPAR  "cb4:FIELDGET(nFldNum|cFldName)"
#define CB4_FILDPUTPAR  "cb4:FIELDPUT(nFldNum|cFldName|expB, expStored|expB)"
#define CB4_SKIPPAR     "cb4:SKIP([nCount],[lRelat])"
#define CB4_GOTOPAR     "cb4:GOTO(nRecord)"
#define CB4_WRONGARG    "%s;wrong type or missing argument # %d"
#define CB4_FILDPUTSTR  "%s;wrong type (%c) of parameter %d"
#define CB4_ALIASERROR  "%s;cannot create/change ALIAS"
#define CB4_CONCUERROR  "%s;cannot set ConcurrencyControl"

#ifdef CODEBASE6
# define DATA4_BOF      dataPtr->bofFlag
# define DATA4_EOF      dataPtr->eofFlag
#else
# define DATA4_BOF      dataPtr->bof_flag    
# define DATA4_EOF      dataPtr->eof_flag    
#endif

/*--------------------------------------------------------------------------*
 | oCb4:Alias=x  ASSIGN set the alias name                                  |
 |                                                                          |
 | Parameters : ovSelf  = object:self                                       |
 |              assVar  = expC alias name                                   |
 | Returns    : ovSelf  = object:self                                       |
 | Description: retrieves the CodeBase function                             |
 | FS invocat.: oCb4:ALIAS := expC                                          |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_ASSIGN (alias)       /* e.g.  _bb_cb4cdx_alias_as(ovSelf,assVar) */
{
   DATA4    *dataPtr;                            /* CB5 structure           */

   CB4FS_DEBUG_BEGIN ("AliasAss",assVar);
   if (dataPtr = CB4FS_GETDATA4PTR) {
      if (! IS_VAR_CHR(assVar))
         CB4FSudf(callError)(ovSelf, 1, DTMMACH, "ALIAS", "cb4:ALIAS=expC",
                  __FILE__, __LINE__);
      else {
         if (fgsAnnounceAlias(ovSelf, VAR_CHR(assVar),
               dataPtr->alias) < 0) {            /* 0=ok, 1=other, -1=error */
            sprintf (repdum, CB4_ALIASERROR, "cb4:ALIAS=expC");
            CB4FSudf(callError)(ovSelf, 1, RTILLVAL, "ALIAS", repdum,
                     __FILE__, __LINE__);
         }
      }
   }
   return CB4FS_DEBUG_END ("AliasAss", SET_VAR_CHR(VAR_NEW, dataPtr->alias));
}

/*--------------------------------------------------------------------------*
 | x=oCb4:Alias   ACCESS retrieve current alias name                        |
 |                                                                          |
 | Parameters : ovSelf  = object:self                                       |
 | Returns    : retC    = alias name                                        |
 | Description: retrieves the CodeBase function                             |
 | FS invocat.: retC := oCb4:ALIAS                                          |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_ACCESS (alias)              /* e.g.  _bb_cb4cdx_alias_ac(ovSelv) */
{
   FSvar    *cvRet = VAR_NEW;                    /* return value            */
   DATA4    *dataPtr;                            /* CB5 structure           */

   CB4FS_DEBUG_BEGIN ("AliasAcc",ovSelf);
   SET_VAR_CHR (cvRet, "");                      /* init return value       */
   if (dataPtr = CB4FS_GETDATA4PTR)
      SET_VAR_CHR (cvRet, (char *)d4alias(dataPtr));   /* CB5 -> char var  */
   return CB4FS_DEBUG_END ("AliasAcc",cvRet);
}

/*--------------------------------------------------------------------------*
 | x=oCb4:Bof   ACCESS retrieve the BOF() status                            |
 |                                                                          |
 | Parameters : ovSelf  = object:self                                       |
 | Returns    : retL    = true if BOF set                                   |
 |              On error or not open WA, FALSE is returned, same as CA/VO   |
 |              and FoxPro. Clipper 5.2 returns TRUE in such a case.        |
 | Description: retrieves the CodeBase function                             |
 | FS invocat.: retL := oCb4:BOF                                            |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_ACCESS (bof)                  /* e.g.  _bb_cb4cdx_bof_ac(ovSelv) */
{
   DATA4    *dataPtr;                            /* CB5 structure           */

   CB4FS_DEBUG_BEGIN ("Bof",ovSelf);
   if (dataPtr = CB4FS_GETDATA4PTR)
      return CB4FS_DEBUG_END ("Bof",
         d4bof(dataPtr) ? TRUE_VAR : FALSE_VAR); /* CB5 */
   return CB4FS_DEBUG_END ("Bof",FALSE_VAR);
}

/*--------------------------------------------------------------------------*
 | x=oCb4:Eof   ACCESS retrieve the EOF() status                            |
 |                                                                          |
 | Parameters : ovSelf  = object:self                                       |
 | Returns    : retL    = true if EOF set                                   |
 |              On error or not open WA, FALSE is returned, same as CA/VO   |
 |              and FoxPro. Clipper 5.2 returns TRUE in such a case.        |
 | Description: retrieves the CodeBase function                             |
 | FS invocat.: retL := oCb4:EOF                                            |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_ACCESS (eof)                  /* e.g.  _bb_cb4cdx_eof_ac(ovSelv) */
{
   DATA4    *dataPtr;                            /* CB5 structure           */

   CB4FS_DEBUG_BEGIN ("Eof",ovSelf);
   if (dataPtr = CB4FS_GETDATA4PTR)
      return CB4FS_DEBUG_END("Eof",
            d4eof(dataPtr) ? TRUE_VAR : FALSE_VAR); /* CB5 */
   return CB4FS_DEBUG_END ("Eof",FALSE_VAR);
}

/*--------------------------------------------------------------------------*
 | oCb4:ConcurrencyControl ACCESS retrieve the current state                |
 |                                                                          |
 | Parameters : ovSelf  = object:self                                       |
 | Returns    : num value                                                   |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_ACCESS (concurrenc)    /* e.g.  _bb_cb4cdx_concurrenc_ac(ovSelf) */
{
   CB4FS_DEBUG_BEGIN ("ConcurrAcc",ovSelf);
   return CB4FS_DEBUG_END ("ConcurrAcc",VAR_OBJ(ovSelf,CB4OBJ_CONCURRENCY));
}

/*--------------------------------------------------------------------------*
 | oCb4:ConcurrencyControl ASSIGN      set the current state                |
 |                                                                          |
 | Parameters : ovSelf  = object:self                                       |
 |                                                                          |
 | Returns    : num value                                                   |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_ASSIGN (concurrenc)  /* e.g.  _bb_cb4cdx_concurrenc_as(ovSelf,assVar) */
{
   CB4FS_DEBUG_BEGIN ("ConcurrAss",ovSelf);
   if (IS_VAR_NUM(assVar))
      SET_VAR_INT ( VAR_OBJ(ovSelf,CB4OBJ_CONCURRENCY), fgsIntVar(assVar));
   else {
      sprintf (repdum, CB4_CONCUERROR, "cb4:Concurrenc=expN");
      CB4FSudf(callError)(ovSelf, 1, RTILLVAL, "Concurrenc", repdum,
               __FILE__, __LINE__);
   }
   return CB4FS_DEBUG_END ("ConcurrAss",VAR_OBJ(ovSelf,CB4OBJ_CONCURRENCY));
}

/*--------------------------------------------------------------------------*
 | oCb4:FieldGet() METHOD retrieves the specified field                     |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |              argv[1] = expN field no.    -or-                            |
 |                        expC field name                                   |
 | Returns    : expU    = value of the field or NIL on error                |
 | Description: depending on the field type,                                |
 |              retrieves the corresponding CodeBase function               |
 |              Note, the field content is copied to the returning variable,|
 |              as opposite to the "quick" access cb4rdd_qFieldGet() funct. |
 |              (see rddcb4q.c) where the string ptr is only assigned to.   |
 | FS invocat.: retU := oCb4:FIELDGET(expN|expC)                            |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (fieldget)      /* e.g.  _bb_cb4cdx_fieldget_me(argc, argv) */
{
   DATA4  *dataPtr;                              /* CB5 structure           */
   FIELD4 *fieldPtr;                             /* CB5 structure           */
   FSvar  *ovSelf;
   FSint  fieldPos;
   FSvar  **adrFSvar;                            /* array of ptr to FSvar   */
   FSvar  *uvRet;                                /* return value            */

   if (argc < 2 || ! IS_VAR_OBJ(argv[0])) {
      CB4FSudf(callError)(argv[0], 1, RTMISARG, "FIELDGET", CB4_FILDGETPAR,
                          __FILE__, __LINE__);
      return NIL_VAR;
   }

   ovSelf   = argv[0];
   CB4FS_DEBUG_BEGIN ("FieldGet",ovSelf);
   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("FieldGet",NIL_VAR);

   adrFSvar = VAR_SPECIAL( VAR_OBJ(
                 ovSelf,CB4OBJ_FIELDVARS));      /* ptr to C arr of ptr to  */
   if (!adrFSvar)                                /* FSvar corresp. to field */
      return CB4FS_DEBUG_END ("FieldGet",NIL_VAR);

   switch (VAR_ISTYPE(argv[1])) {
   case 'I':
   case 'N':
   case 'F': fieldPos = (int) VAR_INTNUM(argv[1]);
             break;
   case 'E':
   case 'C': fieldPos = d4field_number (dataPtr, VAR_CHR(argv[1]));  /* CB5 */
             break;
   default:
      CB4FSudf(callError)(ovSelf, 1, RTARGTYP, "FIELDGET", CB4_FILDGETPAR,
                          __FILE__, __LINE__);
      return CB4FS_DEBUG_END ("FieldGet",NIL_VAR);
   }
   if (fieldPos <= 0)
      return CB4FS_DEBUG_END ("FieldGet (F)",NIL_VAR);  /* field not found  */

   fieldPtr = d4field_j (dataPtr, fieldPos);          /* CB5 field struct   */

   if (! fieldPtr)                                    /* field name found?  */
      return CB4FS_DEBUG_END ("FieldGet (F)",NIL_VAR);         /* no        */

   uvRet = myFieldGet (adrFSvar[fieldPos -1], fieldPtr, dataPtr);     /* store data  */
   return CB4FS_DEBUG_END ("FieldGet",uvRet);
}

/* #ifdef CB4FS_QUICKFLDACCESS */
/*--------------------------------------------------------------------------*
 | oCb4:QuickFieldGet() METHOD retrieves the specified field and stores its |
 |              value directly into the corresponding FlagShip variable.    |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |              argv[1] = ptr of field-corresponding r/o variable           |
 | Returns    : expU    = value of the field or NIL on error                |
 | Description: This hidden method cannot be invoked from the user's .prg.  |
 |              It allows FlagShip's variable system a faster access to     |
 |              the database field. Instead of searching for the field,     |
 |              it receives the corresponding field variable.               |
 |              If not specified or announced, standard FIELDGET() is used. |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (quickfieldget)  /*  _bb_cb4cdx_quickfieldget_me(argc,argv) */
{
   FSvar  *ovSelf, *uvFieldVar;
   FSint  fieldPos;
   DATA4  *dataPtr;                              /* CB5 structure           */
   FIELD4 *fieldPtr;                             /* CB5 structure           */

   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("QuickFieldGet",ovSelf);
   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("QuickFieldGet",NIL_VAR);

   uvFieldVar = argv[1];
   if (! uvFieldVar)                                       /* announced ?   */
      return CB4FS_DEBUG_END ("QuickFieldGet",NIL_VAR);    /* no            */

   fieldPos = VAR_ISFLDPOS(uvFieldVar);          /* from fgsAnnounceField() */
   fieldPtr = d4field_j (dataPtr, fieldPos);     /* CB5 field struct        */
   if (! fieldPtr)                               /* field name found?       */
      return CB4FS_DEBUG_END ("QuickFieldGet",NIL_VAR);          /* no      */

   myFieldGet (uvFieldVar, fieldPtr, dataPtr);   /* store data in FSvar     */
   return CB4FS_DEBUG_END ("QuickFieldGet",uvFieldVar);
}
/* #endif */

/*--------------------------------------------------------------------------*
 | oCb4:FieldPut() METHOD replaces  the specified field                     |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |              argv[1] = expN field no.                          -or-      |
 |                        expC field name                         -or-      |
 |                        expB block evaluated to expN or expC              |
 |              argv[2] = exp  value to be stored into the field  -or-      |
 |                        expB block evaluated to exp  value                |
 | Returns    : exp     = value of the field or NIL on error                |
 | Description: depending on the field and variable type,                   |
 |              retrieves the corresponding CodeBase function to data store.|
 |              The record is not flushed here, use the usual DBCOMMIT().   |
 | Multiuser  : In multiuser mode, at least RECORD locking is required,     |
 |              or SET AUTOLOCK ON must be set                              |
 | FS invocat.: retU := oCb4:FIELDPUT(expN|expC, expU)                      |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (fieldput)      /* e.g.  _bb_cb4cdx_fieldput_me(argc, argv) */
{
   FSvar  *ovSelf;
   DATA4  *dataPtr;                              /* CB5 structure           */
   FIELD4 *fieldPtr;                             /* CB5 structure           */
   FSint  fieldPos = 0;
   FSvar  *uvRet;                                /* return value            */
   FSvar  *uvField;                              /* field name/nr/block     */
   FSvar  *uvStore;                              /* stored variable         */
   char   *tmp;
   FSint  autoLock = 0;                          /* 1=user, -1=autoRlock    */

   if (argc < 3 || ! IS_VAR_OBJ(argv[0]) ) {
      CB4FSudf(callError)(argv[0], 1, RTMISARG, "FIELDPUT", CB4_FILDPUTPAR,
                          __FILE__, __LINE__);
      return NIL_VAR;
   }
   ovSelf  = argv[0];
   uvField = argv[1];
   uvStore = argv[2];
   CB4FS_DEBUG_BEGIN ("FieldPut",ovSelf);

   if (! (dataPtr = CB4FS_GETDATA4PTR)) {             /* dbf open/writable? */
      myRlockCheck (0, 0);                            /* no, RTE error      */
      return CB4FS_DEBUG_END ("FieldPut",NIL_VAR);
   }

   if (IS_VAR_BLK(argv[1])) {                         /* EVAL(argv[1])      */
      uvField = VAR_NEW;
      SET_VAR_COPY (uvField, VAR_EVAL_BLOCK (argv[1]));
   }
   if (IS_VAR_BLK(argv[2])) {                         /* EVAL(argv[2])      */
      uvStore = VAR_NEW;
      SET_VAR_COPY (uvStore, VAR_EVAL_BLOCK (argv[2]));
   }

   switch (VAR_ISTYPE(uvField)) {
   case 'I':
   case 'N':
   case 'F':
      fieldPos = VAR_INTNUM(uvField);
      fieldPtr = d4field_j (dataPtr, fieldPos);       /* CB5 field struct   */
      break;
   case 'C':
   case 'E':
   case 'M':
      fieldPtr = d4field (dataPtr, VAR_CHR(uvField)); /* CB5 field struct   */
      break;
   default:
      sprintf (repdum, CB4_FILDPUTSTR,
            CB4_FILDPUTPAR, VAR_ISTYPE(uvField), 1);  /* RTE: wrong par 1   */
      CB4FSudf(callError)(ovSelf, 1, RTARGTYP, "FIELDPUT", repdum,
                          __FILE__, __LINE__);
      return CB4FS_DEBUG_END ("FieldPut",NIL_VAR);
   }

   if (! (autoLock = myRlockCheck (ovSelf,fieldPtr))) /* dbf writable,lock? */
      return CB4FS_DEBUG_END ("FieldPut",NIL_VAR);    /* no, error          */

   uvRet = myFieldPut (ovSelf,                        /* replace field      */
               f4type(fieldPtr),                      /* CB5 field type     */
               fieldPtr, uvStore);

   if (autoLock < 0)                                  /* autolock ?         */
      CB4FSudf(myAutoUnlock)(ovSelf, dataPtr);
   return CB4FS_DEBUG_END ("FieldPut",uvRet);

}

/* #ifdef CB4FS_QUICKFLDACCESS */
/*--------------------------------------------------------------------------*
 | oCb4:QuickFieldPut() METHOD replacess the specified field by the passed  |
 |              value.                                                      |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |              argv[1] = ptr of field-corresponding r/o variable           |
 |              argv[2] = value to be stored into the field                 |
 | Returns    : expU    = value of the field field-corresponding variable   |
 | Description: This hidden method cannot be invoked from the user's .prg.  |
 |              It allows FlagShip's variable system a faster access to     |
 |              the database field. Instead of ...                          |
 |              via the return variable, this method stores the field value |
 |              (or string ptr) directly into the corresponding variable.   |
 |              If not specified or announced, standard FIELDGET() is used. |
 | Multiuser  : In multiuser mode, at least RECORD locking is required,     |
 |              or SET AUTOLOCK ON must be set                              |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (quickfieldput)  /*  _bb_cb4cdx_quickfieldput_me(argc,argv) */
{
   FSvar  *ovSelf, *uvFieldVar;                  /* passed variables        */
   FSvar  *uvStore, *uvRet;                      /* stored variable         */
   FSint  fieldPos;
   DATA4  *dataPtr;                              /* CB5 structure           */
   FIELD4 *fieldPtr;                             /* CB5 structure           */
   FSint  autoLock = 0;                          /* 1=user, -1=autoRlock    */

   ovSelf     = argv[0];
   uvFieldVar = argv[1];
   uvStore    = argv[2];

   CB4FS_DEBUG_BEGIN ("QuickFieldPut",ovSelf);
   if (! (dataPtr = CB4FS_GETDATA4PTR) || (! uvFieldVar)) {
      myRlockCheck (0, 0);                       /* RTE error               */
      return CB4FS_DEBUG_END ("QuickFieldPut",NIL_VAR);
   }

   fieldPos = VAR_ISFLDPOS(uvFieldVar);          /* from fgsAnnounceField() */
   fieldPtr = d4field_j (dataPtr, fieldPos);     /* CB5 field struct        */

   if (! (autoLock = myRlockCheck (ovSelf, fieldPtr)))
      return CB4FS_DEBUG_END ("QuickFieldPut",NIL_VAR);          /* error   */

   uvRet = myFieldPut (ovSelf,                   /* replace field           */
               (int) VAR_ISTYPE(uvFieldVar),
               fieldPtr, uvStore);

   if (autoLock < 0)                             /* autolock ?              */
      CB4FSudf(myAutoUnlock)(ovSelf, dataPtr);
   return CB4FS_DEBUG_END ("QuickFieldPut",uvRet);
}
/* #endif */

/*--------------------------------------------------------------------------*
 | oCb4:GoTo()  METHOD positions dbf cursor on specified record             |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |              argv[1] = expN record no. (+/-)                             |
 | Returns    : TRUE if possible, FALSE if record no. not in dbf scope      |
 | Description: If not successful, goes to lastrec()+1 and both EOF, BOF    |
 |              are set .T.                                                 |
 |              SET DELETE, SET FILTER are NOT considered                   |
 |              SET REALAT is considered                                    |
 | FS invocat.: retL := oCb4:GOTO(expN)                                     |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (goto)              /* e.g.  _bb_cb4cdx_goto_me(argc, argv) */
{
   FSvar  *ovSelf;
   DATA4  *dataPtr;                                   /* CB5 structure      */
   int    ok;
   long   recNo, oldRec;

   if (argc < 2 || ! IS_VAR_OBJ(argv[0])) {
      CB4FSudf(callError)(argv[0], 1, RTMISARG, "GOTO", CB4_GOTOPAR,
                          __FILE__, __LINE__);
      return FALSE_VAR;                               /* RTE: missing...    */
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("GoTo",ovSelf);
   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("GoTo",FALSE_VAR);

   switch (VAR_ISTYPE(argv[1])) {
   case 'I':
   case 'N':
   case 'F':
      recNo = (long) VAR_INTNUM(argv[1]); break;
   default:
      CB4FSudf(callError)(ovSelf, 1, RTARGTYP, "GOTO", CB4_GOTOPAR,
                          __FILE__, __LINE__);
      return CB4FS_DEBUG_END ("GoTo",FALSE_VAR);      /* RTE: wrong par 1   */
   }

   oldRec = d4recno(dataPtr);                         /* CB5 curr.recno     */
   if(DATA4_BOF || DATA4_EOF)
      dataPtr->recordChanged = 0;   // otherwise CB6.54 error
   ok = !d4go (dataPtr, recNo);                       /* CB5 goto function  */

   if (ok) {
#ifndef CB4FS_MINIMAL_RDD
      if (oldRec != recNo)
         CB4FSudf(moveChildRelation)(ovSelf, dataPtr); /* move relat childs */
#endif
      return CB4FS_DEBUG_END ("GoTo (T)",TRUE_VAR);   /* success            */
   }

   d4go_eof (dataPtr);                                /* goto EOF           */
   DATA4_EOF = 1 ;                                    /* set EOF flag       */
   DATA4_BOF = 1 ;                                    /* set BOF flag       */
   return CB4FS_DEBUG_END ("GoTo (F)",FALSE_VAR);     /* failed             */
}

/*--------------------------------------------------------------------------*
 | oCb4:GoBottom()  METHOD positions dbf cursor on last logical record      |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 | Returns    : TRUE if possible, FALSE on error                            |
 | Description: If dbf is empty, both EOF, BOF are set .T.                  |
 |              SET DELETE, SET FILTER and SET RELATION are considered      |
 | FS invocat.: retL := oCb4:GOBOTTOM()                                     |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (gobottom)      /* e.g.  _bb_cb4cdx_gobottom_me(argc, argv) */
{
   FSvar  *ovSelf;
   DATA4  *dataPtr;                                   /* CB5 structure      */
   FSvar  *bvFilter;
   FSint  ok, isFilter = 0, relok = 0;

   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("GoBottom",ovSelf);
   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("GoBottom",FALSE_VAR);

   bvFilter = VAR_OBJ(ovSelf,CB4OBJ_FILTERBLOCK);
   isFilter = IS_VAR_BLK (bvFilter);

   ok = !d4bottom (dataPtr);                          /* CB5 function       */

   if (ok && (isFilter || fgsGetDelete())) {          /* SET DELE or FILT ? */
      if ((fgsGetDelete() && d4deleted(dataPtr)) ||
          (isFilter && !(IS_VAR_TRUE (VAR_EVAL_BLOCK(bvFilter))) )) {
         ok = CB4FSudf(CSkip)(ovSelf,dataPtr, -1L, 1);         /*  skip -1  */
         relok = 1;
      }
   }

   if (ok) {                                          /* o.k. ?             */
#ifndef CB4FS_MINIMAL_RDD
      if (! relok)
        CB4FSudf(moveChildRelation)(ovSelf, dataPtr); /* move related childs*/
#endif
      return CB4FS_DEBUG_END ("GoBottom",TRUE_VAR);   /* success            */
   }

   d4go_eof (dataPtr);                                /* goto EOF           */
   DATA4_EOF = 1 ;                                    /* set EOF flag       */
   DATA4_BOF = 1 ;                                    /* set BOF flag       */
   return CB4FS_DEBUG_END ("GoBottom (F)",FALSE_VAR); /* failed             */
}

/*--------------------------------------------------------------------------*
 | oCb4:GoTop()  METHOD positions dbf cursor on first logical record        |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 | Returns    : TRUE if possible, FALSE on error                            |
 | Description: If dbf is empty, both EOF, BOF are set .T.                  |
 |              SET DELETE, SET FILTER and SET RELATION are considered      |
 | FS invocat.: retL := oCb4:GOTOP()                                        |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (gotop)            /* e.g.  _bb_cb4cdx_gotop_me(argc, argv) */
{
   FSvar  *ovSelf;
   DATA4  *dataPtr;                                   /* CB5 structure      */
   FSvar  *bvFilter;
   FSint  ok, isFilter = 0, relok = 0;

   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("GoTop",ovSelf);
   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("GoTop",FALSE_VAR);

   bvFilter = VAR_OBJ(ovSelf,CB4OBJ_FILTERBLOCK);
   isFilter = IS_VAR_BLK (bvFilter);

   ok = !d4top (dataPtr);                             /* CB5 function       */

   if (ok && (isFilter || fgsGetDelete())) {          /* SET DELE or FILT ? */
      if ((fgsGetDelete() && d4deleted(dataPtr)) ||
          (isFilter && !(IS_VAR_TRUE (VAR_EVAL_BLOCK(bvFilter))) )) {
         ok = CB4FSudf(CSkip)(ovSelf,dataPtr, 1L, 1);           /*  skip 1  */
         relok = 1;
      }
   }

   if (ok) {                                          /* o.k. ?             */
#ifndef CB4FS_MINIMAL_RDD
      if (! relok)
        CB4FSudf(moveChildRelation)(ovSelf, dataPtr); /* move related childs*/
#endif
      return CB4FS_DEBUG_END ("GoTop (T)",TRUE_VAR);  /* success            */
   }

   d4go_eof (dataPtr);                                /* goto EOF           */
   DATA4_EOF = 1 ;                                    /* set EOF flag       */
   DATA4_BOF = 1 ;                                    /* set BOF flag       */
   return CB4FS_DEBUG_END ("GoTop (F)",FALSE_VAR);    /* failed             */
}

/*--------------------------------------------------------------------------*
 | oCb4:Skip()  METHOD move dbf cursor N logical records forwards/backwards |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |         opt: argv[1] = expN count (+/-) default=1                        |
 |         opt: argv[2] = expL consider RELATION movement, default=.T.   (1)|
 | Note/compat: (1) is not supported by CA/VO.                              |
 | Returns    : TRUE if sucessful, FALSE otherwise and EOF or BOF is set    |
 | Description: skip according to index if such is active                   |
 |              SET DELE, SET FILTER and SET RELAT are considered,          |
 |              except on SKIP 0                                            |
 | Compatib.  : FlagShip and Clipper 5 does not change the EOF/BOF flag     |
 |              on SKIP 0, this is simulated here too                       |
 | FS invocat.: retL := oCb4:SKIP([expN],[expL])                            |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (skip)              /* e.g.  _bb_cb4cdx_skip_me(argc, argv) */
{
   FSvar  *ovSelf;
   DATA4  *dataPtr;                                   /* CB5 structure      */
   long   count = 1L;
   int    err = 0;
   int    isRelat = 1;

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTARGCNT, CB4_SKIPPAR);
      return FALSE_VAR;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("Skip",ovSelf);

/*
 *  Parameter checking
 */
   switch (argc) {
   case 3:
      if (IS_VAR_LOG(argv[2]))
         isRelat = IS_VAR_TRUE (argv[2]);
      else if (! IS_VAR_NIL(argv[2]))
         err = 2;
   case 2:
      if (IS_VAR_NUM(argv[1]))
         count = (long) VAR_INTNUM(argv[1]);
      else if (! IS_VAR_NIL(argv[1]))
         err = 1;
   }
   if (err) {
      sprintf (repdum, CB4_WRONGARG, CB4_SKIPPAR, err);
      CB4FSudf(callError)(ovSelf, 1, RTARGTYP, "SKIP", repdum,
                          __FILE__, __LINE__);
      return CB4FS_DEBUG_END ("Skip",FALSE_VAR);      /* RTE: wrong type    */
   }

   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("Skip",FALSE_VAR);

   if (CB4FSudf(CSkip)(ovSelf, dataPtr, count, isRelat))
      return CB4FS_DEBUG_END ("Skip",TRUE_VAR);
   else
      return CB4FS_DEBUG_END ("Skip",FALSE_VAR);
}

/*--------------------------------------------------------------------------*
 | mid level SKIP function called by C params, w/o par checking             |
 | returns 1 on success, 0 otherwise                                        |
 *--------------------------------------------------------------------------*/
FSint CB4FSudf(CSkip)(ovSelf, dataPtr, count, isRelat)
FSvar  *ovSelf;
DATA4  *dataPtr;                                    /* CB5 structure        */
long   count;                                       /*  +/- records to skip */
int    isRelat;                                     /*  1 = consider realt  */

{
   FSvar  *bvFilter;
   long   skipped = 0L, sign = 1L;
   int    eofFlag, bofFlag, isFilter = 0;
   int    ok, imark, imark0 = VAR_DEL_MARK();

   if (count < 0L) {
      sign  = -1L;
      count = -count;
   }

   eofFlag  = DATA4_EOF;                       /* CB6 structure      */
   bofFlag  = DATA4_BOF;                       /* CB6 structure      */

   bvFilter = VAR_OBJ(ovSelf,CB4OBJ_FILTERBLOCK);
   isFilter = IS_VAR_BLK (bvFilter);

   /* avoid CB error: skip must be made from valid record number
    */
   if(eofFlag) {
      dataPtr->recordChanged = 0;   // otherwise CB6.54 error
      OBJ_METHEXEC("gobottom",-1, 1, &ovSelf);
      if((count * sign) >= 0) {
         DATA4_EOF = 1;
         return(0);
      }
      count--;             // decrease counter (count is always positive)
   } else if(bofFlag) {
      dataPtr->recordChanged = 0;   // otherwise CB6.54 error
      OBJ_METHEXEC("gotop",-1, 1, &ovSelf);
      if((count * sign) <= 0) {
         DATA4_BOF = 1;
         return(0);
      }
   }     

   /* Consider current SET DELETE and SET FILTER settings except for SKIP 0
    */
   if ((!fgsGetDelete() && !isFilter) || (count == 0)) {
      ok = !d4skip (dataPtr, count * sign);           /* CB5 skip function, 1=ok  */
   } else {
      imark = VAR_DEL_MARK();                         /* mark vars garbage  */
      while (skipped < count) {
         ok = !d4skip (dataPtr, sign);                /* CB5 skip function  */
         if (!ok || DATA4_EOF || DATA4_BOF) 
            break;
         if (fgsGetDelete() && d4deleted(dataPtr))
            continue;
         if (isFilter) {
//          VAR_DEL_SINCE_MARK(imark);                /* garbage colection  */
            if ( !IS_VAR_TRUE (VAR_EVAL_BLOCK(bvFilter)) )
               continue;
         }
         skipped++;
      }
   }

   eofFlag  = DATA4_EOF;                       /* CB6 structure      */
   bofFlag  = DATA4_BOF;                       /* CB6 structure      */

   if (!count) {                               /* SKIP 0:            */
      d4go (dataPtr, d4recno(dataPtr));
      DATA4_EOF = eofFlag;                     /* simulate Clipper   */
      DATA4_BOF = bofFlag;                     /* and std FS driver  */
   }
/*
 *  Consider current SET RELATION settings except for SKIP 0
 */
#ifndef CB4FS_MINIMAL_RDD
   if (ok && count && isRelat)
      CB4FSudf(moveChildRelation)(ovSelf, dataPtr);   /* move related childs*/
#endif
// VAR_DEL_SINCE_MARK(imark0);                        /* garbage colection  */
   return (ok);
}

/*--------------------------------------------------------------------------*
 | oCb4:Error()  METHOD std. error handler for all cb4: methods...          |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |              argv[1] = (opt) oErr Error object                           |
 |              argv[2] = (opt) cName Name of the RDD method/access/assign  |
 |                        will be stored to oErr:OPERATION                  |
 | Returns    : NIL                                                         |
 | Description: This error handler is called from all cb4xxx methods when   |
 |              an RTE, I/O or fatal error occur.                           |
 |              The oErr instances are filled by the fillErr() function     |
 |              below. The usual C invocation is:                           |
 |                 initErr (ovSelf,1,DTMMACH,"RddMethodName","Description");|
 |                 fsFnNAME(eval)(1,argv);                                  |
 |              or the equivalent .prg statements are                       |
 |                 LOCAL oMyErr                                             |
 |                 oMyErr:Descriptin := "...." etc.                         |
 |                 EVAL ({|x| _DEFERROR(x)}, oMyErr)                        |
 | Compatib.  : CA/VA requires both arguments and the 2nd argument is "symb"|
 |              This RDD stores the oErr in the object, when par1 and/or    |
 |              par2 are not given, it uses the default oErr.               |
 | FS invocat.: SELF:Error([oMyErr],["seek"])                               |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (error)            /* e.g.  _bb_cb4cdx_error_me(argc, argv) */
{
   FSvar  *ovSelf;
   FSvar  *ovErr;
   FSvar  *bvErrBlock = VAR_NEW;
   FSvar  *parv[2];
   DATA4  *dataPtr;                                   /* CB5 structure      */

   if (argc < 1 || ! IS_VAR_OBJ(argv[0]) ) {
      rt_error(RTMISARG, CB4_ERRORPAR);               /* RTE: missing args  */
      return NIL_VAR;
   }
   ovSelf = argv[0];

   if (argc > 1 && IS_VAR_OBJ(argv[1]))
      ovErr = argv[1];                                /* user specif object */
   else
      ovErr = VAR_OBJ(ovSelf,CB4OBJ_ERROBJECT);       /* default in RDD obj */

   if (! IS_VAR_OBJ(ovErr)) {
      sprintf (repdum, "%s;ErrObject is of type '%c'",
            CB4_ERRORPAR, VAR_ISTYPE(ovErr));
      rt_error(RTARGTYP, repdum);                     /* RTE: wrong arg typ */
      return NIL_VAR;
   }

   if (! IS_VAR_NUM (VAR_OBJ(ovErr, ERROBJ_SEVERITY)) ) {
      sprintf (repdum, "%s;oErr:Severity is not 'N' but '%c'",
            CB4_ERRORPAR, VAR_ISTYPE(VAR_OBJ(ovErr, ERROBJ_SEVERITY)));
      rt_error(RTARGTYP, repdum);                     /* RTE: wrong arg typ */
      return NIL_VAR;
   }

   CB4FS_DEBUG_BEGIN ("Error",ovSelf);
   SET_VAR_COPY (bvErrBlock, fsFnNAME(errorblock)(0,0));

   if (argc > 2 && IS_VAR_CHR(argv[2]))
      SET_VAR_COPY (VAR_OBJ(ovErr,ERROBJ_OPERATION), argv[2]);

   if ((dataPtr = CB4FS_GETDATA4PTR) && argc < 2)     /* err:OSCODE = CB5   */
      SET_VAR_INT (VAR_OBJ(ovErr, ERROBJ_OSCODE),
                   dataPtr->code_base->error_code);   /* CB5 structure      */

   if (! IS_VAR_BLK(bvErrBlock)) {                    /* user handler set ? */
      parv[0] = ovErr;                                /* no, execute datault*/
      fsFnNAME(_deferror)(1,parv);
   } else {
      parv[0] = bvErrBlock;                           /* yes, execute it    */
      parv[1] = ovErr;
      fsFnNAME(eval)(2, parv);
   }

   return CB4FS_DEBUG_END ("Error",NIL_VAR);
}

/*--------------------------------------------------------------------------*
 | internal function for xxxFieldGet() methods above                        |
 | Note: do not use SET_VAR_*() since we assign field vars here             |
 *--------------------------------------------------------------------------*/
static FSvar *myFieldGet (uvFieldVar, fieldPtr, dataPtr)
FSvar  *uvFieldVar;
FIELD4 *fieldPtr;                                /* CB5 structure           */
DATA4  *dataPtr;                                 /* CB5 structure           */
{
   long datum;
   int  eofFlag;
   static char emptyChar[2] = {""};
   
   eofFlag = DATA4_EOF ; 

   switch (f4type(fieldPtr)) {                        /* field/var type     */
   case 'C':
      if(eofFlag) {
         VAR_LCHR(uvFieldVar)  = emptyChar; 
         VAR_ISLEN(uvFieldVar) = 1; // including \0
      } else {
         VAR_LCHR(uvFieldVar)  = (char *)f4ptr(fieldPtr); /* ptr to CB5 field */
         VAR_ISLEN(uvFieldVar) = f4len(fieldPtr) +1;
      }
      break;
   case 'M':
      if(eofFlag) {
         VAR_LCHR(uvFieldVar) = emptyChar; 
         VAR_ISLEN(uvFieldVar) = 1; // including \0
      } else {
         VAR_LCHR(uvFieldVar)  = (char *)f4memo_ptr(fieldPtr); /* CB5 memo field */
         VAR_ISLEN(uvFieldVar) = f4memo_len(fieldPtr) +1;
      }
      break;
   case 'N':
   case 'F':
      if (IS_VAR_INT(uvFieldVar)) {
         if(eofFlag)
            VAR_INT(uvFieldVar) = 0; 
         else
            VAR_INT(uvFieldVar) = (int)f4double(fieldPtr); /* CB5 num -> int   */
      } else {
         if(eofFlag)
            VAR_NUM(uvFieldVar) =  0.0;     
         else
            VAR_NUM(uvFieldVar) = f4double(fieldPtr);      /* CB5 num -> double*/
      }
      break;
   case 'L':
      if(eofFlag)
         VAR_LOG(uvFieldVar) = 0;  
      else
         VAR_LOG(uvFieldVar) = f4true(fieldPtr);         /* CB5 logical        */
      break;
   case 'D':
      if(eofFlag)
         datum = 0L;
      else {
         datum = date4long(f4ptr(fieldPtr));             /* CB5 date -> long   */
         if (datum > 0L)
            datum -= CB4FS_JULIAN_ADJUSTMENT;
         else
            datum = 0L;
      }
      VAR_DATE(uvFieldVar) = datum;
      break;
   }
   return uvFieldVar;
}

/*--------------------------------------------------------------------------*
 | internal function for xxxFieldPut() methods above                        |
 *--------------------------------------------------------------------------*/
static FSvar *myFieldPut (ovSelf, fieldtype, fieldPtr, uvStore)
FSvar  *ovSelf;
int    fieldtype;
FIELD4 *fieldPtr;                                /* CB5 structure           */
FSvar  *uvStore;
{

   switch (fieldtype) {                               /* field/var type     */
   case 'C':
      if (! IS_VAR_CHR(uvStore)) goto error_type;     /* param/field check  */
      f4assign_n (fieldPtr, VAR_CHR(uvStore),
                  VAR_ISLEN(uvStore) - 1);            /* CB5 char of len    */
      break;
   case 'M':
      if (! IS_VAR_CHR(uvStore)) goto error_type;     /* param/field check  */
      f4memo_assign (fieldPtr, VAR_CHR(uvStore));     /* CB5 write memo     */
      break;
   case 'N':
   case 'F':
      if (! IS_VAR_NUM(uvStore)) goto error_type;     /* param/field check  */
      f4assign_double (fieldPtr, VAR_FPNUM(uvStore)); /* CB5 write numeric */
      break;
   case 'L':
      if (! IS_VAR_LOG(uvStore)) goto error_type;     /* param/field check  */
      f4assign_char (fieldPtr,
                 (IS_VAR_TRUE(uvStore) ? 'T':'F'));   /* CB5 write logical  */
      break;
   case 'D':
      if (! IS_VAR_DATE(uvStore)) goto error_type;    /* param/field check  */
      if (VAR_DATE(uvStore) > 0L)
         f4assign_long (fieldPtr,
                        (long) VAR_DATE(uvStore) +
                        CB4FS_JULIAN_ADJUSTMENT );    /* CB5 write date     */
      else
         f4assign_long (fieldPtr, 0L);
      break;
   defaut:
      goto error_type;
   }
   return uvStore;

 error_type:
   sprintf (repdum, CB4_FILDPUTSTR,
            CB4_FILDPUTPAR, VAR_ISTYPE(uvStore), 2);
   CB4FSudf(callError)(ovSelf, 1, DTMMACH, "FIELDPUT", repdum,
                       __FILE__, __LINE__);
   return NIL_VAR;                               /* RTE data type mismatch  */
}

/*--------------------------------------------------------------------------*
 | internal function for xxxFieldPut() methods above                        |
 |                                                                          |
 | ret = 0: error                                                           |
 |       1: user lock or exclusive, no autolock                             |
 |      -1: autoRlock performed                                             |
 *--------------------------------------------------------------------------*/
static FSint myRlockCheck (ovSelf, fieldPtr)
FSvar  *ovSelf;
FIELD4 *fieldPtr;
{
   FSint  currWA = 0, ii;
   DATA4  *dataPtr = 0;                          /* CB5 structure           */
   char   *currAlias = 0;
/* char   repdum[255];    */                     /* global RTE buffer       */
   long   recNo = 0;

   if (fieldPtr)
      dataPtr = fieldPtr->data;
/*
 *  is database open ?
 */
   if (!fieldPtr || !dataPtr) {
      CB4FSudf(callError)(ovSelf, 1, RTNODBF, "FIELDPUT", CB4_REPLNOOPEN,
                          __FILE__, __LINE__);
      return 0;                                  /* RTE: database not open  */
   }
/*
 *  is file open for read only ?
 */
#ifdef CODEBASE6
   if (dataPtr->dataFile->file.isReadOnly) {      /* CB6 structure           */
#else
   if (dataPtr->file.is_read_only) {             /* CB5 structure           */
#endif
      CB4FSudf(getWAdata)(ovSelf, fieldPtr, 0, &currWA, &currAlias);
      sprintf (repdum, CB4_REPLREADON,
               fieldPtr->name, currWA, currAlias);
      CB4FSudf(callError)(ovSelf, 2, IOWRIT, "FIELDPUT", repdum,
                          __FILE__, __LINE__);
      return 0;                                  /* RTE: replace on r/o     */
   }
/*
 *  replace on EOF() ?
 */
   recNo = d4recno (dataPtr);                    /* CB5: record number      */
   if (d4eof(dataPtr) || recNo <= 0L) {
      CB4FSudf(getWAdata)(ovSelf, fieldPtr, 0, &currWA, &currAlias);
      sprintf (repdum, CB4_REPLONEOF,
               fieldPtr->name, currWA, currAlias);
      CB4FSudf(callError)(ovSelf, 1, RTILLEOF, "FIELDPUT", repdum,
                          __FILE__, __LINE__);
      return 0;                                  /* RTE: replace on EOF()   */
   }
/*
 *  is file or record locked ?
 */
   if (d4lock_test_append(dataPtr) > 0)          /* CB5: file locked ||     */
      return 1;                                  /* exclus || append lock ? */

   if (D4LOCK_TEST(dataPtr, recNo) > 0)          /* CB5: record locked ?    */
      return 1;

/*
 *  perform AutoRlock if not disabled
 */
   ii = CB4FSudf(myAutoRFlock)(ovSelf, 1, dataPtr, recNo);
   if (ii != 0)
      return ii;

   CB4FSudf(getWAdata)(ovSelf, fieldPtr, 0, &currWA, &currAlias);
   sprintf (repdum, CB4_REPLNOLOCK,
            fieldPtr->name, currWA, currAlias, recNo);
   CB4FSudf(callError)(ovSelf, 1, RTNOLOCK, "FIELDPUT", repdum,
                       __FILE__, __LINE__);
   return 0;                                     /* RTE: replace w/o lock   */

}

/*--------------------------------------------------------------------------*
 | internal function for AutoUnlock()                                       |
 |                                                                          |
 | ret = 0: error                                                           |
 |       1: user lock or exclusive, no autolock                             |
 |      -1: autoRlock performed                                             |
 |      -2: autoFlock performed                                             |
 *--------------------------------------------------------------------------*/
FSint CB4FSudf(myAutoRFlock)(ovSelf, iRecFile, dataPtr, recNo)
FSvar  *ovSelf;
FSint  iRecFile;                                      /* 1=Rlock, 2=Flock   */
DATA4  *dataPtr;                                      /* CB5 structure      */
long   recNo;                                         /* for Rlock only     */
{
   FSvar  *passArgv[2];
   FSvar  *retVar;
   FSint  ii;

/*
 *  file/record is NOT locked. Check if oRdd:ConcurrencyControl is
 *  enabled. If so and positive (the default), invoke AutoRlock(obj) to
 *  consider the current SET AUTOLOCK status. If negative, the abs(value)
 *  represents the (user set and RDD specific) wait period.
 */
   ii = VAR_INT (VAR_OBJ(ovSelf,CB4OBJ_CONCURRENCY));
   if (ii == 0)
      return 0;                                  /* autolock is disabled    */

#ifdef CB4FS_MINIMAL_RDD
   if (iRecFile == 1) {
      if (!d4lock (dataPtr, recNo))              /* CB5: lock record        */
         return -1;
   } else {
      if (!d4lock_file (dataPtr))                /* CB5: lock file          */
         return -1;
   }
#else
   VAR_NEW_ARGS(passArgv, 2);
   passArgv[0] = ovSelf;

   if (ii >= 1) {                                /* optimistic, AUTOxLOCK() */
      if (iRecFile == 1)
          retVar = FSudfCALL(autorlock)(1, passArgv);  /* call AUTORLOCK()  */
      else
          retVar = FSudfCALL(autoflock)(1, passArgv);  /* call AUTORLOCK()  */
      if (IS_VAR_TRUE (retVar))
         return -1;
   } else {                                      /* user reprocess time set */
      SET_VAR_INT (passArgv[1], -ii);
      if (iRecFile == 1)
          retVar = FSudfCALL(autorlock)(2, passArgv);  /* call AUTORLOCK()  */
      else
          retVar = FSudfCALL(autoflock)(2, passArgv);  /* call AUTORLOCK()  */
      if (IS_VAR_TRUE (retVar))
         return -1;
   }
#endif
   return 0;
}

/*--------------------------------------------------------------------------*
 | internal function for AutoUnlock()                                       |
 |                                                                          |
 *--------------------------------------------------------------------------*/
FSint CB4FSudf(myAutoUnlock)(ovSelf, dataPtr)
FSvar  *ovSelf;
DATA4  *dataPtr;                                      /* CB5 structure      */
{
#ifdef CB4FS_MINIMAL_RDD
    d4unlock (dataPtr);                               /* CB5: unlock all    */
#else
   FSvar  *passArgv[2];

   VAR_NEW_ARGS(passArgv, 2);
   passArgv[0] = ovSelf;
   SET_VAR_INT (passArgv[1], 10);                  /* default: 10 seconds   */
   FSudfCALL(autounlock)(2, passArgv) ;            /* call std AUTOUNLOCK() */
#endif
   return 0;
}

/*--------------------------------------------------------------------------*
 | Internal purpose for all cb4RDDs: init the default ERROR object          |
 |                                                                          |
 | in: ovSelf  = object self                                                |
 |     ovErr   = error object or NIL_VAR for the RDD default object         |
 |     rddmeth = string with the method/access/assign name or ""            |
 |                                                                          |
 | Note: no parameter checking performed                                    |
 *--------------------------------------------------------------------------*/
FSint CB4FSudf(initErr)(ovSelf, ovErr, rddmeth)
FSvar  *ovSelf, *ovErr;
FSchar *rddmeth;
{
   extern FSchar charBuff500[];                           /* global buffer  */

   if (! IS_VAR_OBJ(ovErr))
      ovErr = VAR_OBJ(ovSelf,CB4OBJ_ERROBJECT);

   SET_VAR_NIL (VAR_OBJ(ovErr, ERROBJ_ARGS      ));       /* see FlagShip.h */
   SET_VAR_LOG (VAR_OBJ(ovErr, ERROBJ_CANDEFAULT), 0);
   SET_VAR_LOG (VAR_OBJ(ovErr, ERROBJ_CANRETRY  ), 0);
   SET_VAR_LOG (VAR_OBJ(ovErr, ERROBJ_CANSUBSTIT), 0);
   SET_VAR_NIL (VAR_OBJ(ovErr, ERROBJ_CARGO     ));
   SET_VAR_CHR (VAR_OBJ(ovErr, ERROBJ_DESCRIPTIO), "");

   sprintf (charBuff500, "%s.PRG, module %s line %d",
                  VAR_CHR(fsFnNAME(procfile)(0,0)),
                  VAR_CHR(fsFnNAME(procname)(0,0)),
                  VAR_INTNUM(fsFnNAME(procline)(0,0)));

   SET_VAR_CHR (VAR_OBJ(ovErr, ERROBJ_FILENAME ), charBuff500);
   SET_VAR_INT (VAR_OBJ(ovErr, ERROBJ_GENCODE  ), 0);
   SET_VAR_CHR (VAR_OBJ(ovErr, ERROBJ_OPERATION), rddmeth);
   SET_VAR_INT (VAR_OBJ(ovErr, ERROBJ_OSCODE   ), 0);
   SET_VAR_INT (VAR_OBJ(ovErr, ERROBJ_SEVERITY ), 0);
/* SET_VAR_CHR (VAR_OBJ(ovErr, ERROBJ_SUBSYSTEM), CB4FSrddName);            */
/* SET_VAR_COPY(VAR_OBJ(ovErr, ERROBJ_SUBCODE  ), fsFnNAME(procline)(0,0)); */
   SET_VAR_INT (VAR_OBJ(ovErr, ERROBJ_TRIES    ), 0);

   return 0;
}

/*--------------------------------------------------------------------------*
 | Internal purpose for all cb4RDDs: fill the default error object,         |
 |          call the cb4:ERROR() method with the default or user handler    |
 |                                                                          |
 | in: ovSelf  = object self                                                |
 |     errtype = 1:RTE, 2:I/O, 3:internal                                   |
 |               or negative: RTE errno and errno=DEVWARNING                |
 |     errno   = oERR:GENCODE     num.code according to FSerrors.h          |
 |     rddmeth = oERR:OPERATION   string with the method/access/assign name |
 |     descr   = oERR:DESCRIPTION string with the description               |
 |                                                                          |
 | Note: only minimal parameter checking is performed                       |
 *--------------------------------------------------------------------------*/
FSint CB4FSudf(callError)(ovSelf, errtype, errno, rddmeth, descr, Csrc, Cline)
FSvar  *ovSelf;
FSint  errtype, errno;
FSchar *rddmeth, *descr;
char   *Csrc;
int    Cline;
{
   FSvar *ovErr;
   FSvar *parv[1];
   FSint sever;
   extern FSchar charBuff500[];                           /* global buffer  */
   extern struct __error_report __error_report[];         /* see FSerror.c  */
   char  *xerr = "RTE ";

   if (! IS_VAR_OBJ(ovSelf)) return 0;
   ovErr = VAR_OBJ(ovSelf,CB4OBJ_ERROBJECT);       /* std.cb4: error object */
   if (! IS_VAR_OBJ(ovErr)) return 0;

   switch (errtype) {
   case 0:
   case 1:
      sever = 2; break;                            /* ES_RTERR  in error.fh */
   case 2:
      xerr[0] = 0;
      sever = 1; break;                            /* ES_IOERR  in error.fh */
   default:
      xerr[0] = 0;
      sever = 8; break;                            /* ES_INTERR in error.fh */
   }

   CB4FSudf(initErr)(ovSelf, ovErr, rddmeth);      /* init error object     */

   if ((errtype < 0) && (errno == DEVWARNING)) {   /* special case, DEVWARN */
      sever = 2;
      xerr[0] = 'R';
      errtype = -errtype;
      if ((errtype < MAXERRMSG) && (__error_report[errtype].err_msg))
         sprintf (repdum, "%s;%s%d (%s)", descr, xerr,
                  errtype, __error_report[errtype].err_msg);
      else
         strcpy (repdum, descr);
   } else if (errno == DEVWARNING) {
      strcpy (repdum, descr);
   } else {
      if ((errno < 0) || (errno >= MAXERRMSG) ||       /* see FSerrors.h   */
          (! __error_report[errno].err_msg))
         sprintf (repdum, "%s;%s%d (unknown)", descr, xerr, errno);
      else
         sprintf (repdum, "%s;%s%d (%s)", descr, xerr,
                  errno, __error_report[errno].err_msg);
   }
   SET_VAR_CHR (VAR_OBJ(ovErr, ERROBJ_DESCRIPTIO), repdum);
   SET_VAR_INT (VAR_OBJ(ovErr, ERROBJ_GENCODE   ), errno);
   SET_VAR_INT (VAR_OBJ(ovErr, ERROBJ_SEVERITY  ), sever);

#ifdef CB4FS_ERROUT_CSOURCE
   sprintf (charBuff500, "%s driver (%s, line %d)",
                         CB4FSrddName , Csrc, Cline);
   SET_VAR_CHR (VAR_OBJ(ovErr, ERROBJ_SUBSYSTEM), charBuff500);
   SET_VAR_INT (VAR_OBJ(ovErr, ERROBJ_SUBCODE), -9999);  /* don't display */
#endif

   parv[0] = ovSelf;
   CB4FS_CALL_METH(error)(1,parv);                 /* call cb4:error()      */
   return 1;
}


/**** eof rddcb4b.c ****/

