/****************************************************************************
 * rddcb4e.c     RDD API for the CB4FOX, CB4CDX, CB4NDX, CB4MDX and CB4NTX  *
 *               by using CodeBase 5.1 multiplatform version (not included  *
 *               in the FlagShip package, contact Sequiter Inc.)            *
 *                                                                          *
 *               Note: all CodeBase function & defines are marked by CB5    *
 *                                                                          *
 * (c) 1995..2006 multisoft Datentechnik GmbH, Munich. All rights reserved  *
 * Release $Date: 2006/04/21 07:41:25 $                                     *
 *--------------------------------------------------------------------------*
 * This file contains parts of the full RDD driver:                         *
 *                                                                          *
 *    oCb4:createdb()          METHOD creates DB structure                  *
 *    oCb4:dbstruct()          METHOD retrieves DB structure                *
 *    oCb4:eval()              METHOD evaluate similar to DBEVAL()          *
 *    oCb4:replace()           METHOD general purpose replacing method      *
 *                                                                          *
 *    getGlobalScope()         internal: get global scope settings          *
 *    decodeScope()            internal: scope N/L into two args            *
 *    isScope()                internal: check if scope is set              *
 *                                                                          *
 * For additional methods, see rddcb4a.c to rddcb4?.c                       *
 ****************************************************************************/

#include "rddcb4fs.h"                            /* RDD API for CodeBase    */
#include "FSextend.h"                            /* for malloc (_xalloc)    */

extern FSvar *CB4FS_CALL_METH(fieldput)();       /* cb4:fieldput()          */
extern FSvar *CB4FS_CALL_METH(gotop)();          /* cb4:gotop()             */
extern FSint  CB4FSudf(getGlobalScope)();        /* in rddcb4e.c            */
extern FSint  CB4FSudf(decodeCond)();            /* in rddcb4e.c            */
extern FSint  CB4FSudf(callError)();             /* in rddcb4b.c            */
extern FSint  CB4FSudf(CSkip)();                 /* in rddcb4b.c            */
extern FSint  CB4FSudf(checkFile)();             /* in rddcb4a.c            */

#define CB4_EVALPARAM  "cb4:Eval(block,[for],[while],[scope])"
#define CB4_REPLACPAR  "oCb4:Replace(expr,field,[forCond],[whiCond],[scop])"
#define CB4_EMPTYARR1  "cb4:Replace(expr,field) empty array of expr"
#define CB4_EMPTYARR2  "cb4:Replace(expr,field) empty array of fields"
#define CB4_CREADBPAR  "cb4:CreateDB(expA)"
#define CB4_DBSTRUPAR  "cb4:DBstruct()"
#define CB4_FILDPUTSTR "%s;wrong type (%c) of parameter %d"
#define CB4_WRONGARG   "%s;wrong type or missing argument # %d"
#define CB4_WRONGVAL   "%s;wrong value of argument # %d"
#define CB4_WRONGARTYP "%s;wrong value/type of expA[%d,%d]"
#define CB4_CBERROR    "%s;CodeBase returns error %d"

/*--------------------------------------------------------------------------*
 | oCb4:CreateDB() METHOD creates new DB structure                          |
 |                                                                          |
 | Parameters : argv[0] = object:self  (WorkArea data not considered here)  |
 |              argv[1] = expC name of the database                         |
 |              argv[2] = expA array of field descriptions according to     |
 |                        DBCREATE()                                        |
 |         opt. argv[3] = nAcc open mode (rights), default = umask & 777 (1)|
 | Returns    : expL    = success = .T.                                     |
 | Note/compat: (1) is not supported by CB5                                 |
 | Description: this method is invoked when a new dbd structure is created, |
 |              e.g. DBCREATE(), COPY STRU, COPY TO, TOTAL etc.             |
 | Multiuser  : for full multitasking, this method should take care, that   |
 |              no other user can open the database (by USE, cb4:init(),    |
 |              methodNEW() etc.) during the creation phase. The default FS |
 |              RDD driver locks the file (exclusive) during this period.   |
 |              This CB5 driver does NOT take care for it.                  |
 | Compatib.  : this method is not available in CA/VO                       |
 | FS invocat.: retU := oCb4:CreateDB(expA [,expC])                         |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (createdb)      /* e.g.  _bb_cb4cdx_createdb_me(argc, argv) */
{
   FSvar      *ovSelf;
   FIELD4INFO *f4info;                                     /* CB5 structure */
   DATA4      *dataPtr;                                    /* CB5 structure */
   CODE4      cb4;                                  /* local! CB5 structure */
   FSvar      *subarr;
   FSint      err = 0, ii, arrLen;
   char       type, *cPtr;
   extern FSchar fileNameBuff[];                           /* in rddcb4a.c  */

/*
 *  Parameter checking
 */
   if (argc < 3)
      err = (argc <= 1 ? 1 : argc -1);
   else {
      if (! IS_VAR_CHR(argv[1]) || IS_VAR_EMPTY(argv[1]))    err = 1;
      if (! IS_VAR_ARR(argv[2]) || (VAR_ISDIM(argv[2]) < 1)) err = 2;
   }
   if (err) {
      sprintf (repdum, CB4_WRONGARG, CB4_CREADBPAR, err);
      CB4FSudf(callError)(argv[0], 1, RTARGTYP,
         "CREATEDB", repdum, __FILE__, __LINE__);    /* RTE: wrong arg type */
      return FALSE_VAR;
   }

   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("CreateDB",ovSelf);
/*
 *  Pre-check of array elements accord.to DBCREATE()
 */
   arrLen = VAR_ISDIM(argv[2]);
   for (ii = 1; ii <= arrLen; ii++) {
      subarr = VAR_ARRELEM (argv[2], ii);
      if ((! IS_VAR_ARR(subarr)) || (VAR_ISDIM(subarr) < 4))
         err = 4;
      else {
         if (! IS_VAR_CHR  (VAR_ARRELEM (subarr,1)) ||
               IS_VAR_EMPTY(VAR_ARRELEM (subarr,1)))
            err = 1;
         if (! IS_VAR_CHR  (VAR_ARRELEM (subarr,2)) ||
               IS_VAR_EMPTY(VAR_ARRELEM (subarr,2)))
            err = 2;
         type = VAR_CHR (VAR_ARRELEM (subarr,2))[0];
         if (!strchr ("CMLDNF", toupper(type)))   /* support. by CB5 and FS */
            err = 2;
         if (! IS_VAR_NUM  (VAR_ARRELEM (subarr,3)) ||
               VAR_INTNUM(VAR_ARRELEM (subarr,3)) <= 0L )
            err = 3;
         if (! IS_VAR_NUM  (VAR_ARRELEM (subarr,4)) ||
               VAR_INTNUM(VAR_ARRELEM (subarr,4)) <  0L )
            err = 4;
      }
      if (err) {
         sprintf (repdum, CB4_WRONGARTYP, CB4_CREADBPAR, ii, err);
         CB4FSudf(callError)(ovSelf, 1, RTARGTYP, "CREATEDB", repdum,
                  __FILE__, __LINE__);
         return CB4FS_DEBUG_END ("CreateDB",FALSE_VAR);
      }
   }
/*
 *  Form the file name incl. std. extention, upper/lower according
 *  to FS_SET ("upp"/"low"...) and consider the SET DEFAULT setting
 */
   if (! CB4FSudf(checkFile)
        (VAR_CHR(argv[1]), fileNameBuff, &cPtr,
        VAR_CHR( VAR_OBJ(ovSelf,CB4OBJ_DBF_EXT)), "", -2)) {
      sprintf (repdum, CB4_WRONGVAL, CB4_CREADBPAR, 1);
      CB4FSudf(callError)(ovSelf, 1, RTARGTYP, "CREATEDB", repdum,
               __FILE__, __LINE__);
      return CB4FS_DEBUG_END ("CreateDB",FALSE_VAR);
   }
/*
 *  Allocate CB5 structure, transfer data into
 */
   f4info = (FIELD4INFO *) _xalloc (sizeof(FIELD4INFO) * (arrLen +1));

   for (ii = 1; ii <= arrLen; ii++) {
      subarr = VAR_ARRELEM (argv[2], ii);
      type = VAR_CHR (VAR_ARRELEM (subarr,2))[0];
      f4info[ii-1].name = (char *) VAR_CHR (VAR_ARRELEM (subarr,1));
#ifdef CODEBASE6
      f4info[ii-1].type = (short) toupper(type);
      f4info[ii-1].len  = (unsigned short) VAR_INTNUM (VAR_ARRELEM (subarr,3));
      f4info[ii-1].dec  = (unsigned short) VAR_INTNUM (VAR_ARRELEM (subarr,4));
#else
      f4info[ii-1].type = (int) toupper(type);
      f4info[ii-1].len  = (unsigned int) VAR_INTNUM (VAR_ARRELEM (subarr,3));
      f4info[ii-1].dec  = (unsigned int) VAR_INTNUM (VAR_ARRELEM (subarr,4));
#endif
   }
   f4info[arrLen].name = 0;
   f4info[arrLen].type = 0;
   f4info[arrLen].len  = 0;
   f4info[arrLen].dec  = 0;
/*
 *  Create the database
 */
   d4init (&cb4);                                         /* CB5 function   */
   cb4.safety           = 0;                     /* set SET SAFETY OFF      */
   cb4.auto_open        = 0;                     /* don't open index autom  */
   cb4.field_name_error = 0;                     /* don't print errr fget() */
   cb4.read_lock        = 0;                     /* don't auto-readlock     */
   cb4.exclusive        = 1;                     /* create exclusive        */

   dataPtr = d4create (&cb4, cPtr, f4info, 0);            /* CB5 function   */
   err = cb4.error_code;                                  /* CB5 error code */

   _xfree (f4info);                                       /* extend C API   */

   if (dataPtr)
      d4close (dataPtr);                                  /* CB5 function   */
   d4init_undo (&cb4);                                    /* free local CB5 */

   if (err) {
      sprintf (repdum, CB4_CBERROR, CB4_CREADBPAR, err);
      CB4FSudf(callError)(ovSelf, 1, RTARGTYP, "CREATEDB", repdum,
               __FILE__, __LINE__);
/***  CB4FSudf(callError)(ovSelf, 1, DEVWARNING, "CREATEDB",
              repdum, __FILE__, __LINE__);                          ***/
   }
   return CB4FS_DEBUG_END ("CreateDB",err ? FALSE_VAR : TRUE_VAR);
}

/*--------------------------------------------------------------------------*
 | oCb4:DBstruct() METHOD creates array corresponding to DB structure       |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 | Returns    : expA    = array according to DBSTRUCT()                     |
 |                        or empty arr on error                             |
 | FS invocat.: retA := oCb4:DBstruct()                                     |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (dbstruct)      /* e.g.  _bb_cb4cdx_dbstruct_me(argc, argv) */
{
   DATA4  *dataPtr;                                   /* CB5 structure      */
   FIELD4 *fieldPtr;                                  /* CB5 structure      */
   FSvar  *ovSelf;
   FSint  err = 0, ii, nFields;
   FSvar  *avRet = VAR_NEW, *subarr = VAR_NEW;
   char   type[2];

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, CB4_DBSTRUPAR);
      VAR_NEW_ARRAY (avRet, 0);
      return avRet;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("DBstruct",ovSelf);

   if (! (dataPtr = CB4FS_GETDATA4PTR)) {
      VAR_NEW_ARRAY (avRet, 0);
      return CB4FS_DEBUG_END ("DBstruct",avRet);      /* database not open  */
   }

#ifdef CODEBASE6
   nFields = dataPtr->dataFile->nFields;              /* CB6 structure      */
#else
   nFields = dataPtr->n_fields;                       /* CB5 structure      */
#endif
   VAR_NEW_ARRAY (avRet, nFields);                    /* array of n elem.   */

   type[1] = 0;
   for (ii=1 ; ii <= nFields; ii++) {
      fieldPtr = d4field_j (dataPtr, ii);             /* CB5 field struct   */
      type[0] = (char) fieldPtr->type;
	  VAR_NEW_ARRAY (subarr, 4);                      /* array of 4 elem.   */
      SET_VAR_CHR (VAR_ARRELEM(subarr,1), fieldPtr->name);
      SET_VAR_CHR (VAR_ARRELEM(subarr,2), type);
      SET_VAR_INT (VAR_ARRELEM(subarr,3), fieldPtr->len);
      SET_VAR_INT (VAR_ARRELEM(subarr,4), fieldPtr->dec);
      SET_VAR_COPY(VAR_ARRELEM(avRet, ii), subarr);
   }
   return CB4FS_DEBUG_END ("DBstruct",avRet);
}

/*--------------------------------------------------------------------------*
 | oCb4:Eval()  METHOD evaluates all/matching records of the database       |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |              argv[1] = cbCodeBlock to evaluate                           |
 |         opt. argv[2] = cForCond   | cbForCond                            |
 |         opt. argv[3] = cWhileCond | cbWhileCond                          |
 |         opt. argv[4] = nScope     | lScope                               |
 | Returns    : TRUE on success, FALSE otherwise.                           |
 | Description: -If only the Block parameter is specified, the previously   |
 |               set scope/condition of the RDD object is used.             |
 |               If not set, the default is ALL records.                    |
 |              -The dbf pointer remains on the last record of the scope    |
 |              -If relations are set, each skip moves also related childs. |
 |               For disabling this behaviour and to increase speed, use:   |
 |               xx := oCb4:Relation ; oCb4:Average() ; oCb4:Relation := xx |
 | Compatib.  : CA/VO ignores the evaluation of obj:EVAL() with wrong params|
 |              without any message, wrong block body is also ignored.      |
 |              This RDD generates a developer warning.                     |
 | FS invocat.: retL := oCb4:EVAL(block, [for], [while], [scope])           |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (eval)              /* e.g.  _bb_cb4cdx_eval_me(argc, argv) */
{
   DATA4  *dataPtr;                                   /* CB5 structure      */
   FSvar  *ovSelf;
   FSvar  *uvForCond = VAR_NEW;
   FSvar  *uvWhiCond = VAR_NEW;
   FSvar  *bvBlock;
   long   iScope = 0L;
   FSint  err = 0;
   FSint  imark, isFor, isWhile;

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, "cb4:Eval()");
      return FALSE_VAR;
   }
   ovSelf = argv[0];
   if (argc < 2 || argc > 5 || ! IS_VAR_BLK(argv[1]) ) {
      CB4FSudf(callError)(ovSelf, 1, RTMISARG, "EVAL", CB4_EVALPARAM,
                          __FILE__, __LINE__);
      return FALSE_VAR;
   }

   CB4FS_DEBUG_BEGIN ("Eval",ovSelf);

   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("Eval",FALSE_VAR);

   bvBlock   = argv[1];

   err = CB4FSudf(getGlobalScope)
                 (argc, 2, argv, LONG_MAX, &uvForCond, &uvWhiCond, &iScope);

   if (err) {
      sprintf (repdum, CB4_WRONGARG, CB4_EVALPARAM, err);
      CB4FSudf(callError)(ovSelf, 1, DEVWARNING, "EVAL",
               repdum, __FILE__, __LINE__);      /* Dev.RTE: wrong argument */
      return CB4FS_DEBUG_END ("Eval",FALSE_VAR);
   }

   isFor   = IS_VAR_BLK(uvForCond) && !IS_VAR_NIL(uvForCond);
   isWhile = IS_VAR_BLK(uvWhiCond) && !IS_VAR_NIL(uvWhiCond);

/*
 *  loop for the WHILE / FOR / Scope condition
 */

   imark = VAR_DEL_MARK();                       /* mark next vars garbage  */
   while (!d4eof(dataPtr) &&                     /* while !eof()   .and. ;  */
          (!isWhile ||
           IS_VAR_EVAL_TRUE(uvWhiCond)) &&       /*   eval(bWhile) .and. ;  */
          (iScope > 0L) )                        /*   scope=ok              */
   {
      if (!isFor || IS_VAR_EVAL_TRUE(uvForCond)) /* if eval(bFor)           */
         VAR_EVAL_BLOCK(bvBlock);                /*    eval(bBlock)         */
      if (--iScope > 0L)                         /* scope --                */
         CB4FSudf(CSkip)(ovSelf,dataPtr,1L,1);   /*   skip 1 incl.Relat.    */
      VAR_DEL_SINCE_MARK(imark);                 /* garbage colection of    */
   }                                             /*   temp vars in block    */
   return CB4FS_DEBUG_END ("Eval",TRUE_VAR);
}

/*--------------------------------------------------------------------------*
 | oCb4:Replace()    METHOD general purpose REPLACE method, here performed  |
 |                          by cb4:fieldput() similarily to cb4:eval()      |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |              argv[1] = uValue|bValue|aValue  the value of replacement    |
 |              argv[2] = nFiled|cField|aField  field number or name        |
 |         opt. argv[3] = cFor  |bFor   - the for   condition               |
 |         opt. argv[4] = cWhile|bWhile - the while condition               |
 |         opt. argv[5] = lScope|nScope - the scope                         |
 | Returns    : TRUE on success, FALSE otherwise                            |
 | Description: This method is similar to the FS preprocessor translation   |
 |              of the REPLACE command and DBEVAL()                         |
 |              -The first two parameters accepts also an array of standard |
 |              variables or code blocks.                                   |
 |              -If only two parameters are specified, cb4:REPLACE() is     |
 |              subject to the server scope set with the DBServer ForBlock, |
 |              WhileBlock and Scope access/assign methods. If not set,     |
 |              this server scope defaults to "no scope," and the method    |
 |              processes the current record.                               |
 |              -The scope ends also when fieldput() returns NIL_VAR        |
 |                                                                          |
 | FS invocat.: retL := oCb4:REPLACE(expr,field)  -or-                      |
 |              retL := oCb4:REPLACE(expr,field,[forCond],[whiCond],[scop]) |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (replace)         /* e.g.  _bb_cb4cdx_replace_me(argc,argv) */
{
   FSvar  *ovSelf;
   DATA4  *dataPtr;                                   /* CB5 structure      */
   FSvar  *uvForCond = VAR_NEW;
   FSvar  *uvWhiCond = VAR_NEW;
   FSvar  *uvValue, *uvField;
   FSint  lenValueArr = 0, lenFieldArr = 0, lenMaxArr = 1;
   long   iScope = 0L;
   FSint  ii, err = 0;
   FSint  imark, isFor, isWhile;

/*
 *  parameter checking, set defaults
 */
   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, CB4_REPLACPAR);
      return FALSE_VAR;
   }

   ovSelf = argv[0];
   if (argc < 6) {
      CB4FSudf(callError)(ovSelf, 1, RTMISARG, "REPLACE",
                          CB4_REPLACPAR, __FILE__, __LINE__);
      return FALSE_VAR;
   }

   CB4FS_DEBUG_BEGIN ("Replace",ovSelf);
   if (! (dataPtr = CB4FS_GETDATA4PTR))               /* database open ?    */
      return CB4FS_DEBUG_END ("Replace",FALSE_VAR);

   err = CB4FSudf(getGlobalScope)
                 (argc, 3, argv, 1, &uvForCond, &uvWhiCond, &iScope);
   if (err) {
      sprintf (repdum, CB4_WRONGARG,             /* repdum is global buffer */
                  CB4_REPLACPAR, err);
      CB4FSudf(callError)(ovSelf, 1, RTARGTYP, "REPLACE",
               repdum, __FILE__, __LINE__);        /* RTE: wrong argument   */
      return CB4FS_DEBUG_END ("Replace",FALSE_VAR);
   }

   uvValue = argv[1];
   uvField = argv[2];

/*
 *  If par1 and/or par2 is an array: the shorter dimension is evaluated
 *   for any record within the scope.
 *  Each of them may also be an usual or block variable
 */
   if (IS_VAR_ARR(uvValue)) {
      lenValueArr = VAR_ISDIM (uvValue);
      if (! lenValueArr) {
         CB4FSudf(callError)(ovSelf, 1, RTARGVAL, "REPLACE",
               CB4_EMPTYARR1, __FILE__, __LINE__); /* RTE: empty expr array */
         return CB4FS_DEBUG_END ("Replace",FALSE_VAR);
      }
      lenMaxArr = lenValueArr;
   }

   if (IS_VAR_ARR(uvField)) {
      lenFieldArr = VAR_ISDIM (uvField);
      if (! lenFieldArr) {
         CB4FSudf(callError)(ovSelf, 1, RTARGVAL, "REPLACE",
             CB4_EMPTYARR2, __FILE__, __LINE__); /* RTE: empty field array  */
         return CB4FS_DEBUG_END ("Replace",FALSE_VAR);
      }
      if (lenMaxArr > 1)
         lenMaxArr = (lenMaxArr > lenFieldArr ? lenFieldArr : lenMaxArr);
      else
         lenMaxArr = lenFieldArr;
   }

   isFor   = IS_VAR_BLK(uvForCond);              /* FOR   available ?       */
   isWhile = IS_VAR_BLK(uvWhiCond);              /* WHILE available ?       */

/*
 *  loop for the WHILE / FOR / Scope condition
 */

   imark = VAR_DEL_MARK();                       /* mark next vars garbage  */
   while (! d4eof(dataPtr)                       /* while !eof()   .and. ;  */
          && (! isWhile ||
              IS_VAR_EVAL_TRUE(uvWhiCond))       /*   eval(bWhile) .and. ;  */
          && (iScope > 0L) )                     /*   scope=ok              */
   {
      if (!isFor || IS_VAR_EVAL_TRUE(uvForCond)) {   /* perform for all     */
         for (ii=1; ii <= lenMaxArr; ii++) {         /*  elem of the array  */
            if (lenValueArr >= ii)
               uvValue = VAR_ARRELEM(argv[1], ii);   /* value = arr[elem]   */
            if (lenFieldArr >= ii)
               uvField = VAR_ARRELEM(argv[2], ii);   /* field = arr[elem]   */

            if (IS_VAR_NIL (CB4FS_CALL_METH(fieldput)
                            (uvField,uvValue) ))     /* call cb4:fieldput() */
               return CB4FS_DEBUG_END ("Replace (F)",FALSE_VAR);
         }
      }
      if (--iScope > 0L)                         /* scope -1                */
         CB4FSudf(CSkip)(ovSelf,dataPtr,1L,1);   /*   skip 1 incl.Relat.    */
      VAR_DEL_SINCE_MARK(imark);                 /* garbage colection of    */
   }                                             /*   temp vars in block    */
   return CB4FS_DEBUG_END ("Replace (T)",TRUE_VAR);
}

/*--------------------------------------------------------------------------*
 | internal function getGlobalScope: returns the For/While/Scope settings   |
 |                          from passed parameter or the global storage     |
 |                                                                          |
 | Parameters : argc    = no.of passed arguments in argv[]                  |
 |              startc  = n pos (0..) in argv[] = argv[n] of "for" cond.    |
 |         opt. argv[n+0] = cFor  |bFor  |NIL_VAR  the for   condition      |
 |         opt. argv[n+1] = cWhile|bWhile|NIL_VAR  the while condition      |
 |         opt. argv[n+2] = lScope|nScope|NIL_VAR  the scope                |
 |              defScop   = default scope: 1 or LONG_MAX for ALL            |
 |                                                                          |
 | Returns    : *bForCond = valid For   block/NIL already init.FSvar        |
 |              *bWhiCond = valid While block/NIL already init.FSvar        |
 |              *iScope   = rest scope                                      |
 |              fn return = 0 ok, >0 wrong argv #                           |
 *--------------------------------------------------------------------------*/
FSint CB4FSudf(getGlobalScope)(argc, startc, argv, defScope, bForCond, bWhiCond, iScope)
int argc;
int startc;
FSvar *argv[];
long  defScope;
FSvar **bForCond;
FSvar **bWhiCond;
long  *iScope;
{
   int pos, diff, err = 0;
   FSvar  *uvForCond = NIL_VAR;
   FSvar  *uvWhiCond = NIL_VAR;
   FSvar  *uvScope   = FALSE_VAR;
   FSvar  *ovSelf    = argv[0];

   diff = argc - startc;
   if (diff > 3) diff = 3;
   if (diff < 0) diff = 0;

   switch (diff) {
   case 3:                                                 /* scope L/N     */
      pos = startc + 2;
      if (IS_VAR_LOG(argv[pos]) || IS_VAR_NUM(argv[pos]))
         uvScope = argv[pos];
      else
         if (!IS_VAR_NIL(argv[pos])) err = pos;
   case 2:                                                 /* while C/B     */
      pos = startc + 1;
      if (IS_VAR_CHR(argv[pos]) || IS_VAR_BLK(argv[pos])) {
         uvWhiCond = argv[pos];
		 if(! IS_VAR_NUM(uvScope)) uvScope = TRUE_VAR;
	  }
	  else
         if (!IS_VAR_NIL(argv[pos])) err = pos;
   case 1:                                                 /* for   C/B     */
      pos = startc;
      if (IS_VAR_CHR(argv[pos]) || IS_VAR_BLK(argv[pos]))
         uvForCond = argv[pos];
      else
         if (!IS_VAR_NIL(argv[pos])) err = pos;
      break;
   case 0:
      uvForCond = VAR_OBJ(ovSelf,CB4OBJ_FOR_COND);
      uvWhiCond = VAR_OBJ(ovSelf,CB4OBJ_WHI_COND);
      uvScope   = VAR_OBJ(ovSelf,CB4OBJ_SCOPE);

      if (   IS_VAR_NIL(uvForCond)
		  && IS_VAR_NIL(uvWhiCond)
		  && IS_VAR_NIL(uvScope)  ) {            /* if no scope is defined  */
         *iScope = defScope;
         if (defScope == LONG_MAX)               /* == ALL ?                */
		    CB4FS_CALL_METH(gotop)(1,argv);      /*   goto TOP              */
	  }
   }
   if (err)
      return err;

   switch (VAR_ISTYPE(uvScope)) {
   case 'U':
	  if(IS_VAR_NIL(uvWhiCond))
	     CB4FS_CALL_METH(gotop)(1,argv);         /*   goto TOP              */
      *iScope = LONG_MAX;                        /* new scope = rest        */
      break;
   case 'I':
   case 'N':
   case 'F':
      if(VAR_INTNUM(uvScope) == 0 && VAR_ISNIL(uvScope)) {
         *iScope = defScope;
         if (defScope == LONG_MAX)               /* == ALL ?                */
		      CB4FS_CALL_METH(gotop)(1,argv);      /*   goto TOP              */
      } else
         *iScope = (long) VAR_INTNUM(uvScope);   /* next N records          */
      break;
   case 'L':
      if (! IS_VAR_TRUE(uvScope))                /* FALSE = DBSCOPEALL      */
         CB4FS_CALL_METH(gotop)(1,argv);         /*   goto TOP              */
      *iScope = LONG_MAX;                        /* new scope = rest        */
      break;
   default:
      *iScope = 0L;
      break;
   }

   if (IS_VAR_CHR(uvForCond)) {                  /* compile char -> block   */
      SET_VAR_COPY (*bForCond, COMPILE_BLOCK (uvForCond));
      if(IS_VAR_EMPTY(uvForCond))
         VAR_ISNIL(*bForCond) = 1;
   } else {
      SET_VAR_COPY (*bForCond, uvForCond);
      if(IS_VAR_NIL(uvForCond))
         VAR_ISNIL(*bForCond) = 1;
   }  
          
   if (IS_VAR_CHR(uvWhiCond)) {                   /* compile char -> block   */
      SET_VAR_COPY (*bWhiCond, COMPILE_BLOCK (uvWhiCond));
      if(IS_VAR_EMPTY(uvWhiCond))
         VAR_ISNIL(*bWhiCond) = 1;
   } else {
      SET_VAR_COPY (*bWhiCond, uvWhiCond);
      if(IS_VAR_NIL(uvWhiCond))
         VAR_ISNIL(*bWhiCond) = 1;
   }
   return 0;
}

/*--------------------------------------------------------------------------*
 | internal function: determines if any global scope (For, While, Scope)    |
 |                    is set.                                               |
 |                                                                          |
 | Returns    : (int) 1..3= some of the glocal scope is active, 0 otherwise |
 *--------------------------------------------------------------------------*/
FSint CB4FSudf(isScope)(ovSelf)      /* e.g.  _bb_cb4cdx_isScope_fn(ovSelf) */
FSvar *ovSelf;
{

   switch (VAR_ISTYPE(
             VAR_OBJ(ovSelf,CB4OBJ_FOR_COND))) {   /* 'For' Char or Block   */
   case 'C':
   case 'E':
      if (VAR_ISLEN( VAR_OBJ(ovSelf,CB4OBJ_FOR_COND)) > 1)
         return 1;
   case 'B':
   case 'b':
         return 1;
   }

   switch (VAR_ISTYPE(
             VAR_OBJ(ovSelf,CB4OBJ_WHI_COND))) {   /* 'While' Char or Block */
   case 'C':
   case 'E':
      if (VAR_ISLEN( VAR_OBJ(ovSelf,CB4OBJ_WHI_COND)) > 1)
         return 2;
   case 'B':
   case 'b':
         return 2;
   }

   switch (VAR_ISTYPE( VAR_OBJ(ovSelf,CB4OBJ_SCOPE))) { /* Next/rest/record */
   case 'I':
   case 'N':
   case 'F':
      if (VAR_INTNUM( VAR_OBJ(ovSelf,CB4OBJ_SCOPE)) != 0L)
         return 3;
   }

   return 0;
}

/*--------------------------------------------------------------------------*
 | Internal purpose for all cb4RDDs: decode the object scope to 3 param     |
 |                                                                          |
 | in: scope   = object scope (N,L)                                         |
 |     storage = array of FSvars to store the results                       |
 |     next    = position (0..n) in storage for NEXT clause                 |
 |     rest    = position (0..n) in storage for the REST/ALL clause         |
 | ret: 0 = ok, 1 = error                                                   |
 *--------------------------------------------------------------------------*/
FSint CB4FSudf(decodeScope)(scope, storage, next, rest)
FSvar *scope;
FSvar *storage[];
FSint next, rest;
{
   int retErr = 0;

   switch (VAR_ISTYPE(scope)) {
   case 'I':
   case 'N':
   case 'F': SET_VAR_COPY (storage[next], scope);   /* NEXT <n> clause   */
             break;
   case 'L': if (IS_VAR_TRUE(scope))
                SET_VAR_LOG (storage[rest], 1);       /* REST clause       */
             else
                SET_VAR_LOG (storage[rest], 0);       /* ALL  clause       */
   case 'U': break;
   default : retErr = 1;
   }
   return retErr;
}


/*--------------------------------------------------------------------------*
 | Internal purpose for all cb4RDDs: decode the For or While condition      |
 |                                                                          |
 | in : entered  condition (C/B/U)                                          |
 | out: resultig condition (B/nil), must be initilized FSvar                |
 | ret: 0 = ok, 1 = error                                                   |
 *--------------------------------------------------------------------------*/
FSint CB4FSudf(decodeCond)(inVar, outVar)
FSvar *inVar;
FSvar *outVar;
{
   int ok = 0;

   switch (VAR_ISTYPE(inVar)) {
     case 'E':
     case 'C': SET_VAR_COPY(outVar, COMPILE_BLOCK(inVar));
               break;
     case 'B':
     case 'b':
     case 'U': SET_VAR_COPY(outVar, inVar);
               break;
     default : SET_VAR_NIL(outVar);
               ok = 1;
   }
   return ok;
}

/**** eof rddcb4e.c ****/

