/****************************************************************************
 * rddcb4f.c     RDD API for the CB4FOX, CB4CDX, CB4NDX, CB4MDX and CB4NTX  *
 *               by using CodeBase 5.1 multiplatform version (not included  *
 *               in the FlagShip package, contact Sequiter Inc.)            *
 *                                                                          *
 *               Note: all CodeBase function & defines are marked by CB5    *
 *                                                                          *
 * (c) 1995..2004 multisoft Datentechnik GmbH, Munich. All rights reserved  *
 * Release $Date: 2004/10/27 09:52:02 $                                     *
 *--------------------------------------------------------------------------*
 * This file contains parts of the full RDD driver:                         *
 *    oCb4:filter            ACCESS object-wide filter                      *
 *    oCb4:filter            ASSIGN object-wide filter                      *
 *    oCb4:getlocate         ACCESS the LOCATE string/block                 *
 *                                                                          *
 *    oCb4:clearfilter()     METHOD clears the current filter               *
 *    oCb4:clearlocate()     METHOD clears the locate scope                 *
 *    oCb4:continue()        METHOD continue pending locate()               *
 *    oCb4:locate()          METHOD locate for record                       *
 *    oCb4:setfilter()       METHOD set object-wide filter block/string     *
 *                                                                          *
 * For additional methods, see rddcb4a.c to rddcb4?.c                       *
 ****************************************************************************/

#include "rddcb4fs.h"                            /* RDD API for CodeBase    */

extern FSvar *CB4FS_CALL_ASSIGN(filter)();
extern FSint  CB4FSudf(callError)();             /* in rddcb4b.c            */
extern FSint  CB4FSudf(CSkip)();                 /* in rddcb4b.c            */
extern FSint  CB4FSudf(getGlobalScope)();        /* in rddcb4e.c            */
extern FSint  CB4FSudf(moveChildRelation)();     /* in rddcb4r.c            */

#define CB4_ASSFILTPAR   "cb4:Filter := exprC|exprB|NIL"
#define CB4_SETFILTPAR   "cb4:SetFilter([bFilter],[cFilter])"
#define CB4_SETFILTARG   "%s argument %d"
#define CB4_SETFILTBLK   "%s;block does not return logical value"
#define CB4_SETFILTSTR   "%s;string evaluation does not return logical value"
#define CB4_LOCATEPAR    "cb4:Locate([c|bFor],[c|bWhile],[n|lScope])"
#define CB4_LOCATESTR    "%s;wrong type of argument no. %d"
#define CB4_LOCATEFOR    "%s;'For' code block (arg/global) required"
#define CB4_CONTINPAR    "cb4:Continue()"
#define CB4_CLEARFILT    "cb4:ClearFilter()"
#define CB4_CLEARLOCAT   "cb4:ClearLocate()"

/*--------------------------------------------------------------------------*
 | x=oCb4:Filter  ACCESS retrieve the FILTER string                         |
 |                                                                          |
 | Parameters : ovSelf  = object:self                                       |
 | Returns    : retC    = the current Filter string or "" if not set        |
 | Description: retrieves the object storage                                |
 | FS invocat.: retC := oCb4:FILTER                                         |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_ACCESS (filter)            /* e.g.  _bb_cb4cdx_filter_ac(ovSelv) */
{
   FSvar    *cvRetVar = VAR_NEW;

   if (IS_VAR_CHR ( VAR_OBJ(ovSelf,CB4OBJ_FILTERSTRING)))
      SET_VAR_COPY (cvRetVar, VAR_OBJ(ovSelf,CB4OBJ_FILTERSTRING));
   else
      SET_VAR_CHR (cvRetVar, "");
   return cvRetVar;
}

/*--------------------------------------------------------------------------*
 | x=oCb4:Filter  ASSIGN set the FILTER string (and the block)              |
 |                                                                          |
 | Parameters : ovSelf  = object:self                                       |
 |              assVar  = new filter string (and block)                     |
 | Returns    : ovSelf  = object:self                                       |
 | Description: sets the object storage                                     |
 | Compatib.  : CA/VO 1.0 allows to set char only - as opposite to descript.|
 | FS invocat.: oCb4:FILTER := exprC                                        |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_ASSIGN (filter)     /* e.g.  _bb_cb4cdx_filter_as(ovSelv,assVar) */
{
   CB4FS_DEBUG_BEGIN ("Filter-Ass",ovSelf);

   switch (VAR_ISTYPE(assVar)) {
   case 'E':
   case 'C':
      SET_VAR_COPY (VAR_OBJ(ovSelf,CB4OBJ_FILTERSTRING), assVar);
      if (VAR_ISLEN (assVar) < 2)
         SET_VAR_NIL  (VAR_OBJ(ovSelf,CB4OBJ_FILTERBLOCK));
      else
         SET_VAR_COPY (VAR_OBJ(ovSelf,CB4OBJ_FILTERBLOCK),
                           COMPILE_BLOCK (assVar));
      break;
   case 'B':
   case 'b':
      SET_VAR_CHR  (VAR_OBJ(ovSelf,CB4OBJ_FILTERSTRING), "");
      SET_VAR_COPY (VAR_OBJ(ovSelf,CB4OBJ_FILTERBLOCK ), assVar);
      break;
   case 'U':
      SET_VAR_CHR  (VAR_OBJ(ovSelf,CB4OBJ_FILTERSTRING), "");
      SET_VAR_NIL  (VAR_OBJ(ovSelf,CB4OBJ_FILTERBLOCK));
      break;
   default:
      CB4FSudf(callError)(ovSelf, 1, RTARGTYP, "FILTER",
               CB4_ASSFILTPAR, __FILE__, __LINE__);   /* RTE wrong arg type */
   }

/*
 *  check the code block if evaluates to logical
 */
   if (! IS_VAR_NIL(VAR_OBJ(ovSelf,CB4OBJ_FILTERBLOCK))) {
      if (! IS_VAR_LOG (VAR_EVAL_BLOCK(
            VAR_OBJ(ovSelf,CB4OBJ_FILTERBLOCK))) ) { /* eval to logical ?   */
         sprintf (repdum, CB4_SETFILTBLK, CB4_ASSFILTPAR);
         CB4FSudf(callError)(ovSelf, 1, RTARGVAL, "FILTER",
              repdum, __FILE__, __LINE__);           /* RTE: wrong arg val  */
      }
   }
   return CB4FS_DEBUG_END ("Filter-Ass", assVar);
}

/*--------------------------------------------------------------------------*
 | x=oCb4:GetLocate  ACCESS retrieve the LOCATE code block                  |
 |                                                                          |
 | Parameters : ovSelf  = object:self                                       |
 | Returns    : retB    = the current Locate code block or NIL if not set   |
 | Description: retrieves the object storage                                |
 | FS invocat.: retB := oCb4:GETLOCATE                                      |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_ACCESS (getlocate)      /* e.g.  _bb_cb4cdx_getlocate_ac(ovSelv) */
{
   FSvar    *cvRetVar = VAR_NEW;

   SET_VAR_COPY (cvRetVar, VAR_OBJ(ovSelf,CB4OBJ_LOCATEBLOCK));
   return cvRetVar;
}

/*--------------------------------------------------------------------------*
 | oCb4:ClearFilter()  METHOD clears the filter condition                   |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 | Returns    : TRUE on success, FALSE otherwise.                           |
 | Description: equivalent to cb4:FILTER := NIL                             |
 | FS invocat.: retL := oCb4:CLEARFILTER()                                  |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (clearfilter) /* e.g.  _bb_cb4cdx_clearfilter_me(argc,argv) */
{
   FSvar  *ovSelf;

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, CB4_CLEARFILT);              /* RTE: missing args  */
      return FALSE_VAR;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("ClearFilter",ovSelf);
   CB4FS_CALL_ASSIGN(filter)(argv[0], NIL_VAR);
   return CB4FS_DEBUG_END ("ClearFilter",TRUE_VAR);
}

/*--------------------------------------------------------------------------*
 | oCb4:ClearLocate()  METHOD clears the locate condition                   |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 | Returns    : TRUE on success, FALSE otherwise.                           |
 | Description: clears to condition from cb4:locate()                       |
 | FS invocat.: retL := oCb4:CLEARLOCATE()                                  |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (clearlocate) /* e.g.  _bb_cb4cdx_clearlocate_me(argc,argv) */
{
   FSvar  *ovSelf;

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, CB4_CLEARLOCAT);             /* RTE: missing args  */
      return FALSE_VAR;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("ClearLocate",ovSelf);
   SET_VAR_NIL (VAR_OBJ(ovSelf,CB4OBJ_LOCATEBLOCK));
   SET_VAR_NIL (VAR_OBJ(ovSelf,CB4OBJ_LOCATERECNO));
   return CB4FS_DEBUG_END ("ClearLocate",TRUE_VAR);
}

/*--------------------------------------------------------------------------*
 | oCb4:SetFilter()  METHOD sets the filter block and/or filter string      |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |              argv[1] = (opt) Filter code block (used for processing)     |
 |              argv[2] = (opt) Filter string     (info only)               |
 | Returns    : TRUE on success, FALSE otherwise.                           |
 | Description: The code block is used for the execution. When the string   |
 |              is given only, it will be compiled to code block.           |
 |              The string itself is used for info only and is accessabe    |
 |              by the cb4:FILTER access.                                   |
 | FS invocat.: retL := oCb4:SETFILTER( [bFilter],[cFilter] )               |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (setfilter)    /* e.g.  _bb_cb4cdx_setfilter_me(argc, argv) */
{
   FSvar  *ovSelf;
   FSvar  *bvTemp;

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, CB4_SETFILTPAR);             /* RTE: missing args  */
      return FALSE_VAR;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("SetFilter",ovSelf);

/*
 * reset filter ?
 */
   switch (argc) {
   case 3:
      if (! IS_VAR_NIL (argv[3])) break;
   case 2:
      if (! IS_VAR_NIL (argv[2])) break;
   case 1:
      SET_VAR_NIL  (VAR_OBJ(ovSelf,CB4OBJ_FILTERBLOCK));
      SET_VAR_CHR  (VAR_OBJ(ovSelf,CB4OBJ_FILTERSTRING), "");
      return CB4FS_DEBUG_END ("SetFilter (T)",TRUE_VAR);
   }
/*
 * filter block specified ?
 */
   if (argc > 1 && IS_VAR_BLK(argv[1])) {
      if (! IS_VAR_LOG (VAR_EVAL_BLOCK(argv[1])) ) { /* eval to logical ?   */
         sprintf (repdum, CB4_SETFILTBLK,
                    CB4_SETFILTPAR);
         CB4FSudf(callError)(ovSelf, 1, RTARGVAL,
               "SETFILTER", repdum,
               __FILE__, __LINE__);                  /* RTE: wrong arg val  */
         return CB4FS_DEBUG_END ("SetFilter (F)",FALSE_VAR);
      }
      SET_VAR_COPY (VAR_OBJ(ovSelf,CB4OBJ_FILTERBLOCK ), argv[1]);
      if (argc > 2 && IS_VAR_CHR(argv[1]))
         SET_VAR_COPY (VAR_OBJ(ovSelf,CB4OBJ_FILTERSTRING), argv[2]);
      else
         SET_VAR_CHR  (VAR_OBJ(ovSelf,CB4OBJ_FILTERSTRING), "");
      return CB4FS_DEBUG_END ("SetFilter (T)",TRUE_VAR);
   }
/*
 * ony filter string specified ?
 */
   if (argc > 2 && IS_VAR_CHR(argv[2])) {
      if (VAR_ISLEN(argv[2]) < 2) {
         SET_VAR_NIL  (VAR_OBJ(ovSelf,CB4OBJ_FILTERBLOCK));
         SET_VAR_CHR  (VAR_OBJ(ovSelf,CB4OBJ_FILTERSTRING), "");
         return CB4FS_DEBUG_END ("SetFilter (T)",TRUE_VAR);
      }
      bvTemp = VAR_NEW;
      SET_VAR_COPY (bvTemp, COMPILE_BLOCK (argv[2]));
      if (! IS_VAR_LOG (VAR_EVAL_BLOCK(bvTemp)) ) {  /* eval to logical ?   */
         sprintf (repdum, CB4_SETFILTSTR,
                    CB4_SETFILTPAR);                 /* no,                 */
         CB4FSudf(callError)(ovSelf, 1, RTARGVAL,
               "SETFILTER", repdum,
               __FILE__, __LINE__);                  /* RTE: wrong arg val  */
         return CB4FS_DEBUG_END ("SetFilter (F)",FALSE_VAR);
      }
      SET_VAR_COPY (VAR_OBJ(ovSelf,CB4OBJ_FILTERSTRING), argv[2]);
      SET_VAR_COPY (VAR_OBJ(ovSelf,CB4OBJ_FILTERBLOCK ), bvTemp);
      return CB4FS_DEBUG_END ("SetFilter (T)",TRUE_VAR);
   }

   sprintf (repdum, CB4_SETFILTARG, CB4_SETFILTPAR, 1);
   CB4FSudf(callError)(ovSelf, 1, RTARGTYP,
            "SETFILTER", repdum,
            __FILE__, __LINE__);                     /* RTE: wrong arg typ  */
   return CB4FS_DEBUG_END ("SetFilter (F)",FALSE_VAR);
}

/*--------------------------------------------------------------------------*
 | oCb4:Locate()     METHOD search sequentially for the specif.condition    |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |              argv[1] = (opt) searched for condition    (C/B)             |
 |              argv[2] = (opt) while condition           (C/B)             |
 |              argv[3] = (opt) scope                     (N/L)             |
 |              argv[4] = (opt) consider relations on EACH skip,def=.F.  (1)|
 | Returns    : TRUE if record was found, FALSE otherwise.                  |
 |              The return value is stored to cb4:FOUND access variable.    |
 | Note       : (1) is not supported by CA/VO                               |
 |              You must set the 4th param TRUE if the condition or scope   |
 |              affect any related database, but it may slow-down           |
 |              the execution significantly.                                |
 | Description: -If NO parameters are specified, the object-wide settings   |
 |               of the for-block, while-block and scope is used.           |
 |              -The for-condition is stored to locate-block var, which is  |
 |               used in subsequent cb4:Continue()                          |
 |              -If no parameters or object-cond is set, no Locate is perf. |
 |              -If relations are set, each skip moves also related childs. |
 |               when par3 is set TRUE (default = FALSE).                   |
 | FS invocat.: retL := oCb4:LOCATE([cbFor],[cbWhile],[nlScope])            |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (locate)          /* e.g.  _bb_cb4cdx_locate_me(argc, argv) */
{
   DATA4  *dataPtr;                                   /* CB5 structure      */
   FSvar  *ovSelf;
   FSvar  *uvForCond = VAR_NEW;
   FSvar  *uvWhiCond = VAR_NEW;
   long   iScope = 0L;
   FSint  imark, isFor, isWhile, isRel = 0;
   FSint  err = 0, found = 0;

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, CB4_LOCATEPAR);              /* RTE: missing args  */
      return FALSE_VAR;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("Locate",ovSelf);

   SET_VAR_LOG (VAR_OBJ(ovSelf,CB4OBJ_FOUNDFLAG), 0);      /* not found yet */

   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("Locate (F)",FALSE_VAR);

   err = CB4FSudf(getGlobalScope)
                 (argc, 1, argv, 1, &uvForCond, &uvWhiCond, &iScope);
   if (err) {
      sprintf (repdum, CB4_LOCATESTR,            /* repdum is global buffer */
                  CB4_LOCATEPAR, err);
      CB4FSudf(callError)(ovSelf, 1, RTARGTYP,
               "LOCATE", repdum,
               __FILE__, __LINE__);              /* RTE: wrong argument     */
      return CB4FS_DEBUG_END ("Locate",FALSE_VAR);
   }

   isFor   = IS_VAR_BLK(uvForCond);              /* FOR   available ?       */
   isWhile = IS_VAR_BLK(uvWhiCond);              /* WHILE available ?       */
   isRel   = ((argc > 4) && IS_VAR_TRUE (argv[4]));

   if (! isFor) {
      sprintf (repdum, CB4_LOCATEFOR, CB4_LOCATEPAR);
      CB4FSudf(callError)(ovSelf, 1, RTARGVAL,
               "LOCATE", repdum,
               __FILE__, __LINE__);              /* RTE: for required       */
      return CB4FS_DEBUG_END ("Locate",FALSE_VAR);
   }

/*
 *  loop for the WHILE / FOR / Scope condition
 */

   imark = VAR_DEL_MARK();                       /* mark next vars garbage  */
   while (!d4eof(dataPtr)                        /* while !eof()   .and. ;  */
          && (!isWhile ||
              IS_VAR_EVAL_TRUE(uvWhiCond))       /*   eval(bWhile) .and. ;  */
          && (iScope > 0L) )                     /*   scope=ok              */
   {
      if (IS_VAR_EVAL_TRUE(uvForCond)) {         /* if eval(bFor)           */
         found = 1;                              /* found !                 */
         break;
      }
      if (--iScope > 0L)                         /* scope --                */
         CB4FSudf(CSkip)(ovSelf,dataPtr,1L,isRel); /*  skip 1, NO relation  */
      VAR_DEL_SINCE_MARK(imark);                 /* garbage colection of    */
   }                                             /*   temp vars in block    */

/*
 *  set object vars
 */
   SET_VAR_COPY (VAR_OBJ(ovSelf,CB4OBJ_LOCATEBLOCK), uvForCond);
   SET_VAR_LOG  (VAR_OBJ(ovSelf,CB4OBJ_FOUNDFLAG),   found);
   SET_VAR_INT  (VAR_OBJ(ovSelf,CB4OBJ_LOCATERECNO),
                 (int) dataPtr->rec_num);                    /* long == int */

   if (! isRel)
      CB4FSudf(moveChildRelation)(ovSelf, dataPtr);  /* move related childs */

   return CB4FS_DEBUG_END ("Locate",found ? TRUE_VAR : FALSE_VAR);
}

/*--------------------------------------------------------------------------*
 | oCb4:Continue()   METHOD continue locate() on next record                |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |              argv[1] = (opt) consider relations on EACH skip,def=.F.  (1)|
 | Note       : (1) is not supported by CA/VO                               |
 |              You must set the 4th param TRUE if the condition or scope   |
 |              affect any related database, but it may slow-down           |
 |              the execution significantly.                                |
 | Returns    : TRUE if record was found, FALSE otherwise.                  |
 |              The return value is stored to cb4:FOUND access variable.    |
 | Description: The for-block of previously executed locate() is used.      |
 | FS invocat.: retL := oCb4:CONTINUE()                                     |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (continue)      /* e.g.  _bb_cb4cdx_continue_me(argc, argv) */
{
   DATA4  *dataPtr;                                   /* CB5 structure      */
   FSvar  *ovSelf;
   FSvar  *uvForCond = NIL_VAR;
   FSint  imark;
   FSint  found = 0, iRel = 0;
   long   recNo;

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, CB4_CONTINPAR);              /* RTE: missing args  */
      return FALSE_VAR;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("Continue",ovSelf);

   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("Continue (F)",FALSE_VAR);

   SET_VAR_LOG (VAR_OBJ(ovSelf,CB4OBJ_FOUNDFLAG), 0);      /* not found yet */

   uvForCond = VAR_OBJ(ovSelf,CB4OBJ_LOCATEBLOCK);
   if (! IS_VAR_BLK(uvForCond) ||
       ! IS_VAR_INT (VAR_OBJ(ovSelf,CB4OBJ_LOCATERECNO)) ) {
      sprintf (repdum, CB4_LOCATEFOR, CB4_CONTINPAR);
      CB4FSudf(callError)(ovSelf, 1, RTUNVAR,
            "LOCATE", repdum,
            __FILE__, __LINE__);                 /* RTE: undef.var 'for'    */
      return CB4FS_DEBUG_END ("Continue (F)",FALSE_VAR);
   }
/*
 *  skip to next record, if still on the last locate() or continue() record
 */
   recNo = dataPtr->rec_num;
   if ((int) recNo == VAR_INT (VAR_OBJ(ovSelf,CB4OBJ_LOCATERECNO)))
      CB4FSudf(CSkip)(ovSelf,dataPtr,1L,iRel);   /*   skip 1, NO relation   */

/*
 *  loop for the FOR condition
 */
   imark = VAR_DEL_MARK();                       /* mark next vars garbage  */
   while (!d4eof(dataPtr) ) {                    /* while !eof()            */
      if (IS_VAR_EVAL_TRUE(uvForCond)) {         /* if eval(bFor)           */
         found = 1;                              /* found !                 */
         break;
      }
      CB4FSudf(CSkip)(ovSelf,dataPtr,1L,iRel);   /*   skip 1, NO relation   */
      VAR_DEL_SINCE_MARK(imark);                 /* garbage colection of    */
   }                                             /*   temp vars in block    */

/*
 *  set object vars
 */
   SET_VAR_COPY (VAR_OBJ(ovSelf,CB4OBJ_LOCATEBLOCK), uvForCond);
   SET_VAR_LOG  (VAR_OBJ(ovSelf,CB4OBJ_FOUNDFLAG),   found);
   SET_VAR_INT  (VAR_OBJ(ovSelf,CB4OBJ_LOCATERECNO),
                 (int) dataPtr->rec_num);                    /* long == int */

   CB4FSudf(moveChildRelation)(ovSelf, dataPtr);     /* move related childs */

   return CB4FS_DEBUG_END ("Continue",found ? TRUE_VAR : FALSE_VAR);
}

/**** eof rddcb4f.c ****/

