/****************************************************************************
 * rddcb4g.c     RDD API for the CB4FOX, CB4CDX, CB4NDX, CB4MDX and CB4NTX  *
 *               by using CodeBase 5.1 multiplatform version (not included  *
 *               in the FlagShip package, contact Sequiter Inc.)            *
 *                                                                          *
 *               Note: all CodeBase function & defines are marked by CB5    *
 *                                                                          *
 * (c) 1995..2007 multisoft Datentechnik GmbH, Munich. All rights reserved  *
 * Release $Date: 2007/01/25 15:14:46 $                                     *
 *--------------------------------------------------------------------------*
 * This file contains parts of the full RDD driver:                         *
 *    oCb4:clearindex()        METHOD clear index                           *
 *    oCb4:createindex()       METHOD create index file with order          *
 *    oCb4:createorder()       METHOD create order within existing index    *
 *    oCb4:deleteorder()       METHOD destroy order                         *
 *    oCb4:reindex()           METHOD reindex                               *
 *    oCb4:setordercondition() METHOD set order condition                   *
 *                                                                          *
 * For additional methods, see rddcb4a.c to rddcb4?.c                       *
 ****************************************************************************/

/* #define TEST_OUT */

#include "rddcb4fs.h"                            /* RDD API for CodeBase    */
#include "FSextend.h"                            /* for malloc (_xalloc)    */

#ifdef TEST_ON
extern FSvar *CB4FSudf(rddTestOut)();   /* test only */
#endif

extern FSvar *CB4FS_CALL_METH(clearindex)();
extern FSvar *CB4FS_CALL_METH(createindex)();
extern FSvar *CB4FS_CALL_METH(setordercondition)();
extern FSvar *CB4FS_CALL_METH(setorder)();                  /* rddcb4h.c    */

extern FSint fgsExtractName();
extern FSvar *fsFnNAME(aadd)();                             /* std. AADD()  */
extern FSint CB4FSudf(checkFile)();                         /* rddcb4a.c    */
extern FSint CB4FSudf(setIndexArr)();                       /* rddcb4g.c    */
extern FSint CB4FSudf(callError)();                         /* rddcb4b.c    */
extern FSint CB4FSudf(myAutoRFlock)();                      /* rddcb4b.c    */
extern FSint CB4FSudf(myAutoUnlock)();                      /* rddcb4b.c    */

static FSint myFlockCheck ();

#define CB4_ARGUMENTSTR   "%s;argument %d"
#define CB4_ARGNOTSUPP    "%s;argument %d ignored by this RDD"
#define CB4_NOTSUPPORT    "%s;not supported by this RDD"
#define CB4_FLOCKREQUIR   "%s;at least Flock() required"
#define CB4_EXCLREQUIR    "%s;exclusive open required"
#define CB4_READONLY      "%s;not possible on ReadOnly file"
#define CB4_CBFUNCTFAIL   "%s;CodeBase function fails with error %d"
#define CB4_CREINDEXPAR   "cb4:CreateIndex(cFileName,cExpr,[bExpr],[lUniq],[lNoFlock])"
#define CB4_CREORDERPAR   "cb4:CreateOrder(cFileName,cOrder,cExpr,[bExpr],[lUniq],[lNoFlock])"
#define CB4_CLEARINDEX    "cb4:ClearIndex()"
#define CB4_DELORDERPAR   "cb4:DeleteOrder(c|nOrder,cFile)"
#define CB4_REINDEXPAR    "cb4:Reindex()"
#define CB4_SETORDCONPAR  "cb4:SetOrderCondition ([cFor],[bFor],[lAllOrders],[bWhile],;[bEval],[nStep],[nStart],[nCount],[nRecno],[lRest],[lDesc])"
#define CB4_NAMETOOLONG   "%s;file name too long"

/*--------------------------------------------------------------------------*
 | oCb4:ClearIndex()  METHOD clears all indexes associated to this WA       |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 | Returns    : TRUE on success, FALSE otherwise.                           |
 | Description: This RDD supports an unlimited number of indicies           |
 | FS invocat.: retL := oCb4:CLEARINDEX()                                   |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (clearindex)   /* e.g.  _bb_cb4cdx_clearindex_me(argc,argv) */
{
   FSvar    *ovSelf;
   DATA4    *dataPtr;                                 /* CB5 structure      */
   TAG4INFO tagInfo[2];                               /* CB5 structure      */
   INDEX4   *indexPtr;                                /* CB5 structure      */
   FSint    ii;
   FSvar    *aIndexArr;
   FSvar    *aSubArr;

/*  each SET INDEX entry occupies following subarray
 *  stored in object element oCb4[CB4OBJ_INDEX4ARR]
 *  aIndexArray :=
 *      [n, {cFullFileName,
 *           cTagName,
 *           iTagNo,
 *           lNewMalloc,
 *           specIndexPtr,
 *           specTagPtr} ]
 */

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, CB4_CREINDEXPAR);            /* RTE: missing args  */
      return FALSE_VAR;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("ClearIndex",ovSelf);

   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("ClearIndex",FALSE_VAR); /* database not open */
   aIndexArr = CB4FS_GETINDEXARR;
   if (IS_VAR_NIL(aIndexArr))
      return CB4FS_DEBUG_END ("ClearIndex",FALSE_VAR); /* indexes  not open */
/*
 *  close indexes,
 *  frees allocated memory
 */
   d4flush (dataPtr);
   for (ii = 1; ii <= VAR_ISDIM(aIndexArr); ii++) {
      aSubArr = VAR_ARRELEM(aIndexArr,ii);
      if ((VAR_ISDIM(aSubArr) == CB4OBJ_INDEX_T4PTR) &&
          (IS_VAR_TRUE (VAR_ARRELEM(aSubArr,CB4OBJ_INDEX_I4NEW))) ) {
         if (IS_VAR_SPECIAL( VAR_ARRELEM( aSubArr,CB4OBJ_INDEX_I4PTR))) {
            indexPtr = VAR_SPECIAL (VAR_ARRELEM(aSubArr,CB4OBJ_INDEX_I4PTR));
            if (indexPtr) {
               if (i4close (indexPtr)) {                     /* CB5 fn      */
                  sprintf (repdum, CB4_CBFUNCTFAIL,
                           CB4_CLEARINDEX, dataPtr->code_base->error_code);
                  CB4FSudf(callError)(ovSelf, 1, DEVWARNING,
                           "CLEARINDEX", repdum,
                           __FILE__, __LINE__);              /* Devel: fail */
                  return CB4FS_DEBUG_END ("ClearIndex",FALSE_VAR);

               }
/**            _xfree (indexPtr);   i4close should free it **/
            }
         }
         SET_VAR_NIL (VAR_ARRELEM(aSubArr,CB4OBJ_INDEX_I4PTR));
         SET_VAR_NIL (VAR_ARRELEM(aSubArr,CB4OBJ_INDEX_T4PTR));
      }
   }
/*
 *  now, frees the whole array by the FS variable system
 */
  VAR_CRE_ARRAY (aIndexArr,0);         /* reset to NULL_ARRAY */
  return CB4FS_DEBUG_END ("ClearIndex (T)",TRUE_VAR);
}

/*--------------------------------------------------------------------------*
 | oCb4:CreateIndex() METHOD creates an index file with an order in it      |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |              argv[1] = (C) index file name                               |
 |              argv[2] = (C) index expression stored in order header       |
 |              argv[3] = (B opt) index expression code block for indexing *|
 |                        used during the creation process only, not stored |
 |              argv[4] = (L opt) unique flag                               |
 |              argv[5] = (L opt) TRUE if no FLOCK required, default=FALSE  |
 | Returns    : TRUE on success, FALSE otherwise.                           |
 | Description: This method is works like CreateOrder() which also supports |
 |               multiple-order-bags and should therefore be preffered.     |
 |              Additional flags/conditions may be set by using the         |
 |               cb4:SetOrderCondition() method.                            |
 |              * Since CodeBase 5.1 does NOT support code blocks, and      |
 |                i4create() has no parameter for the "create" CB/expr,     |
 |                the 3rd parameter is checked but ignored here.            |
 |              When CB4FS_CBLOCK_SUPPORT is enabled, retrive the data      |
 |                from the Object stored there by SetOrderCondition().      |
 | Compatib.  : see note in SetOrderCondition() below.                      |
 |              argv[5] is not supported by CA/VO                           |
 | Multiuser  : In multiuser mode, at least FLOCK() must be set,            |
 |              or par5 must be set .T.,                                    |
 |              or SET AUTOLOCK ON must be set                              |
 | FS invocat.: retL := oCb4:CREATEINDEX(cName,cExpr,...)                   |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (createindex) /* e.g.  _bb_cb4cdx_createindex_me(argc,argv) */
{
   FSvar    *ovSelf;
   DATA4    *dataPtr;                                 /* CB5 structure      */
   TAG4     *tagPtr;                                  /* CB5 structure      */
   TAG4INFO tagInfo[2];                               /* CB5 structure      */
   INDEX4   *indexPtr;                                /* CB5 structure      */
   FSint    err = 0;
   FSint    autoLock;
   char     *cDummy;
   char     *tagName;
   extern   FSchar fileNameBuff[];
   FSvar    *retVar, *param[2];

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, CB4_CREINDEXPAR);            /* RTE: missing args  */
      return FALSE_VAR;
   }
   ovSelf = argv[0];
   if (argc < 3 ) {
      CB4FSudf(callError)(ovSelf, 1, RTMISARG,
         "CREATEINDEX", CB4_CREINDEXPAR,
         __FILE__, __LINE__);                         /* RTE: missing args  */
      return FALSE_VAR;
   }
   CB4FS_DEBUG_BEGIN ("CreateIndex",ovSelf);

   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("CreateIndex",FALSE_VAR);    /* dbf not open  */
/*
 *  parameter checking
 */
   if (argc > 6) argc = 6;
   switch (argc) {
   case 6: if (!IS_VAR_LOG(argv[5]) && !IS_VAR_NIL(argv[5])) err = 5;
   case 5: if (!IS_VAR_LOG(argv[4]) && !IS_VAR_NIL(argv[4])) err = 4;
   case 4: if (!IS_VAR_BLK(argv[3]) && !IS_VAR_NIL(argv[3])) err = 3;
   }
   if (! IS_VAR_CHR(argv[2])) err = 2;
   if (! IS_VAR_CHR(argv[1])) err = 1;
   if (err) {
      sprintf (repdum, CB4_ARGUMENTSTR, CB4_CREINDEXPAR, err);
      CB4FSudf(callError)(ovSelf, 1, RTARGTYP,
         "CREATEINDEX", repdum,
         __FILE__, __LINE__);                        /* RTE: wrong arg type */
      return CB4FS_DEBUG_END ("CreateIndex",FALSE_VAR);
   }
   if (IS_VAR_EMPTY(argv[2])) err = 2;
   if (IS_VAR_EMPTY(argv[1])) err = 1;
   if (err) {
      sprintf (repdum, CB4_ARGUMENTSTR, CB4_CREINDEXPAR, err);
      CB4FSudf(callError)(ovSelf, 1, RTARGVAL,
         "CREATEINDEX", repdum,
         __FILE__, __LINE__);                        /* RTE: wrong arg val  */
      return CB4FS_DEBUG_END ("CreateIndex",FALSE_VAR);
   }
/*
 *  check the locking
 */
   autoLock = myFlockCheck (ovSelf, (argc >= 6 ? argv[5] : NIL_VAR),
                            dataPtr, CB4_CREINDEXPAR, "CREATEINDEX");
   if (!autoLock)
      return CB4FS_DEBUG_END ("CreateIndex",FALSE_VAR);
/*
 *  clear all open indicies first
 */
   CB4FS_CALL_METH(clearindex)(1, argv);

/*
 *  add default extention, transform upper/lower case according to fs_set()
 *  new full file name = fileNameBuff
 */
   if (! CB4FSudf (checkFile)
         (VAR_CHR(argv[1]), fileNameBuff, &cDummy,
          VAR_CHR(VAR_OBJ (ovSelf,CB4OBJ_IDX_EXT)), CB4_CREINDEXPAR, -2) )
      return CB4FS_DEBUG_END ("CreateIndex",FALSE_VAR);
/*
 *  prepare indexing
 */
   tagName = VAR_CHR(argv[1]);

   tagInfo[0].name       = fileNameBuff;
   tagInfo[0].expression = VAR_CHR(argv[2]);
   tagInfo[0].filter     = (IS_VAR_EMPTY(VAR_OBJ(ovSelf,CB4OBJ_INDEXFORSTR)) ?
                            0 : VAR_CHR(VAR_OBJ(ovSelf,CB4OBJ_INDEXFORSTR)));
#ifdef CODEBASE6
   tagInfo[0].unique     = (short) (argc >= 5 ? IS_VAR_TRUE(argv[4]) : 0);
   tagInfo[0].descending = (unsigned short) IS_VAR_TRUE(VAR_OBJ(ovSelf,
														   CB4OBJ_INDEXDESC));
#else
   tagInfo[0].unique     = (argc >= 5 ? IS_VAR_TRUE(argv[4]) : 0);
   tagInfo[0].descending = IS_VAR_TRUE(VAR_OBJ(ovSelf,CB4OBJ_INDEXDESC));
#endif

   tagInfo[1].name       = 0;
   tagInfo[1].expression = 0;
   tagInfo[1].filter     = 0;
   tagInfo[1].unique     = 0;
   tagInfo[1].descending = 0;

/*
 *  create index by CodeBase
 */
   indexPtr = i4create (dataPtr, fileNameBuff, tagInfo);     /* CB5 function */

/* optional: fileNameBuff should be passed only if we want to create index group files
 *	   otherwise fileNameBuff should be an empty string
 */
//   indexPtr = i4create (dataPtr, (char *)0, tagInfo);      /* CB5 function */

   if (! indexPtr) {
      sprintf (repdum, CB4_CBFUNCTFAIL,
               CB4_CREINDEXPAR, dataPtr->code_base->error_code);
      CB4FSudf(callError)(ovSelf, 1, DEVWARNING,
         "CREATEINDEX", repdum, __FILE__, __LINE__);         /* Devel: fail */
      return CB4FS_DEBUG_END ("CreateIndex",FALSE_VAR);
   }

/*
 *  reset flags if not for ALL orders
 */
   if (! IS_VAR_NIL (VAR_OBJ(ovSelf,CB4OBJ_INDEXEFFECT)) &&
       ! IS_VAR_TRUE(VAR_OBJ(ovSelf,CB4OBJ_INDEXEFFECT)) )
      CB4FS_CALL_METH(setordercondition)(1, argv);           /* reset flags */

/*
 *  store data in object, since index remains open
 */
   tagPtr = d4tag_default (dataPtr);
   CB4FSudf(setIndexArr)
           (ovSelf, 0, fileNameBuff, tagName, 1, 1, indexPtr, tagPtr);

#ifdef TEST_ON
   CB4FSudf(rddTestOut)(__FILE__, __LINE__);   /* test only */
#endif

   if (autoLock < 0)
      CB4FSudf(myAutoUnlock)(ovSelf, dataPtr);

   param[0] = ovSelf;
   param[1] = SET_VAR_INT (VAR_NEW, 1);
   retVar = CB4FS_CALL_METH(setorder)(2, param);         /* cb4:setorder(1) */

   return CB4FS_DEBUG_END ("CreateIndex",retVar);
}

/*--------------------------------------------------------------------------*
 | oCb4:CreateOrder() METHOD creates order within existing index file       |
 |                           or create new index file if such not exists    |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |              argv[1] = (C) name of the existing index file               |
 |              argv[2] = (C) order name                                    |
 |              argv[3] = (C) the indexing expression                       |
 |              argv[4] = (B opt) index expression code block for indexing *|
 |                        used during the creation process only, not stored |
 |              argv[5] = (L opt) unique flag                               |
 |              argv[6] = (L opt) TRUE if no FLOCK required, default=FALSE  |
 | Returns    : TRUE on success, FALSE otherwise.                           |
 | Description: This method is works like CreateIndex() but because of the  |
 |               support of multiple-order-bags with CreateOrder(), this    |
 |               one shold be preffered.                                    |
 |              Additional flags/conditions may be set by using the         |
 |               cb4:SetOrderCondition() method.                            |
 |              * Since CodeBase 5.1 does NOT support code blocks, and      |
 |                i4create() has no parameter for the "create" CB/expr,     |
 |                the 4rd parameter is checked but ignored here.            |
 | Compatib.  : see note in SetOrderCondition() below.                      |
 |              argv[6] is not supported by CA/VO                           |
 | Multiuser  : In multiuser mode, at least FLOCK() must be set,            |
 |              or par6 must be set .T.,                                    |
 |              or SET AUTOLOCK ON must be set                              |
 | FS invocat.: retL := oCb4:CREATEORDER(cName,cOrder,cExpr,...)            |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (createorder) /* e.g.  _bb_cb4cdx_createorder_me(argc,argv) */
{
   FSvar    *ovSelf;
   FSvar    *aIndexArr;
   FSvar    *xArr, *retVar;
   FSvar    *param[6];
   DATA4    *dataPtr;                                 /* CB5 structure      */
   INDEX4   *indexPtr = 0;                            /* CB5 structure      */
   TAG4     *tagPtr;
   TAG4INFO tagInfo[2];                               /* CB5 structure      */
   TAG4INFO *tagInfoPtr = tagInfo;                    /* CB5 structure ptr  */
   FSint    ii, err = 0, add = 1, must_free = 0;
   FSint    autoLock;
   char     *cPtr, *tagName;
   extern   FSchar fileNameBuff[];

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, CB4_CREORDERPAR);            /* RTE: missing args  */
      return FALSE_VAR;
   }
   ovSelf = argv[0];
   if (argc < 4) {
      CB4FSudf(callError)(ovSelf, 1, RTMISARG,
         "CREATEORDER", CB4_CREORDERPAR,
         __FILE__, __LINE__);                         /* RTE: missing args  */
      return FALSE_VAR;
   }
   CB4FS_DEBUG_BEGIN ("CreateOrder",ovSelf);

   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("CreateOrder",FALSE_VAR);    /* dbf not open  */
/*
 *  parameter checking
 */
   if (argc > 7) argc = 7;
   switch (argc) {
   case 7:
      err = 6;
      if (! IS_VAR_LOG(argv[6]) && ! IS_VAR_NIL(argv[6])) goto err_par;
   case 6:
      err = 5;
      if (! IS_VAR_LOG(argv[5]) && ! IS_VAR_NIL(argv[5])) goto err_par;
   case 5:
      err = 4;
      if (! IS_VAR_BLK(argv[4]) && ! IS_VAR_NIL(argv[4])) goto err_par;
#ifndef CB4FS_CBLOCK_SUPPORT
      if (IS_VAR_BLK(argv[4])) {                           /* unsupp.by CB5 */
         sprintf (repdum, CB4_ARGNOTSUPP, CB4_CREORDERPAR, 4);
         CB4FSudf(callError)(ovSelf, 1, DEVWARNING,
               "CREATEORDER", repdum,
               __FILE__, __LINE__);                        /* Devel: unsupp */
      }
#endif
   case 4:
   case 3:
   case 2:
      err = 3; if (! IS_VAR_CHR(argv[3])) goto err_par;
      err = 2; if (IS_VAR_NIL(argv[2]) && IS_VAR_CHR(argv[1]))
                  argv[2] = argv[1];
               else if (! IS_VAR_CHR(argv[2])) goto err_par;
      err = 1; if (! IS_VAR_CHR(argv[1])) goto err_par;
      err = 0;
      break;
   default:
      err = 6; goto err_par;
      break;
   }

#if (CB4FScanUseIndexOrder == 1)

/* -------------------------------- only if orders (tags) are supported
 *  prepare CodeBase structure
 */
   tagName = VAR_CHR(argv[2]);
/*   c4upper(tagName); */

   tagInfo[0].name       = tagName;
   tagInfo[0].expression = VAR_CHR(argv[3]);
   tagInfo[0].filter     = (IS_VAR_EMPTY(VAR_OBJ(ovSelf,CB4OBJ_INDEXFORSTR)) ?
                            0 : VAR_CHR(VAR_OBJ(ovSelf,CB4OBJ_INDEXFORSTR)));
#ifdef CODEBASE6
   tagInfo[0].unique     = (short) (argc >= 6 ? IS_VAR_TRUE(argv[5]) : 0);
   tagInfo[0].descending = (unsigned short) IS_VAR_TRUE(VAR_OBJ(ovSelf,
														   CB4OBJ_INDEXDESC));
#else
   tagInfo[0].unique     = (argc >= 6 ? IS_VAR_TRUE(argv[5]) : 0);
   tagInfo[0].descending = IS_VAR_TRUE(VAR_OBJ(ovSelf,CB4OBJ_INDEXDESC));
#endif

   tagInfo[1].name       = 0;
   tagInfo[1].expression = 0;
   tagInfo[1].filter     = 0;
   tagInfo[1].unique     = 0;
   tagInfo[1].descending = 0;

/*
 *  check the locking
 */
   autoLock = myFlockCheck (ovSelf, (argc >= 7 ? argv[6] : NIL_VAR),
                            dataPtr, CB4_CREORDERPAR, "CREATEORDER");
   if (!autoLock)
      return CB4FS_DEBUG_END ("CreateOrder (F)",FALSE_VAR);

   aIndexArr = VAR_OBJ (ovSelf, CB4OBJ_INDEX4ARR);

/*
 *  add default extention, transform upper/lower case according to fs_set()
 *  Check if file exists. If so, check if the index file is already open.
 *  If none, create new index file.
 */
   if (! CB4FSudf (checkFile)
         (VAR_CHR(argv[1]), fileNameBuff, &cPtr,
          VAR_CHR(VAR_OBJ (ovSelf,CB4OBJ_IDX_EXT)), CB4_CREORDERPAR, 0) )

   if (cPtr && *cPtr) {                          /* file exists             */
      strcpy (fileNameBuff, cPtr);               /* path incl. file.ext     */
      if (IS_VAR_ARR (aIndexArr)) {
         for (ii=1; ii <= VAR_ISDIM(aIndexArr); ii++) {
            xArr = VAR_ARRELEM(aIndexArr,ii);
            if (IS_VAR_ARR(xArr) &&
                !strcmp (fileNameBuff,
                  VAR_CHR(VAR_ARRELEM(xArr,CB4OBJ_INDEX_FNAME)) )) {
               indexPtr = VAR_SPECIAL (VAR_ARRELEM(xArr,CB4OBJ_INDEX_I4PTR));
               break;
            }
         }
      }
   }

   if (!indexPtr) {                              /* index not open          */
      if (cPtr && *cPtr) {                       /* file exists, not open   */
		 indexPtr = i4open (dataPtr, fileNameBuff);  /* CB5 function: open  */
		 if (! indexPtr)
            goto err_cb5;
		 goto do_checktag;
	  }

      else {
 do_create:
         indexPtr = i4create (dataPtr,           /* index file not exists,  */
                         fileNameBuff, tagInfoPtr); /* CB5 function: create */
		 if (! indexPtr)
            goto err_cb5;
         add = 0;
      }
   }

   else {
 do_checktag:

/*
 *  the required tag already exists in the file, we have to build
 *  the file anew with the changed tag since CB doesn't have provisions
 *  to delete a tag from the file
 */
	  tagInfoPtr = i4tag_info (indexPtr);   /* CB5 fun: get all tags */
	  must_free  = 1;

	  for (ii = 0;
		   tagInfoPtr[ii].name
		   && strcmp(tagName, tagInfoPtr[ii].name);
		   ii++);

	  if (tagInfoPtr[ii].name) {
/*
 *  the tag already exists in the file, must create file anew
 */

		 tagInfoPtr[ii].name       = tagInfo[0].name       ;
		 tagInfoPtr[ii].expression = tagInfo[0].expression ;
		 tagInfoPtr[ii].filter     = tagInfo[0].filter     ;
		 tagInfoPtr[ii].unique     = tagInfo[0].unique     ;
		 tagInfoPtr[ii].descending = tagInfo[0].descending ;

		 i4close (indexPtr);
		 goto do_create;
	  }
   }

   if (must_free) u4free(tagInfoPtr);
   if (! indexPtr) goto err_cb5;

/*
 *  add new order into existing index file
 */
   if (add) {
      if (i4add_tag(indexPtr, tagInfo))          /* CB5 undocummented fn    */
         goto err_cb5;                           /* failed                  */
   }
/*
 *  reset flags if not for ALL orders
 */
   if ( ! IS_VAR_NIL (VAR_OBJ(ovSelf,CB4OBJ_INDEXEFFECT)) &&
        ! IS_VAR_TRUE(VAR_OBJ(ovSelf,CB4OBJ_INDEXEFFECT)) )
      CB4FS_CALL_METH(setordercondition)(1, argv);           /* reset flags */

/*
 *  store data in object, since index remains open
 */
   tagPtr = d4tag_default (dataPtr);
   CB4FSudf(setIndexArr)
           (ovSelf, 1, fileNameBuff, tagName, 1, 1, indexPtr, tagPtr);

#ifdef TEST_ON
   fprintf(stderr, "\nCreateOrder/%d: tagPtr= %x = %x",
        __LINE__, tagPtr, d4tag_selected(dataPtr));
   CB4FSudf(rddTestOut)(__FILE__, __LINE__);   /* test only */
#endif

   if (autoLock < 0)
      CB4FSudf(myAutoUnlock)(ovSelf, dataPtr);

   retVar = TRUE_VAR;
   goto set_order;

#else

/* -------------------------------- only if orders (tags) are NOT supported
 *  call cb4:CreateIndex()
 */
   param[0] = argv[0];
   param[1] = argv[1];
   param[2] = argv[3];
   param[3] = (argc > 4 ? argv[4] : 0);
   param[4] = (argc > 5 ? argv[5] : 0);
   param[5] = (argc > 6 ? argv[6] : 0);
   retVar = CB4FS_CALL_METH(createindex)(argc-1, param);
   goto set_order;

#endif

/*  -------------------------------- call cb4:setorder(1)
 */
 set_order:
   if (! IS_VAR_TRUE(retVar))
      return CB4FS_DEBUG_END ("CreateOrder (F)",FALSE_VAR);

   param[0] = argv[0];
   param[1] = SET_VAR_INT (VAR_NEW, 1);
   retVar = CB4FS_CALL_METH(setorder)(2, param);          /* set order to 1 */
#ifdef TEST_ON
   fprintf(stderr, "\nCreateOrder/%d: tagPtr= %x = %x recno=%d",
        __LINE__, tagPtr, d4tag_selected(dataPtr), (int)d4recno(dataPtr));
   CB4FSudf(rddTestOut)(__FILE__, __LINE__);   /* test only */
#endif
   return CB4FS_DEBUG_END ("CreateOrder",retVar);

/*  -------------------------------- errors
 */
 err_par:
   sprintf (repdum, CB4_ARGUMENTSTR, CB4_CREORDERPAR, err);
   CB4FSudf(callError)(ovSelf, 1, RTARGTYP,
            "CREATEORDER", repdum,
            __FILE__, __LINE__);                     /* RTE: wrong arg type */
   return CB4FS_DEBUG_END ("CreateOrder",FALSE_VAR);

 err_cb5:
   sprintf (repdum, CB4_CBFUNCTFAIL,
            CB4_CREORDERPAR, dataPtr->code_base->error_code);
   CB4FSudf(callError)(ovSelf, 2, IOOPEN,
            "CREATEORDER", repdum,
            __FILE__, __LINE__);                     /* RTE: CB5 error      */
   return CB4FS_DEBUG_END ("CreateOrder",FALSE_VAR);

}

/*--------------------------------------------------------------------------*
 | oCb4:DeleteOrder() METHOD delete order from an index file                |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |              argv[1] = (CN) order name or number to be deleted           |
 |              argv[2] = (CN) index file name                              |
 | Returns    : TRUE on success, FALSE otherwise.                           |
 | Description: This method is supported only on multiple-orders.           |
 | FS invocat.: retL := oCb4:DELETEORDER(cOrder,cFile)                      |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (deleteorder) /* e.g.  _bb_cb4cdx_deleteorder_me(argc,argv) */
{

   sprintf (repdum, CB4_NOTSUPPORT, CB4_DELORDERPAR);
   CB4FS_DEBUG_BEGIN ("DeleteOrder",argv[0]);

#if (CB4FScanUseIndexOrder != 1)
   CB4FSudf(callError)(argv[0], 1, RTILLMET,
      "DELETEORDER", repdum, __FILE__, __LINE__);
   return CB4FS_DEBUG_END ("DeleteOrder",FALSE_VAR);
#else

/* -------------------------------- only if orders (tags) ARE supported
 *  sorry, not supported by CodeBase
 */
   CB4FSudf(callError)(argv[0], 1, DEVWARNING,
      "DELETEORDER", repdum, __FILE__, __LINE__);             /* Dev.Warn.  */
   return CB4FS_DEBUG_END ("DeleteOrder",FALSE_VAR);
#endif

}

/*--------------------------------------------------------------------------*
 | oCb4:Reindex()  METHOD reindex all open index files                      |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 | Returns    : TRUE on success, FALSE otherwise.                           |
 | Multiuser  : EXCLUSIVE is required                                       |
 | FS invocat.: retL := oCb4:REINDEX()                                      |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (reindex)         /* e.g.  _bb_cb4cdx_reindex_me(argc,argv) */
{
   DATA4    *dataPtr;                                 /* CB5 structure      */
   INDEX4   *indexPtr = 0;                            /* CB5 structure      */
   FSvar    *ovSelf;
   FSvar    *aIndexArr, *xArr;
   FSint    ii, err = 0;

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, CB4_REINDEXPAR);             /* RTE: missing args  */
      return FALSE_VAR;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("Reindex",ovSelf);
   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("Reindex",FALSE_VAR);   /* database not open  */

   aIndexArr = VAR_OBJ (ovSelf, CB4OBJ_INDEX4ARR);
   if (! IS_VAR_ARR (aIndexArr))
      return CB4FS_DEBUG_END ("Reindex (F)",FALSE_VAR);   /* no index open  */
/*
 *  check if dbf is open exclusive
 */
#ifdef CODEBASE6
   if (! dataPtr->dataFile->file.lowAccessMode == OPEN4DENY_RW) {
#else
   if (! dataPtr->file.is_exclusive) {
#endif
      sprintf (repdum, CB4_EXCLREQUIR, "oRdd:Reindex()");
      CB4FSudf(callError)(ovSelf, 1, RTNOLOCK,
               "REINDEX", repdum,
               __FILE__, __LINE__);              /* RTE: replace w/o lock   */
      return CB4FS_DEBUG_END ("Reindex",FALSE_VAR);
   }

   for (ii=1; ii <= VAR_ISDIM(aIndexArr); ii++) {
      xArr = VAR_ARRELEM(aIndexArr,ii);
      if (IS_VAR_ARR(xArr) &&
          IS_VAR_TRUE (VAR_ARRELEM(xArr,CB4OBJ_INDEX_I4NEW)) ) {
         indexPtr = VAR_SPECIAL (VAR_ARRELEM(xArr,CB4OBJ_INDEX_I4PTR));
         err = i4reindex(indexPtr) || err;                   /* CB5 function*/
      }
   }
   return CB4FS_DEBUG_END ("Reindex",err ? FALSE_VAR : TRUE_VAR);
}

/*--------------------------------------------------------------------------*
 | oCb4:SetOrderCondition() METHOD sets conditions that is applied to       |
 |              index and order creation.                                   |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |              argv[1] = (C opt) for condition string                      |
 |              argv[2] = (B opt) for condition block                      *|
 |              argv[3] = (L opt) TRUE if all orders are affected=default   |
 |              argv[4] = (B opt) while condition block                    *|
 |              argv[5] = (B opt) code block evaluated for ea rec          *|
 |              argv[6] = (N opt) step count to modify eval block          *|
 |              argv[7] = (N opt) starting rec no, default=top=0           *|
 |              argv[8] = (N opt) count of proceed records, default=all=0  *|
 |              argv[9] = (N opt) recno to be processed                    *|
 |              argv[10]= (L opt) rest = TRUE, all=default=FALSE           *|
 |              argv[11]= (L opt) descending order if TRUE, default=FALSE   |
 | Returns    : TRUE on success, FALSE otherwise.                           |
 | Description: The settings for subsequent index creation are stored in    |
 |              the object.                                                 |
 | Compatib.  : -CA/VO documents - but does NOT follows it:                 |
 |               "If SetOrderCondition() has not been called, order creation|
 |               is subject to the general server scope. If neither has been|
 |               set, orders are not conditional."                          |
 |               Therefore, the global For condition is not considered here.|
 |              -Since CodeBase 5.1 does NOT support code blocks, the       |
 |               optional code blocks in the condition are not supported.   |
 |              -The * marked params are unsupported by CB5.                |
 |               It's your choice to modify the CB5 i4create.c file by      |
 |               using FlagShip's Open C API to support code blocks.        |
 |                                                                          |
 | FS invocat.: retL := oCb4:SETORDERCONDITION(...)                         |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (setordercondition)   /* e.g.  _bb_cb4cdx_setordercondition_me(argc,argv) */
{
   DATA4    *dataPtr;                                 /* CB5 structure      */
   FSvar    *ovSelf;
   FSint    err = 0;

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, CB4_SETORDCONPAR);           /* RTE: missing args  */
      return FALSE_VAR;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("SetOrderCondition",ovSelf);

   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("SetOrderCondition",FALSE_VAR);
/*
 *  parameter checking, full
 */
   if (argc > 12) argc = 12;
   switch (argc) {
   case 12:
      err = 11;
      if (! IS_VAR_LOG(argv[11]) && ! IS_VAR_NIL(argv[11])) goto err_par;
   case 11:
      err = 10;
      if (! IS_VAR_LOG(argv[10]) && ! IS_VAR_NIL(argv[10])) goto err_par;
   case 10:
      err = 9;
      if (! IS_VAR_NUM(argv[9]) && ! IS_VAR_NIL(argv[9])) goto err_par;
   case 9:
      err = 8;
      if (! IS_VAR_NUM(argv[8]) && ! IS_VAR_NIL(argv[8])) goto err_par;
   case 8:
      err = 7;
      if (! IS_VAR_NUM(argv[7]) && ! IS_VAR_NIL(argv[7])) goto err_par;
   case 7:
      err = 6;
      if (! IS_VAR_NUM(argv[6]) && ! IS_VAR_NIL(argv[6])) goto err_par;
   case 6:
      err = 5;
      if (! IS_VAR_BLK(argv[5]) && ! IS_VAR_NIL(argv[5])) goto err_par;
   case 5:
      err = 4;
      if (! IS_VAR_BLK(argv[4]) && ! IS_VAR_NIL(argv[4])) goto err_par;
   case 4:
      err = 3;
      if (! IS_VAR_LOG(argv[3]) && ! IS_VAR_NIL(argv[3])) goto err_par;
   case 3:
      err = 2;
      if (! IS_VAR_BLK(argv[2]) && ! IS_VAR_NIL(argv[2])) goto err_par;
   case 2:
      err = 1;
      if (! IS_VAR_CHR(argv[1]) && ! IS_VAR_NIL(argv[1])) goto err_par;
   case 1:
      err = 0;
      break;
   default:
      err = 12; goto err_par;
      break;
   }
/*
 *  set index conditions for this RDD (almost unsupported by CB5)
 */
   switch (argc) {
   case 12:
	  if (! IS_VAR_NIL(argv[11]))
	     SET_VAR_COPY (VAR_OBJ(ovSelf,CB4OBJ_INDEXDESC), argv[11]);
#ifdef CB4FS_CBLOCK_SUPPORT
   case 11:
	  if (! IS_VAR_NIL(argv[10]))
	     SET_VAR_COPY (VAR_OBJ(ovSelf,CB4OBJ_INDEXREST   ), argv[10]);
   case 10:
	  if (! IS_VAR_NIL(argv[ 9]))
	     SET_VAR_COPY (VAR_OBJ(ovSelf,CB4OBJ_INDEXRECORD ), argv[ 9]);
   case  9:
	  if (! IS_VAR_NIL(argv[ 8]))
	      SET_VAR_COPY (VAR_OBJ(ovSelf,CB4OBJ_INDEXRECNEXT), argv[ 8]);
   case  8:
	  if (! IS_VAR_NIL(argv[ 7]))
	      SET_VAR_COPY (VAR_OBJ(ovSelf,CB4OBJ_INDEXRECSTAR), argv[ 7]);
   case  7:
	  if (! IS_VAR_NIL(argv[ 6]))
	      SET_VAR_COPY (VAR_OBJ(ovSelf,CB4OBJ_INDEXSTEP   ), argv[ 6]);
   case  6:
	  if (! IS_VAR_NIL(argv[ 5]))
	      SET_VAR_COPY (VAR_OBJ(ovSelf,CB4OBJ_INDEXRECBLK ), argv[ 5]);
   case  5:
	  if (! IS_VAR_NIL(argv[ 4]))
	      SET_VAR_COPY (VAR_OBJ(ovSelf,CB4OBJ_INDEXWHIBLK ), argv[ 4]);
   case  4:
	  if (! IS_VAR_NIL(argv[ 3]))
	      SET_VAR_COPY (VAR_OBJ(ovSelf,CB4OBJ_INDEXEFFECT ), argv[ 3]);
   case  3:
	  if (! IS_VAR_NIL(argv[ 2]))
	      SET_VAR_COPY (VAR_OBJ(ovSelf,CB4OBJ_INDEXFORBLK ), argv[ 2]);
#else
   case 11: if (! IS_VAR_NIL(argv[10])) err = 10;         /* unsupp. by CB5 */
   case 10: if (! IS_VAR_NIL(argv[9]))  err = 9;          /* unsupp. by CB5 */
   case  9: if (! IS_VAR_NIL(argv[8]))  err = 8;          /* unsupp. by CB5 */
   case  8: if (! IS_VAR_NIL(argv[7]))  err = 7;          /* unsupp. by CB5 */
   case  7: if (! IS_VAR_NIL(argv[6]))  err = 6;          /* unsupp. by CB5 */
   case  6: if (! IS_VAR_NIL(argv[5]))  err = 5;          /* unsupp. by CB5 */
   case  5: if (! IS_VAR_NIL(argv[4]))  err = 4;          /* unsupp. by CB5 */
   case  4:
	  if (! IS_VAR_NIL(argv[ 3]))
	     SET_VAR_COPY (VAR_OBJ(ovSelf,CB4OBJ_INDEXEFFECT), argv[3]);
   case  3: if (! IS_VAR_NIL(argv[2]))  err = 2;          /* unsupp. by CB5 */
#endif
   case  2:
	  if (! IS_VAR_NIL(argv[ 1]))
	     SET_VAR_COPY (VAR_OBJ(ovSelf,CB4OBJ_INDEXFORSTR), argv[1]);
      break;
   case  1:
      err = 0;
      SET_VAR_LOG (VAR_OBJ(ovSelf,CB4OBJ_INDEXDESC), 0);
      SET_VAR_LOG (VAR_OBJ(ovSelf,CB4OBJ_INDEXEFFECT), 0);
      SET_VAR_CHR (VAR_OBJ(ovSelf,CB4OBJ_INDEXFORSTR), "");
      break;
   }
   if (err) {
      sprintf (repdum, CB4_ARGNOTSUPP, CB4_SETORDCONPAR, err);
      CB4FSudf(callError)(ovSelf, 1, DEVWARNING,
               "SETORDERCONDITION", repdum,
               __FILE__, __LINE__);                  /* Dev.Warn: unsupport */
   }
   return CB4FS_DEBUG_END ("SetOrderCondition (T)",TRUE_VAR);

 err_par:
   sprintf (repdum, CB4_ARGUMENTSTR, CB4_SETORDCONPAR, err);
   CB4FSudf(callError)(ovSelf, 1, RTARGTYP,
               "SETORDERCONDITION", repdum,
               __FILE__, __LINE__);                  /* RTE: wrong arg type */
   return CB4FS_DEBUG_END ("SetOrderCondition",FALSE_VAR);
}


/*--------------------------------------------------------------------------*
 | internal function: set index list                                        |
 |                                                                          |
 | in:  ovSelf      object                                                  |
 |      add         1 = add new, 0 = replace first                          |
 |      fileName    full file name, already converted                       |
 |      tagName     tag name, if "" determine tag from file name            |
 |      tagNum      ordinal numb of tag in index file, 0=self               |
 |      newIndex    1 if index memory is stored, 0 for subseq. same index   |
 |      indexPtr    ptr to INDEX4 storage                                   |
 |      tagPtr      ptr to TAG4   storage or NULL                           |
 | ret: len of the array (1..n)                                             |
 *--------------------------------------------------------------------------*/
FSint CB4FSudf(setIndexArr)(ovSelf, add, fileName, tagName, tagNum, newIndex, indexPtr, tagPtr)
FSvar  *ovSelf;
FSint   add;
char   *fileName, *tagName;
FSint   tagNum;
FSint   newIndex;
INDEX4 *indexPtr;
TAG4   *tagPtr;
{
   FSvar   *aIndexArr;
   FSvar   *aSubArr;
   FSvar   *param[2];
   FSchar  newTagName[11];

   aIndexArr = VAR_OBJ (ovSelf, CB4OBJ_INDEX4ARR);

   if (!tagNum && add) {
      if (IS_VAR_ARR(aIndexArr))
         tagNum = VAR_ISDIM(aIndexArr) +1;
      else
         tagNum = 1;
   }
   newTagName[0] = '\0';
   if (tagName && *tagName)
      strncpy (newTagName, tagName, 10);
   else {
#ifdef CODEBASE6
      if (! tagPtr->index->accessName )
#else
      if (! tagPtr->alias )
#endif
         fgsExtractName (fileName, newTagName, 11);
      else
         strcpy (newTagName, t4alias(tagPtr));
   }
   fgs2Upper (newTagName);                                     /* uppercase */

   VAR_NEW_ARRAY (aSubArr, CB4OBJ_INDEX_T4PTR);

   SET_VAR_CHR (VAR_ARRELEM(aSubArr,CB4OBJ_INDEX_FNAME), fileName);
   SET_VAR_CHR (VAR_ARRELEM(aSubArr,CB4OBJ_INDEX_TNAME), newTagName);
   SET_VAR_INT (VAR_ARRELEM(aSubArr,CB4OBJ_INDEX_TNUMB), tagNum);
   SET_VAR_LOG (VAR_ARRELEM(aSubArr,CB4OBJ_INDEX_I4NEW), newIndex);
   SET_VAR_SPECIAL (VAR_ARRELEM(aSubArr,CB4OBJ_INDEX_I4PTR), indexPtr);
   SET_VAR_SPECIAL (VAR_ARRELEM(aSubArr,CB4OBJ_INDEX_T4PTR), tagPtr);

   if (add && IS_VAR_ARR(aIndexArr)) {
      param[0] = aIndexArr;
      param[1] = aSubArr;
      fsFnNAME(aadd)(2, param);                             /* call AADD()  */
   } else {
      VAR_NEW_ARRAY (aIndexArr, 1);
      SET_VAR_COPY  (VAR_ARRELEM(aIndexArr, 1), aSubArr);
      SET_VAR_COPY  (VAR_OBJ (ovSelf, CB4OBJ_INDEX4ARR), aIndexArr);
   }

   return VAR_ISDIM(aIndexArr);
}

/*--------------------------------------------------------------------------*
 | internal function for flock()                                            |
 | returns 0 = error, 1 = ok, -1 = ok, auto Flocked                         |
 *--------------------------------------------------------------------------*/
static FSint myFlockCheck (ovSelf, noFlock, dataPtr, fnCallStr, module)
FSvar  *ovSelf, *noFlock;
DATA4  *dataPtr;                                 /* CB5 structure           */
char   *fnCallStr, *module;
{

   FSvar  *passArgv[2];
   FSint  ii;

/*
 *  is file open for read only ?
 */
#ifdef CODEBASE6
   if (dataPtr->dataFile->file.isReadOnly) {     /* CB6 structure           */
#else
   if (dataPtr->file.is_read_only) {             /* CB5 structure           */
#endif
      sprintf (repdum, CB4_READONLY, fnCallStr);
      CB4FSudf(callError)(ovSelf, 2, IOWRIT,
               module, repdum,
               __FILE__, __LINE__);              /* RTE: replace on r/o     */
      return 0;
   }
/*
 *  is file open exclusivelly or FLOCKed ?
 */
   if (IS_VAR_TRUE(noFlock))
      return 1;
#ifdef CODEBASE6
# if S4VERSION == 6502
  if (dataPtr->dataFile->file.lowAccessMode == OPEN4DENY_RW  ||
	   dataPtr->dataFile->fileServerWriteLock                 ||
      dataPtr->dataFile->fileClientWriteLock )
      return 1;
# else
   if (dataPtr->dataFile->file.lowAccessMode == OPEN4DENY_RW  ||
	   dataPtr->dataFile->fileClientLock                       ||
	   dataPtr->dataFile->fileServerLock                        )
      return 1;
#endif
#else
   if (dataPtr->file.is_exclusive || dataPtr->file_lock)
      return 1;
#endif

/*
 *  perform AutoFlock if not disabled
 */
   ii = CB4FSudf(myAutoRFlock)(ovSelf, 2, dataPtr, 0);
   if (ii != 0)
      return ii;

   sprintf (repdum, CB4_FLOCKREQUIR, fnCallStr);
   CB4FSudf(callError)(ovSelf, 1, RTNOLOCK,
               module, repdum,
               __FILE__, __LINE__);              /* RTE: replace w/o lock   */
   return 0;
}


/**** eof rddcb4g.c ****/

