/****************************************************************************
 * rddcb4i.c     RDD API for the CB4FOX, CB4CDX, CB4NDX, CB4MDX and CB4NTX  *
 *               by using CodeBase 5.1 multiplatform version (not included  *
 *               in the FlagShip package, contact Sequiter Inc.)            *
 *                                                                          *
 *               Note: all CodeBase function & defines are marked by CB5    *
 *                                                                          *
 * (c) 1995..2007 multisoft Datentechnik GmbH, Munich. All rights reserved  *
 * Release $Date: 2007/01/25 15:25:43 $                                     *
 *--------------------------------------------------------------------------*
 * This file contains parts of the full RDD driver:                         *
 *    oCb4:info()            METHOD retrieve info about the server          *
 *    oCb4:fieldinfo()       METHOD retrieve info about the specified field *
 *    oCb4:recordinfo()      METHOD returns the req. record status info     *
 *                                                                          *
 * For additional methods, see rddcb4a.c to rddcb4?.c                       *
 ****************************************************************************/

#include "rddcb4fs.h"                            /* RDD API for CodeBase    */

extern FSvar *CB4FS_CALL_ASSIGN(alias)();
extern FSvar *CB4FS_CALL_ASSIGN(filter)();

extern FSvar *CB4FS_CALL_ACCESS(alias)();
extern FSvar *CB4FS_CALL_ACCESS(bof)();
extern FSvar *CB4FS_CALL_ACCESS(eof)();
extern FSvar *CB4FS_CALL_ACCESS(fcount)();
extern FSvar *CB4FS_CALL_ACCESS(filter)();
extern FSvar *CB4FS_CALL_ACCESS(found)();
extern FSvar *CB4FS_CALL_ACCESS(header)();
extern FSvar *CB4FS_CALL_ACCESS(lupdate)();
extern FSvar *CB4FS_CALL_ACCESS(rlocklist)();
extern FSvar *CB4FS_CALL_ACCESS(recsize)();
extern FSvar *CB4FS_CALL_ACCESS(shared)();

#define CB4_WRONGREQ     "unknown request"
#define CB4_NOTSUPPRTED  "not supported by this RDD"
#define CB4_ACCESSONLY   "access only, unsupported assign"
#define CB4_INFOPARAM    "cb4:Info(nWhat, [uNewValue])"
#define CB4_INFOSTRING   "cb4:Info(%d) %s"
#define CB4_INFOASSTYPE  "wrong assing type, not set"
#define CB4_FIELDINFPAR  "cb4:FieldInfo(nRequest, nFldNo|cFldName)"
#define CB4_RECINFPAR    "cb4:RecordInfo(nRequest, [nRecNo])"
#define CB4_PARAMSTRING  "%s argument %d"

/*--------------------------------------------------------------------------*
 | oCb4:Info() METHOD retrieve info about the server                        |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |              argv[1] = requested info according to 'rddsys.fh'           |
 |                 DBI_ALIAS           == alias ACCESS/ASSIGN     <--> char |
 |                 DBI_BLOB_HANDLE     not supported               --> NIL  |
 |                 DBI_BLOB_INTEGRITY  not supported               --> NIL  |
 |                 DBI_BLOB_RECOVER    not supported               --> NIL  |
 |                 DBI_BOF             == bof()                    --> log  |
 |                 DBI_CANPUTREC       replace supported ?         --> log  |
 |                 DBI_CHILDCOUNT      no.of open relation         --> num  |
 |                 DBI_DBFILTER        == filter ACCESS/ASSIGN    <--> char |
 |                 DBI_DB_VERSION      release of the RDD (4.40)   --> num  |
 |                 DBI_EOF             == eof()                    --> log  |
 |                 DBI_FCOUNT          == fcount ACCESS            --> num  |
 |                 DBI_FILEHANDLE      not supported               --> NIL  |
 |                 DBI_FOUND           == found ACCESS/ASSIGN     <--> log  |
 |                 DBI_FULLPATH        dbf name incl.full path     --> char |
 |                 DBI_GETDELIMITER    default delimiter           --> num  |
 |                 DBI_GETHEADERSIZE   == header()                 --> num  |
 |                 DBI_GETLOCKARRAY    == rlocklist ACCESS         --> arr  |
 |                 DBI_GETRECSIZE      == recsize ACCESS           --> num  |
 |                 DBI_GETSCOPE        locate condition           <--> block|
 |                 DBI_ISANSI          .dbf = IBM-PC8 chars?       --> log  |
 |                 DBI_ISDBF           .dbf format supported?      --> log  |
 |                 DBI_ISFLOCK         is FLOCK() active ?         --> log  |
 |                 DBI_LASTUPDATE      == lupdate ACCESS           --> date |
 |                 DBI_LOCKCOUNT       number or RLOCKed records   --> num  |
 |                 DBI_LOCKOFFSET      not supported               --> NIL  |
 |                 DBI_MEMOBLOCKSIZE   block size of memo file     --> num  |
 |                 DBI_MEMOEXT         default extension = ".dbt" <--> char |
 |                 DBI_MEMOHANDLE      handle of memo file         --> int  |
 |                 DBI_RDD_VERSION     rdd version (4.40)          --> num  |
 |                 DBI_SETDELIMITER    default delimiter          <--  num  |
 |                 DBI_SHARED          == shared ACCESS            --> log  |
 |                 DBI_TABLEEXT        default extension = ".dbf" <--> char |
 |                 DBI_VALIDBUFFER     buffer valid ?              --> log  |
 |              argv[2] = (opt) new value                                   |
 | Returns    : the requested info or NIL on error                          |
 |              If argv[2] is not specified, DBServer:RDDInfo() returns the |
 |              current setting.  If argv[2] is specified, the previous     |
 |              setting is returned.                                        |
 | FS invocat.: ret := oCb4:INFO(DBI_xxx, [uNewValue])                      |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (info)              /* e.g.  _bb_cb4cdx_info_me(argc, argv) */
{
   FSvar  *ovSelf;
   DATA4  *dataPtr;                                   /* CB5 structure      */
   FIELD4 *fieldPtr;                                  /* CB5 structure      */
   FSvar  *uvRetVar = VAR_NEW;                        /* = NIL              */
   FSint  what;

   if (argc < 2 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTARGCNT, CB4_INFOPARAM);              /* RTE param          */
      return NIL_VAR;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("Info",ovSelf);
   dataPtr = CB4FS_GETDATA4PTR;

   switch (VAR_ISTYPE(argv[1])) {
   case 'I':
   case 'N':
   case 'F':
      what = VAR_INTNUM (argv[1]);
      break;
   default:
      rt_error (RTARGTYP, CB4_INFOPARAM);             /* RTE: illegal type  */
      return CB4FS_DEBUG_END ("Info",NIL_VAR);
   }

/*
 *  Access
 */
   switch (what) {
   case DBI_ALIAS:
      if (dataPtr)
         SET_VAR_COPY (uvRetVar, CB4FS_CALL_ACCESS(alias)(ovSelf));
      else
         return CB4FS_DEBUG_END ("Info",NULLSTR_VAR);
      break;
   case DBI_BOF:
      if (dataPtr)
         SET_VAR_COPY (uvRetVar, CB4FS_CALL_ACCESS(bof)(ovSelf));
      else
         return CB4FS_DEBUG_END ("Info",FALSE_VAR);
      break;
   case DBI_CANPUTREC:
      if (dataPtr)
         SET_VAR_LOG  (uvRetVar, 1);
      else
         return CB4FS_DEBUG_END ("Info",FALSE_VAR);
      break;
   case DBI_CHILDCOUNT:
      if (IS_VAR_NUM (VAR_OBJ(ovSelf,CB4OBJ_RELCOUNT)))
         SET_VAR_COPY (uvRetVar, VAR_OBJ(ovSelf,CB4OBJ_RELCOUNT));
      else
         return CB4FS_DEBUG_END ("Info",ZERO_VAR);
      break;
   case DBI_DBFILTER:
      if (dataPtr)
         SET_VAR_COPY (uvRetVar, CB4FS_CALL_ACCESS(filter)(ovSelf));
      else
         return CB4FS_DEBUG_END ("Info",FALSE_VAR);
      break;
   case DBI_DB_VERSION:
      SET_VAR_INT (uvRetVar, S4VERSION );     /* CB5 in d4all.h     */
      break;
   case DBI_EOF:
      if (dataPtr)
         SET_VAR_COPY (uvRetVar, CB4FS_CALL_ACCESS(eof)(ovSelf));
      else
         return CB4FS_DEBUG_END ("Info",FALSE_VAR);
      break;
   case DBI_FCOUNT:
      if (dataPtr)
         SET_VAR_COPY (uvRetVar, CB4FS_CALL_ACCESS(fcount)(ovSelf));
      else
         return CB4FS_DEBUG_END ("Info",ZERO_VAR);
      break;
   case DBI_FOUND:
      if (dataPtr)
         SET_VAR_COPY (uvRetVar, CB4FS_CALL_ACCESS(found)(ovSelf));
      else
         return CB4FS_DEBUG_END ("Info",FALSE_VAR);
      break;
   case DBI_FULLPATH:
      if (dataPtr)
         SET_VAR_COPY (uvRetVar, VAR_OBJ(ovSelf,CB4OBJ_FILEFULLNAME));
      else
         return CB4FS_DEBUG_END ("Info",NULLSTR_VAR);
      break;
   case DBI_GETDELIMITER:
   case DBI_SETDELIMITER:
      SET_VAR_COPY (uvRetVar, VAR_OBJ(ovSelf,CB4OBJ_DELIMITER));
      break;
   case DBI_GETHEADERSIZE:
      if (dataPtr)
         SET_VAR_COPY (uvRetVar, CB4FS_CALL_ACCESS(header)(ovSelf));
      else
         return CB4FS_DEBUG_END ("Info",ZERO_VAR);
      break;
   case DBI_GETLOCKARRAY:
      if (dataPtr)
         SET_VAR_COPY (uvRetVar, CB4FS_CALL_ACCESS(rlocklist)(ovSelf));
      else
         return CB4FS_DEBUG_END ("Info",ZERO_VAR);
      break;
   case DBI_GETRECSIZE:
      if (dataPtr)
         SET_VAR_COPY (uvRetVar, CB4FS_CALL_ACCESS(recsize)(ovSelf));
      else
         return CB4FS_DEBUG_END ("Info",ZERO_VAR);
      break;
   case DBI_GETSCOPE:
      SET_VAR_COPY (uvRetVar, VAR_OBJ(ovSelf,CB4OBJ_LOCATEBLOCK));
      break;
   case DBI_ISANSI:
      SET_VAR_LOG  (uvRetVar, 0);
      break;
   case DBI_ISDBF:
      SET_VAR_LOG  (uvRetVar, 1);
      break;
   case DBI_ISFLOCK:
      if (dataPtr)
#ifdef CODEBASE6
# if S4VERSION == 6502
         SET_VAR_LOG  (uvRetVar,
					     dataPtr->dataFile->fileServerWriteLock
					   | dataPtr->dataFile->fileClientWriteLock);/* CB6.54 structure*/
# else
         SET_VAR_LOG  (uvRetVar,
					     dataPtr->dataFile->fileServerLock
					   | dataPtr->dataFile->fileClientLock);/* CB6 structure*/
# endif
#else
         SET_VAR_LOG  (uvRetVar, dataPtr->file_lock); /* CB5 structure      */
#endif
      else
         return CB4FS_DEBUG_END ("Info",FALSE_VAR);
      break;
   case DBI_LASTUPDATE:
      if (dataPtr)
         SET_VAR_COPY (uvRetVar, CB4FS_CALL_ACCESS(lupdate)(ovSelf));
      else
         return CB4FS_DEBUG_END ("Info",NULLDATE_VAR);
      break;
   case DBI_LOCKCOUNT:
      if (dataPtr)
         SET_VAR_INT (uvRetVar,
					  VAR_ISDIM(CB4FS_CALL_ACCESS(rlocklist)(ovSelf)));
      else
         return CB4FS_DEBUG_END ("Info",ZERO_VAR);
      break;
   case DBI_MEMOBLOCKSIZE:
      if (dataPtr)
#ifdef CODEBASE6
         SET_VAR_INT (uvRetVar, dataPtr->dataFile->blockSize); /* CB6 */
#else
         SET_VAR_INT (uvRetVar, dataPtr->block_size);          /* CB5 */
#endif
      else
         return CB4FS_DEBUG_END ("Info",ZERO_VAR);
      break;
   case DBI_MEMOEXT:
      SET_VAR_COPY (uvRetVar, VAR_OBJ (ovSelf,CB4OBJ_DBT_EXT));
      break;
   case DBI_RDD_VERSION:
      SET_VAR_NUM (uvRetVar, CB4FSrddVersion);        /* in rddcb4fs.h      */
      break;
   case DBI_SHARED:
      if (dataPtr)
         SET_VAR_COPY (uvRetVar, CB4FS_CALL_ACCESS(shared)(ovSelf));
      else
         return CB4FS_DEBUG_END ("Info",FALSE_VAR);
      break;
   case DBI_TABLEEXT:
      SET_VAR_COPY (uvRetVar, VAR_OBJ (ovSelf,CB4OBJ_DBF_EXT));
      break;
   case DBI_VALIDBUFFER:
      if (dataPtr)
         SET_VAR_LOG  (uvRetVar, 1);
      else
         return CB4FS_DEBUG_END ("Info",FALSE_VAR);
      break;

   case DBI_FILEHANDLE:
      if (dataPtr) {
#ifdef CODEBASE6
	 #if defined(WIN32) || defined(FS_WIN32)
	      if(dataPtr->dataFile->file.hand == INVALID_HANDLE_VALUE)
		      SET_VAR_INT  (uvRetVar, -1);
	      else
		      SET_VAR_INT (uvRetVar, 10000000);
	 #else
         SET_VAR_INT  (uvRetVar, dataPtr->dataFile->file.hand);
    #endif
#else
         SET_VAR_INT  (uvRetVar, dataPtr->file.hand);
#endif
      } else
         return CB4FS_DEBUG_END ("Info", IZERO_VAR);
      break;

/* case DBI_BLOB_HANDLE:             equivalent #define with DBI_MEMOHANDLE */
   case DBI_BLOB_INTEGRITY:
   case DBI_BLOB_RECOVER:
   case DBI_LOCKOFFSET:
   case DBI_MEMOHANDLE:
      sprintf  (repdum, CB4_INFOSTRING, what, CB4_NOTSUPPRTED);
      rt_error (DEVWARNING, repdum);                  /* Devel.Warning      */
      return CB4FS_DEBUG_END ("Info",NIL_VAR);
//    break;
   default:
      sprintf  (repdum, CB4_INFOSTRING, what, CB4_WRONGREQ);
      rt_error (RTARGVAL, repdum);                    /* RTE: illegal value */
      return CB4FS_DEBUG_END ("Info",NIL_VAR);
   }

/*
 *  Assign
 */
   if (argc > 2) {
      switch (what) {
      case DBI_ALIAS:                                 /* set alias          */
         CB4FS_CALL_ASSIGN(alias)(ovSelf, argv[2]);
         break;
      case DBI_DBFILTER:                              /* set filter         */
         CB4FS_CALL_ASSIGN(filter)(ovSelf, argv[2]);
         break;
      case DBI_FOUND:
         if (IS_VAR_LOG (argv[2]))
            SET_VAR_COPY (VAR_OBJ(ovSelf,CB4OBJ_FOUNDFLAG), argv[2]);
         else
            goto error_assign;
         break;
      case DBI_GETSCOPE:
         switch (VAR_ISTYPE(argv[2])) {
         case 'B':
         case 'b':
            SET_VAR_COPY (VAR_OBJ(ovSelf,CB4OBJ_LOCATEBLOCK), argv[2]);
            break;
         case 'C':
         case 'E':
         case 'M':
            SET_VAR_COPY (VAR_OBJ(ovSelf,CB4OBJ_LOCATEBLOCK),
                          COMPILE_BLOCK(argv[2]));
            break;
         case 'U':
            SET_VAR_NIL (VAR_OBJ(ovSelf,CB4OBJ_LOCATEBLOCK));
            break;
         default:
            goto error_assign;
//          break;
         }
         break;
      case DBI_SETDELIMITER:                          /* set delimiter      */
         if (IS_VAR_NUM(argv[2]))
            SET_VAR_COPY (VAR_OBJ(ovSelf,CB4OBJ_DELIMITER), argv[2]);
         else if (IS_VAR_NIL(argv[2]))
            SET_VAR_NIL (VAR_OBJ(ovSelf,CB4OBJ_DELIMITER));
         else
            goto error_assign;
         break;
      case DBI_MEMOEXT:
         if (IS_VAR_CHR(argv[2]))
            SET_VAR_COPY (VAR_OBJ (ovSelf,CB4OBJ_DBT_EXT), argv[2]);
      case DBI_TABLEEXT:
         if (IS_VAR_CHR(argv[2]))
            SET_VAR_COPY (VAR_OBJ (ovSelf,CB4OBJ_DBF_EXT), argv[2]);
      default:
         sprintf  (repdum, CB4_INFOSTRING,
                      what, CB4_ACCESSONLY);          /* unsupp ASSIGN      */
         rt_error (DEVWARNING, repdum);               /* Devel.Warning      */
      }
   }

   return CB4FS_DEBUG_END ("Info",uvRetVar);

 error_assign:
   sprintf  (repdum, CB4_INFOSTRING,
                what, CB4_INFOASSTYPE);               /* wrong assign       */
   rt_error (DEVWARNING, repdum);                     /* RTE: devel.warning */
   return CB4FS_DEBUG_END ("Info",uvRetVar);

}

/*--------------------------------------------------------------------------*
 | oCb4:FieldInfo() METHOD returns the info about specified field           |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |              argv[1] = requested info according to 'rddsys.fh'           |
 |                        DBS_DEC    = number of decimals    --> num        |
 |                        DBS_LEN    = field size            --> num        |
 |                        DBS_NAME   = field mame            --> char       |
 |                        DBS_TYPE   = field type            --> char       |
 |              argv[2] = nFieldNo | cFieldName                             |
 | Returns    : the requested info or NIL on error                          |
 | FS invocat.: ret := oCb4:FIELDINFO(DBS_xxx, nFld|cFld)                   |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (fieldinfo)    /* e.g.  _bb_cb4cdx_fieldinfo_me(argc, argv) */
{
   FSvar  *ovSelf;
   DATA4  *dataPtr;                                   /* CB5 structure      */
   FIELD4 *fieldPtr;                                  /* CB5 structure      */
   FSvar  *uvRetVar = VAR_NEW;
   FSint  info, fieldPos;
   char   str[2];

   if (argc < 3 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTARGCNT, CB4_FIELDINFPAR);            /* RTE param          */
      return NIL_VAR;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("FieldInfo",ovSelf);
   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("FieldInfo",NIL_VAR);

   switch (VAR_ISTYPE (argv[1])) {
   case 'I':
   case 'N':
   case 'F':
      info = VAR_INTNUM(argv[1]);
      break;
   default:
      sprintf  (repdum, CB4_PARAMSTRING, CB4_FIELDINFPAR, 1);
      rt_error (RTARGTYP, repdum);                    /* RTE: illegal type  */
      return CB4FS_DEBUG_END ("FieldInfo",NIL_VAR);
   }

   switch (VAR_ISTYPE(argv[2])) {
   case 'I':
   case 'N':
   case 'F':
      fieldPos = VAR_INTNUM(argv[2]);
      break;
   case 'E':
   case 'C':
      fieldPos = d4field_number (dataPtr, VAR_CHR(argv[2]));         /* CB5 */
      break;
   default:
      sprintf  (repdum, CB4_PARAMSTRING, CB4_FIELDINFPAR, 2);
      rt_error (RTARGTYP, repdum);                    /* RTE: illegal type  */
      return CB4FS_DEBUG_END ("FieldInfo",NIL_VAR);
   }
   if (fieldPos <= 0)
      return CB4FS_DEBUG_END ("FieldInfo",NIL_VAR);   /* field not found    */
   fieldPtr = d4field_j (dataPtr, fieldPos);          /* CB5 field struct   */
   if (! fieldPtr)                                    /* field name found?  */
      return CB4FS_DEBUG_END ("FieldInfo",NIL_VAR);   /* no                 */

   switch (info) {
   case DBS_NAME:
      SET_VAR_CHR (uvRetVar, (char *)f4name (fieldPtr));                /* CB5 fn   */
      break;
   case DBS_TYPE:
      str[0] = (char) f4type (fieldPtr);                        /* CB5 fn   */
      str[1] = '\0';
      SET_VAR_CHR (uvRetVar, str);
      break;
   case DBS_LEN:
      SET_VAR_INT (uvRetVar, f4len (fieldPtr));                 /* CB5 fn   */
      break;
   case DBS_DEC:
      SET_VAR_INT (uvRetVar, f4decimals (fieldPtr));            /* CB5 fn   */
      break;
   default:
      sprintf  (repdum, CB4_PARAMSTRING, CB4_FIELDINFPAR, 1);
      rt_error (RTARGVAL, repdum);                    /* RTE: illegal value */
      return CB4FS_DEBUG_END ("FieldInfo",NIL_VAR);
   }
   return CB4FS_DEBUG_END ("FieldInfo",uvRetVar);
}

/*--------------------------------------------------------------------------*
 | oCb4:RecordInfo()   METHOD returns the info about specified record       |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |              argv[1] = requested info according to 'rddsys.fh'           |
 |                        DBRI_DELETED is record deleted?    --> log        |
 |                        DBRI_LOCKED  is record locked ?    --> log        |
 |                        DBRI_RECSIZE the record length     --> num        |
 |                        DBRI_RECNO   currect record no.    --> num        |
 |              argv[2] = (opt) nRecNo                                      |
 | Returns    : the requested info or NIL on error                          |
 | Description: If nRecNo==0 or omitted, status of the curr.record returned |
 | FS invocat.: ret := oCb4:RECINFO(DBS_xxx, [nRecNo])                      |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (recordinfo)  /* e.g.  _bb_cb4cdx_recordinfo_me(argc, argv) */
{
   FSvar  *ovSelf;
   DATA4  *dataPtr;                                   /* CB5 structure      */
   FSvar  *uvRetVar = VAR_NEW;
   FSint  info;
   long   recNoCurr, recNoReq = 0;

   if (argc < 2 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTARGCNT, CB4_RECINFPAR);              /* RTE param          */
      return NIL_VAR;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("RecordInfo",ovSelf);
   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("RecordInfo",NIL_VAR);

   switch (VAR_ISTYPE (argv[1])) {
   case 'I':
   case 'N':
   case 'F':
      info = VAR_INTNUM(argv[1]);
      break;
   default:
      sprintf  (repdum, CB4_PARAMSTRING, CB4_RECINFPAR, 1);
      rt_error (RTARGTYP, repdum);                    /* RTE: illegal type  */
      return CB4FS_DEBUG_END ("RecordInfo",NIL_VAR);
   }

   switch (VAR_ISTYPE (argv[2])) {
   case 'I':
   case 'N':
   case 'F':
      recNoReq = (long) VAR_INTNUM(argv[2]);
   case 'U':
      break;
   default:
      sprintf  (repdum, CB4_PARAMSTRING, CB4_RECINFPAR, 2);
      rt_error (RTARGTYP, repdum);                    /* RTE: illegal type  */
      return CB4FS_DEBUG_END ("RecordInfo",NIL_VAR);
   }

   recNoCurr = d4recno (dataPtr);                     /* CB5 function       */
   if (recNoReq > 0L) {
      if (d4go (dataPtr, recNoReq)) {                 /* CB5 function       */
         d4go (dataPtr, recNoCurr);
         return CB4FS_DEBUG_END ("RecordInfo",NIL_VAR);  /* error           */
      }
   } else
     recNoReq = recNoCurr;

   switch (info) {
   case DBRI_DELETED:
      SET_VAR_LOG (uvRetVar, d4deleted(dataPtr));          /* CB5 function  */
      break;
   case DBRI_LOCKED:
#ifdef CODEBASE6
      SET_VAR_LOG (uvRetVar,
				   ((dataPtr->dataFile->file.lowAccessMode == OPEN4DENY_RW) ||
					(D4LOCK_TEST(dataPtr,recNoReq))                         ||
					(d4lock_test_append(dataPtr)) ));      /* CB6 stru/fn   */
#else
      SET_VAR_LOG (uvRetVar,
				   ((dataPtr->file.is_exclusive)            ||
					(D4LOCK_TEST(dataPtr,recNoReq))         ||
					(d4lock_test_append(dataPtr)) ));      /* CB5 stru/fn   */
#endif
      break;
   case DBRI_RECSIZE:
      SET_VAR_INT (uvRetVar, d4record_width(dataPtr));     /* CB5 stru/fn   */
      break;
   case DBRI_RECNO:
      SET_VAR_INT (uvRetVar, recNoReq);
      break;
   default:
      sprintf  (repdum, CB4_PARAMSTRING, CB4_RECINFPAR, 1);
      rt_error (RTARGVAL, repdum);                    /* RTE: illegal value */
      SET_VAR_NIL (uvRetVar);
   }
   if (recNoCurr != recNoReq)
      d4go (dataPtr, recNoCurr);
   return CB4FS_DEBUG_END ("RecordInfo",uvRetVar);
}

/**** eof rddcb4i.c ****/

