/****************************************************************************
 * rddcb4j.c     RDD API for the CB4FOX, CB4CDX, CB4NDX, CB4MDX and CB4NTX  *
 *               by using CodeBase 5.1 multiplatform version (not included  *
 *               in the FlagShip package, contact Sequiter Inc.)            *
 *                                                                          *
 *               Note: all CodeBase function & defines are marked by CB5    *
 *                                                                          *
 * (c) 1995..2007 multisoft Datentechnik GmbH, Munich. All rights reserved  *
 * Release $Date: 2007/03/05 19:09:30 $                                     *
 *--------------------------------------------------------------------------*
 * This file contains parts of the full RDD driver:                         *
 *                                                                          *
 *    oCb4:appenddb()        METHOD APPEND FROM ...                         *
 *    oCb4:appenddelimited() METHOD APPEND FROM ... DELIMITED               *
 *    oCb4:appendsdf()       METHOD APPEND FROM ... SDF                     *
 *    oCb4:copydb()          METHOD COPY TO ...                             *
 *    oCb4:copydelimited()   METHOD COPY TO ... DELIMITED                   *
 *    oCb4:copysdf()         METHOD COPY TO ... SDF                         *
 *    oCb4:copystructure()   METHOD COPY STRUCTURE ... TO                   *
 *    oCb4:pack()            METHOD PACK                                    *
 *    oCb4:zap()             METHOD ZAP                                     *
 *                                                                          *
 * For additional methods, see rddcb4a.c to rddcb4?.c                       *
 ****************************************************************************/

#include "rddcb4fs.h"                            /* RDD API for CodeBase    */

extern FSint  CB4FSudf(callError)();             /* in rddcb4b.c            */
extern FSint  CB4FSudf(decodeScope)();           /* in rddcb4e.c            */
extern FSint  CB4FSudf(decodeCond)();            /* in rddcb4e.c            */
extern FSvar *CB4FS_CALL_METH(gotop)();          /* cb4:gotop()  rddcb4c.c  */
extern FSvar *fsFnNAME(__dbapp)();               /* == APPEND FROM...       */
extern FSvar *fsFnNAME(__dbappsdf)();            /* == APPEND FROM...SDF    */
extern FSvar *fsFnNAME(__dbappdel)();            /* == APPEND FROM...DELIM  */
extern FSvar *fsFnNAME(__dbcopy)();              /* == COPY TO ...          */
extern FSvar *fsFnNAME(__dbcopysd)();            /* == COPY TO ... SDF      */
extern FSvar *fsFnNAME(__dbcopyde)();            /* == COPY TO ... DELIM    */
extern FSvar *fsFnNAME(__dbcopyst)();            /* == COPY STRU TO ...     */
extern FSvar *fsFnNAME(__dbzap)();               /* == ZAP                  */
extern FSvar *fsFnNAME(__dbpack)();              /* == PACK                 */

static int checkAppPar  ();                      /* forward declaration     */
static int checkAppDeliPar ();                   /* forward declaration     */

#define CB4_WRONGARG     "%s;wrong type or missing argument # %d"
#define CB4_APPENDDBPAR  "cb4:AppendDB(cFile,[acFields,cbFor,cbWhile,nlScop,cVia])"
#define CB4_APPENDSDFPAR "cb4:AppendSDF(cFile,[acFields,cbFor,cbWhile,nlScop])"
#define CB4_APPENDDELPAR "cb4:AppendDelimited(cFile,[cDelim,acFields,cbFor,cbWhile,nlScop])"
#define CB4_COPYDBPAR    "cb4:CopyDB(cFile,[acFields,cbFor,cbWhile,nlScop,cVia])"
#define CB4_COPYSDFPAR   "cb4:CopySDF(cFile,[acFields,cbFor,cbWhile,nlScop])"
#define CB4_COPYDELPAR   "cb4:CopyDelimited(cFile,[cDelim,acFields,cbFor,cbWhile,nlScop])"
#define CB4_COPYSTRPAR   "cb4:CopyStructure(cFile,acFields,[cVia])"
#define CB4_PACKPAR      "cb4:Pack()"
#define CB4_ZAPPAR       "cb4:Zap()"

#define CB4_CBFUNCTFAIL   "%s;CodeBase function fails with error %d"
#define CB4_EXCLREQUIR    "%s;exclusive open required"

/*--------------------------------------------------------------------------*
 | oCb4:appenddb()    METHOD appends new records from another DBF file      |
 |                                                                          |
 | Parameters : argv[ 0] = object:self                                      |
 |              argv[ 1] (C)   source file name                          (1)|
 |              argv[ 2] (aC)  opt, array of fields from the src file    (1)|
 |              argv[ 3] (C|B) opt, "for" condition of the src file         |
 |              argv[ 4] (C|B) opt, "while" condition of the src file       |
 |              argv[ 5] (N|L) opt, scope (.F. all,def; .T. rest; n records)|
 |              argv[ 6] (C)   opt, via RDD driver                       (2)|
 | Notes      : (1) object values of VO are unsupported                     |
 |              (2) FS only, n/a in VO                                      |
 | Returns    : TRUE on success, FALSE otherwise.                           |
 | Description: compatible to APPEND FROM ... command                       |
 |              This fn uses the default fn __DBAPP(...), which itself      |
 |              uses the low level cb4:methods of minimal driver only.      |
 | FS invocat.: retL := oCb4:AppendDB(...)                                  |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (appenddb)       /* e.g.  _bb_cb4cdx_appenddb_me(argc,argv) */
{
   FSvar    *ovSelf;
   DATA4    *dataPtr;                                 /* CB5 structure      */
   FSvar    *passArgv[8];

   VAR_NEW_ARGS(passArgv, 8);

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, CB4_APPENDDBPAR);            /* RTE: missing args  */
      return FALSE_VAR;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("AppendDB",ovSelf);
   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("AppendDB (F)",FALSE_VAR);   /* dbf not open  */
/*
 *  Parameter checking, fill passArgv[0..7]
 */
   if (! checkAppPar (argc, argv, passArgv, CB4_APPENDDBPAR, "APPENDDB"))
      return CB4FS_DEBUG_END ("AppendDB (F)",FALSE_VAR);

/*  call generic append function
 *       FUNCTION __dbapp (  f_name         AS Char , ;
 *                         [@fld_a          AS Arra], ;
 *                         [@for_b          AS Code], ;
 *                         [@while_b        AS Code], ;
 *                         [@next_e         AS Intv], ;
 *                         [@rec_e          AS Intv], ;
 *                         [@rest_l         AS Logi], ;
 *                         [ cVia           AS Char]  ) AS Logi
 */
   fsFnNAME(__dbapp)(argc -1, passArgv);              /* call __DBAPP(...)  */

   return CB4FS_DEBUG_END ("AppendDB (T)",TRUE_VAR);
}

/*--------------------------------------------------------------------------*
 | oCb4:appendsdf()    METHOD appends new records from another SDF file     |
 |                                                                          |
 | Parameters : argv[ 0] = object:self                                      |
 |              argv[ 1] (C)   source file name                          (1)|
 |              argv[ 2] (aC)  opt, array of fields from the src file    (1)|
 |              argv[ 3] (C|B) opt, "for" condition of the src file         |
 |              argv[ 4] (C|B) opt, "while" condition of the src file       |
 |              argv[ 5] (N|L) opt, scope (.F. all,def; .T. rest; n records)|
 | Note       : (1) object values of VO are unsupported                     |
 | Returns    : TRUE on success, FALSE otherwise.                           |
 | Description: compatible to APPEND FROM ... SDF command                   |
 |              This fn uses the default fn __DBappSDF(...), which itself   |
 |              uses the low level cb4:methods of minimal driver only.      |
 | FS invocat.: retL := oCb4:AppendSDF(...)                                 |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (appendsdf)     /* e.g.  _bb_cb4cdx_appendsdf_me(argc,argv) */
{
   FSvar    *ovSelf, *vRet;
   DATA4    *dataPtr;                                 /* CB5 structure      */
   FSvar    *passArgv[9];
   int      ii, iArg;

   UDF_PROT(__dbappgen);

   VAR_NEW_ARGS(passArgv, 9);

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, CB4_APPENDSDFPAR);           /* RTE: missing args  */
      return FALSE_VAR;
   }
   if (argc < 2 || ! IS_VAR_CHR(argv[1]) || IS_VAR_EMPTY(argv[1]) ) {
      rt_error(RTARGVAL, "cb4:AppendSDF() missing or wrong file name");   /* RTE: wgong arg2  */
      return FALSE_VAR;
   }

   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("AppendSDF",ovSelf);
   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("AppendSDF",FALSE_VAR); /* database not open  */

/*  fill passArgv[0..8] and call generic append SDF/Delimited function
 *      function __dbappgener(  oSelf   AS Object, ;              passArgv[0]
 *                              f_name  AS String, ;              passArgv[1]
 *                              delim_e AS String, ;              passArgv[2]
 *                            [@fld_a   AS Array ],;              passArgv[3]
 *                            [@for_b   AS CodeBl],;              passArgv[4]
 *                            [@while_b AS CodeBl],;              passArgv[5]
 *                            [@next_e  AS Intvar],;              passArgv[6]
 *                            [@rec_e   AS Intvar],;              passArgv[7]
 *                            [@rest_l  AS Logical]) AS Logical   passArgv[8]
 */
   passArgv[0] = argv[0];
   passArgv[1] = argv[1];
   passArgv[2] = NULLSTR_VAR;
   iArg = 3;

   for(ii = 2; ii <= 4 && argc > ii; ii++) {
      passArgv[ii +1] = argv[ii];   // array of fields, for condition, while condition
      iArg++;
   }
   if(argc >= 6 && !IS_VAR_NIL(argv[5])) {
      if(CB4FSudf(decodeScope)(argv[5], passArgv, 6, 8) == 0)  // next or rest
         iArg = 9;
   }

   vRet = UDF_EXEC(__dbappgen)(iArg, passArgv);       /* call __dbappgen(...)  */
   return CB4FS_DEBUG_END ("AppendSDF",vRet);
}

/*--------------------------------------------------------------------------*
 | oCb4:appenddelim()  METHOD appends new records from delim. ASCII file    |
 |                                                                          |
 | Parameters : argv[ 0] = object:self                                      |
 |              argv[ 1] (C)   source file name                          (1)|
 |              argv[ 2] (C)   opt, delimiter                               |
 |              argv[ 3] (aC)  opt, array of fields from the src file    (1)|
 |              argv[ 4] (C|B) opt, "for" condition of the src file         |
 |              argv[ 5] (C|B) opt, "while" condition of the src file       |
 |              argv[ 6] (N|L) opt, scope (.F. all,def; .T. rest; n records)|
 | Note       : (1) object values of VO are unsupported                     |
 | Returns    : TRUE on success, FALSE otherwise.                           |
 | Description: compatible to APPEND FROM ... DELIMITED command             |
 |              This fn uses the default fn __DBappDelim(...), which itself |
 |              uses the low level cb4:methods of minimal driver only.      |
 | FS invocat.: retL := oCb4:AppendDelimited(...)                           |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (appenddelimited)   /* e.g.  _bb_cb4cdx_appenddelimited_me(argc,argv) */
{
   FSvar    *ovSelf, *vRet;
   DATA4    *dataPtr;                                 /* CB5 structure      */
   FSvar    *passArgv[9];
   int      ii, iArg;

   UDF_PROT(__dbappgen);

   VAR_NEW_ARGS(passArgv, 9);

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, CB4_APPENDDELPAR);           /* RTE: missing args  */
      return FALSE_VAR;
   }
   if (argc < 2 || ! IS_VAR_CHR(argv[1]) || IS_VAR_EMPTY(argv[1]) ) {
      rt_error(RTARGVAL, "cb4:AppendDelimited() missing or wrong file name");   /* RTE: wgong arg2  */
      return FALSE_VAR;
   }

   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("AppendDelimited",ovSelf);
   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("AppendDelimited",FALSE_VAR); /* dbf not open */

/*  fill passArgv[0..8] and call generic append SDF/Delimited function
 *      function __dbappgener(  oSelf   AS Object, ;              passArgv[0]
 *                              f_name  AS String, ;              passArgv[1]
 *                              delim_e AS String, ;              passArgv[2]
 *                            [@fld_a   AS Array ],;              passArgv[3]
 *                            [@for_b   AS CodeBl],;              passArgv[4]
 *                            [@while_b AS CodeBl],;              passArgv[5]
 *                            [@next_e  AS Intvar],;              passArgv[6]
 *                            [@rec_e   AS Intvar],;              passArgv[7]
 *                            [@rest_l  AS Logical]) AS Logical   passArgv[8]
 */
   passArgv[0] = argv[0];
   passArgv[1] = argv[1];
   if(argc >= 3 && IS_VAR_CHR(argv[2]) && !IS_VAR_EMPTY(argv[2]) )
      passArgv[2] = argv[2];     // delimiter
   else
      passArgv[2] = SET_VAR_CHR(VAR_NEW, ",");  // default delimiter = comma
   iArg = 3;

   for(ii = 3; ii <= 5 && argc > ii; ii++) {
      passArgv[ii] = argv[ii];   // array of fields, for condition, while condition
      iArg++;
   }
   if(argc >= 7 && !IS_VAR_NIL(argv[6])) {
      if(CB4FSudf(decodeScope)(argv[6], passArgv, 6, 8) == 0)  // next or rest
         iArg = 9;
   }

   vRet = UDF_EXEC(__dbappgen)(iArg, passArgv);       /* call __dbappgen(...)  */
   return CB4FS_DEBUG_END ("AppendDelimited",vRet);
}

/*--------------------------------------------------------------------------*
 | oCb4:copydb()    METHOD copies records to another DBF file               |
 |                                                                          |
 | Parameters : argv[ 0] = object:self                                      |
 |              argv[ 1] (C)   target file name                          (1)|
 |              argv[ 2] (aC)  opt, array of fields from the src file    (1)|
 |              argv[ 3] (C|B) opt, "for" condition of the src file         |
 |              argv[ 4] (C|B) opt, "while" condition of the src file       |
 |              argv[ 5] (N|L) opt, scope (.F. all,def; .T. rest; n records)|
 |              argv[ 6] (C)   opt, via RDD driver                       (2)|
 | Notes      : (1) object and symbol values of VO are unsupported          |
 |              (2) FS only, n/a in VO                                      |
 | Returns    : TRUE on success, FALSE otherwise.                           |
 | Description: compatible to COPY TO ... command                           |
 |              This fn uses the default fn __DBCOPY(...), which itself     |
 |              uses the low level cb4:methods of minimal driver only.      |
 | FS invocat.: retL := oCb4:CopyDB(...)                                    |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (copydb)           /* e.g.  _bb_cb4cdx_copydb_me(argc,argv) */
{
   FSvar    *ovSelf;
   DATA4    *dataPtr;                                 /* CB5 structure      */
   FSvar    *passArgv[8];

   VAR_NEW_ARGS(passArgv, 8);

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, CB4_COPYDBPAR);              /* RTE: missing args  */
      return FALSE_VAR;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("CopyDB",ovSelf);
   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("CopyDB",FALSE_VAR);    /* database not open  */
/*
 *  Parameter checking, fill passArgv[0..7]
 */
   if (! checkAppPar (argc, argv, passArgv, CB4_COPYDBPAR, "COPYDB"))
      return CB4FS_DEBUG_END ("CopyDB",FALSE_VAR);

/* function __dbcopy (cName, aFld,   bFor,  bWhile,
 *                    nNext, nRecno, lRest, cVia )
 */
   fsFnNAME(__dbcopy)(8, passArgv);                  /* call __DBCOPY(...)  */
   return CB4FS_DEBUG_END ("CopyDB",TRUE_VAR);
}

/*--------------------------------------------------------------------------*
 | oCb4:copysdf()    METHOD copies records to a SDF file                    |
 |                                                                          |
 | Parameters : argv[ 0] = object:self                                      |
 |              argv[ 1] (C)   target file name                          (1)|
 |              argv[ 2] (aC)  opt, array of fields from the src file    (1)|
 |              argv[ 3] (C|B) opt, "for" condition of the src file         |
 |              argv[ 4] (C|B) opt, "while" condition of the src file       |
 |              argv[ 5] (N|L) opt, scope (.F. all,def; .T. rest; n records)|
 | Note       : (1) object values of VO are unsupported                     |
 | Returns    : TRUE on success, FALSE otherwise.                           |
 | Description: compatible to COPY TO ... SDF command                       |
 |              This fn uses the default fn __DBcopySDF(...), which itself  |
 |              uses the low level cb4:methods of minimal driver only.      |
 | FS invocat.: retL := oCb4:CopySDF(file,...)                              |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (copysdf)         /* e.g.  _bb_cb4cdx_copysdf_me(argc,argv) */
{
   FSvar    *ovSelf;
   DATA4    *dataPtr;                                 /* CB5 structure      */
   FSvar    *passArgv[8];

   VAR_NEW_ARGS(passArgv, 8);

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, CB4_COPYSDFPAR);             /* RTE: missing args  */
      return FALSE_VAR;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("CopySDF",ovSelf);
   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("CopySDF",FALSE_VAR);   /* database not open  */
/*
 *  Parameter checking, fill passArgv[0..7]
 */
   if (! checkAppPar (argc, argv, passArgv, CB4_COPYSDFPAR, "COPYSDF"))
      return CB4FS_DEBUG_END ("CopySDF",FALSE_VAR);

   fsFnNAME(__dbcopysd) (7, passArgv);              /* call __DBCOPYSDF(...)*/
/* function __dbcopySDF (cName, aFld,   bFor,  bWhile,
 *                       nNext, nRecno, lRest )
 */
   return CB4FS_DEBUG_END ("CopySDF",TRUE_VAR);
}

/*--------------------------------------------------------------------------*
 | oCb4:copydelimited()  METHOD copies records to delim. ASCII file         |
 |                                                                          |
 | Parameters : argv[ 0] = object:self                                      |
 |              argv[ 1] (C)   target file name                          (1)|
 |              argv[ 2] (C)   opt, delimiter                               |
 |              argv[ 3] (aC)  opt, array of fields from the src file    (1)|
 |              argv[ 4] (C|B) opt, "for" condition of the src file         |
 |              argv[ 5] (C|B) opt, "while" condition of the src file       |
 |              argv[ 6] (N|L) opt, scope (.F. all,def; .T. rest; n records)|
 | Note       : (1) object values of VO are unsupported                     |
 | Returns    : TRUE on success, FALSE otherwise.                           |
 | Description: compatible to COPY TO ... DELIMITED command                 |
 |              This fn uses the default fn __DBcopyDelim(...), which itself|
 |              uses the low level cb4:methods of minimal driver only.      |
 | FS invocat.: retL := oCb4:CopyDelimited(...)                             |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (copydelimited)   /* e.g.  _bb_cb4cdx_copydelimited_me(argc,argv) */
{
   FSvar    *ovSelf;
   DATA4    *dataPtr;                                 /* CB5 structure      */
   FSvar    *passArgv[8];

   VAR_NEW_ARGS(passArgv, 8);

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, CB4_COPYDELPAR);             /* RTE: missing args  */
      return FALSE_VAR;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("CopyDelimited",ovSelf);
   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("CopyDelimited",FALSE_VAR);  /* dbf not open  */
/*
 *  Parameter checking, fill passArgv[0..8]
 */
   if (! checkAppDeliPar (argc, argv, passArgv, CB4_COPYDELPAR, "COPYDELIMITED"))
      return CB4FS_DEBUG_END ("CopyDelimited",FALSE_VAR);

   fsFnNAME(__dbcopyde)   (8, passArgv);         /* call __DBcopyDelim(...) */
/* function __dbcopyDelim (cName, cDelim, aFld, bFor, bWhile,
 *                         nNext, nRecno, lRest )
 */
   return CB4FS_DEBUG_END ("CopyDelimited",TRUE_VAR);
}

/*--------------------------------------------------------------------------*
 | oCb4:copystructure()  METHOD copies records to another DBF file          |
 |                                                                          |
 | Parameters : argv[ 0] = object:self                                      |
 |              argv[ 1] (C)   target file name                          (1)|
 |              argv[ 2] (aC)  opt, array of fields from the src file    (1)|
 |              argv[ 3] (C)   opt, via RDD driver                       (2)|
 | Notes      : (1) object and symbol values of VO are unsupported          |
 |              (2) FS only, n/a in VO                                      |
 | Returns    : TRUE on success, FALSE otherwise.                           |
 | Description: compatible to COPY STRUCT TO ... command                    |
 |              This fn uses the default fn __DBCOPYSTRU(...), which itself |
 |              uses the low level cb4:methods of minimal driver only and   |
 |              the cb4:createdb() method.                                  |
 | FS invocat.: retL := oCb4:CopyStructure(...)                             |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (copystructure)  /* e.g.  _bb_cb4cdx_copystructure_me(argc,argv) */
{
   FSvar    *ovSelf;
   DATA4    *dataPtr;                                 /* CB5 structure      */
   FSvar    *passArgv[3];
   FSvar    *avArr = VAR_NEW;
   FSint    err = 0;

   VAR_NEW_ARGS(passArgv, 3);

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, CB4_COPYSTRPAR);             /* RTE: missing args  */
      return FALSE_VAR;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("CopyStructure",ovSelf);
   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("CopyStructure",FALSE_VAR);
/*
 *  Parameter checking, fill passArgv[0..2]
 */
   if (argc > 4) argc = 4;
   switch (argc) {
   case 4:
      if (! IS_VAR_CHR(argv[3]) || IS_VAR_EMPTY(argv[3])) err = 3;
      SET_VAR_COPY (passArgv[2], argv[3]);
   case 3:
      if (IS_VAR_CHR(argv[2])) {
         VAR_NEW_ARRAY (avArr, 1);
         SET_VAR_COPY  (VAR_ARRELEM(avArr,1), argv[2]);
         SET_VAR_COPY  (passArgv[1], avArr);
      } else if (IS_VAR_ARR(argv[2]))
         SET_VAR_COPY  (passArgv[1], argv[2]);
      else if (! IS_VAR_EMPTY(argv[2]))
         err = 2;
   case 2:
      if (! IS_VAR_CHR(argv[1])) err = 1;
      SET_VAR_COPY (passArgv[0], argv[1]);
   }
   if (err) {
      sprintf (repdum, CB4_WRONGARG, CB4_COPYSTRPAR, err);
      CB4FSudf(callError)(ovSelf, 1, RTARGTYP,
         "COPYSTRUCT", repdum, __FILE__, __LINE__);  /* RTE: wrong arg type */
      return CB4FS_DEBUG_END ("CopyStructure",FALSE_VAR);
   }

/* function __dbcopystruct (cName, aFld, cVia)
 */
   return CB4FS_DEBUG_END ("CopyStructure",fsFnNAME(__dbcopyst)(3,passArgv));
}


/*--------------------------------------------------------------------------*
 | oCb4:pack()       METHOD removes all deleted records from the .dbf,.dbt  |
 |                          re-indexes all open indicies.                   |
 |                                                                          |
 | Parameters : argv[ 0] = object:self                                      |
 | Returns    : TRUE on success, FALSE otherwise.                           |
 | Description: compatible to PACK command                                  |
 |              This fn uses the default fn __DBPACK(...), which itself     |
 |              uses the low level cb4:methods but not the cb4:pack()       |
 | Multiuser  : The database must be open EXCLUSIVE                         |
 | FS invocat.: retL := oCb4:Pack()                                         |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (pack)               /* e.g.  _bb_cb4cdx_pack_me(argc,argv) */
{
   FSvar    *ovSelf;
   DATA4    *dataPtr;                                 /* CB5 structure      */

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, CB4_PACKPAR);                /* RTE: missing args  */
      return FALSE_VAR;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("Pack",ovSelf);
   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("Pack",FALSE_VAR);      /* database not open  */

/*
 *  check if dbf is open exclusive
 */
#ifdef CODEBASE6
   if (! dataPtr->dataFile->file.lowAccessMode == OPEN4DENY_RW) {
#else
   if (! dataPtr->file.is_exclusive) {
#endif
      sprintf (repdum, CB4_EXCLREQUIR, "Pack");
      CB4FSudf(callError)(ovSelf, 1, RTNOLOCK,
               "PACK", repdum,
               __FILE__, __LINE__);              /* RTE:  pack w/o exclusive*/
      return CB4FS_DEBUG_END ("Pack",FALSE_VAR);
   }

   if (d4pack (dataPtr) < 0) {
	  sprintf (repdum, CB4_CBFUNCTFAIL,
			   CB4_PACKPAR, dataPtr->code_base->error_code);
	  CB4FSudf(callError)(ovSelf, 1, DEVWARNING,
						  "PACK", repdum,
						  __FILE__, __LINE__);               /* Devel: fail */
      return CB4FS_DEBUG_END ("Pack",FALSE_VAR);
   }

   return CB4FS_DEBUG_END ("Pack", CB4FS_CALL_METH(gotop)(1, argv));
}

/*--------------------------------------------------------------------------*
 | oCb4:zap()        METHOD removes all deleted records from the .dbf,.dbt  |
 |                          re-indexes all open indicies.                   |
 |                                                                          |
 | Parameters : argv[ 0] = object:self                                      |
 | Returns    : TRUE on success, FALSE otherwise.                           |
 | Description: compatible to PACK command                                  |
 |              This fn uses the default fn __DBZAP(...), which itself      |
 |              uses the low level cb4:methods but not the cb4:zap()        |
 | Multiuser  : The database must be open EXCLUSIVE                         |
 | FS invocat.: retL := oCb4:zap()                                          |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (zap)                 /* e.g.  _bb_cb4cdx_zap_me(argc,argv) */
{
   FSvar    *ovSelf;
   DATA4    *dataPtr;                                 /* CB5 structure      */

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, CB4_ZAPPAR);                 /* RTE: missing args  */
      return FALSE_VAR;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("Zap",ovSelf);
   if (! (dataPtr = CB4FS_GETDATA4PTR))
      return CB4FS_DEBUG_END ("Zap",FALSE_VAR);       /* database not open  */

   if (d4zap(dataPtr, 1, d4reccount(dataPtr)))
      return CB4FS_DEBUG_END ("Zap",FALSE_VAR);       /* could no zap       */

   if (d4memo_compress(dataPtr))
      return CB4FS_DEBUG_END ("Zap",FALSE_VAR);       /* could not compress */

   return CB4FS_DEBUG_END ("Zap", TRUE_VAR);
}

/*--------------------------------------------------------------------------*
 | internal function for append... and copy... methods above                |
 *--------------------------------------------------------------------------*/
static int checkAppPar (argc, argv, passArgv, parsyntax, fnname)
int   argc;
FSvar *argv[], *passArgv[];
char  *parsyntax, *fnname;
{
   FSint errparam = 0;

   switch (argc) {
   case 7:                                             /* via clause        */
      SET_VAR_COPY (passArgv[7], argv[6]);
      switch (VAR_ISTYPE(argv[6])) {
      case 'C':
      case 'E':
      case 'U': break;
      default : errparam = 6;
      }
   case 6:                                             /* scope log,num     */
      if (CB4FSudf(decodeScope)(argv[5], passArgv, 4, 6))
         errparam = 5;
   case 5:
      if (CB4FSudf(decodeCond)(argv[4], passArgv[3]))  /* while cond        */
         errparam = 4;
   case 4:
      if (CB4FSudf(decodeCond)(argv[3], passArgv[2]))  /* for   cond        */
         errparam = 3;
   case 3:                                             /* array of fields   */
      SET_VAR_COPY (passArgv[1], argv[2]);
      switch (VAR_ISTYPE(argv[2])) {
      case 'A': if (VAR_ISDIM(argv[2]) < 1) errparam = 2;
      case 'U': break;
      default : errparam = 2;
      }
   case 2:                                             /* file name         */
      if (! IS_VAR_CHR(argv[1]) || IS_VAR_EMPTY(argv[1]))
         errparam = 1;
      SET_VAR_COPY (passArgv[0], argv[1]);
      break;
   default:
      errparam = 1;
   }
   if (errparam) {
      sprintf (repdum, CB4_WRONGARG, parsyntax, errparam);
      CB4FSudf(callError)(argv[0], 1, RTARGTYP, fnname,
               repdum, __FILE__, __LINE__);                         /* RTE  */
      return 0;
   }
   return 1;
}

/*--------------------------------------------------------------------------*
 | internal function for append... and copy... methods above                |
 *--------------------------------------------------------------------------*/
static int checkAppDeliPar (argc, argv, passArgv, parsyntax, fnname)
int   argc;
FSvar *argv[], *passArgv[];
char  *parsyntax, *fnname;
{
   FSint errparam = 0;

   switch (argc) {
   case 7:                                             /* scope log,num     */
      if (CB4FSudf(decodeScope)(argv[6], passArgv, 5, 7))
         errparam = 6;                                 /* NEXT, REST|ALL    */
   case 6:
      if (CB4FSudf(decodeCond)(argv[5], passArgv[4]))  /* while cond        */
         errparam = 5;
   case 5:
      if (CB4FSudf(decodeCond)(argv[4], passArgv[3]))  /* for   cond        */
         errparam = 4;
   case 4:                                             /* array of fields   */
      SET_VAR_COPY (passArgv[2], argv[3]);
      switch (VAR_ISTYPE(argv[3])) {
      case 'A': if (VAR_ISDIM(argv[3]) < 1) errparam = 3;
      case 'U': break;
      default : errparam = 3;
      }
   case 3:                                             /* delimiter         */
      SET_VAR_COPY (passArgv[1], argv[2]);
      switch (VAR_ISTYPE(argv[2])) {
      case 'C':
      case 'E':
      case 'U': break;
      default : errparam = 2;
      }
   case 2:                                             /* file name         */
      if (! IS_VAR_CHR(argv[1]) || IS_VAR_EMPTY(argv[1]))
         errparam = 1;
      SET_VAR_COPY (passArgv[0], argv[1]);
      break;
   default:
      errparam = 1;
   }
   if (errparam) {
      sprintf (repdum, CB4_WRONGARG, parsyntax, errparam);
      CB4FSudf(callError)(argv[0], 1, RTARGTYP, fnname,
               repdum, __FILE__, __LINE__);                         /* RTE  */
      return 0;
   }
   return 1;
}

/**** eof rddcb4j.c ****/

