/****************************************************************************
 * rddcb4l.c     RDD API for the CB4FOX, CB4CDX, CB4NDX, CB4MDX and CB4NTX  *
 *               by using CodeBase 5.1 multiplatform version (not included  *
 *               in the FlagShip package, contact Sequiter Inc.)            *
 *                                                                          *
 *               Note: all CodeBase function & defines are marked by CB5    *
 *                                                                          *
 * (c) 1995..2004 multisoft Datentechnik GmbH, Munich. All rights reserved  *
 * Release $Date: 2004/10/27 10:01:11 $                                     *
 *--------------------------------------------------------------------------*
 * This file contains parts of the full RDD driver:                         *
 *                                                                          *
 *    oCb4:getarray()          METHOD retrieves 1-dim array of fields       *
 *    oCb4:getarrfields()      METHOD retrieves n-dim array of fields       *
 *    oCb4:getlookuptable()    METHOD retrieves 2-dim array of fields       *
 *                                                                          *
 * For additional methods, see rddcb4a.c to rddcb4?.c                       *
 ****************************************************************************/

#include "rddcb4fs.h"                            /* RDD API for CodeBase    */
#include "FSextend.h"                            /* for malloc (_xalloc)    */

extern FSint  CB4FSudf(callError)();             /* in rddcb4b.c            */
extern FSint  CB4FSudf(CSkip)();                 /* in rddcb4b.c            */
extern FSvar *CB4FS_CALL_METH(seek)();           /* in rddcb4h.c            */
extern FSvar *CB4FS_CALL_METH(indexkey)();       /* in rddcb4h.c            */
extern FSvar *CB4FS_CALL_METH(fieldget)();       /* in rddcb4b.c            */
extern FSvar *CB4FS_CALL_METH(getarrfields)();   /* in rddcb4L.c            */
extern FSvar *fsFnNAME(aadd)();                  /* std. AADD()             */

#define CB4_WRONGARG     "%s;wrong type or missing argument # %d"
#define CB4_WRONGARRVAL  "%s;wrong value of array[%d]"
#define CB4_GETARRPAR    "cb4:GetArray([nLen,cnField,uSearch])"
#define CB4_GETARRFLDPAR "cb4:GetArrFields([nLen,aFields,uSearch])"
#define CB4_GETLTABPAR   "oCb4:GetLookupTable([nLen,cnFld1,cnFld2,uSearch])"

static char errMsg[100];                         /* parameter syntax        */
static char errFun[20];                          /* invoked method name     */

/*--------------------------------------------------------------------------*
 | oCb4:getArray()  METHOD retrieves n-times specif.field, optional SEEK    |
 |                  This method is a special case of oCb4:getArrFilds().    |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |         opt. argv[1] = nRecords = no of records = out array size, def=100|
 |         opt. argv[2] = cnField = no/name of the field, def=1          (1)|
 |         opt. argv[3] = uSearValue = the SEEK value to search the 1st rec |
 | Note/compat: (1) object values and symbols of VO are unsupported         |
 | Returns    : 1-dim array of values of fields, corresponding to the search|
 |              value (if given), or of the first par1 values otherwise.    |
 |              Empty array on error, eof, not found etc.                   |
 | Description: If par3 is not specified, the array is filed starting by the|
 |              current record position. After completion, the dbf pointer  |
 |              remains on the last record of the scope.                    |
 | FS invocat.: retA := oCb4:GetArray([nLen,cnField,uSearch])               |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (getarray)      /* e.g.  _bb_cb4cdx_getarray_me(argc, argv) */
{
   strcpy (errMsg, CB4_GETARRPAR);
   strcpy (errFun, "GETARRAY");
   return CB4FS_CALL_METH (getarrfields)(argc,argv);
}

/*--------------------------------------------------------------------------*
 | oCb4:getLookupTable()  METHOD retrieves n-times specif.field, opt. SEEK  |
 |                  This method is a special case of oCb4:getArrFilds().    |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |         opt. argv[1] = nRecords = no of records = out array size, def=100|
 |         opt. argv[2] = cnField = no/name of the 1st field, def=1      (1)|
 |         opt. argv[3] = cnField = no/name of the 2nd field, def=2      (1)|
 |         opt. argv[4] = uSearValue = the SEEK value to search the 1st rec |
 | Note/compat: (1) object values and symbols of VO are unsupported         |
 | Returns    : 1-dim array of values of fields, corresponding to the search|
 |              value (if given), or of the first par1 values otherwise.    |
 |              Empty array on error, eof, not found etc.                   |
 | Description: If par3 is not specified, the array is filed starting by the|
 |              current record position. After completion, the dbf pointer  |
 |              remains on the last record of the scope.                    |
 | FS invocat.: retA := oCb4:GetLookupTable([nLen,cnFld1,cnFld2,uSearch])   |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (getlookuptable)  /* e.g.  _bb_cb4cdx_getlookuptable_me(argc, argv) */
{
   FSvar *passArg[4];
   FSvar *subarr = VAR_NEW;

   VAR_NEW_ARGS  (passArg, 4);
   VAR_NEW_ARRAY (subarr,  2);
   SET_VAR_INT (VAR_ARRELEM(subarr,1), 1);                  /* set defaults */
   SET_VAR_INT (VAR_ARRELEM(subarr,2), 2);

   if (argc > 5) argc = 5;
   switch (argc) {                        /* convert to oCb4:getArrFields() */
      case 5: SET_VAR_COPY (passArg[3], argv[4]);
      case 4: SET_VAR_COPY (VAR_ARRELEM(subarr,2), argv[3]);
      case 3: SET_VAR_COPY (VAR_ARRELEM(subarr,1), argv[2]);
      case 2: SET_VAR_COPY (passArg[1], argv[1]);
      case 1: SET_VAR_COPY (passArg[0], argv[0]);
   }
   SET_VAR_COPY (passArg[2], subarr);

   strcpy (errMsg, CB4_GETLTABPAR);
   strcpy (errFun, "GETLOOKUPTABLE");
   return CB4FS_CALL_METH (getarrfields)(4, passArg);
}

/*--------------------------------------------------------------------------*
 | oCb4:getArrFields() METHOD retrieves n-times specif.fields, opt. SEEK    |
 |                     This method is similar to oCb4:getArray(), but       |
 |                     accepts an array of fields and returns n-dim.array   |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |         opt. argv[1] = nRecords = no of records = max aRet[] size,def=100|
 |         opt. argv[2] = acnField = [array with] no/name of the field      |
 |                        If the array element is NIL, field=arrElem        |
 |                        If par2 = NIL, field=1 (== oCb4:getArray() meth.) |
 |         opt. argv[3] = uSearValue = the SEEK value to search the 1st rec |
 | Returns    : n-dim array of values of fields, corresponding to the search|
 |              value (if given), or of the first par1 values otherwise.    |
 |              The array size is 1...argv[1], or an                        |
 |              empty array on error, eof, not found etc.                   |
 | Description: If par3 is not specified, the array is filed starting by the|
 |              current record position. After completion, the dbf pointer  |
 |              remains on the last record of the scope.                    |
 | Compatibil.: this method is not available in CA/VO                       |
 | FS invocat.: retA := oCb4:GetArrFields([nLen,acnField,uSearch])          |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (getarrfields)   /* e.g.  _bb_cb4cdx_getarrfields_me(argc, argv) */
{
   DATA4  *dataPtr;                                   /* CB5 structure      */
   FSvar  *ovSelf;
   FSvar  *aRet     = VAR_NEW;
   FSvar  *subarr   = VAR_NEW;
   FSvar  *uvSearch = VAR_NEW;
   FSvar  *uTmpVar ;
   FSvar  *uIndexKey, *bIndexKey;
   FSvar  *argFGet[2];
   FSvar  *argAadd[2];
   FSint  ii, iTmp, err = 0, isSeek = 0, imark, imark0;
   FSint  nSize = 100, iFields = 0;
   FSint  *iaFieldNo = 0;

   if (! errMsg[0]) {
      strcpy (errMsg, CB4_GETARRFLDPAR);
      strcpy (errFun, "GETARRFIELDS");
   }

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, errMsg);
      VAR_NEW_ARRAY (aRet, 0);
      errMsg[0] = 0;
      return aRet;
   }

   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("GetArrFields",ovSelf);

   imark0   = VAR_DEL_MARK();                         /* mark garbage coll. */

   if (! (dataPtr = CB4FS_GETDATA4PTR))
      goto err_return;                                /* dbf not open       */

/*
 *  Parameter checking,
 *  allocate C-array <iaFieldNo> and fill it with ordinal field numbers.
 */

   if (argc > 4) argc = 4;
   switch (argc) {
   case 4:
      if (strchr ("CNFILDU", VAR_ISTYPE(argv[3])))
         SET_VAR_COPY (uvSearch, argv[3]);
      else
         err = 3;
   case 3:
      if (IS_VAR_ARR (argv[2])) {
         iFields = VAR_ISDIM(argv[2]);
         VAR_NEW_ARRAY (subarr, iFields);           /* sub-array of results */
      } else
         iFields = 1;

      if (iFields > 0)
         iaFieldNo = (FSint *)
            _xalloc (sizeof(FSint) * (iFields));            /* extend C API */

      for (ii=1; ii <= iFields; ii++) {
         if (IS_VAR_ARR (argv[2]))
            uTmpVar = VAR_ARRELEM (argv[2],ii);
         else
            uTmpVar = argv[2];

         switch (VAR_ISTYPE(uTmpVar)) {
           case 'I':
           case 'N':
           case 'F': iaFieldNo[ii-1] = VAR_INTNUM(uTmpVar);
                break;
           case 'E':
           case 'C': iaFieldNo[ii-1] = d4field_number(
                                     dataPtr, VAR_CHR(uTmpVar));     /* CB5 */
                     break;
           default : iaFieldNo[ii-1] = ii;
         }
      }
   case 2:
      if (IS_VAR_NUM(argv[1]))
         nSize = VAR_INTNUM(argv[1]);
   }

   if (argc < 3) {
      iFields = 1;
      iaFieldNo = (FSint *)
            _xalloc (sizeof(FSint) * (iFields));            /* extend C API */
      iaFieldNo[0] = 1;
   }

   if (iFields < 1) err = 2;
   if (nSize   < 1) err = 1;

   if (err) {
      sprintf (repdum, CB4_WRONGARG, errMsg, err);
      CB4FSudf(callError)(ovSelf, 1, RTARGTYP, errFun,
               repdum, __FILE__, __LINE__);                          /* RTE */
      goto err_return;
   }
   iTmp = d4num_fields(dataPtr);                        /* CB5 no.of fields */
   for (ii=0; ii < iFields; ii++) {
      if ((iaFieldNo[ii] <= 0) || (iaFieldNo[ii] > iTmp)) {
         sprintf (repdum, CB4_WRONGARRVAL, errMsg, ii+1);
         CB4FSudf(callError)(ovSelf, 1, RTARGTYP, errFun, repdum,
                  __FILE__, __LINE__);
         goto err_return;
      }
   }

/*
 *  Seek for the first occurence, create IndexKey codeBlock
 */
   if (! IS_VAR_NIL(uvSearch)) {
      argFGet[0] = ovSelf;
      argFGet[1] = uvSearch;
      uTmpVar = CB4FS_CALL_METH(seek)(2,argFGet);         /* cb4:SEEK()     */
      if (! IS_VAR_TRUE(uTmpVar))
         goto err_return;
      uIndexKey = CB4FS_CALL_METH(indexkey)(1,argv);
      if (! IS_VAR_EMPTY (uIndexKey)) {
         isSeek = 1;
         bIndexKey = VAR_NEW;
         SET_VAR_COPY (bIndexKey, COMPILE_BLOCK(uIndexKey));
      }
   }

   argAadd[0] = aRet;                                     /* for AADD()     */
   argFGet[0] = ovSelf;                                   /* for fieldget() */
   argFGet[1] = VAR_NEW;                                  /* for fieldget() */

/*
 *  Loop for n-records or within SEEK scope
 */
   imark = VAR_DEL_MARK();                            /* start garb collect */
   while ((nSize > 0) && (! d4eof(dataPtr))) {            /* CB5 eof() ?    */
      for (ii=1; ii <= iFields; ii++) {
         SET_VAR_INT (argFGet[1], iaFieldNo[ii-1]);
         uTmpVar= CB4FS_CALL_METH(fieldget)(2,argFGet);   /* cb4:FieldGet() */
         if (IS_VAR_NIL (uTmpVar))                        /* failed ?       */
            goto err_return;
         if (iFields == 1)
            SET_VAR_COPY(subarr, uTmpVar);
         else
            SET_VAR_COPY(VAR_ARRELEM(subarr,ii), uTmpVar);
      }
      argAadd[1] = subarr;                                /* for AADD()     */
      fsFnNAME(aadd)(2, argAadd);                         /* call AADD()    */

      if (! CB4FSudf(CSkip)(ovSelf,dataPtr, 1L, 1))       /* SKIP 1 incl.rel*/
         break;
      nSize--;
      VAR_DEL_SINCE_MARK(imark);                          /* garb.colection */
      if (isSeek && (nSize > 0) &&                        /* key = seek ?   */
          (! IS_VAR_EQ (uvSearch, VAR_EVAL_BLOCK(bIndexKey))) )
         break;
   }

   _xfree (iaFieldNo);                                    /* extend C API   */
   goto std_return;

 err_return:
   if (iaFieldNo)
      _xfree (iaFieldNo);                                 /* extend C API   */
   SET_VAR_NIL   (aRet);
   VAR_NEW_ARRAY (aRet, 0);

 std_return:
   VAR_DEL_SINCE_MARK(imark0);                        /* full garb.colection*/
   errMsg[0] = 0;
   return CB4FS_DEBUG_END ("GetArrFields",aRet);

}



/**** eof rddcb4l.c ****/

