/****************************************************************************
 * rddcb4r.c     RDD API for the CB4FOX, CB4CDX, CB4NDX, CB4MDX and CB4NTX  *
 *               by using CodeBase 5.1 multiplatform version (not included  *
 *               in the FlagShip package, contact Sequiter Inc.)            *
 *                                                                          *
 *               Note: all CodeBase function & defines are marked by CB5    *
 *                                                                          *
 * (c) 1995..2004 multisoft Datentechnik GmbH, Munich. All rights reserved  *
 * Release $Date: 2004/10/27 10:01:49 $                                     *
 *--------------------------------------------------------------------------*
 * This file contains parts of the full RDD driver:                         *
 *                                                                          *
 *    oCb4:isrelation          ACCESS relation activating status            *
 *    oCb4:isrelation          ASSIGN activate/deactivate relations         *
 *    oCb4:clearrelation()     METHOD clear relation                        *
 *    oCb4:relation()          METHOD return the relation string            *
 *    oCb4:relationobject()    METHOD return the object of relation N       *
 *    oCb4:setrelation()       METHOD set relation                          *
 *                                                                          *
 *    moveChildRelation()      internal: process child relation positioning *
 *                                                                          *
 * For additional methods, see rddcb4a.c to rddcb4?.c                       *
 ****************************************************************************/

#include "rddcb4fs.h"                            /* RDD API for CodeBase    */

extern FSint  CB4FSudf(moveChildRelation)();     /* in rddcb4r.c            */
extern FSint  CB4FSudf(callError)();             /* in rddcb4b.c            */
extern FSvar *CB4FS_CALL_METH(clearrelation)();  /* in rddcb4r.c            */
extern FSvar *fsFnNAME(aadd)();                  /* std. AADD()             */
extern FSvar *fsFnNAME(adel)();                  /* std. ADEL()             */
extern FSvar *fsFnNAME(asize)();                 /* std. ASIZE()            */

#define CB4_WRONGARG      "%s;wrong type or missing argument # %d"
#define CB4_WRONGVALDEF   "%s;wrong value of argument # %d, default set"
#define CB4_RELATIONPAR   "cb4:Relation([nRelation])"
#define CB4_RELOBJNPAR    "cb4:RelationObject([nRelation])"
#define CB4_SETRELATPAR   "oCb4:SetRelation(oRDD,[a]c|bRel,[[a]cRel])"
#define CB4_CLRRELATPAR   "oCb4:ClearRelation([nRelat])"
#define CB4_RELOBJPAR     "oCb4:RelationObject([nRelat])"

/*--------------------------------------------------------------------------*
 | x=oCb4:isrelation  ACCESS returns .T. is relation(s) are set             |
 |                    required for COMMIT ALL to avoid recursions           |
 |                                                                          |
 | Parameters : ovSelf  = object:self                                       |
 | Returns    : retL    = .T. if relation(s) are set, .F. otherwise         |
 | Compatib.  : not available in CA/VO                                      |
 | FS invocat.: retL := oCb4:Relation                                       |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_ACCESS (isrelation)      /* e.g.  _bb_cb4cdx_isrelation_ac(ovSelv) */
{
   CB4FS_DEBUG_BEGIN ("IsRelation",ovSelf);
   if (VAR_OBJ (ovSelf,CB4OBJ_RELCOUNT) > 0)
      return CB4FS_DEBUG_END ("IsRelation (T)",TRUE_VAR);
   return CB4FS_DEBUG_END ("IsRelation (F)",FALSE_VAR);
}

/*--------------------------------------------------------------------------*
 | x=oCb4:isrelation ASSIGN activates/deactivates the active relations      |
 |                                                                          |
 | Parameters : ovSelf  = object:self                                       |
 |              assVar  = .T. to activate relations (if any),               |
 |                        .F. to deactivate them (will be not unset)        |
 | Returns    : retL    = the current relat.status after executing the assg |
 | Compatib.  : not available in CA/VO                                      |
 | FS invocat.: oCb4:Relation := lValue                                     |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_ASSIGN (isrelation) /* e.g.  _bb_cb4cdx_isrelation_as(ovSelv,assVar) */
{
   FSint count;

   CB4FS_DEBUG_BEGIN ("IsRelationAss",ovSelf);
   count = VAR_INT (VAR_OBJ (ovSelf,CB4OBJ_RELCOUNT));
   if (count == 0)
      return CB4FS_DEBUG_END ("IsRelationAss",FALSE_VAR);

   if (IS_VAR_TRUE(assVar))
      count = abs(count);
   else
      count = -(abs(count));
   SET_VAR_INT (VAR_OBJ(ovSelf,CB4OBJ_RELCOUNT), count);
   return CB4FS_DEBUG_END ("IsRelationAss",count > 0 ? TRUE_VAR : FALSE_VAR);
}

/*--------------------------------------------------------------------------*
 | oCb4:Relation()    METHOD retrieves the relation string                  |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |              argv[1] = (N) opt, number of the relation (1..n)            |
 |                            if not specified or < 1, 1 is default         |
 | Returns    : string on success, Null string on failure.                  |
 | FS invocat.: retC := oCb4:RELATION([nExp])                               |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (relation)       /* e.g.  _bb_cb4cdx_relation_me(argc,argv) */
{
   FSvar    *ovSelf;
   FSvar    *cvRet = VAR_NEW;
   FSvar    *avRelat;
   FSint    count, nRelat = 0;

   SET_VAR_CHR (cvRet, "");

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, CB4_RELATIONPAR);            /* RTE: missing args  */
      return cvRet;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("Relation",ovSelf);

   if (argc > 1 && IS_VAR_NUM(argv[1]))
      nRelat = VAR_INTNUM(argv[1]);
   if (nRelat <= 0)
      nRelat = 1;

   count = abs(VAR_INT (VAR_OBJ (ovSelf,CB4OBJ_RELCOUNT)));

   if (nRelat > count)
      return CB4FS_DEBUG_END ("Relation",cvRet);      /* return null string */

   avRelat = VAR_OBJ (ovSelf,CB4OBJ_RELATIONS);
   if (IS_VAR_ARR(avRelat) && ((int) VAR_ISDIM(avRelat) >= nRelat))
      SET_VAR_COPY (cvRet, VAR_ARRELEM(avRelat, nRelat));

   return CB4FS_DEBUG_END ("Relation",cvRet);
}

/*--------------------------------------------------------------------------*
 | oCb4:RelationObject()  METHOD retrieves the object in n-th relation      |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |              argv[1] = (N) relat.No, default = 1                         |
 | Returns    : object of the child relation, NIL on error                  |
 | FS invocat.: retO := oCb4:RelationObject([nExp])                         |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (relationobject)  /* e.g.  _bb_cb4cdx_relationobject_me(argc,argv) */
{
   FSvar    *ovSelf;
   FSint    nRelat = 1;

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, CB4_RELOBJPAR);              /* RTE: missing args  */
      return NIL_VAR;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("RelationObject",ovSelf);

   if (argc > 1 && IS_VAR_NUM(argv[1]))
      nRelat = VAR_INTNUM(argv[1]);
   if (nRelat <= 0)
      nRelat = 1;

   if (nRelat > abs(VAR_INT (VAR_OBJ (ovSelf,CB4OBJ_RELCOUNT))))
      return CB4FS_DEBUG_END ("RelationObject",NIL_VAR);

   return CB4FS_DEBUG_END ("RelationObject",
                          VAR_ARRELEM (
                          VAR_ARRELEM (
                              VAR_OBJ (ovSelf,CB4OBJ_RELATIONS), /* adr arr */
                                nRelat),      /* sub = nth elem of array    */
                                    1) );     /* obj = 1st elem of subarray */
}

/*--------------------------------------------------------------------------*
 | oCb4:SetRelation() METHOD set the child relation                         |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |              argv[1] = (O) opt, object RDD child                         |
 |                            if not specified, all relations of the        |
 |                            current object are removed = ClearRelation()  |
 |              argv[2] = (aCB) code block or string specifying the relat.  |
 |              argv[3] = (aC) string for retrieving the relation           |
 | Returns    : TRUE on success, FALSE otherwise                            |
 | Description: The object may points to ANY DbServer (also other of cb4:)  |
 |              Array of Chars for argv[2] and [3] is supported; the array  |
 |              elements representing each field are formed by the plus     |
 |              sign and the resulting code block is created therefrom.     |
 |              The data for the child relation are stored in the           |
 |              parent object.                                              |
 | Compat.    : This RDD supports any count of relations, Clipper max 8     |
 | FS invocat.: retL := oCb4:SetRelation(oRDD,bRel,[cRel])                  |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (setrelation) /* e.g.  _bb_cb4cdx_setrelation_me(argc,argv) */
{
   FSvar    *ovSelf;
   DATA4    *dataPtr;                                 /* CB5 structure      */
   FSvar    *bvRelat = VAR_NEW;
   FSvar    *cvDescr = VAR_NEW;
   FSvar    *uvTemp  = VAR_NEW;
   FSvar    *cvTemp2 = VAR_NEW;
   FSvar    *aObjRelat;
   FSvar    *passArg[2];
   FSvar    *subArr;
   FSint    ii, err = 0;

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, CB4_SETRELATPAR);            /* RTE: missing args  */
      return FALSE_VAR;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("SetRelation",ovSelf);

/*
 *  Parameter checking
 */
   if (argc > 4) argc = 4;
   switch (argc) {
   case 4:
      if (IS_VAR_CHR(argv[3]) || IS_VAR_ARR(argv[3]))
         SET_VAR_COPY (cvDescr, argv[3]);
      else
         err = 3;
   case 3:
      if (IS_VAR_CHR(argv[2]) || IS_VAR_BLK(argv[2]) || IS_VAR_ARR(argv[2]))
         SET_VAR_COPY (bvRelat, argv[2]);
      else
         err = 2;
   case 2:
   case 1:
      if (argc == 1 || IS_VAR_NIL(argv[1]))
         return CB4FS_DEBUG_END ("SetRelation",
            CB4FS_CALL_METH(clearrelation)(1,argv) ); /* cb4:CearRelation() */
      else if (! IS_VAR_OBJ(argv[1]))
         err = 1;
   }

/*
 *  If par2 is an array, concatenate to string
 */
   if (!err && IS_VAR_ARR(bvRelat)) {
      SET_VAR_CHR (cvTemp2, "");
      for (ii = 1; ii <= VAR_ISDIM(bvRelat); ii++) {
         SET_VAR_COPY (uvTemp, VAR_ARRELEM(bvRelat, ii));
         if (!IS_VAR_CHR(uvTemp) || IS_VAR_EMPTY(uvTemp)) {
            err = 2;
            break;
         }
         SET_VAR_ADD (cvTemp2, uvTemp);
      }
      if (!err)
         SET_VAR_COPY (bvRelat, cvTemp2);
   }
   if (err) {
      sprintf (repdum, CB4_WRONGARG, CB4_SETRELATPAR, err);
      CB4FSudf(callError)(ovSelf, 1, RTARGTYP,
         "SETRELATION", repdum, __FILE__, __LINE__); /* RTE: wrong arg type */
      return CB4FS_DEBUG_END ("SetRelation",FALSE_VAR);
   }

/*
 *  If par3 is an array, concatenate to string
 */
   if (IS_VAR_ARR(cvDescr)) {
      SET_VAR_CHR (cvTemp2, "");
      for (ii = 1; ii <= VAR_ISDIM(cvDescr); ii++) {
         SET_VAR_COPY (uvTemp, VAR_ARRELEM(cvDescr, ii));
         if (!IS_VAR_CHR(uvTemp) || IS_VAR_EMPTY(uvTemp)) {
            err = 3;
            break;
         }
         SET_VAR_ADD (cvTemp2, uvTemp);
      }
      if (!err)
         SET_VAR_COPY (cvDescr, cvTemp2);
   }

/*
 *  If par2 is a string, create code block.
 *  If par3 is not given, par3 = par2 string
 */
   if (IS_VAR_CHR(bvRelat)) {
      if (! IS_VAR_CHR(cvDescr) || IS_VAR_EMPTY(cvDescr))
         SET_VAR_COPY (cvDescr, bvRelat);
      SET_VAR_COPY (bvRelat, COMPILE_BLOCK(bvRelat));
   }

/*
 *  Add the relation data to a new array of relations.
 *  ovself[CB4OBJ_RELATIONS] == {{oChild, bRelat, cDescr},...}
 */
   VAR_NEW_ARRAY (subArr, 3);                                /* == ARRAY(3) */
   SET_VAR_COPY (VAR_ARRELEM(subArr,1), argv[1]);
   SET_VAR_COPY (VAR_ARRELEM(subArr,2), bvRelat);
   SET_VAR_COPY (VAR_ARRELEM(subArr,3),
                 (IS_VAR_CHR(cvDescr) ? cvDescr : NULLSTR_VAR));

   aObjRelat = VAR_OBJ(ovSelf, CB4OBJ_RELATIONS);
   if (! IS_VAR_ARR (aObjRelat))
      VAR_NEW_ARRAY (aObjRelat, 0);
   passArg[0] = aObjRelat;
   passArg[1] = subArr;
   fsFnNAME(aadd)(2, passArg);                 /* == AADD(aObjRelat,subArr) */

   return CB4FS_DEBUG_END ("SetRelation (T)",TRUE_VAR);
}

/*--------------------------------------------------------------------------*
 | oCb4:ClearRelation()   METHOD removes all child relations                |
 |                                                                          |
 | Parameters : argv[0] = object:self                                       |
 |              argv[1] = (N) opt, number of the relation (1..n)         (1)|
 |                            if not specified, or 0, all is default        |
 | Note/Compat: (1) is not supported by CA/VO                               |
 | Returns    : string on success, Null string on failure.                  |
 | FS invocat.: retC := oCb4:RELATION([nExp])                               |
 *--------------------------------------------------------------------------*/
CB4FS_DECL_METH (clearrelation) /* e.g.  _bb_cb4cdx_clearrelation_me(argc,argv) */
{
   FSvar    *ovSelf;
   FSvar    *aObjRelat;
   FSvar    *passArg[2];
   FSint    is, count = 0;

   if (argc < 1 || ! IS_VAR_OBJ(argv[0])) {
      rt_error(RTMISARG, CB4_CLRRELATPAR);            /* RTE: missing args  */
      return FALSE_VAR;
   }
   ovSelf = argv[0];
   CB4FS_DEBUG_BEGIN ("ClearRelation",ovSelf);
/*
 *  Parameter checking
 */
   if (argc > 1) {
      if (IS_VAR_NUM(argv[1]))
         count = VAR_INTNUM(argv[1]);
      else {
         sprintf (repdum, CB4_WRONGARG, CB4_CLRRELATPAR, 1);
         CB4FSudf(callError)(ovSelf, 1, RTARGTYP, "CLEARRELATION",
                 repdum, __FILE__, __LINE__);        /* RTE: wrong arg type */
         return CB4FS_DEBUG_END ("ClearRelation",FALSE_VAR);
      }
   }
   is        = VAR_INT(VAR_OBJ(ovSelf, CB4OBJ_RELCOUNT));
   aObjRelat = VAR_OBJ(ovSelf, CB4OBJ_RELATIONS);    /* array of subarrays  */

   if (count > 0 && count > abs(is)) {
      sprintf (repdum, CB4_WRONGVALDEF, CB4_CLRRELATPAR, 1);
      CB4FSudf(callError)(ovSelf, 1, DEVWARNING, "CLEARRELATION",
              repdum, __FILE__, __LINE__);           /* RTE: wrong arg type */
      count = 0;
   }

/*
 *  Clear (selective) relations
 */
   if (count <= 0 || abs(is) <= 1) {           /* clear all relations       */
      VAR_CRE_ARRAY (aObjRelat, 0);
      is = 0;
   } else {
      passArg[0] = aObjRelat;
      passArg[1] = VAR_NEW;
      SET_VAR_INT (passArg[1], count);
      fsFnNAME(adel)(2, passArg);              /* == ADEL(aObjRelat,count)  */

      is += (is > 0 ? -1 : 1);
      SET_VAR_INT (passArg[1], abs(is));
      fsFnNAME(asize)(2, passArg);             /* == ASIZE(aObjRelat,len-1) */
   }
   SET_VAR_INT (VAR_OBJ(ovSelf,CB4OBJ_RELCOUNT), is);
   return CB4FS_DEBUG_END ("ClearRelation", (is ? TRUE_VAR : FALSE_VAR));
}

/*--------------------------------------------------------------------------*
 | moveChildRelation()  move dbf pointer of the child relation              |
 |                                                                          |
 | Parameters : objRel  = object of the child relation                      |
 |              dataPtr = CB5 structure                                     |
 | Returns    : 1 = ok, 0 = failure                                         |
 | Description: The movement of related childs has to be done by this RDD   |
 |              on any dbf poiner movement, since relations are not         |
 |              automatically included in CB5 low-level d4* functions.      |
 |              This RDD driver also supports code blocks, which CB5        |
 |              doesn't by using the relate4* functions.                    |
 *--------------------------------------------------------------------------*/
FSint CB4FSudf(moveChildRelation)(ovSelf, dataPtr)
FSvar *ovSelf;
DATA4  *dataPtr;                                      /* CB5 structure      */
{

#ifndef CB4FS_MINIMAL_RDD

   FSvar  *aObjRelat, *ovChild;
   FSvar  *passArg[3];
   FSvar  *nvInfoPar = VAR_NEW;
   FSvar  *nvOrder;
   FSint  ii, count;

/*
 *  Consider current SET RELATION settings, if any.
 *  Has to be done here, because CB5 does not support it directly.
 */
   if ((count = VAR_INT(VAR_OBJ(ovSelf,CB4OBJ_RELCOUNT))) <= 0)
      return 1;

   aObjRelat = VAR_OBJ(ovSelf, CB4OBJ_RELATIONS);    /* array of subarrays  */

/*
 *  The parent dbf is on EOF() : move all relations to EOF
 */
#ifdef CODEBASE6
   if (dataPtr->eofFlag) {
#else
   if (dataPtr->eof_flag) {
#endif
      for (ii=1; ii <= count; ii++) {
         passArg[0] = VAR_ARRELEM ( VAR_ARRELEM (
                          VAR_OBJ (ovSelf,CB4OBJ_RELATIONS), ii), 1);
         OBJ_METHEXEC ("goBottom",-1,1,passArg); /* oRel:GoBottom() run-time */
         OBJ_METHEXEC ("skip",    -1,1,passArg); /* oRel:Skip()     run-time */
      }
      return 1;
   }
/*
 *  execute SEEK on the child .dbf when index available,
 *  or GOTO <recno> if index is not set and the code block returns numer.var
 */
   SET_VAR_INT (nvInfoPar, DBOI_ORDERCOUNT);
   for (ii=1; ii <= count; ii++) {
      passArg[0] = VAR_ARRELEM ( VAR_ARRELEM (                     /* child */
                       VAR_OBJ (ovSelf,CB4OBJ_RELATIONS), ii), 1); /* obj   */
      passArg[1] = nvInfoPar;
      nvOrder = OBJ_METHEXEC ("OrderInfo",-1,2,passArg);        /* run-time */

      passArg[1] = VAR_EVAL_BLOCK(                                 /* eval  */
                     VAR_ARRELEM ( VAR_ARRELEM (                   /* child */
                       VAR_OBJ (ovSelf,CB4OBJ_RELATIONS),ii), 2)); /* block */

/*
 *    Child Order no. > 0 ? If so: child -> SEEK <cbValue>
 */
      if (IS_VAR_NUM (nvOrder) && (VAR_INTNUM(nvOrder) > 0)) {
         passArg[2] = FALSE_VAR;
         OBJ_METHEXEC ("Seek",-1, 3,passArg);       /* oRel:Seek() run-time */
/*
 *    Block return numeric value ? If so: child -> GO TO <cbValue>
 */
      } else if (IS_VAR_NUM (passArg[1])) {
         OBJ_METHEXEC ("GoTo",-1, 2,passArg);       /* oRel:Goto() run-time */
/*
 *    Otherwise: child -> GOTO EOF()
 */
      } else {
         OBJ_METHEXEC ("GoBottom",-1, 1,passArg); /* oRel:GoBott() run-time */
         OBJ_METHEXEC ("Skip",    -1, 1,passArg); /* oRel:skip()   run-time */
      }
   }
#endif

   return 1;
}

/**** eof rddcb4r.c ****/

