#include "cb4gener.fh"

* --------------------------------------------------------------------------
* Test for user-defined RDD drivers.
* Part 1 - minimal RDD methods
*
* (c) 1996..2004 by multisoft Datentechnik GmbH, Munich. All rights reserved.
* Release $Date: 2004/11/03 15:03:55 $
*
* You may specify the tasks to be tested at the command line, e.g.
*    rddtest1            = test all
*    rddtest1 all        = test all (don't create new database)
*    rddtest1 1 3 5      = test tasks 1, 3, and 5
*    rddtest1 all -2 -4  = test all except tasks 2 and 4
*    rddtest1 file ....  = redirect test results to file 'rddtest.doc'
*
* Requires: rddtest1.dbf + rddtest1.dbt - may be created with
*    rddtest1 1 [...]
*    rddtest1 all 1 [...]
*
* compile: FlagShip rddtest1.prg -na -Mmain -L. -lcb4cdx 
*      or: CLIPPER  rddtest1 /n ; RTLINK fi rddtest1 lib dbfcdx
* --------------------------------------------------------------------------
* Task 1: create empty database with the default RDD driver (on request only)
*      2: minimal driver: open, skip, goto, eof, bof, access fields
*      3: minimal driver: replace and fieldput()
*      4: minimal driver: use the same database in two WA's EXCL + READONLY
*      5: minimal driver: use the same database in two WA's EXCL + SHARED
*    6,8: minimal driver: use the same database SHARED from two users
*    7,9: minimal driver: use the same database EXCLUSIVE from two users

#define MIN_REC1  8
#define ALLTASKS  " 234567  "
#define FILENAME  "rddtest1.doc"
#define CRLF      chr(13)+chr(10)
* #define MEMOINCLUDE                     /* applic.for cb4ntx only */
#ifdef FlagShip
# define CANFIELDPUTNAME
#endif

#ifdef FlagShip
  static rddstd  := "DbfIdx"
  static rddnew  := "cb4cdx"             // both entries
  extern cb4cdx                          // has to match
* static rddnew  := "dbfidx"             // both entries
#else
  extern dbfntx
  static rddstd  := "DbfNtx"             // Clipper 5.2x
  static rddnew  := "dbfcdx"
  extern dbfcdx
# define secondscpu seconds
# define LOCAL_INT LOCAL
#endif

STATIC tasks  := "          "
STATIC isFile := .F.
STATIC isok   := ""
STATIC cRec5Mem1 := "", cRec5Mem2 := ""

#ifdef MEMOINCLUDE
STATIC struct := {{"Num",   "N",  5, 0}, ;
                  {"Deci",  "N", 15, 7}, ;
                  {"Char",  "C", 23, 0}, ;
                  {"Log",   "L",  1, 0}, ;
                  {"Date",  "D",  8, 0}, ;
                  {"Memo1", "M", 10, 0}, ;
                  {"Memo2", "M", 10, 0}}
#else
STATIC struct := {{"Num",   "N",  5, 0}, ;
                  {"Deci",  "N", 15, 7}, ;
                  {"Char",  "C", 23, 0}, ;
                  {"Log",   "L",  1, 0}, ;
                  {"Date",  "D",  8, 0}}
#endif

**************************************************************************
FUNCTION main (p1,p2,p3,p4,p5,p6,p7,p8,p9,p10,p11,p12,p13,p14,p15)
**************************************************************************
local ii, xx
memvar idx

* check and process incoming parameters from the command line

if !checkpar (pcount(), p1,p2,p3,p4,p5,p6,p7,p8,p9,p10,p11,p12,p13,p14,p15)
   quit
endif

set font "courier"

for ii := 1 to len(tasks)
   idx := ltrim(str(ii,2))
   if strpeek (@tasks, ii) != asc(" ")
      cls
      ? "*** Only methods of the minimal RDD driver are used ***"
      do task&idx
      ? "...task " + idx + " done, press any key (or 'R' to repeat, 'Q' to abort)..."
      xx := upper(chr(inkey(0)))
      if xx == "R"
         ii--
      elseif xx == 'Q' .or. lastkey() == 27
         exit
      endif
   endif
next
? "*** RDDtest1 done ***"
quit
return 0

**************************************************************************
static function checkpar
**************************************************************************
local par
local ii, jj, kk := 1, xx
memvar idx
parameters param, p1,p2,p3,p4,p5,p6,p7,p8,p9,p10,p11,p12,p13,p14,p15

p1 := if (p1 == NIL, "", p1)
p2 := if (p2 == NIL, "", p2)

if param = 0
   tasks := ALLTASKS
elseif upper(p1) == "FILE"
   kk := 2
   isFile := .T.
   set printer to FILENAME
   set printer on
   if val(p2) < 0 .or. param < kk
      tasks := ALLTASKS
   endif
elseif val(p1) < 0
   tasks := ALLTASKS
endif
for ii := kk to param                    // what to test
   idx := ltrim(str(ii,2))
   par := p&idx
   if upper(par) == "ALL"
      tasks := ALLTASKS
   else
      jj := val(par)
      strpoke (@tasks, abs(jj), if (jj < 0, asc(" "), asc("x")))
   endif
next
return .T.

**************************************************************************
procedure task1
* create new database by the default RDD, fill it with data
**************************************************************************
local temp, start, ii

? "--- task1, standard " + dbsetdriver() + " driver"
start := secondscpu()
dbcreate ("rddtest1", struct, rddstd)                   // FS 4.4
use rddtest1 via (rddstd)

?? " alias= " + alias()
while lastrec() < MIN_REC1
   ? "adding record "
   append blank
   ?? ii := recno()
   temp := ltrim(str(ii)) + "-"
   replace num  with ii, deci with ii * 10000 + ii / 10000 + 0.0000012345
   replace Char with replicate(temp,13), log with (ii % 2 == 0)
   replace date with ctod("05/31/95") + ii -1
#ifdef MEMOINCLUDE
*  memo 1 filled in record 1, 4, 5, 8
*  memo 2 filled in record 2, 4, 6, 8
*  no memo       in record 3, 7
   if (ii % 4) <= 1
      replace memo1 with "Memo data 1 for record " + ltrim(str(recno()))
   endif
   if (ii % 4) == 0 .or. (ii % 4) == 2
      replace Memo2 with "Memo data 2" + CRLF + "for record " + ;
						  ltrim(str(recno()))
   endif
#endif
enddo
use
? "--- end task1,", secondscpu() - start, "seconds"
return

**************************************************************************
procedure task2
* minimal driver: open, skip, goto, access fields by name & fieldget()
**************************************************************************
LOCAL_INT ii
LOCAL jj, start, oldErr, loop, cTmp, nTmp
FIELD num, deci, char, log, date, Memo1, Memo2

isok := ""

copy file "rddtest1.dbf" to "rddtest2.dbf"
#ifdef MEMOINCLUDE
 copy file "rddtest1.dbt" to "rddtest2.dbt"
#endif

use rddtest2 via (rddnew)
? "--- task2, tested driver " + rddnew + " driver, alias= " + alias()
printUsed()
if !used()
   ?? chr(7)
   ? "*** ERROR, USE/OPEN database failed"
   return
endif
display0()
start := secondscpu()
goto 1
display1("go 1:")

loop := 0
while !eof() .and. loop < MIN_REC1
   skip
   display1("skip:")
   loop++
enddo
if !eof()
   ? "*** ERROR, EOF not set"
endif

loop := 0
while !bof() .and. loop < MIN_REC1
   skip -2
   display1("sk-2:")
   loop++
enddo
if !bof()
   ? "*** ERROR, BOF not set"
endif

goto 5
display1("go 5:")
#ifdef MEMOINCLUDE
 checknempty (memo2, "Memo2 field" )
#endif

goto 4
display1("go 4:")

checkempty (num  , "Num"  + " field")
checkempty (Deci , "Deci" + " field")
checkempty (Char , "Char" + " field")
checkempty (Log  , "Log"  + " field")
checkempty (Date , "Date" + " field")
#ifdef MEMOINCLUDE
 checkempty (memo1, "Memo1"+ " field")
 checkempty (MEMO2, "Memo2"+ " field")
#endif

for ii := 1 to len(struct)
   cTmp = struct[ii,1]
   checkempty (fieldget(ii),          "FieldGet(" + ltrim(str(ii)) + ")")
# ifdef CANFIELDPUTNAME
   checkempty (fieldget(upper(cTmp)), 'FieldGet("' + cTmp + '")')
# endif
next

jj := 4
cTmp := ltrim(str(jj)) + "-"
cTmp := left(replicate(cTmp, 13), struct[3,3])
nTmp := jj * 10000 + jj / 10000 + 0.0000012345
nTmp := round(nTmp, struct[2,4])

?
checkval (num,  jj,   "Num" , .T.)
checkval (deci, nTmp, "Deci", .T.)
checkval (Char, cTmp, "Char", .T.)
checkval (log,  .T.,  "Log" , .T.)
checkval (date, ctod("05/31/95") + jj -1, "Date", .T.)
#ifdef MEMOINCLUDE
 checkval (memo1, "Memo data 1 for record 4", "Memo1", .T.)
 checkval (memo2, "Memo data 2" + CRLF + "for record 4", "Memo2", .T.)
#endif

goto 99
display1("go99:")
if !eof()
   ? "*** ERROR, EOF not set"
   isok := "NOT "
endif
checknempty (num  , "Num"  + " field")
checknempty (Deci , "Deci" + " field")
checknempty (Char , "Char" + " field")
checknempty (Log  , "Log"  + " field")
checknempty (Date , "Date" + " field")
#ifdef MEMOINCLUDE
 checknempty (MEMO1, "Memo1"+ " field")
 checknempty (memo2, "Memo2"+ " field")
#endif

for ii := 1 to len(struct)
   cTmp = struct[ii,1]
   checknempty (fieldget(ii),          "FieldGet(" + ltrim(str(ii)) + ")")
# ifdef CANFIELDPUTNAME
   checknempty (fieldget(lower(cTmp)), 'FieldGet("' + cTmp + '")')
# endif
next

use
if used()
   ? "*** ERROR, USE/CLOSE database failed, database still open"
   isok := "NOT "
endif
?
? "--- end task2,", secondscpu() - start, "seconds"
? "--- x=<field> access " + isok + "successfully tested."
return

**************************************************************************
procedure task3
* minimal driver: SET PATH
*                 replace fields by name and by fieldput()
**************************************************************************
LOCAL start, jj, cTmp, nTmp, cMem1, cMem2
FIELD num, deci, char, log, date, Memo1, Memo2

#ifdef TEST_ON
cls      
#endif

set path to ..\

isok := ""
if file("rddtest2.dbf")
   delete file ("rddtest2.dbf")
# ifdef MEMOINCLUDE
   delete file ("rddtest2.dbt")
# endif
endif

copy file "rddtest1.dbf" to "..\rddtest2.dbf"
#ifdef MEMOINCLUDE
 copy file "rddtest1.dbt" to "..\rddtest2.dbt"
#endif
? "rddtest2.db? copied to upper ../ directory"

select 10
use rddtest2 via (rddnew)
? "--- task3, tested driver " + upper(rddnew) + ", alias " + alias()

select rddtest2

if select() != 10 .or. !(alias() == "RDDTEST2")
   ? "*** ERROR, Alias should be 10 / [RDDTEST2], is", ;
     ltrim(str(select())), "/ [" + alias() + "]"
   return
endif
printUsed()
if !used()
   ?? chr(7)
   ? "*** ERROR, USE/OPEN database failed"
   return
endif

start := secondscpu()

goto 5
jj   := 98765
nTmp := -987654.7654321
cTmp := left("abcdefghijklmnopqrstuvwxyz", struct[3,3])
#ifdef MEMOINCLUDE
 cRec5Mem1:= memo1 + CRLF + replicate("x",2500) + "end memo1"
 cRec5Mem2:= memo2 + CRLF + replicate("X",400)  + "end memo2"
#endif

? "Database replacement in alias = ", alias()

if fieldput(1, jj) != jj
   ? "*** ERROR: field 'Num' NOT correctly replaced"
   isok := "NOT "
endif
replace deci  with nTmp
#ifdef CANFIELDPUTNAME
   if ! (fieldput("Char", cTmp) == cTmp)
      ? "*** ERROR: field 'Char' NOT correctly replaced"
      isok := "NOT "
   endif
#else
   if ! (fieldput(3, cTmp) == cTmp)
      ? "*** ERROR: field 'Char' NOT correctly replaced"
      isok := "NOT "
   endif
#endif
replace log   with .t.
replace date  with date()

#ifdef MEMOINCLUDE
  replace memo1 with cRec5Mem1
# ifdef CANFIELDPUTNAME
   if ! (fieldput ("memo2", cRec5Mem2) == cRec5Mem2)
      ? "*** ERROR: field 'Memo2' NOT correctly replaced"
      isok := "NOT "
   endif
# else
   if ! (fieldput (7, cRec5Mem2) == cRec5Mem2)
      ? "*** ERROR: field 'Memo2' NOT correctly replaced"
      isok := "NOT "
   endif
# endif
#endif

skip 0
checkval (num,   jj,     "Num field")

goto 1
checkval (num,   1,      "Num field")
goto 5
checkval (num,   jj,     "Num" , .T.)
checkval (deci,  nTmp,   "Deci", .T.)
checkval (Char,  cTmp,   "Char", .T.)
checkval (log,   .T.,    "Log" , .T.)
checkval (date,  date(), "Date", .T.)
#ifdef MEMOINCLUDE
 checkval (memo1, cRec5Mem1,  "Memo1", .T.)
 checkval (memo2, cRec5Mem2,  "Memo2", .T.)
#endif

** Database is NOT closed (for task4)

? "--- end task3,", secondscpu() - start, "seconds"
? "--- x=<field>, FIELDGET(), REPLACE, FIELDPUT() " + isok + "successful tested."
return

**************************************************************************
procedure task4
* minimal driver: SET PATH
*                 multiple database usage (different aliases) EXCL/READONLY
**************************************************************************
select 10
if !used()
   ? "*** Perform task 3 first"
   return
endif

? "--- task4"
? "The SAME database rddtest2 (opened EXCLUSIVE-ly in WA 10) will be now"
? "opened READONLY in WA 20 by using another alias name"
? "This SHOULD FAIL for correct multiuser/multitasking handling"
wait
?
select 20
use rddtest2 via (rddnew) alias second readonly

do test45 with "task4"
return

**************************************************************************
procedure task5
* minimal driver: multiple database usage (different aliases) EXCL/SHARED
**************************************************************************
select 10
if !used()
   ? "*** Perform task 3 first"
   return
endif

? "--- task5"
? "The SAME database rddtest2 (open EXCLUSIVE in WA 10) will be now"
? "opened SHARED in WA 20 by using another alias name"
? "This SHOULD FAIL for correct multiuser/multitasking handling"
wait
?
select 20
use rddtest2 via (rddnew) alias second shared

do test45 with "task5"

return


*---------------------------------------------------------------------------
procedure test45 (task)
*---------------------------------------------------------------------------
LOCAL start, jj, cTmp, nTmp
FIELD num, deci, char, log, date, Memo1, Memo2

select 10
? "--- " + task + ", tested driver " + upper(rddnew) + ", alias= " + alias()

select 20
? "           tested driver " + upper(rddnew) + ", alias= " + alias()

if !used()
   ?? chr(7)
   ? "*** ERROR, second USE/OPEN of the database failed"
   return
endif
if select() != 20 .or. !(alias() == "SECOND")
   ? "*** ERROR, Alias " + alias() + " should be 20"
   return
endif

select rddtest2

if select() != 10 .or. !(alias() == "RDDTEST2")
   ? "*** ERROR, Alias " + alias() + " should be 10"
   return
endif
printUsed()
if !used()
   ?? chr(7)
   ? "*** ERROR, USE/OPEN database failed"
   return
endif

start := secondscpu()

select 10
goto 5

jj   := 98765
nTmp := -987654.7654321
cTmp := left("abcdefghijklmnopqrstuvwxyz", struct[3,3])

?
? "--- data of record 5 in WA10, alias =", alias()
checkval (num,   jj,     "Num" , .T.)
checkval (deci,  nTmp,   "Deci", .T.)
checkval (Char,  cTmp,   "Char", .T.)
checkval (log,   .T.,    "Log" , .T.)
checkval (date,  date(), "Date", .T.)
#ifdef MEMOINCLUDE
 checkval (memo1, cRec5Mem1,  "Memo1", .T.)
 checkval (memo2, cRec5Mem2,  "Memo2", .T.)
#endif

select 20
?
? "--- the same database and record in WA20, alias =", alias()
goto 2
checkval (num,   2,     "Num field")

goto 5
checkval (num,   jj,     "Num" , .T.)
checkval (deci,  nTmp,   "Deci", .T.)
checkval (Char,  cTmp,   "Char", .T.)
checkval (log,   .T.,    "Log" , .T.)
checkval (date,  date(), "Date", .T.)
#ifdef MEMOINCLUDE
 checkval (memo1, cRec5Mem1,  "Memo1", .T.)
 checkval (memo2, cRec5Mem2,  "Memo2", .T.)
#endif

?
select 20
use
if used()
   ? "*** ERROR, USE/CLOSE of the second database failed, database still open"
endif

? "--- end " + task, secondscpu() - start, "seconds"
return

**************************************************************************
procedure task6
* minimal driver: multiple database usage (multiuser) SHARED
**************************************************************************
task67 (" 6", .T., " 8")
return

**************************************************************************
procedure task7
* minimal driver: multiple database usage (multiuser) EXCLUSIVE
**************************************************************************
task67 (" 7", .F., " 9")
return


**************************************************************************
procedure task8
* minimal driver: multiple database usage (multiuser) SHARED
* same as task6, but invoked from task6
**************************************************************************
task89 (" 8", .T.)
return

**************************************************************************
procedure task9
* minimal driver: multiple database usage (multiuser) EXCLUSIVE
* same as task7, but invoked from task7
**************************************************************************
task89 (" 9", .F.)
return

*---------------------------------------------------------------------------
static function task67 (istask, shared, calltask)
local txt, start, runcmd := execname() + calltask
set path to

select 10
use
select 20
use

copy file "rddtest1.dbf" to "rddtest2.dbf"
# ifdef MEMOINCLUDE
  copy file "rddtest1.dbt" to "rddtest2.dbt"
# endif

txt := if (shared, "SHARED", "EXCLUSIVE")

? "--- task" + ltrim(istask) + ", tested driver " + upper(rddnew)

start := secondscpu()

select 10
if shared
   use rddtest2 via (rddnew) SHARED
else
   use rddtest2 via (rddnew) EXCLUSIVE
endif

if !shared
   ? "replacing NUM with 11111, date with", date()
   display0()
   replace num with 11111, date with date()
else
   display0()
endif
display1("Rec1:")
?
? "The SAME database rddtest2 (open " + txt + " in WA 10) is now"
? "tried to open " + txt + " by other user "

#ifdef FlagShip
  ? "by automaticaly executing this applic. using RUN '" + runcmd + "'"
  ? "You may also try to execute '" + runcmd + "' from other terminal now..."
  save screen
#else
  ? 'On DOS, try to execute "RDDTEST1' + calltask + ;
    '" from another user now ...'
#endif

wait

#ifdef FlagShip
  RUN (execname() + calltask)
  restore screen
#endif

if file ("rddtest8.txt")
   ? "------- output from task" + ltrim(calltask) + " -----------------"
   ? memoread ("rddtest8.txt")
   ?
   ? "------- end of output from task" + ltrim(calltask) + " ----------"
   delete file rddtest8.txt
endif
if !shared
   skip 0
   display0()
   display1("Rec1:")
endif
use

? "--- end task" + ltrim(istask), secondscpu() - start, "seconds"
return

*---------------------------------------------------------------------------
static function task89 (istask, shared)
local txt, start
set path to

select 10
use
select 20
use

set print to rddtest8.txt
set print on

txt := if (shared, "SHARED", "EXCLUSIVE")

? "--- task" + ltrim(istask) + ", tested driver " + upper(rddnew)
?
? "The already " + txt + " open database rddtest2 will be now"
? "opended by second user"

start := secondscpu()

select 10
if shared
   use rddtest2 via (rddnew) SHARED
else
   use rddtest2 via (rddnew) EXCLUSIVE
endif

if !used()
   ? "*** Error, database not open"
else
   printused()
   ? "--- Database was opened succesfully. Available data:"
   display0()
   display1("Rec1:")
   if !shared
      ? "      --- Replacing NUM with 22222, date with", date() +10
      replace num with 22222, date with date() +10
      display1("Rec1:")
   endif
endif
use

? "--- end task" + ltrim(istask), secondscpu() - start, "seconds"

set print to
quit
return

*---------------------------------------------------------------------------
static function display0()
?
? "      ..num ....num.deci... ........char........... log mm/dd/yy eof bof M1 M2"
return NIL

*---------------------------------------------------------------------------
static function display1(prefix)
FIELD num, deci, char, log, date, Memo1, Memo2
if !used()
   ? "database not open"
   return NIL
endif
if eof()
   ? "record:",ltrim(recno())," is eof()"
   return NIL
endif
? prefix, num, deci, char, log, date, eof(), bof()
#ifdef MEMOINCLUDE
?? "", str(len(Memo1),2), str(len(Memo2),2)
#endif
return nil

*---------------------------------------------------------------------------
static function checkempty (value, name)
if (empty(value))
   ? "*** ERROR, " + name + " returns: EMPTY"
   return .F.
endif
return .T.

*---------------------------------------------------------------------------
static function checknempty (value, name)
if (! empty(value))
   ? "*** ERROR, " + name + " returns: NOT EMPTY"
   return .F.
endif
return .T.

*---------------------------------------------------------------------------
static function checkval (fldval, value, name, out)
local disp1 := fldval, disp2 := value

if (valtype(fldval) $ "MC" .and. len(fldval) > 35) .or. ;
   (valtype(value)  $ "MC" .and. len(value)  > 35)
   if len(fldval) > 70
      disp1 := left(fldval,34) + "..." + right(fldval,34)
   endif
   if len(value) > 70
      disp2 := left(value, 34) + "..." + right(value, 34)
   endif
endif
if valtype(fldval) $ "MC"
   disp1 := strtran (disp1, CRLF, "<CRLF>")
   disp2 := strtran (disp2, CRLF, "<CRLF>")
endif
if valtype(disp1) == "C" .and. len(disp1 + disp2) > 35
   if fldval == value
      if valtype(out) == "L" .and. out
         ? "dbf: " + padr(name,10) + " correct value [ LEN() = " + ;
            ltrim(str(len(fldval))) + " ] :"
         if col() + len(disp1) + 3 > 79
            ?
         endif
         ?? " [" + disp1 + "]"
      endif
      return .T.
   else
      isok := "NOT "
      ? "*** ERROR, dbf: " + name + " is/should be [ LEN() = " + ;
        ltrim(str(len(fldval))) + "/" + ltrim(str(len(value))) + " ] :"
      ? " [" + disp1 + "]"
      ? " [" + disp2 + "]"
   endif
   return .F.
endif

if fldval == value
   if valtype(out) == "L" .and. out
      ? "dbf: " + padr(name,10) + " correct value [", disp1, "]"
   endif
else
   isok := "NOT "
   ? "*** ERROR, dbf: " + name + " [", disp1, "] should be [", disp2, "]"
   return .F.
endif
return .T.

*---------------------------------------------------------------------------
#ifndef FlagShip
function strpeek (str,pos)
return asc(substr(str,pos,1))
function strpoke (str,pos,char)
str = stuff(str,pos,1,chr(char))
return str
FUNCTION ExecName()
RETURN "rddtest2.exe"
#endif

*---------------------------------------------------------------------------
function PrintUsed()
local currArr := select()
local ii, jj := 0
for ii := 1 to 250
   select (ii)
   if used()
      ? "Select WA " + str(ii,3) + ":", alias()
      jj++
   endif
next
select (currArr)
return jj


*** eof rddtest1.prg ***

