* --------------------------------------------------------------------------
* Test for user-defined RDD drivers.
* Part 2 - all RDD basic methods (see RDDTEST1 for minimal driver)
*          available in rddcb4[a,b].c source files
*
 * (c) 1995..2004 multisoft Datentechnik GmbH, Munich. All rights reserved  
 * Release $Date: 2004/11/03 15:04:06 $                                     
*
* You may specify the tasks to be tested at the command line, e.g.
*    rddtest2            = test all
*    rddtest2 all        = test all
*    rddtest2 1 3 5      = test tasks 1, 3, and 5
*    rddtest2 all -1     = avoid to create the database
*    rddtest2 all -2 -4  = test all except tasks 2 and 4
*    rddtest2 file ....  = redirect test results to file 'rddtest2.out'
*
* compile: FlagShip rddtest2.prg -na -Mmain -L. -lcb4cdx
*      or: CLIPPER  rddtest2 /n ; RTLINK fi rddtest2 lib dbfcdx
* --------------------------------------------------------------------------
* Task 1: create empty database with this RDD driver
*      2: minimal driver: open, skip, goto, eof, bof, access fields
*      3: minimal driver: replace and fieldput()
*      4: minimal driver: use the same database in two WA's EXCL + READONLY
*      5: minimal driver: use the same database in two WA's EXCL + SHARED
*    6,8: minimal driver: use the same database SHARED from two users
*    7,9: minimal driver: use the same database EXCLUSIVE from two users

#define REC_ORIG  8
#define ALLTASKS  "1234567  "
#define FILEOUT   "rddtest2.out"
#define FILETRASF "rddt2_08.out"
#define ORIG_DBF  "rddt2_01.dbf"
#define ORIG_DBT  "rddt2_01" + memoext()
#define WORK_NAME "rddt2_02"
#define WORK_DBF  "rddt2_02.dbf"
#define WORK_DBT  WORK_NAME + memoext()
#define CRLF      chr(13)+chr(10)
* #define MEMOINCLUDE
* #define OUTPUT_OBJ_TO_STDERR

#ifdef FlagShip
# define CANFIELDPUTNAME
# define DRIVER_CREATOR Cb4CdxNew
#endif

#ifdef FlagShip
  static rddstd  := "DbfIdx"             /*  std. FS driver   */
* static rddnew  := "dbfidx"             /*  \ both entries   */
  static rddnew  := "cb4cdx"             /*  \ both entries   */
  extern cb4cdx                          /*  / has to match   */
#else
  extern dbfntx
  static rddstd  := "DbfNtx"             // Clipper 5.2x
  static rddnew  := "dbfcdx"
  extern dbfcdx
# define secondscpu seconds
# define LOCAL_INT LOCAL
#endif

STATIC tasks  := "          "
STATIC isFile := .F.
STATIC isok   := ""
STATIC cRec5Mem1 := "", cRec5Mem2 := ""

#ifdef MEMOINCLUDE
STATIC struct := {{"Num",   "N",  5, 0}, ;
                  {"Deci",  "N", 15, 7}, ;
                  {"Char",  "C", 23, 0}, ;
                  {"Log",   "L",  1, 0}, ;
                  {"Date",  "D",  8, 0}, ;
                  {"Memo1", "M", 10, 0}, ;
                  {"Memo2", "M", 10, 0}}
#else
STATIC struct := {{"Num",   "N",  5, 0}, ;
                  {"Deci",  "N", 15, 7}, ;
                  {"Char",  "C", 23, 0}, ;
                  {"Log",   "L",  1, 0}, ;
                  {"Date",  "D",  8, 0}}
#endif

**************************************************************************
FUNCTION main (p1,p2,p3,p4,p5,p6,p7,p8,p9,p10,p11,p12,p13,p14,p15)
**************************************************************************
local ii, xx
memvar idx

* check and process incoming parameters from the command line

if !checkpar (pcount(), p1,p2,p3,p4,p5,p6,p7,p8,p9,p10,p11,p12,p13,p14,p15)
   quit
endif

set font "courier"

rddsetdefault (rddnew)             // set the tested driver for all operat.

#ifdef OUTPUT_OBJ_TO_STDERR
   ? 'invoke: sh -c "' + execname() + ' 2>testdata"'
   wait
#endif

for ii := 1 to len(tasks)
   idx := ltrim(str(ii,2))
   if strpeek (@tasks, ii) != asc(" ")
      cls
      ? "*** Methods of the basic RDD driver are used ***"
      do task&idx
      ? "...task " + idx + " done, press any key (or 'R' to repeat it, 'Q' to abort)..."
      xx := upper(chr(inkey(0)))
      if xx == "R"
         ii--
      elseif xx == 'Q' .or. lastkey() == 27
         exit
      endif
   endif
next
? "*** RDDtest2 done ***"
quit
return 0

**************************************************************************
static function checkpar
**************************************************************************
local par
local ii, jj, kk := 1, xx
memvar idx
parameters param, p1,p2,p3,p4,p5,p6,p7,p8,p9,p10,p11,p12,p13,p14,p15

p1 := if (p1 == NIL, "", p1)
p2 := if (p2 == NIL, "", p2)

if param = 0
   tasks := ALLTASKS
elseif upper(p1) == "FILE"
   kk := 2
   isFile := .T.
   set printer to FILEOUT
   set printer on
   if val(p2) < 0 .or. param < kk
      tasks := ALLTASKS
   endif
elseif val(p1) < 0
   tasks := ALLTASKS
endif
for ii := kk to param                    // what to test
   idx := ltrim(str(ii,2))
   par := p&idx
   if upper(par) == "ALL"
      tasks := ALLTASKS
   else
      jj := val(par)
      strpoke (@tasks, abs(jj), if (jj < 0, asc(" "), asc("x")))
   endif
next
return .T.

**************************************************************************
procedure task1
* create new database by the default RDD, fill it with data
**************************************************************************
local temp, start, ii

? "--- task1, " + dbsetdriver() + " driver"
start := secondscpu()
? "    creating " +  ORIG_DBF
dbcreate (ORIG_DBF, struct)
? "    opening " +  ORIG_DBF
use (ORIG_DBF)

?? " alias= " + alias()
while lastrec() < REC_ORIG
   ? "adding record "
   append blank
   ?? ii := recno()
   temp := ltrim(str(ii)) + "-"
   replace num  with ii, deci with ii * 10000 + ii / 10000 + 0.0000012345
   replace Char with replicate(temp,13), log with (ii % 2 == 0)
   replace date with ctod("05/31/95") + ii -1
#ifdef MEMOINCLUDE
*  memo 1 filled in record 1, 4, 5, 8
*  memo 2 filled in record 2, 4, 6, 8
*  no memo       in record 3, 7
   if (ii % 4) <= 1
      replace memo1 with "Memo data 1 for record " + ltrim(str(recno()))
   endif
   if (ii % 4) == 0 .or. (ii % 4) == 2
      replace Memo2 with "Memo data 2" + CRLF + "for record " + ltrim(str(recno()))
   endif
#endif
enddo
use
? "--- end task1,", secondscpu() - start, "seconds"
return

**************************************************************************
procedure task2
* basic driver: open, skip, goto, access fields by name & fieldget()
**************************************************************************
LOCAL_INT ii
LOCAL jj, start, oldErr, loop, cTmp, nTmp
FIELD num, deci, char, log, date, Memo1, Memo2

isok := ""

copy file ORIG_DBF to WORK_DBF
#ifdef MEMOINCLUDE
 copy file (ORIG_DBT) to (WORK_DBT)
#endif

use WORK_DBF
? "--- task2, tested driver " + rddnew + ", alias= " + alias(), " dbf= " + dbf()
printUsed()
if !used()
   ?? chr(7)
   ? "*** ERROR, USE/OPEN database failed"
   return
endif

#ifdef OUTPUT_OBJ_TO_STDERR
cb4cdxtest("task2, open WORK in WA " + ltrim(str(select())))
#endif

display0()
start := secondscpu()
goto 1
display1("go 1:")

loop := 0
while !eof() .and. loop < REC_ORIG
   skip
   display1("r# " + ltrim(str(recno())) + ":")                // RECNO()
   loop++
enddo
display1("e# " + ltrim(str(recno())) + ":")
if !eof()
   ? "*** ERROR, EOF not set"
endif

loop := 0
while !bof() .and. loop < REC_ORIG
   skip -2
   display1("sk-2:")
   loop++
enddo
display1("b# " + ltrim(str(recno())) + ":")
if !bof()
   ? "*** ERROR, BOF not set"
   isok := "NOT "
endif
if recno() # 1 .or. !(num == 1)
   ? "*** ERROR, recno() should be 1"
   isok := "NOT "
endif

goto 5
display1("go 5:")
#ifdef MEMOINCLUDE
 checknempty (memo2, "Memo2 field" )
#endif

goto 4
display1("go 4:")

checkempty (num  , "Num"  + " field")
checkempty (Deci , "Deci" + " field")
checkempty (Char , "Char" + " field")
checkempty (Log  , "Log"  + " field")
checkempty (Date , "Date" + " field")
#ifdef MEMOINCLUDE
 checkempty (memo1, "Memo1"+ " field")
 checkempty (MEMO2, "Memo2"+ " field")
#endif

for ii := 1 to len(struct)
   cTmp = struct[ii,1]
   checkempty (fieldget(ii),          "FieldGet(" + ltrim(str(ii)) + ")")
# ifdef CANFIELDPUTNAME
   checkempty (fieldget(upper(cTmp)), 'FieldGet("' + cTmp + '")')
# endif
next

jj := 4
cTmp := ltrim(str(jj)) + "-"
cTmp := left(replicate(cTmp, 13), struct[3,3])
nTmp := jj * 10000 + jj / 10000 + 0.0000012345
nTmp := round(nTmp, struct[2,4])

?
checkval (num,  jj,   "Num" , .T.)
checkval (deci, nTmp, "Deci", .T.)
checkval (Char, cTmp, "Char", .T.)
checkval (log,  .T.,  "Log" , .T.)
checkval (date, ctod("05/31/95") + jj -1, "Date", .T.)
#ifdef MEMOINCLUDE
 checkval (memo1, "Memo data 1 for record 4", "Memo1", .T.)
 checkval (memo2, "Memo data 2" + CRLF + "for record 4", "Memo2", .T.)
#endif

goto bottom
if recno() # REC_ORIG
   ? "*** ERROR, Record no. should be " + ltrim(str(REC_ORIG)) + ;
     ", but is " + ltrim(str(recno()))
   isok := "NOT "
endif
skip
display1("gEOF:")
if !eof()
   ? "*** ERROR, EOF not set"
   isok := "NOT "
endif
checknempty (num  , "Num"  + " field")
checknempty (Deci , "Deci" + " field")
checknempty (Char , "Char" + " field")
checknempty (Log  , "Log"  + " field")
checknempty (Date , "Date" + " field")
#ifdef MEMOINCLUDE
 checknempty (MEMO1, "Memo1"+ " field")
 checknempty (memo2, "Memo2"+ " field")
#endif

for ii := 1 to len(struct)
   cTmp = struct[ii,1]
   checknempty (fieldget(ii),          "FieldGet(" + ltrim(str(ii)) + ")")
# ifdef CANFIELDPUTNAME
   checknempty (fieldget(lower(cTmp)), 'FieldGet("' + cTmp + '")')
# endif
next

use
if used()
   ? "*** ERROR, USE/CLOSE database failed, database still open"
   isok := "NOT "
endif
?
? "--- end task2,", secondscpu() - start, "seconds"
? "--- x=<field> access, RECNO(), GO TOP, GO BOTT " + isok + "successful tested."
return

**************************************************************************
procedure task3
* basic driver: SET PATH, RLOCK()
*               replace fields by name and by fieldput()
**************************************************************************
LOCAL start, ii, jj, cTmp, nTmp, cMem1, cMem2
FIELD num, deci, char, log, date, Memo1, Memo2

set path to ..\

isok := ""
if file(WORK_DBF)
   delete file (WORK_DBF)
# ifdef MEMOINCLUDE
   delete file (WORK_DBT)
# endif
endif

copy file ORIG_DBF to ("..\" + WORK_DBF)
#ifdef MEMOINCLUDE
 copy file (ORIG_DBT) to ("..\" + WORK_DBT)
#endif
? WORK_DBF + " copied to upper ../ directory"

select 10
use WORK_DBF SHARED
? "--- task3, tested driver " + upper(rddnew) + ", alias " + alias()
? "    dbf = " + dbf() + " SHARED "
# ifdef MEMOINCLUDE
  ?? " dbt = " + WORK_DBT
#endif
#ifdef OUTPUT_OBJ_TO_STDERR
  cb4cdxtest("task3, open WORK in WA 10")
#endif

select WORK_NAME

if select() != 10 .or. !(alias() == upper(WORK_NAME))
   ? "*** ERROR, Alias should be 10 / [" + upper(WORK_NAME) + "], is", ;
     ltrim(str(select())), "/ [" + alias() + "]"
   return
endif
printUsed()
if !used()
   ?? chr(7)
   ? "*** ERROR, USE/OPEN database failed"
   return
endif

start := secondscpu()

goto 5
jj   := 98765
nTmp := -987654.7654321
cTmp := left("abcdefghijklmnopqrstuvwxyz", struct[3,3])
#ifdef MEMOINCLUDE
 cRec5Mem1:= memo1 + CRLF + replicate("x",2500) + "end memo1"
 cRec5Mem2:= memo2 + CRLF + replicate("X",400)  + "end memo2"
#endif

? "Database replacement on SHARED dbf in alias = ", alias(), "... trying RLOCK()"
testrlock(@isok)

if fieldput(1, jj) != jj
   ? "*** ERROR: field 'Num' NOT correctly replaced"
   isok := "NOT "
endif
replace deci  with nTmp
#ifdef CANFIELDPUTNAME
   if ! (fieldput("Char", cTmp) == cTmp)
      ? "*** ERROR: field 'Char' NOT correctly replaced"
      isok := "NOT "
   endif
#else
   if ! (fieldput(3, cTmp) == cTmp)
      ? "*** ERROR: field 'Char' NOT correctly replaced"
      isok := "NOT "
   endif
#endif
replace log   with .t.
replace date  with date()

#ifdef MEMOINCLUDE
  replace memo1 with cRec5Mem1
# ifdef CANFIELDPUTNAME
   if ! (fieldput ("memo2", cRec5Mem2) == cRec5Mem2)
      ? "*** ERROR: field 'Memo2' NOT correctly replaced"
      isok := "NOT "
   endif
# else
   if ! (fieldput (7, cRec5Mem2) == cRec5Mem2)
      ? "*** ERROR: field 'Memo2' NOT correctly replaced"
      isok := "NOT "
   endif
# endif
#endif

testunlock(@isok)

skip 0
checkval (num,   jj,     "Num field")

goto 1
checkval (num,   1,      "Num field")
goto 5
checkval (num,   jj,     "Num" , .T.)
checkval (deci,  nTmp,   "Deci", .T.)
checkval (Char,  cTmp,   "Char", .T.)
checkval (log,   .T.,    "Log" , .T.)
checkval (date,  date(), "Date", .T.)
#ifdef MEMOINCLUDE
 checkval (memo1, cRec5Mem1,  "Memo1", .T.)
 checkval (memo2, cRec5Mem2,  "Memo2", .T.)
#endif

** Database is NOT closed (for task4)

? "--- end task3,", secondscpu() - start, "seconds"
? "--- x=<field>, FIELDGET(), RLOCK(), REPLACE, FIELDPUT() " + isok + "successful tested."
return

**************************************************************************
procedure task4
* basic driver: SET PATH, NEW
*               multiple database usage (different aliases) SHARED/SHARED
**************************************************************************
select 10
if !used()
   ? "*** Perform task 3 first"
   return
endif
#ifdef OUTPUT_OBJ_TO_STDERR
  cb4cdxtest("task4, selected 1st WORK in WA 10")
#endif

? "--- task4"
? "The SAME database " + WORK_NAME + " (open SHARED in WA 10) will be now"
? "opened SHARED in the next free WA by using another alias name"
? "This SHOULD succeed for correct multiuser/multitasking handling"
wait
?

** use WORK_NAME alias second shared NEW
** if used()
if dbusearea(.T., rddnew, WORK_NAME, "second", .T.)
#ifdef OUTPUT_OBJ_TO_STDERR
  cb4cdxtest("task4, open 2nd WORK in WA " + ltrim(str(select())))
#endif
   do test45 with "task4", select(), 55, date()
endif
return

**************************************************************************
procedure task5
* basic driver: multiple database usage (different aliases) SHARED/EXCL
*               object usage
**************************************************************************
local oRDD := NIL
select 10
if !used()
   ? "*** Perform task 3 first"
   return
endif
#ifdef OUTPUT_OBJ_TO_STDERR
  cb4cdxtest("task4, selected 1st WORK in WA 10")
#endif

? "--- task5"
? "The SAME database " + WORK_NAME + " (open EXCLUSIVE in WA 10) will be now"
? "opened EXCLUSIVE in WA 20 by using another alias name"
? "This SHOULD FAIL for correct multiuser/multitasking handling"
wait
?
select 20
** #ifdef FlagShip
**    oRDD := DRIVER_CREATOR (NIL, WORK_DBF, .F.)
**    oRDD:ALIAS := "second"
** #else
   use WORK_DBF via (rddnew) alias second exclusive
** #endif
if used()
#ifdef OUTPUT_OBJ_TO_STDERR
  cb4cdxtest("task5, open 2nd WORK in WA 20")
#endif
   do test45 with "task5", 20, 555, date() + 60
else
   ? "*** ERROR on open"
endif
return


*---------------------------------------------------------------------------
procedure test45 (task, selected, newnum, newdate)
*---------------------------------------------------------------------------
LOCAL start, ii, jj, cTmp, nTmp
FIELD num, deci, char, log, date, Memo1, Memo2

select 10
? "--- " + task + ", tested driver " + upper(rddnew) + ", alias= " + alias()

#ifdef OUTPUT_OBJ_TO_STDERR
   cb4cdxtest("test45 after select 1st")
#endif

select (selected)
? "           tested driver " + upper(rddnew) + ", alias= " + alias()

#ifdef OUTPUT_OBJ_TO_STDERR
   cb4cdxtest("test45 after select 2nd")
#endif

if !used()
   ?? chr(7)
   ? "*** ERROR, second USE/OPEN of the database failed"
   return
endif
if selected == 10 .or. select() != selected .or. !(alias() == "SECOND")
   ? "*** ERROR, Alias " + alias() + " should be " + ltrim(str(selected)) + ;
     " but # 10"
   return
endif

select (WORK_NAME)

if select() != 10 .or. !(alias() == upper(WORK_NAME))
   ? "*** ERROR, Alias " + alias() + " should be 10"
   return
endif
printUsed()
if !used()
   ?? chr(7)
   ? "*** ERROR, USE/OPEN database failed"
   return
endif

start := secondscpu()

select 10
goto 5

jj   := 98765
nTmp := -987654.7654321
cTmp := left("abcdefghijklmnopqrstuvwxyz", struct[3,3])

?
? "--- data of record 5 in WA10, alias =", alias()
checkval (num,   jj,     "Num" , .T.)
checkval (deci,  nTmp,   "Deci", .T.)
checkval (Char,  cTmp,   "Char", .T.)
checkval (log,   .T.,    "Log" , .T.)
checkval (date,  date(), "Date", .T.)
#ifdef MEMOINCLUDE
 checkval (memo1, cRec5Mem1,  "Memo1", .T.)
 checkval (memo2, cRec5Mem2,  "Memo2", .T.)
#endif

select (selected)
?
? "--- the same database and record in WA" + ltrim(str(selected)) + ;
  ", alias =", alias()
goto 2
checkval (num,   2,     "Num field")

goto 5
checkval (num,   jj,     "Num" , .T.)
checkval (deci,  nTmp,   "Deci", .T.)
checkval (Char,  cTmp,   "Char", .T.)
checkval (log,   .T.,    "Log" , .T.)
checkval (date,  date(), "Date", .T.)
#ifdef MEMOINCLUDE
 checkval (memo1, cRec5Mem1,  "Memo1", .T.)
 checkval (memo2, cRec5Mem2,  "Memo2", .T.)
#endif

? "-- replacing NUM with " + ltrim(str(newnum)) + ", DATE with", ;
  newdate, ", trying FLOCK()..."
testflock(@isok)
field->num := newnum
replace date with newdate
#ifdef OUTPUT_OBJ_TO_STDERR
   cb4cdxtest("test45 after replace 2nd")
#endif
testunlock()

checkval (num,   newnum,  "Num" , .T.)
checkval (date,  newdate, "Date", .T.)

?
select (selected)
use
if used()
   ? "*** ERROR, USE/CLOSE of the second database failed, database still open"
endif

? "--- end " + task, secondscpu() - start, "seconds"
return

**************************************************************************
procedure task6
* basic driver: multiple database usage (multiuser) SHARED
**************************************************************************
task67 (" 6", .T., " 8")
return

**************************************************************************
procedure task7
* basic driver: multiple database usage (multiuser) SHARED, FLOCK
**************************************************************************
task67 (" 7", .F., " 9")
return


**************************************************************************
procedure task8
* basic driver: multiple database usage (multiuser) SHARED, RLOCK
* same as task6, but invoked from task6
**************************************************************************
task89 (" 8", .T.)
return

**************************************************************************
procedure task9
* minimal driver: multiple database usage (multiuser) SHARED, FLOCK
* same as task7, but invoked from task7
**************************************************************************
task89 (" 9", .F.)
return

*---------------------------------------------------------------------------
static function task67 (istask, rlock, calltask)
local txt, start, runcmd := execname() + calltask
set path to

select 10
use
if select("second") > 0
   select ("second")
   use
endif

copy file ORIG_DBF to WORK_DBF
# ifdef MEMOINCLUDE
  copy file (ORIG_DBT) to (WORK_DBT)
# endif

txt := if (rlock, "RLOCK", "FLOCK")

? "--- task" + ltrim(istask) + ", tested driver " + upper(rddnew)

start := secondscpu()

select 10
use WORK_DBF via (rddnew) SHARED
if !used()
   ? "** could not open", WORK_DBF
   return
endif

if rlock
   testrlock()
else
   testflock()
endif

? "replacing NUM with 11111, date with", date()
display0()
display1("old :")
replace num with 11111, date with date()
display1("new :")
testunlock()

?
? "The SAME database " + WORK_NAME + " (open " + txt + " in WA 10) is now"
? "tried to open " + txt + " by other user "

#ifdef FlagShip
  ? "by automaticaly executing this applic. using RUN '" + runcmd + "'"
  ? "You may also try to execute '" + runcmd + "' from other terminal now..."
  save screen
#else
  ? 'On DOS, try to execute "RDDTEST1' + calltask + ;
    '" from another user now ...'
#endif

wait

#ifdef FlagShip
  RUN (execname() + calltask)
  restore screen
#endif

if file (FILETRASF)
   ? "------- output from task" + ltrim(calltask) + " -----------------"
   ? memoread (FILETRASF)
   ?
   ? "------- end of output from task" + ltrim(calltask) + " ----------"
   delete file FILETRASF
endif

commit
display0()

display1("NEW :")
use

? "--- end task" + ltrim(istask), secondscpu() - start, "seconds"
return

*---------------------------------------------------------------------------
static function task89 (istask, rlock)
local txt, start
set path to

select 10
use
select 20
use

set print to FILETRASF
set print on

txt := if (rlock, "RLOCK", "FLOCK")

? "--- task" + ltrim(istask) + ", tested driver " + upper(rddnew)
?
? "The already " + txt + " open database " + WORK_NAME + " will be now"
? "opended by second user"

start := secondscpu()

select 10
use WORK_NAME via (rddnew) SHARED

if !used()
   ? "*** Error, database not open"
else
   printused()
   ? "--- Database was opened succesfully. Available data:"
   display0()
   display1("Rec1:")
   ? "      --- Replacing NUM with 22222, date with", date() +10
   if rlock
      testrlock()
   else
      testflock()
   endif
   replace num with 22222, date with date() +10
   testunlock()
   skip 0
   display1("Rec1:")
endif
use

? "--- end task" + ltrim(istask), secondscpu() - start, "seconds"

set print to
quit
return

*---------------------------------------------------------------------------
static function display0()
?
? "      ..num ....num.deci... ........char........... log mm/dd/yy eof bof M1 M2"
return NIL

*---------------------------------------------------------------------------
static function display1(prefix)
FIELD num, deci, char, log, date, Memo1, Memo2
? prefix, num, deci, char, log, date, eof(), bof()
#ifdef MEMOINCLUDE
?? "", str(len(Memo1),2), str(len(Memo2),2)
#endif
return nil

*---------------------------------------------------------------------------
static function checkempty (value, name)
if (empty(value))
   ? "*** ERROR, " + name + " returns: EMPTY"
   return .F.
endif
return .T.

*---------------------------------------------------------------------------
static function checknempty (value, name)
if (! empty(value))
   ? "*** ERROR, " + name + " returns: NOT EMPTY"
   return .F.
endif
return .T.

*---------------------------------------------------------------------------
static function checkval (fldval, value, name, out)
local disp1 := fldval, disp2 := value

if (valtype(fldval) $ "MC" .and. len(fldval) > 35) .or. ;
   (valtype(value)  $ "MC" .and. len(value)  > 35)
   if len(fldval) > 70
      disp1 := left(fldval,34) + "..." + right(fldval,34)
   endif
   if len(value) > 70
      disp2 := left(value, 34) + "..." + right(value, 34)
   endif
endif
if valtype(fldval) $ "MC"
   disp1 := strtran (disp1, CRLF, "<CRLF>")
   disp2 := strtran (disp2, CRLF, "<CRLF>")
endif
if valtype(disp1) == "C" .and. len(disp1 + disp2) > 35
   if fldval == value
      if valtype(out) == "L" .and. out
         ? "dbf: " + padr(name,10) + " correct value [ LEN() = " + ;
            ltrim(str(len(fldval))) + " ] :"
         if col() + len(disp1) + 3 > 79
            ?
         endif
         ?? " [" + disp1 + "]"
      endif
      return .T.
   else
      isok := "NOT "
      ? "*** ERROR, dbf: " + name + " is/should be [ LEN() = " + ;
        ltrim(str(len(fldval))) + "/" + ltrim(str(len(value))) + " ] :"
      ? " [" + disp1 + "]"
      ? " [" + disp2 + "]"
   endif
   return .F.
endif

if fldval == value
   if valtype(out) == "L" .and. out
      ? "dbf: " + padr(name,10) + " correct value [", disp1, "]"
   endif
else
   isok := "NOT "
   ? "*** ERROR, dbf: " + name + " [", disp1, "] should be [", disp2, "]"
   return .F.
endif
return .T.

*---------------------------------------------------------------------------
function testrlock(ok)
local ii
for ii = 1 to 10
   if rlock()
      exit
   endif
   if ii == 10
      ? "*** ERROR: Rlock() failed, tried 10x within 10 seconds"
      if valtype(ok) == "C"
         ok := "NOT "
      endif
   endif
   inkey(1)
next
#ifdef OUTPUT_OBJ_TO_STDERR
cb4cdxtest("after RLOCK() from " + procname(1)+"/"+ltrim(str(procline(1)))+;
            ", recno=" + ltrim(str(recno())))
#endif
return

*---------------------------------------------------------------------------
function testflock(ok)
local ii
for ii = 1 to 10
   if flock()
      exit
   endif
   if ii == 10
      ? "*** ERROR: Flock() failed, tried 10x within 10 seconds"
      if valtype(ok) == "C"
         ok := "NOT "
      endif
   endif
   inkey(1)
next
#ifdef OUTPUT_OBJ_TO_STDERR
cb4cdxtest("after FLOCK() from " + procname(1)+"/"+ltrim(str(procline(1)))+;
            ", recno=" + ltrim(str(recno())))
#endif
return

*---------------------------------------------------------------------------
function testunlock(ok)
unlock
if neterr()
   ? "*** ERROR: unlock failed"
   isok := "NOT "
endif
#ifdef OUTPUT_OBJ_TO_STDERR
cb4cdxtest("after UNLOCK from " + procname(1)+"/"+ltrim(str(procline(1)))+;
            ", recno=" + ltrim(str(recno())))
#endif

*---------------------------------------------------------------------------
function PrintUsed()
local currArr := select()
local ii, jj := 0
for ii := 1 to 250
   select (ii)
   if used()
      ? "Select WA " + str(ii,3) + ":", alias()
      jj++
   endif
next
select (currArr)
return jj

*----------------------------------Clipper -> FS----------------------------
#ifndef FlagShip
function strpeek (str,pos)
return asc(substr(str,pos,1))
function strpoke (str,pos,char)
str = stuff(str,pos,1,chr(char))
return str
function memoext()
return ".dbt"
FUNCTION ExecName()
RETURN "rddtest2.exe"
#endif

*** eof rddtest1.prg ***

