#ifdef COMMENT_ONLY
* For FlagShip 5 and 6
* Copyright (c) 2003..2007 by multisoft, Munich, Germany
*
* You may modify this source according to your needs and distribute the
* object code freely with your executable. You may distribute also the
* modified source code when your application is distributed in source,
* but may not remove the copyright notice.
*
* Release $Date: 2007/12/12 16:47:33 $
*
* FoxPro interface for FlagShip 5 and 6, available in the FS library.
* May be used via stdfoxpro.fh, i.e. with the -fox compiler switch.
*
* If you wish to modify this file,
*   Compile without (!) the -fox switch: FlagShip -c -na -m foxpro_api.prg
*   Link on Linux/Unix: FlagShip myapp*.prg -fox foxpro_api.o
*   Link in MS-Windows: FlagShip myapp*.prg -fox foxpro_api.obj
#endif

#Cinline
# include "FSopenc.h"
#endCinline

#include "set.fh"

static _fox_n_onPage := 0  as intvar
static _fox_c_doPage := "" as char

/*---------------------------------------------------------------------------
 * fox_AtMenu(<r>,<c>,<arr>,<size1>,<size2>,<title>) -> arr[6]
 *
 * stores the data  from @..MENU arr... into array for Achoice()
 * triggered by READ MENU
 * returns an array: [1..4] = row1, col1, row2, col2
 *                   [5]    = array of strings
 *                   [6]    = title or NIL
 */
FUNCTION fox_AtMenu(row,col,arr,elem,[height],[title])
   local retArr[6], ii, nSize, newArr

   retArr[1] := minmax(0, row, maxrow() -3)        // top row
   retArr[2] := minmax(0, col, maxcol() -3)        // left column
   retArr[6] := title                              // title or NIL
   // re-assign array if not equivalent
   if elem > 0 .and. elem < len(arr)
      newArr := array(elem)
      for ii := 1 to elem
         newArr[ii] := arr[ii]
      next
      retArr[5] := aclone(newArr)
   else
      retArr[5] := aclone(arr)
   endif
   // calculate rigth column from text and title length
   nSize := if(valtype(title) == "C", StrLen2col(title), 0)
   for ii := 1 to len(retArr[5])
      nSize := max(nSize, StrLen2col(retArr[5,ii]))
   next
   retArr[4] := round(retArr[2] + nSize + 0.9, 0)
   // calculate bottom row considering <height> and <title>
   retArr[3] := retArr[1] + ;
                if(valtype(height) == "N" .and. height > 0, ;
                   height -1, len(retArr[5]) -1)
   if AppIoMode() == "G"
      if valtype(title) == "C"
         retArr[3] += 2.7     // fix bottom size
      endif
      if valtype(height) == "N" .and. height > 0
         retArr[4] += 2       // fix columns for scrollbar
      endif
   endif
   return retArr

/* --------------------------------------------------------------------------
 * fox_Browse() -> NIL
 *
 * command BROWSE, equivalent to Browse() function
 */
FUNCTION fox_Browse()
   Browse()
   return NIL

/*---------------------------------------------------------------------------
 * fox_DelMenu(oTop1 [, oTop2, oTop3, ...] )
 * = DEACTIVATE MENU <menu1> [, <menu2> ...]
 */
FUNCTION fox_DelMenu(objTopBar)     // parameters dynamically encountered
   local ii, iStep, iMax, obj

   for iStep := 1 to pcount()
      obj := param(iStep)
      if valtype(obj) == "O" .and. ascan(IsObjClass(obj), {|x| x == "_TOPBAR"}) > 0
         iMax := obj:ItemCount
         for ii := iMax to 1 step -1
            obj:DelItem(ii)
         next
      endif
   next
   // oTopBar:Display()  // this is in stdfoxpro.fh
   return

/* --------------------------------------------------------------------------
 * fox_FieldPut(cField, uValue) -> NIL
 *
 * similar to fieldput(), but accepts alias in <cField>
 */
FUNCTION fox_FieldPut(cField, uValue)
   cField := strtran(@cField, ".", "->")
   if "->" $ cField
      &(cField) := uValue
   else
      fieldput(cField, uValue)
   endif
   return NIL

/* --------------------------------------------------------------------------
 * fox_MdiOpen(r1,c1,r2,c2,capt) -> nWindow
 *
 * check if FS2 Tools is available (i.e. Wopen() called directly, or
 * EXTERN Wopen specified). If so, use Wopen() otherwise MdiOpen()
 */
FUNCTION fox_MdiOpen(r1,c1,r2,c2,capt)
   local retVar := -1 as intVar
   local cMacro as usual

   if AppMdiMode()
      retVar := MdiOpen(r1,c1,r2,c2,capt)
   elseif IsFunction("WOPEN")
      /*
       * Cannot call here WOpen(r1,c1,r2,c2,NIL,capt) directly,
       * since the linker would report "unresolved externals"
       * when the FS2 Toolbox is not available. Therefore
       * this more complicated call by maco:
       */
      cMacro := "WOpen(" + ltrim(r1) + "," + ;
                           ltrim(c1) + "," + ;
                           ltrim(r2) + "," + ;
                           ltrim(c2) + "," + ;
                           "NIL," + ;
                           if(valtype(capt)=="C","'" + capt + "')","NIL)")
      retVar := &(cMacro)  // == WOpen(r1,c1,r2,c2,NIL,capt)
   else
      alert("To be able to use the DEFINE WINDOW command;" + ;
            "(in FUNCTION " + procname(1) + " " + procfile(1) + ".prg/" + ltrim(procline(1)) + ;
            ");you either need to compile GUI based application with -mdi switch,;" + ;
            "or use the FS2 Toolbox and use -DFS2 switch;" + ;
            "Ignoring windowing now.")
   endif
   return retVar

/* --------------------------------------------------------------------------
 * fox_MdiClose(id) -> 0 or -1
 *
 * check if FS2 Tools is available (i.e. Wopen() called directly, or
 * EXTERN Wopen specified). If so, use Wclose() otherwise MdiClose()
 */
FUNCTION fox_MdiClose(idNum)
   local retVar := -1 as intVar
   local cMacro as usual

   if AppMdiMode()
      retvar := MdiClose(idNum)
   elseif IsFunction("WOPEN")
      /*
       * Cannot call here WClose(idNum) directly,
       * since the linker would report "unresolved externals"
       * when the FS2 Toolbox is not available. Therefore
       * this more complicated call by maco:
       */
      cMacro := "WClose(" + ltrim(idNum) + ")"
      retVar := &(cMacro)  // == Wclose(idNum)
   endif
   return retVar

/* --------------------------------------------------------------------------
 * fox_MdiSelect(idNum) -> 0 or -1
 *
 * check if FS2 Tools is available (i.e. Wopen() called directly, or
 * EXTERN Wopen specified). If so, use Wclose() otherwise MdiClose()
 */
FUNCTION fox_MdiSelect(idNum)
   local retVar := -1 as intVar
   local cMacro as usual

   if AppMdiMode()
      retvar := MdiSelect(idNum)
   elseif IsFunction("WOPEN")
      /*
       * Cannot call here WClose(idNum) directly,
       * since the linker would report "unresolved externals"
       * when the FS2 Toolbox is not available. Therefore
       * this more complicated call by maco:
       */
      cMacro := "WSelect(" + ltrim(idNum) + ")"
      retVar := &(cMacro)  // == Wselect(idNum)
   endif
   return retVar

/* --------------------------------------------------------------------------
 * fox_menuBar(cCapt, cnItem, nRow, nCol, lSkip) -> oMenuItem
 * = helper for DEFINE BAR with AT... or SKIP [FOR...]
 */
FUNCTION fox_menuBar(cCapt, cnItem, nRow, nCol, lSkip) 
   return fox_menuItem(cCapt, cnItem, nRow, nCol, lSkip) 

/* --------------------------------------------------------------------------
 * fox_menuPad(cCapt, cnItem, nRow, nCol, lSkip) -> oMenuItem
 * = helper for DEFINE PAD with AT... or SKIP [FOR...]
 */
FUNCTION fox_menuPad(cCapt, cnItem, nRow, nCol, lSkip) 
   return fox_menuItem(cCapt, cnItem, nRow, nCol, lSkip) 

/* --------------------------------------------------------------------------
 * fox_menuItem(cCapt, cnItem, nRow, nCol, lSkip) -> oMenuItem
 * 
 * note: lSkip = .T. means disable item
 */
FUNCTION fox_menuItem(cCapt, cnItem, nRow, nCol, lSkip) 
   local oMenuItem as usual
   oMenuItem := MenuItem{cCapt, cnItem}
   if valtype(nRow) $ "NI" .and. nRow >= 0
      oMenuItem:Row(nRow)
   endif
   if valtype(nCol) $ "NI" .and. nCol >= 0
      oMenuItem:Column(nCol)
   endif
   if valtype(lSkip) == "L"
      oMenuItem:Enabled := !lSkip
   endif
   return oMenuItem
 
/*---------------------------------------------------------------------------
 * fox_OnPage([nLine], [cProc]) -> NIL
 * = ON PAGE [AT LINE nLine] [DO cProc as codeblock]
 */
FUNCTION fox_OnPage(nLine, cbProc)
   if pcount() == 0 .or. (empty(nLine) .and. empty(cbProc))
      _fox_n_onPage := 0
      _fox_c_doPage := ""
   else
      if valtype(nLine) == "N"
         _fox_n_onPage := nLine
      endif
      if valtype(cbProc) == "C"
         _fox_c_doPage := cbProc
      endif
   endif
   return

/*---------------------------------------------------------------------------
 * fox_ReadMenu(arr, [@saveVar]) -> nRet
 * = READ MENU TO <var> [SAVE]
 */
FUNCTION fox_ReadMenu(arr, [saveVar])
   local retVar, r1, r2, c1, c2 as intVar
   local cColor

   // draw border (with title) for Terminal i/o with "border" color
   if AppIoMode() == "T"
      cColor := SetColor()
      ColorSelect(2)       // use border (third pair) color
      r1 := arr[1] ++
      c1 := arr[2] ++
      r2 := ++arr[3]
      c2 := arr[4]

      @ r1, c1 TO r2 +1, c2 +1
      if valtype(arr[6]) == "C"
         @ r1, c1 + 1 + (c2 - c1 - len(arr[6])) / 2 SAY arr[6]
      endif
      SetColor(cColor)
   endif

   if pcount() == 2
      return Achoice(arr[1], arr[2], arr[3], arr[4], arr[5], ;
                     , , , ,  @saveVar, , arr[6])
   endif
   retVar := Achoice(arr[1], arr[2], arr[3], arr[4], arr[5], ,,,,,, arr[6])
   return retVar

/*---------------------------------------------------------------------------
 * fox_PopupExec(obj) -> nSelect
 *
 * Exec/process PopUp - for FoxPro ACTIVATE POPUP only
 * simulated by achoice()
 */
FUNCTION fox_PopupExec(obj)
   local ii, oMenuItem, iLngTxtPix := 0 as usual
   local iRet                           as intvar
   local aAllow, aOut                   as array

   m->fox_caption := ""

   if valtype(obj) != "O" .or. !IsObjClass(obj,"_popup")
      return 0
   endif

   obj:Display()
   /*
    * simulate popup by Achoice() if no TopBar available
    */
   if obj:oParent != NIL
      return 0
   endif

   for ii := 1 to obj:ItemCount
      oMenuItem := obj:GetItem(ii)
      if valtype(oMenuItem) != "O"
         aadd(aOut, chr(1))
      endif
      if oMenuItem:IsSeparator()
         aadd(aOut, chr(1))
      else
         aadd(aOut, oMenuItem:Caption)
         iLngTxtPix := max(iLngTxtPix, StrLen2space(oMenuItem:Caption))
      endif
   next
   aAllow := array(len(aOut))
   afill(aAllow, .T.)

   for ii := 1 to len(aOut)
      if aOut[ii] == chr(1)
         aOut[ii] := space(iLngTxtPix)
         aAllow[ii] := .F.
      endif
   next

   if valtype(obj:Top) != "N" .or. obj:Top < 0
      obj:Top := 0
   endif
   if valtype(obj:Left) != "N" .or. obj:Left < 0
      obj:Left := 0
   endif
   if valtype(obj:Bottom) != "N" .or. obj:Bottom < 0
      obj:Bottom := minmax(3, len(aOut), maxrow(.T.) - obj:Top) + obj:Top
   endif
   if valtype(obj:Right) != "N" .or. obj:Right < 0
      obj:Right := minmax(3, iLngTxtPix, maxcol(.T.) - obj:Left) + obj:Left
   endif

   iRet := achoice(obj:Top, obj:Left, obj:Bottom, obj:Right, ;
                   aOut, aAllow,,,,@obj:FoxWidget, .T., ;
                   obj:FoxHeader, obj:FoxFooter )   // coord in pixels
   if iRet > 0
      oMenuItem := obj:GetItem(iRet)
      if valtype(obj:FoxDbRec) == "A" .and. len(obj:FoxDbRec) > 0 .and. iRet <= len(obj:FoxDbRec)
         goto (obj:FoxDbRec[iRet])
      endif
      if valtype(oMenuItem:Data) == "B"
         eval(oMenuItem:Data)
      endif
      m->fox_caption := oMenuItem:Caption
   endif
   return iRet

/*---------------------------------------------------------------------------
 * fox_Prompt() -> cMenu
 *
 * return caption of currently selected prompt
 */
FUNCTION fox_Prompt()
   local retVar as char
   #Cinline
   {
      extern FSvar *_foxPrompt;     /* in inglobal.c, set in omenu_g.prg  */
      if(_foxPrompt && IS_VAR_CHR(_foxPrompt))
         SET_VAR_COPY(VAR_NAME_LOCAL(retvar), _foxPrompt);
   }
   #endCinline
   return retVar

/* --------------------------------------------------------------------------
 * fox_printjob(onOff) -> nCopies or NIL
 *
 * onOff = .T. or NIL
 *    This function is invoked from PRINTJOB command.
 *    It turn output to printer and returns number of copies from _PCOPIES
 * onOff = .F.
 *    This function is invoked from ENDPRINTJOB command.
 *    It restores the console and printer status
 */
FUNCTION fox_PrintJob(onOff)
   static pr_console := SET(_SET_CONSOLE) as usual
   static pr_device  := SET(_SET_DEVICE)  as usual
   static pr_print   := SET(_SET_PRINTER) as usual
   local retVar := 1 as intVar

   if valtype(onOff) == "L" .and. !onOff
      SET(_SET_CONSOLE, pr_console)
      SET(_SET_DEVICE,  pr_device)
      SET(_SET_PRINTER, pr_print)
      return NIL
   endif

   if type("_PCOPIES") == "N"
      retVar := max(1, m->_PCOPIES)
   endif

   pr_console := SET(_SET_CONSOLE, .F.)
   pr_device  := SET(_SET_DEVICE,  "printer")
   pr_print   := SET(_SET_PRINTER, .T.)

   if type("_PSCODE") == "C"
      ?? m->_PSCODE        // printer init sequence
   endif
   if type("_PEJECT") == "C"
      if upper(m->_PEJECT) == "BEFORE" .or. upper(m->_PEJECT) == "BOTH"
         eject
      endif
   endif
   @ 0,0 SAY ""   // sets pcol() and prow() to 0
   return retVar

/* --------------------------------------------------------------------------
 * fox_Qout(...) -> NIL
 *
 * FoxPro's ? command incl. AT, PICTURE, FUNCTION, FONT, STYLE clauses
 * and considers ON PAGE AT LINE DO ...
 * sets FoxPro internal variables
 */
FUNCTION fox_Qout(col, pict, fun, color, guicolor, cFont, nFont, cnStyle, exp1)
   local iCurLine

   ? ""   // print new line

   if type("_PageNo") == "U"
      public _PageNo             // FoxPro internal
   endif
   if valtype(m->_PageNo) != "N"
      _PageNo := 0
   endif

   if type("_PLineNo") == "U"
      public _PLineNo            // FoxPro internal
   endif
   if valtype(m->_PLineNo) != "N"
      _PLineNo := 0
   endif
   _PLineNo ++

   if type("_PColNo") == "U"
      public _PColNo             // FoxPro internal
   endif
   _PColNo := 0

   if !empty(_fox_c_doPage)         // consider ON PAGE AT LINE DO ...
      iCurLine := if(set(_SET_PRINTER), prow(), row())
      if iCurLine >= _fox_n_onPage
         eval(&(_fox_c_doPage))     // code block set by ON PAGE DO ...
      endif
   endif
#Cinline
   // need to consider and pass variable parameter count
   UDF_EXEC(fox_qqout)(parno, parptr);
#endCinline
   return NIL

/* --------------------------------------------------------------------------
 * fox_Qqout(...) -> NIL
 *
 * FoxPro's ?? command incl. AT, PICTURE, FUNCTION, FONT, STYLE clauses
 * sets FoxPro internal variables
 */
FUNCTION fox_Qqout(col, pict, fun, color, guicolor, cFont, nFont, cnStyle, exp1)
   local ii, iCurCol, oFont, cPict := ""
   local aExp := {}

   if type("_PColNo") == "U"
      public _PColNo             // FoxPro internal
   endif
   if valtype(m->_PColNo) != "N"
      _PColNo := 0
   endif

   if pcount() < 9
      return NIL  // nothing to print, no exp1... given
   endif
   if valtype(col) == "N"              // AT clause
      iCurCol := if(set(_SET_PRINTER), pcol(), col())
      if iCurCol < col
         ?? space(col - iCurCol)
      endif
   endif
   if valtype(pict) == "C"             // PICTURE clause
      cPict := pict
   elseif valtype(fun) == "C"          // FUNCTION clause
      cPict := "@" + fun
   endif
   if valtype(cFont) == "C" .or. ;
         valtype(nFont) == "N" .or. ;
         !empty(cnStyle)               // FONT or STYLE clause
      oFont := FontNew(cFont, nFont)
      if valtype(cnStyle) == "C"
         cnStyle := upper(alltrim(cnStyle))
         if "B" $ cnStyle
            oFont:Bold := .T.
         endif
         if "I" $ cnStyle
            oFont:Italic := .T.
         endif
         if "U" $ cnStyle
            oFont:Underline := .T.
         endif
         if "-" $ cnStyle
            oFont:StrikeThru := .T.
         endif
      endif
   endif
   if !empty(cPict)
      for ii := 9 to pcount()
         aadd(aExp, transform(param(ii), cPict))
      next
   endif
#Cinline
   {
      // need to consider and pass variable parameter count
      int   ii, iExp, myArgc;
      FSvar *myArgv[23];         // here: max. exp1...exp20 (or extend)

      iExp   = parno - 8;        // number of expressions to print
      myArgc = 3;                // first 3 param of QQout3()
      myArgv[0] = parptr[3];     // VAR_NAME_LOCPAR(color);
      myArgv[1] = parptr[4];     // VAR_NAME_LOCPAR(guicolor);
      myArgv[2] = VAR_NAME_LOCAL(ofont);

      for(ii=0; ii < iExp && ii < 20; ii++) {
         myArgc++;
         if(IS_VAR_EMPTY(VAR_NAME_LOCAL(cpict)))
            myArgv[3 + ii] = parptr[8 + ii];
         else
            myArgv[3 + ii] = VAR_ARRELEM(VAR_NAME_LOCAL(aexp), ii+1);
      }
      // call QQout3(color,guicolor,font,exp1,[exp2,exp3,...,exp20])
      UDF_EXEC(qqout3)(myArgc, myArgv);
   }
#endCinline

   _PColNo := if(set(_SET_PRINTER), pcol(), col())

   return NIL

/* --------------------------------------------------------------------------
 * fox_EjectPage() -> NIL
 *
 * FoxPro's EJECT PAGE, sets FoxPro variables
 */
FUNCTION fox_EjectPage()
   local iCurLine

   EJECT   // std.eject

   if type("_PageNo") == "U"
      public _PageNo             // FoxPro internal
   endif
   if valtype(m->_PageNo) != "N"
      _PageNo := 0
   endif
   _PageNo ++

   if type("_PLineNo") == "U"
      public _PLineNo            // FoxPro internal
   endif
   _PLineNo := 0

   if type("_PColNo") == "U"
      public _PColNo             // FoxPro internal
   endif
   _PColNo := 0
   return NIL

/* --------------------------------------------------------------------------
 * fox_seleMax() -> num
 *    simulates the FoxPro SELECT(1) function (seldomly used) to
 *    determine highest unused working area
 */
FUNCTION fox_seleMax()
   local currSel := DbSelectArea()
   local ret := 65535

   while ret > 0
      DbSelectArea(ret)
      if !used()
         exit
      endif
      ret--
   enddo
   DbSelectArea(currSel)
   return ret  // 65535..1, 0 signals failure

/* --------------------------------------------------------------------------
 * fox_set(cType[,1])    -> retValue
 * fox_set(nType [,val]) -> retValue
 *
 * simulates the most of FoxPro's SET() functions
 */
FUNCTION fox_set(p1, p2)
   local p2set, cType4
   static allOff := ",AUTO,BELL,BLIN,BRST,CARR,CLEA,CLOC,COMP,DEBU,DOHI" + ;
                    ",ECHO,FIEL,FULL,HEAD,HELP,HIST,LOGE,MENU,NEAR,NOTI" + ;
                    ",ODOM,OPTI,PALE,RESO,SAFE,SHAD,STAT,STEP,STIC,SYSM" + ;
                    ",TALK,TEXT,"

   if valtype(p1) == "N"   // FlagShip set(...)
      if p2 == NIL
         return set(p1)
      endif
      return set(p1,p2)
   endif

   if valtype(p1) != "C" .or. empty(p1)
      return .F.
   endif
   cType4 := left(upper(alltrim(p1)),4)

   if ("," + cType4 + ",") $ allOff
      return "OFF"
   endif
   p2set := valtype(p2) == "N" .and. p2 == 1

   do case
   case cType4 == "ALTE"
      if p2set
         return set(_SET_ALTFILE)
      endif
      return if(set(_SET_ALTERNATE), "ON", "OFF")
   case cType4 == "ANSI"
      return if(set(_SET_ANSI), "ON", "OFF")
   case cType4 == "BLOC"
      return ""
   case cType4 == "CENT"
      return if(set(_SET_CENTURY), "ON", "OFF")
   case cType4 == "COLO"
      return set(_SET_COLOR)
   case cType4 == "CONF"
      return set(_SET_CONFIRM)
   case cType4 == "CONS"
      return set(_SET_CONSOLE)
   case cType4 == "CURR"
      return "LEFT"
   case cType4 == "CURS"
      return if(set(_SET_CURSOR) == 0, "OFF", "ON")
   case cType4 == "DATE"
      return set(_SET_DATEFORMAT)
   case cType4 == "DECI"
      return set(_SET_DECIMALS)
   case cType4 == "DEFA"
      return set(_SET_DEFAULT)
   case cType4 == "DELE"
      return if(set(_SET_DELETED), "ON", "OFF")
   case cType4 == "DELI"
      if p2set
         return set(_SET_DELIMCHARS)
      endif
      return if(set(_SET_DELIMITERS), "ON", "OFF")
   case cType4 == "DEVE"
      return if(fs_set("devel"), "ON", "OFF")
   case cType4 == "DEVI"
      return set(_SET_DEVICE)
   case cType4 == "ESCA"
      return if(set(_SET_ESCAPE), "ON", "OFF")
   case cType4 == "EXAC"
      return if(set(_SET_EXACT), "ON", "OFF")
   case cType4 == "EXCL"
      return if(set(_SET_EXCLUSIVE), "ON", "OFF")
   case cType4 == "FILT"
      return if(empty(dbSetFilter()), "OFF", "ON")
   case cType4 == "FIXE"
      return if(set(_SET_FIXED), "ON", "OFF")
   case cType4 == "HOUR"
      return 24
   case cType4 == "INTE"
      return if(set(_SET_INTENSITY), "ON", "OFF")
   case cType4 == "KEYC"
      return "WINDOWS"
   case cType4 == "LOCK"
      return if(set(_SET_AUTOLOCK), "ON", "OFF")
   case cType4 == "MARG"
      return set(_SET_MARGIN)
   case cType4 == "MARK"
      return ""
   case cType4 == "MEMO"
      return 80
   case cType4 == "MESS"
      return set(_SET_MESSAGE)
   case cType4 == "MOUS"
      return if(AppIoMode() == "G", "ON", "OFF")
   case cType4 == "MULT"
      return if(set(_SET_MULTILOCKS), "ON", "OFF")
   case cType4 == "PRIN"
      if p2set
         return set(_SET_PRINTFILE)
      endif
      return if(set(_SET_PRINTER), "ON", "OFF")
   case cType4 == "PROC"
      return procfile()
   case cType4 == "SCOR"
      return if(set(_SET_SCOREBOARD), "ON", "OFF")
   case cType4 == "SEPA"
      return "."
   case cType4 == "SPAC"
      return "ON"
   case cType4 == "TOPI"
      return ""
   case cType4 == "TYPE"
      return set(_SET_TYPEAHEAD)
   case cType4 == "UDFP"
      return "REFERENCE"  // when called by DO ..., otherwise "VALUE"
   case cType4 == "UNIQ"
      return if(set(_SET_UNIQUE), "ON", "OFF")
   endcase
   return .F.

/* --------------------------------------------------------------------------
 * fox_sys(num [,par2 [,par3]]) -> retValue or .F. on error
 *
 * simulates the most of FoxPro's SYS() functions
 */
FUNCTION fox_sys(par1, par2, par3)
   static nStoneAge := 2415021
   local  dDate := ctod(""), aTmp := {}
   local  cTmp, iPos, iTmp

   do case
   case valtype(par1) != "N"
      return .F.
   case par1 == 0                                  // SYS(0)
      return NetName()
   case par1 == 1                                  // SYS(1)
      /* Microsoft's idea of "julian date" starts in
       * stone age, approx in year 4700 B.C. :-)
       */
      return int(date() - ctod("01/01/1900") + nStoneAge)
   case par1 == 2                                  // SYS(2)
      return Seconds()
   case par1 == 3                                  // SYS(3)
      // it is better to use standard FlagShip function:
      * return TempFileName(,"_")

      /* This is exact FoxPro sys(3) algorithm, it returns digits only, but
       * the first char is here  underscore to allow it use for dbf/alias.
       * If you wish to be 100% FoxPro compatible, use sys(3,"") or
       * modify the code below. You may pass any prefix string of your
       * choice in parameter 2, e.g. sys(3,"tmp") which adds the prefix
       * to the returned value.
       */
      par2 := if(valtype(par2) != "C", "_", alltrim(par2))
      * par2 := ""    // uncomment for 100% FoxPro compatibility
      iTmp := int(seconds() * 1000)
      while file(par2 + ltrim(iTmp))
         iTmp++
      enddo
      return par2 + ltrim(iTmp)
   case par1 == 5                                  // SYS(5)
      return Set(_SET_DEFAULT)
   case par1 == 6                                  // SYS(6)
      return Set(_SET_PRINTFILE)
   case par1 == 10 .and. valtype(par2) == "N"      // SYS(10,n)
      return dtoc(par2 - nStoneAge + ctod("01/01/1900"))
   case par1 == 11                                 // SYS(11,c|d)
      if valtype(par2) == "C"
         dDate := ctod(par2)
      elseif valtype(par2) == "D"
         dDate := par2
      endif
      if empty(dDate)
         return 0          // error return on invalid date
      endif
      return int(dDate - ctod("01/01/1900") + nStoneAge)
   case par1 == 12                                 // SYS(12)
      return 655360        // default MS value
   case par1 == 13                                 // SYS(13)
      return if(IsPrinter(),"ONLINE","OFFLINE")
   case par1 == 14                                 // SYS(14,[c|n])
      if !empty(par2)
         return (par2)->(IndexKey())
      endif
      return IndexKey()
   case par1 == 16                                 // SYS(16,[n])
      /* MS-FoxPro stack starts with 0 or 1 in main, 2 is the next depth.
       * FlagShip's starts with 0 in current procedure, 1 is one above
       * (caller) etc., so we emulate here the raising FoxPro callstack
       * and return string "PROCEDURE <proc> in <file>" like FoxPro.
       */
      if valtype(par2) != "N" .or. par2 < 0
         aadd(aTmp, {procName(1), procFile(1)} )
         iPos := 1
      else
         iPos := 1  // skip this fox_sys() function, 0 will include it
         cTmp := ProcName(iPos)
         while !empty(cTmp)
            aadd(aTmp, {cTmp, procFile(iPos)} )
            cTmp := ProcName(++iPos)
         enddo
         if par2 == 0
            par2 := 1   // 0 and 1 are equivalent
         endif
         if par2 > len(aTmp)
            return ""
         endif
         iPos := minmax(1, len(aTmp) - par2 +1, len(aTmp))  // reverse
      endif
      return "PROCEDURE " + upper(aTmp[iPos,1]) + ;
             " in " + aTmp[iPos,2] + ".prg"
   case par1 == 18                                 // SYS(18)
      return ReadVar()
   case par1 == 21                                 // SYS(21)
      if !used()
         return 0
      endif
      return IndexOrd()
   case par1 == 22                                 // SYS(22,[n])
      if !empty(par2) .and. valtype(par2) == "N"
         return IndexKey(par2)
      endif
      return IndexKey(0)
   case par1 == 23                                 // SYS(23)
      return 0             // default MS value
   case par1 == 30                                 // SYS(30)
      cTmp := NetName()
      return left(cTmp, at("@",cTmp) -1)
   case par1 == 100                                // SYS(100)
      return if(Set(_SET_CONSOLE),"ON","OFF")
   case par1 == 101                                // SYS(101)
      return Set(_SET_DEVICE)
   case par1 == 102                                // SYS(102)
      return if(Set(_SET_PRINTER),"ON","OFF")
   case par1 == 103                                // SYS(103)
      return "OFF"         // default
   case par1 == 1001                               // SYS(1001)
      return 655360        // default MS value, same as sys(12)
   case par1 == 2001                               // SYS(2001,c)
      do case
      case valtype(par2) != "C"
         return "-error par2-"
      case left(upper(par2),4) == "ALTE"
         if valtype(par3) == "N"
            return set(_SET_ALTFILE)
         endif
         return if(set(_SET_ALTERNATE), "ON", "OFF")
      case left(upper(par2),4) == "CENT"
         return if(set(_SET_CENTURY), "ON", "OFF")
      case left(upper(par2),4) == "CONF"
         return if(set(_SET_CONFIRM), "ON", "OFF")
      case left(upper(par2),4) == "CONS"
         return if(set(_SET_CONSOLE), "ON", "OFF")
      case left(upper(par2),4) == "DEFA"
         return set(_SET_DEFAULT)
      case left(upper(par2),4) == "DEVI"
         return set(_SET_DEVICE)
      case left(upper(par2),4) == "EPOC"
         return set(_SET_EPOCH)
      case left(upper(par2),4) == "ESCA"
         return if(set(_SET_ESCAPE), "ON", "OFF")
      case left(upper(par2),4) == "EXAC"
         return if(set(_SET_EXACT), "ON", "OFF")
      case left(upper(par2),4) == "EXCL"
         return if(set(_SET_EXCLUSIVE), "ON", "OFF")
      case left(upper(par2),4) == "FIXE"
         return if(set(_SET_FIXED), "ON", "OFF")
      case left(upper(par2),4) == "PATH"
         return set(_SET_PATH)
      case left(upper(par2),4) == "PRIN"
         if valtype(par3) == "N"
            return set(_SET_PRINTFILE)
         endif
         return if(set(_SET_PRINTER), "ON", "OFF")
      case left(upper(par2),4) == "UNIQ"
         return if(set(_SET_UNIQUE), "ON", "OFF")
      case left(upper(par2),4) == "WRAP"
         return if(set(_SET_WRAP), "ON", "OFF")
      endcase
      alert(procstack() + ";unsupported SYS(2001,'" + par2 + "') option" )
      return ""
   case par1 == 2002                               // SYS(2002,[c|n])
      if !empty(par2)
         return SetCursor(1)
      endif
      return SetCursor(0)
   case par1 == 2003                               // SYS(2003)
      return CurDir()
   case par1 == 2004                               // SYS(2004)
      alert(procstack(1) + ;
            ";Assign CURDIR() to Public or Private variable" + ;
            ";at the begin of your main program, then use  " + ;
            ";this variable instead of SYS(2004)")
      return ""
   case par1 == 2010                               // SYS(2010)
      return 255           // default MS value
   case par1 == 2011                               // SYS(2011)
      return if(IsDbExcl(),"Exclusive", ;
                  if(IsDbFlock(),"File Locked", ;
                     if(IsDbRlock(),"Record locked","Not locked")))

   endcase
   alert(procstack() + ";unsupported SYS(" + ltrim(par1) + ") option")
   return .F.

/* --------------------------------------------------------------------------
 * fox_showget(cVarName, aGetList, [lEnable], [lDisable], [cColor]) -> ok
 *
 * simulates SHOW GET
 */
FUNCTION fox_showget(cVarName, aGetList, lEnable, lDisable, cColor)
   local ii, iPos := 0
   if empty(cVarName) .or. empty(aGetList)
      return .F.
   endif
   for ii := 1 to len(aGetList)
      if upper(aGetList[ii]:Name) == upper(cVarName)
         iPos := ii
         exit
      endif
   next
   if iPos > 0 .and. iPos <= len(aGetList)
      if valtype(lEnable) == "L"
         aGetList[iPos]:GetEnabled := lEnable
      endif
      if valtype(lDisable) == "L"
         aGetList[iPos]:GetEnabled := !lDisable
      endif
      if valtype(cColor) == "C" .and. "," $ cColor 
         aGetList[iPos]:ColorSpec := cColor
         if AppIoMode() == "G"
            aGetList[iPos]:GuiColor := cColor
         endif
      endif
      return .T. 
   endif
   return .F.
      
/* --------------------------------------------------------------------------
 * fox_Warning(cText) -> NIL
 *
 * used in stdfoxpro.fh to display unsupported commands
 */
FUNCTION fox_Warning(cText)
   alert("*** Compatibility warning in;FUNCTION " + ;
         procname(1) + " " + procfile(1) + ".prg/" + ltrim(procline(1)) + ;
         ";" + cText)
   return NIL

* eof *
