#ifdef COMMENT_ONLY
/* For FlagShip 5 and 6.
 * Copyright (c) 1994..2007 by multisoft GmbH, Munich, Germany
 * You may modify this source according to your needs and distribute the
 * object code freely with your executable. You may distribute also the
 * modified source code when your application is distributed in source,
 * but may not remove the copyright notice.
 *
 * Release $Date: 2007/11/29 15:52:49 $
 */
#endif

****************************************************************************
*  User modifiable functions used in NT2/CA3 for handling serial           *
*  communication. Note, FlagShip support 16 (or more) ports instead of 4   *
*  in CA3, but usually there are only 2-4 ports initialized by the system. *
*                                                                          *
*  Note: When FS_SET("devel", .T.) is active, you will receive additional  *
*  iformation on unsupported functions.                                    *
*                                                                          *
*  If you need to change this file, copy it into your local directory,     *
*  make changes and recompile by using                                     *
*     FlagShip -c -m -na -delc fs2_com.prg                                 *
*  and then link it with your application                                  *
*     Unix/Linux: FlagShip -delc myprog*.prg fs2_com.o                     *
*     MS-Windows: FlagShip -delc myprog*.prg fs2_com.obj                   *
*                                                                          *
*  Debugging your changes: simply add -d compiler switch when compiling    *
*  the modified fs2_com.prg and your mytest.prg, invoke the executable,    *
*  load fs2_com.prg in GUI debugger, set breakpoints and execute...        *
*                                                                          *
*  C debugging in Unix/Linux: if you want to debug the C inline parts by   *
*  Unix C debugger like gdb, cv, xldb etc, your should:                    *
*     cd <my_working_dir>                                                  *
*     cp /usr/local/FlagShip6/system/fs2_com.prg .                         *
*     FlagShip fs2_com.prg [-DTESTPRINT] -c -m -na -g [-d]                 *
*     FlagShip myprog*.prg fs2_com.o [-d]                                  *
*     gdb a.out (or: ddd a.out)                                            *
*     list _bb_fs2_termio  ...etc  to debug the fs2_termio() function      *
*     br 3218              ...etc, see help                                *
*     r   [c, n, s...]                                                     *
*     quit                                                                 *
*                                                                          *
*  C debugging in MS-Windows: if you wish to debug C inline parts by       *
*  MS-VC6 debugger, you may either                                         *
*  - compile your main and fs2_com sources to .c by using -b FS compiler   *
*    switch and add the .c files (including the mytest_m.c start-up        *
*    source) into new MSVC project, or simply                              *
*  - create the executable as usual by FlagShip with -g switch, e.g.       *
*       FlagShip fs2_com.prg [-DTESTPRINT] -c -m -na -g [-d]               *
*       FlagShip mytest.prg -m -g fs2_com.obj [-d]                         *
*    then invoke "msdev mytest.exe", load fs2_com.c, set corresponding     *
*    breakpoints there, and execute to breakpoint...                       *
*                                                                          *
****************************************************************************
* Differences and extentions to CA3 (see also text in the FN header below)
*
*                     FS2/Windows                 FS2/Linux
*
*  com_break ()       supported                   returns fix .F.
*  com_close ()       supported                   supported
*  com_count ()       supported                   supported
*  com_crc   ()       supported/differs           supported/differs
*  com_cts   ()       supported                   returns fix .T.
*  com_dcd   ()       supported                   returns fix .T.
*  com_dsr   ()       supported                   returns fix .T.
*  com_dtr   ()       supported                   returns fix .T.
*  com_doscon()       returns fix ""              returns fix ""
*  com_errchr()       supported                   returns fix .F.
*  com_error ()       supported                   supported
*  com_event ()       returns fix -1              returns fix -1
*  com_flush ()       supported                   supported
*  com_getio ()       supported                   supported
*  com_getirq()       returns file handler        returns file handler
*  com_hard  ()       supported                   supported
*  com_init  ()       supported, extended         supported, extended
*  com_init2 ()       FS2 only, n/a CA3           FS2 only, n/a CA3
*  com_isopen()       FS2 only, n/a CA3           FS2 only, n/a CA3
*  com_key   ()       returns fix .F.             returns fix .F.
*  com_lsr   ()       supported                   supported
*  com_mcr   ()       returns fix 0               returns fix 0
*  com_msr   ()       returns fix 0               returns fix 0
*  com_num   ()       supported                   supported
*  com_open  ()       supported, extended         supported, extended
*  com_read  ()       supported, extended         supported, extended
*  com_remote()       returns fix .F.             returns fix .F.
*  com_ring  ()       returns fix .T.             returns fix .T.
*  com_rts   ()       supported                   returns fix .T.
*  com_scount()       supported                   supported
*  com_send  ()       supported                   supported
*  com_setio ()       returns fix .F.             returns fix .F.
*  com_setirq()       returns fix .F.             returns fix .F.
*  com_sflush()       supported                   supported
*  com_skey  ()       returns fix .F.             returns fix .F.
*  com_smode ()       supported                   supported
*  com_soft  ()       supported                   supported
*  com_soft_r()       supported                   supported
*  com_skey_s()       returns fix .F.             returns fix .F.
*  com_timeout()      FS2 only, n/a CA3           fix {0,0,0,0,0}
****************************************************************************

#Cinline
#include "FSopenc.h"
#include <malloc.h>
#include <string.h>

static char *fileID="@@# $RCSfile: fs2_com.prg,v $ $Revision: 1.13 $ $Date: 2007/11/29 15:52:49 $";

/* example
 * test helper to print bits in zeroinsert() or zeroremove()
 * active and available only when compiled with -DTESTPRINT
 */
#ifdef TESTPRINT
void printbuff(char *txt, unsigned char cc, unsigned char *buff, int size)
{
   int ii;
   unsigned char *ptr;

   fprintf(stderr, "\n%s Char=%.2X Buff=", txt, cc);
   for (ptr=buff; ptr < buff + size; ptr++) {
     for (ii=7; ii >= 0; ii--)
        fprintf(stderr, "%c", ((*ptr & (1 << ii)) ? '1' : '0'));
     fprintf(stderr, "\n");
   }
}
#endif
#endCinline

#include "set.fh"

#ifdef FS_WIN32
/* ************************* MS-Windows part only ***************************
 */
#Cinline
#include "FSopenc.h"
#include "windows.h"
#include "winbase.h"

#ifndef FS_WIN32
#define FS_WIN32
#endif
#define MAXPORTS  16                   /* must correspond to MAXPORTS in .prg part */
static HANDLE aHANDLE[MAXPORTS +1];    // used are [1]..[16]

/*------------------------------------------------------------------------
 * nFd := _WinOpen(nPort, cPortName, nBuffSize)
 *   performs OPEN operation for Windows, similar to FOPEN()
 *   note: no parameter check here, must match!
 */
FSstaticudf(_winopen)
{
   int  iPort      = VAR_INTNUM(argv[0]); // port number 1..16
   char *cPortName = VAR_CHR(argv[1]);    // port name e.g. "COM2"
   int  iBuffSize  = VAR_INTNUM(argv[2]); // buffer size
   int  ok, iFd = -1;                     // file descriptor
   COMMTIMEOUTS CommTimeouts;             // (Win32)

   if(iPort < 1 || iPort > MAXPORTS)
      return SET_VAR_INT(VAR_NEW, -1);

   aHANDLE[iPort] = CreateFile(cPortName,                          /* (Win32) port name "COM1".."COMnn" */
                               GENERIC_READ | GENERIC_WRITE,       /* read/write */
                               FILE_SHARE_READ | FILE_SHARE_WRITE, /* 0 for exclusive-access, or: FILE_SHARE_READ | FILE_SHARE_WRITE */
                               NULL,                               /* no security attributes */
                               OPEN_EXISTING,                      /* must use OPEN_EXISTING */
                               0,                                  /* not overlapped i/o */
                               0 );
   if(aHANDLE[iPort] != INVALID_HANDLE_VALUE) {
      /*
       * either no time-out at all (e.g. with direct access)
       */
//    CommTimeouts.ReadIntervalTimeout = MAXDWORD;
//    CommTimeouts.ReadTotalTimeoutMultiplier  = 0;
//    CommTimeouts.ReadTotalTimeoutConstant    = 0;
//    CommTimeouts.WriteTotalTimeoutMultiplier = 0;
//    CommTimeouts.WriteTotalTimeoutConstant   = 0;

      /* or small time-out delays for common speed
       * The ReadTotalTimeoutConstant is also the dalay when
       * no chars are available on port, before returning 0
       *
       * You may change timeouts later by the com_timeout() function
       */
      CommTimeouts.ReadIntervalTimeout         = 100;       // stop if 100 msec elapses between two received bytes
      CommTimeouts.ReadTotalTimeoutMultiplier  = 0;         // no length sensitivity
      CommTimeouts.ReadTotalTimeoutConstant    = 100;       // allow 100 (or 400) msec for a read (reset command can take a while)
      CommTimeouts.WriteTotalTimeoutMultiplier = 0;         // no length sensitivity
      CommTimeouts.WriteTotalTimeoutConstant   = 200;       // allow 200 msec for a write

      ok = SetCommTimeouts(aHANDLE[iPort], &CommTimeouts);  // set timeouts (Win32)
      SetupComm(aHANDLE[iPort], iBuffSize, iBuffSize);      // set buffers  (Win32)

      iFd = _get_osfhandle(aHANDLE[iPort]);   // file descriptor (Win32)
      if(iFd <= 0)                            // used for init check only,
         iFd = 999;                           // don't use it otherwise
   }
   return SET_VAR_INT(VAR_NEW, iFd);
}

/*------------------------------------------------------------------------
 * lOk := _WinClose(nPort)
 *   performs CLOSE operation for Windows, similar to FCLOSE()
 *   note: no parameter check here, must match!
 */
FSstaticudf(_winclose)
{
   int iPort = VAR_INTNUM(argv[0]);          // port number 1..16

   if(iPort < 1 || iPort > MAXPORTS || aHANDLE[iPort] == INVALID_HANDLE_VALUE)
      return FALSE_VAR;
   CloseHandle(aHANDLE[iPort]);              // (Win32)
   aHANDLE[iPort] = INVALID_HANDLE_VALUE;
   return TRUE_VAR;
}

/*------------------------------------------------------------------------
 * nRread := _WinRead(nPort, @cBuff, nSize, @nErr)
 *   performs READ operation for Windows, similar to FREAD()
 *   note: no parameter check here, must match!
 */
FSstaticudf(_winread)
{
   int  iPort  = VAR_INTNUM(argv[0]);  // port number 1..16
   char *cBuff = VAR_CHR(argv[1]);     // ptr to buffer
   int   iSize = VAR_INTNUM(argv[2]);  // bytes to read
   DWORD iRet  = 0;                    // bytes read

   if(iPort >= 1 && iPort <= MAXPORTS && aHANDLE[iPort] != INVALID_HANDLE_VALUE) {
      if(! ReadFile(aHANDLE[iPort], cBuff, iSize, &iRet, 0) ) {   // (Win32)
         iRet = 0;
         SET_VAR_INT(argv[3], GetLastError());                    // (Win32)
      } else
         SET_VAR_INT(argv[3], 0);
   }
   return SET_VAR_INT(VAR_NEW, iRet);
}

/*------------------------------------------------------------------------
 * nWrit := _WinWrite(nPort, @cBuff, nSize, @nErr)
 *   performs WRITE operation for Windows, similar to FWRITE()
 *   note: no parameter check here, must match!
 */
FSstaticudf(_winwrite)
{
   int  iPort  = VAR_INTNUM(argv[0]);  // port number 1..16
   char *cBuff = VAR_CHR(argv[1]);     // ptr to buffer
   int  iSize  = VAR_INTNUM(argv[2]);  // bytes to write
   FSvar *vError = argv[3];            // ptr to error status
   DWORD iRet   = 0;                    // bytes written

   if(iPort >= 1 && iPort <= MAXPORTS && aHANDLE[iPort] != INVALID_HANDLE_VALUE) {
      if(! WriteFile(aHANDLE[iPort], cBuff, iSize, &iRet, 0) ) {   // (Win32)
         iRet = 0;
         SET_VAR_INT(vError, GetLastError());                      // (Win32)
      } else
         SET_VAR_INT(vError, 0);
   }
   return SET_VAR_INT(VAR_NEW, iRet);
}

/*------------------------------------------------------------------------
 * int _winCheckSpeed(int request)
 *   set max available baud rate according to request
 */
static int _winCheckSpeed(int request)
{
   int ret = CBR_9600;

   if(request >= 110   ) ret = CBR_110;
   if(request >= 300   ) ret = CBR_300;
   if(request >= 600   ) ret = CBR_600;
   if(request >= 1200  ) ret = CBR_1200;
   if(request >= 2400  ) ret = CBR_2400;
   if(request >= 4800  ) ret = CBR_4800;
   if(request >= 9600  ) ret = CBR_9600;
   if(request >= 14400 ) ret = CBR_14400;
   if(request >= 19200 ) ret = CBR_19200;
   if(request >= 38400 ) ret = CBR_38400;
   if(request >= 56000 ) ret = CBR_56000;
   if(request >= 57600 ) ret = CBR_57600;
   if(request >= 115200) ret = CBR_115200;
   if(request >= 128000) ret = CBR_128000;
   if(request >= 256000) ret = CBR_256000;
   return ret;
}

/*------------------------------------------------------------------------
 * int _winCheckParity(FSvar *parity)
 *   return windows parity mark
 */
static int _winCheckParity(FSvar *vParity)
{
   int  iParity = NOPARITY;
   char cParity = 'N';

   if(vParity && IS_VAR_CHR(vParity)) {
      cParity = ((char *)VAR_CHR(vParity))[0];
      cParity = toupper(cParity);
      switch(cParity) {
         case 'E' : iParity = EVENPARITY; break;
         case 'O' : iParity = ODDPARITY;  break;
         case 'N' :
         default  : iParity = NOPARITY;   break;
      }
   }
   return iParity;
}

/*------------------------------------------------------------------------
 * void _winDisplayDcb("comment", DCB *dcb)
 *   display DCB structure (test only)
 */
static void _winDisplayDcb(char *txt, DCB *dcb)
{
fprintf(stderr,
"%s: BaudRate=%d ByteSize=%d Parity=%d\n\
  StopBits=%d fBinary=%d fParity=%d fOutxCtsFlow=%d fOutxDsrFlow=%d\n\
  fDtrControl=%d fRtsControl=%d fDsrSensitivity=%d fNull=%d\n\
  fInX=%d fOutX=%d fTXContinueOnXoff=%d XonLim=%d XoffLim=%d\n\
  XonChar=%d XoffChar=%d ErrorChar=%d EofChar=%d EvtChar=%d\n\
  fErrorChar=%d fAbortOnError=%d\n",
txt, dcb->BaudRate, dcb->ByteSize, dcb->Parity,
dcb->StopBits, dcb->fBinary, dcb->fParity, dcb->fOutxCtsFlow, dcb->fOutxDsrFlow,
dcb->fDtrControl, dcb->fRtsControl, dcb->fDsrSensitivity, dcb->fNull,
dcb->fInX, dcb->fOutX, dcb->fTXContinueOnXoff, dcb->XonLim, dcb->XoffLim,
dcb->XonChar, dcb->XoffChar, dcb->ErrorChar, dcb->EofChar, dcb->EvtChar,
dcb->fErrorChar, dcb->fAbortOnError);
}

#endCinline
#endif /* FS_WIN32 */

*************************** common *****************************************
#include "set.fh"

#define MAXPORTS  16
#define BUFFSIZE  1024

static anFileHd[MAXPORTS]            // file handle for COM1 ... COM16
static acBuffer[MAXPORTS]            // for com_count(), com_read()
static anBufLen[MAXPORTS]            // valid chars in cBuffRead    (0, 1..n)
static anBufPos[MAXPORTS]            // current pointer in cBuffRead (0..n-1)
static anLastEr[MAXPORTS]            // last r/w error on this port (if > 0)
static asStatus[MAXPORTS]            // saved port status on com_open()
static anBufSiz[MAXPORTS]            // allocated buffer size or NIL

static MSG_PORT0   := "Port manipulation not possible"
static MSG_PORT    := MSG_PORT0 + ", returning "
static MSG_COMINIT := "Port manipulation not possible, use;" + ;
                      "com_init() or com_init2(). Returning "
static MSG_DOSANSI := "DOS ANSI.SYS is not available"


*****************************************************************************
* com_break(nPort, [nTime]) -> lOk
*
* Creates break on a transmission line
*
* THIS FUNCTION IS NOT APPLICABLE ON UNIX
*   Default value .F. is returned
* In MS-Windows environment, this function is supported. It interrupts the
*   transmission for <nTime> milliseconds, where the default is 100 ms.
*----------------------------------------------------------------------------
FUNCTION com_break(nPort, nTime)
local_int iPort := nPort
local_int iTime := if(valtype(nTime) == "N", nTime, 0)
#ifdef FS_WIN32
   if com_isopen(nPort)
      #Cinline
      {
         int iSleepMs = itime > 0 ? itime : 100;
    if( SetCommBreak(aHANDLE[iport]) ) {
       Sleep(iSleepMs);
       ClearCommBreak(aHANDLE[iport]);
    }
      }
      #endCinline
      return .T.
   endif
#else
   fs2_devWarn(MSG_PORT + ".F.")
#endif
return .F.

*****************************************************************************
* com_close(nPort) -> lOk
*
* Clears the receiving buffer and closes the com port
*----------------------------------------------------------------------------
FUNCTION com_close(nPort)
local ok
if !com_isopen(nPort)
   fs2_devWarn("Port " + ltrim(nPort) + " is not open")
   return .F.
endif
if valtype(asStatus[nPort]) != "U"
   fs2_termio(-2, nPort, @asStatus[nPort])  // restore old status
endif
#ifdef FS_WIN32
   ok := _WinClose(nPort)
#else
   ok := FCLOSE(anFileHd[nPort])
#endif
if ok
   anFileHd[nPort] := 0
   anBufLen[nPort] := 0
   anBufPos[nPort] := 0
   anLastEr[nPort] := 0
   acBuffer[nPort] := NIL
   asStatus[nPort] := NIL
   anBufSiz[nPort] := NIL
endif
return ok

*****************************************************************************
* com_count(nPort) -> nChar
*
* Counts the number of characters in the input buffer
*
* nPort: 1 to 16 for COM1 to COM16
* nChar: >0 data in the device buffer are available,
*        =0 no data in COM buffer available
*        -1 error (port not open)
*
* This function is tightly related to com_read(). It reads
* up to 1k characters from the device into internal buffer to
* be subsequentely be read via com_read().
*----------------------------------------------------------------------------
FUNCTION com_count(nPort)
if !com_isopen(nPort)
   return -1
endif
if anBufLen[nPort] > 0 .and. anBufPos[nPort] < anBufLen[nPort]
   return anBufLen[nPort] - anBufPos[nPort]
endif
// ??## time(1), "Com_Count() start at",procstack()   // -test-
#ifdef FS_WIN32
   anBufLen[nPort] := _WinRead(nPort, @acBuffer[nPort], len(acBuffer[nPort]), @anLastEr[nPort])
#else
   anBufLen[nPort] := FREAD(anFileHd[nPort], @acBuffer[nPort], len(acBuffer[nPort]))
   anLastEr[nPort] := FERROR()
#endif
// ??## time(1), "Com_Count() end, bytes available=",ltrim(anBufLen[nPort])  // -test-

anBufPos[nPort] := 0
return anBufLen[nPort]

*****************************************************************************
* com_crc(cString, [nStart], [nPoly]) -> nCRC
*
* Computes CRC value for the string <cString>
*
* cString : the string for which the CRC value is computed
* nStart  : optional, value added to resulting nCRC
* nPoy    : optional, used Polynomial (any int value > 0)
*           once set, the polynomial remains active until a new nPoly is set
* nCRC    : resulting CRC value
*
* XMODEM (default)      nStart = 0,     nPoly = 40961
* X25                   nStart = 0,     nPoly =  4129
* CCITT                 nStart = 65535, nPoly =  4129
* ARC                   nStart = 0,     nPoly = 32773
*
* CA3: Note, the NT3/CA3 algorithm is compatible only to istelf!
* FS2: FlagShip follows the international CRC16 standard.
*----------------------------------------------------------------------------
* function com_crc(cString, nStart, nPoly)
*     this function is available in the FS2 library and use
*     Crc16(cString, [nStart], [nPoly])

*****************************************************************************
* com_cts(nPort) -> lActive
*
* CA3: Queries the Clear To Send (CTS) status
* FS2: THIS FUNCTION AND MECHANISM IS NOT APPLICABLE ON UNIX
*        Therefore, default value .T. is returned here
*        To set the RTS/CTS auto handshake, use com_init()
*      In MS-Windows environment, this function is supported.
*----------------------------------------------------------------------------
FUNCTION com_cts(nPort)
local lRet := .T.
#ifdef FS_WIN32
   if !com_isopen(nPort)
      return .F.
   else
   #Cinline
   {
      int  iStatus, iPort = VAR_INTNUM(VAR_NAME_LOCPAR(nport));
      DWORD lModemStat = MS_CTS_ON;

      iStatus = GetCommModemStatus(aHANDLE[iPort], &lModemStat);
      lModemStat &= MS_CTS_ON;
      SET_VAR_LOG(VAR_NAME_LOCAL(lret), lModemStat);
   }
   #endCinline
   endif
#else
   fs2_devWarn(MSG_COMINIT + ".T.")
#endif
return lRet

*****************************************************************************
* com_dcd(nPort) -> lActive
*
* CA3: Queries the Data Carrier Detect (DCD) status
* FS2: THIS FUNCTION AND MECHANISM IS NOT APPLICABLE ON UNIX
*        Therefore, default value .T. is returned here
*        To set the RTS/CTS auto handshake, use com_init()
*      In MS-Windows environment, this function is supported.
*----------------------------------------------------------------------------
FUNCTION com_dcd(nPort)
local lRet := .T.
#ifdef FS_WIN32
   if !com_isopen(nPort)
      return .F.
   else
   #Cinline
   {
      int iStatus, iPort = VAR_INTNUM(VAR_NAME_LOCPAR(nport));
      DWORD lModemStat = MS_RLSD_ON;

      iStatus = GetCommModemStatus(aHANDLE[iPort], &lModemStat);
      lModemStat &= MS_RLSD_ON;
      SET_VAR_LOG(VAR_NAME_LOCAL(lret), lModemStat);
   }
   #endCinline
   endif
#else
   fs2_devWarn(MSG_COMINIT + ".T.")
#endif
return lRet

*****************************************************************************
* com_dsr(nPort) -> lActive
*
* CA3: Queries the Data Set Ready (DSR) status
* FS2: THIS FUNCTION AND MECHANISM IS NOT APPLICABLE ON UNIX
*        Therefore, default value .T. is returned here
*        To set the RTS/CTS auto handshake, use com_init()
*      In MS-Windows environment, this function is supported.
*----------------------------------------------------------------------------
FUNCTION com_dsr(nPort)
local lRet := .T.
#ifdef FS_WIN32
   if !com_isopen(nPort)
      return .F.
   else
   #Cinline
   {
      int iStatus, iPort = VAR_INTNUM(VAR_NAME_LOCPAR(nport));
      DWORD lModemStat = MS_DSR_ON;

      iStatus = GetCommModemStatus(aHANDLE[iPort], &lModemStat);
      lModemStat &= MS_DSR_ON;
      SET_VAR_LOG(VAR_NAME_LOCAL(lret), lModemStat);
   }
   #endCinline
   endif
#else
   fs2_devWarn(MSG_COMINIT + ".T.")
#endif
return lRet

*****************************************************************************
* com_dtr(nPort, [lnNew]) -> lActive
*
* CA3: Queries or set the Data Terminal Ready (DTR) status
* FS2: THIS FUNCTION AND MECHANISM IS NOT APPLICABLE ON UNIX
*        Therefore, default value .T. is returned here
*        To set the RTS/CTS auto handshake, use com_init()
*      In MS-Windows environment, this function is supported. If <lnNew>
*        is logical, DTR is set on/off, same as CA3.
*        FS2: if <lnNew> is numeric, 0=off, 1=on, 2 (or other num)=handshake
*----------------------------------------------------------------------------
FUNCTION com_dtr(nPort, lnNew)
local lRet := .T.
#ifdef FS_WIN32
   if !com_isopen(nPort)
      return .F.
   endif
   #Cinline
   {
      DCB dcb;
      int iOldSet = 0, iNew;
      int iPort = VAR_INTNUM(VAR_NAME_LOCPAR(nport));

      if( GetCommState(aHANDLE[iPort], &dcb) ) {
         iOldSet = dcb.fDtrControl != DTR_CONTROL_DISABLE;
         if(IS_VAR_LOG(VAR_NAME_LOCPAR(lnnew))) {
            dcb.fDtrControl = IS_VAR_TRUE(VAR_NAME_LOCPAR(lnnew)) ?
                                 DTR_CONTROL_ENABLE : DTR_CONTROL_DISABLE;
            SetCommState(aHANDLE[iPort], &dcb);
         } else if(IS_VAR_NUM(VAR_NAME_LOCPAR(lnnew))) {
            iNew = VAR_INTNUM(VAR_NAME_LOCPAR(lnnew));
            dcb.fDtrControl = (iNew == 0 ? DTR_CONTROL_DISABLE :
                                 (iNew == 1 ? DTR_CONTROL_ENABLE :
                                    DTR_CONTROL_HANDSHAKE));
            SetCommState(aHANDLE[iPort], &dcb);
         }
      }
      SET_VAR_LOG(VAR_NAME_LOCAL(lret), iOldSet);
   }
   #endCinline
#else
   fs2_devWarn(MSG_COMINIT + ".T.")
#endif
return lRet

*****************************************************************************
* com_doscon(cString,[nLine],[nColumn]) -> ""
*
* CA3: Provides screen output through DOS (ANSI.SYS terminal emulation)
* FS2: THIS FUNCTION AND MECHANISM IS NOT APPLICABLE ON UNIX
*        Therefore, default value .T. is returned here
*        To set the RTS/CTS auto handshake, use com_init()
*      This function is also not applicable in MS-Windows
*----------------------------------------------------------------------------
FUNCTION com_doscon(cString, nLine, nColumn )
fs2_devWarn(MSG_DOSANSI)
return ""

*****************************************************************************
* com_errchr(nPort, [ncChar]) -> lChanged
*
* CA3: Defines replacement for a "trash" character
* FS2: THIS FUNCTION AND MECHANISM IS NOT APPLICABLE ON UNIX
*        Therefore, default value .F. is returned here
*      In MS-Windows environment, this function is supported.
*----------------------------------------------------------------------------
FUNCTION com_errchr(nPort, ncChar)
local lRet := .T.
#ifdef FS_WIN32
   if !com_isopen(nPort)
      return .F.
   else
   #Cinline
   {
      DCB dcb;
      char cNew;
      int iOk = 0, iSet = 1;
      int iPort = VAR_INTNUM(VAR_NAME_LOCPAR(nport));
      FSvar *vPar2 = VAR_NAME_LOCPAR(ncchar);

      if( GetCommState(aHANDLE[iPort], &dcb) ) {
         iOk = 1;
         if(IS_VAR_CHR(vPar2))
            cNew = ((char *)VAR_CHR(vPar2))[0];
         else if(IS_VAR_NUM(vPar2))
            cNew = VAR_INTNUM(vPar2);
         else
            iSet = 0;
         if(iSet) {
            dcb.ErrorChar = cNew;
            SetCommState(aHANDLE[iPort], &dcb);
         }
      }
      SET_VAR_LOG(VAR_NAME_LOCAL(lret), iOk);
   }
   #endCinline
   endif
#else
   fs2_devWarn(MSG_PORT + ".F.")
   lRet = .F.
#endif
return lRet

*****************************************************************************
* com_error(nPort) -> nError
*
* FS2: Reports the last r/w error, if any
*      Not applicable for com_open() and com_close() in Unix, where
*      FERROR() should be used instead.
* CA3: not available, use com_lsr
*
* nError = last error (see the FERROR function) on this port
*----------------------------------------------------------------------------
FUNCTION com_error(nPort)
if !com_isopen(nPort)
   return  0
endif
return anLastEr[nPort]

*****************************************************************************
* com_event(nPort, nMode) -> nRet
*
* CA3: Event triggering of a port
* FS2: THIS FUNCTION AND MECHANISM IS NOT APPLICABLE ON UNIX
*         Therefore, default value -1 is returned here
*      This function is not applicable in MS-Windows
*----------------------------------------------------------------------------
FUNCTION com_event(nPort, nMode)
fs2_devWarn(MSG_PORT + "-1")
return -1

*****************************************************************************
* com_flush(nPort) -> lOk
*
* Flushes the internal buffer of com_count()
* However, it does not flush the rest of incomming chars on the serial line
*----------------------------------------------------------------------------
FUNCTION com_flush(nPort)
local_int iPort := nPort
if !com_isopen(nPort)
   return .F.
endif
#ifdef FS_WIN32
   #Cinline
      FlushFileBuffers(aHANDLE[iport]);
   #endCinline
#endif
anBufLen[nPort] := 0
anBufPos[nPort] := 0
return .T.

*****************************************************************************
* com_getio(nPort) -> nFileHd
*
* FS2: Determines the file handle used for an open <nPort>
* CA3: returns the I/O port address instead
*----------------------------------------------------------------------------
FUNCTION com_getio(nPort)
if !com_isopen(nPort)
   return -1
endif
return anFileHd[nPort]

*****************************************************************************
* com_getirq(nPort) -> nFileHd
*
* FS2: Determines the file handle used for an open <nPort>
* CA3: returns the IRQ address instead
*----------------------------------------------------------------------------
FUNCTION com_getirq(nPort)
if !com_isopen(nPort)
   return -1
endif
return anFileHd[nPort]

*****************************************************************************
* com_hard(nPort, [lOn], [lDtr]) -> lOldRts
*
* CA3: Sets the DTR/DSR or RTS/CTS handshake
* FS2: only RTS/CTS handshake (lDtr == .F.) is supported in Unix & Windows
*----------------------------------------------------------------------------
FUNCTION com_hard(nPort, lOn, lDtr)
local lRet := .F.
if com_isopen(nPort)
#ifdef FS_WIN32
   lRet := fs2_termio(1, nPort, lOn, lDtr)
#else
   lRet := fs2_termio(1, nPort)
   if valtype(lOn) == "L"
      fs2_termio(1, nPort, lOn)
   endif
#endif
endif
return lRet

*****************************************************************************
* com_init(nPort, [nBaud], [cParity], [nDataLng], [nStop], ;
*          [lCtsRts], [lLocal], [lCanon]) -> lOk
*
* Sets the serial port attributtes
*
* nPort   : 1 to 16, corresponding to com_open()
* nBaud   : 300, 1200, 2400, 4800, (9600), 19200, 38400, 57600, 115200, 23040
* cParity : ("N[o]") , "E[ven]", "O[dd]"
* nDataLng: 5, 6, 7 or (8)
* nStop   : 0, (1), 2
* lCtsRts : enable (.T.) or disable [.F.] CTS/RTS
* lLocal  : local  (.T.) or dial-up [.F.] line
* lCanon  : canonical [.T.] inut processing = editeable or non-canoncal (.F.)
*
* If the optional parameter are not given or are NIL, the default values
* given above in parentheses () are used. So when all optional parameters
* are omited, the default setting
*     com_init(nPort, 9600, "N", 8, 1, .T., .T., .F.)
* is used.
*
* CA3: The default speed is 300 baud instead 9600 in FS2. In CA3, only
*      the first five parameters are supported.
*----------------------------------------------------------------------------
FUNCTION com_init(nPort, nBaud, cParity, nDataLng, nStop, lCtsRts, lLocal, lCanon)
local lRet   := .F.

if com_isopen(nPort)
   nBaud    := if(valtype(nBaud) $ "NI", nBaud, 9600)
   cParity  := if(valtype(cParity) == "C", upper(left(cParity,1)), "N")
   cParity  := if(cParity $ "NEO", cParity, "N")
   nDataLng := if(valtype(nDataLng) $ "NI", nDataLng, 8)
   nDataLng := if(ascan({5,6,7,8}, nDataLng) > 0, nDataLng, 8)
   nStop    := if(valtype(nStop) $ "NI", nStop, 1)
   nStop    := if(ascan({0,1,2}, nStop) > 0, nStop, 1)
   lCtsRts  := if(valtype(lCtsRts) == "L", lCtsRts, .T.)
   lLocal   := if(valtype(lLocal)  == "L", lLocal,  .T.)
   lCanon   := if(valtype(lCanon)  == "L", lCanon,  .F.)

   lRet := fs2_termio(0, nPort, nBaud, nDataLng, cParity, nStop, ;
                      lCtsRts, lLocal, lCanon)
endif
return lRet

*****************************************************************************
* com_init2(nPort, [nBaud], [cParity], [nDataLng], [nStop]) -> lOk
*
* Sets the serial port attributtes
*
* nPort   : 1 to 16, corresponding to com_open()
* nBaud   : 300, 1200, 2400, 4800, (9600), 19200, 38400, 57600, 115200, 23040
* cParity : ("N[o]") , "E[ven]", "O[dd]"
* nDataLng: 5, 6, 7 or (8)
* nStop   : 0, (1), 2
*
* As opposed to com_init() where all the port parameters are initialized,
* com_init2() sets only the specified parameters, i.e. not NIL.
*
* CA3: not available
*----------------------------------------------------------------------------
FUNCTION com_init2(nPort, nBaud, cParity, nDataLng, nStop)
local lRet   := .T.

if com_isopen(nPort)
   if valtype(nBaud) == "N" .and. !empty(nBaud)
#ifndef FS_WIN32
       fs2_devWarn("Setting general serial speed. To initialize;" + ;
                   "specific port, use com_init() instead.")
#endif
       lRet  := fs2_termio(2, nPort, nBaud) == nBaud
   endif
   if valtype(cParity) == "C" .and. !empty(cParity) .and. lRet
       lRet  := fs2_termio(3, nPort, if(upper(left(cParity,1)) == "N", 0, 1))
   endif
   if valtype(nDataLng) == "N" .and. !empty(nDataLng) .and. lRet
      lRet  := fs2_termio(4, nPort, nDataLng)
   endif
   if valtype(nStop) == "N" .and. lRet
      lRet  := fs2_termio(5, nPort, nStop)
   endif
endif
return lRet

*****************************************************************************
* com_isopen(nPort) -> lOk
*
* Checks if the port is already open
*
* CA3: This function is not available in CA3
*----------------------------------------------------------------------------
FUNCTION com_isopen(nPort)
if valtype(nPort) $ "NI" .and. ;
   nPort > 0 .and. nPort <= len(anFileHd) .and. ;
   valtype(anFileHd[nPort]) $ "NI" .and. anFileHd[nPort] > 0 .and. ;
   valtype(acBuffer[nPort]) == "C" .and. len(acBuffer[nPort]) > 0
   return .T.
endif
return .F.

*****************************************************************************
* com_key(nPort, [nKey1], [nKey2]) -> lTrap
*
* CA3: Monitors port triggering
* FS2: THIS FUNCTION AND MECHANISM IS NOT APPLICABLE ON UNIX nor in Windows
*      Therefore, default value .F. is returned here
*----------------------------------------------------------------------------
FUNCTION com_key(nPort, nKey1, nKey2)
fs2_devWarn(MSG_PORT + ".F.")
return .F.

*****************************************************************************
* com_lsr(nPort) -> nError
*
* FS2: Reports the last error, if any
* CA3: Reports the Status Register (LSR)
*
* nError = last error (see the FERROR function) on this port
*----------------------------------------------------------------------------
FUNCTION com_lsr(nPort)
return com_error(nPort)

*****************************************************************************
* com_mcr(nPort) -> nRet
*
* CA3: Reads or sets the Modem Control Register (MCR)
* FS2: THIS FUNCTION AND MECHANISM IS NOT APPLICABLE ON UNIX nor in Windows
*      Therefore, default value 0 is returned here
*----------------------------------------------------------------------------
FUNCTION com_mcr(nPort)
fs2_devWarn(MSG_PORT + "0")
return  0

*****************************************************************************
* com_msr(nPort) -> nRet
*
* CA3: Reads or sets the Modem Status Register (MSR)
* FS2: THIS FUNCTION AND MECHANISM IS NOT APPLICABLE ON UNIX nor in Windows
*      Therefore, default value 0 is returned here
*----------------------------------------------------------------------------
FUNCTION com_msr(nPort)
fs2_devWarn(MSG_PORT + "0")
return  0

*****************************************************************************
* com_num() -> nPorts
*
* FS2: Reports the number of supported ports (default = 16)
* CA3: Reports the number of highest available serial port
*----------------------------------------------------------------------------
FUNCTION com_num()
return len(anFileHd)

*****************************************************************************
* com_open(nPort, [nSize], [nOut], [lTrigg], [cDevice]) -> lOk
*
* Opens the port and initializes the buffer
* When the special device <cDevice> is not given, the corresponding
* serial device to nPort will be opened. Since there ae two standards,
* we first try to open /dev/ttyS<n> and then /dev/cua<n>where <n>
* corresponds to <nPort> -1, i.e. 3 stay for COM3 or /dev/ttyS2.
*
* nPort  : 1..16 for port COM1 .. COM16
* nSize  : optional, size of the receiving buffer, default = BUFFSIZE (=1024)
* nOut   : Output buffer for background. Not used by FS2, ignored
* lTrig  : Triggering by interrupt. Not applicable in FS2, ignored.
* cDevice: optional, device if other than /dev/cua<n> or /dev/ttyS<n>
* return : .T. if successfully opened,
*          .F. on error (port is open, or no file handlers available,
*              could not open, permissions etc.). You may check FERROR()
*              after !com_open() for the error reason.
*
* CA3: Only 4 parameters are supported
* FS2: Note, you may use the com_*() functions also on any other devices,
*      not only serial, by using the 5th parameter, e.g. for LPT1: or console
*      com_open(5, , , , "/dev/lp1") or com_open(10, , , , "/dev/tty3") etc.
*      The standard FOPEN() function opens the port in non-blocking mode.
*
* DOS driver         Unix device
* COM1:              /dev/cua0   or  /dev/ttyS0   or user defined
* COM2:              /dev/cua1   or  /dev/ttyS1   or user defined
*  :
* COM16:             /dev/cua15  or  /dev/ttyS15  or user defined
*----------------------------------------------------------------------------
FUNCTION com_open(nPort, nSize, dum3, dum4, cDevice)
if com_isopen(nPort)
   fs2_devWarn("Port " + ltrim(nPort) + " is already open")
   return .F.
endif

if !(valtype(nPort) $ "NI") .or. ;
   (nPort <= 0 .or. nPort > len(anFileHd))
   return .F.
endif

nSize := if(valtype(nSize) $ "NI" .and. nSize > 0, nSize, BUFFSIZE)

anLastEr[nPort] := 0
anBufLen[nPort] := 0
anBufPos[nPort] := 0
acBuffer[nPort] := space(nSize)
anBufSiz[nPort] := nSize

#ifdef FS_WIN32
   if valtype(cDevice) == "C" .and. !empty(cDevice)
      anFileHd[nPort] := _WinOpen(nPort, alltrim(cDevice), nSize)
   else
      anFileHd[nPort] := _WinOpen(nPort, ;
                                  if(nPort < 10, "COM", "\\.\COM") + ;
                                    ltrim(nPort), ;
                                  nSize)
   endif
#else
   if valtype(cDevice) == "C" .and. !empty(cDevice)
      anFileHd[nPort] := FOPEN(alltrim(cDevice), 2)
   else
      // Unix/Linux default devices are  /dev/ttyS0..n  -or-  /dev/cua0..n
      anFileHd[nPort] := FOPEN("/dev/ttyS" + ltrim(nPort -1), 2)
      if anFileHd[nPort] < 1
         anFileHd[nPort] := FOPEN("/dev/cua" + ltrim(nPort -1), 2)
      endif
   endif
#endif
if anFileHd[nPort] > 0
   fs2_termio(-1, nPort, @asStatus[nPort])  // save current status
   return .T.
endif
return .F.

*****************************************************************************
* com_read(nPort, [nSize], [lHold]) -> cString
*
* CA3: Reads characters from the internal buffer.
* FS2: Reads characters from the internal buffer.  Additionally, when <nSize>
*      is larger than the rest of bytes in the internal buffer, the remaining
*      data are also repeatly read from the device, so the
*      com_count() + com_read() loop sequence is not required in your prg.
*      But: when lHold is .T., only the rest of the internal buffer is
*      returned, no additional data are read from the device.
*
* nPort  : corresponds to the <nPort> from com_open()
* nSize  : number of characters to be read. The default is the rest
*          of internal buffer; if this is empty, a new set is read
* lHold  : .F. (default) remove read characters from the buffer
*          .T. hold characters in the buffer
* cString: receiving characters of the size <nSize>, or less if no
*          additional data are available.
*----------------------------------------------------------------------------
FUNCTION com_read(nPort, nSize, lHold)
local nAvail, nLen, lMustRead := .F., cRet := ""
if !com_isopen(nPort)
   return cRet
endif
nAvail := com_count(nPort)     // read into buffer if required
if nAvail <= 0
   return cRet
endif

if !(valtype(nSize) $ "NI") .or. nSize <= 0
   nSize := nAvail
endif
nLen := nSize

if nSize > nAvail
   lMustRead := .T.
   nLen := nAvail
endif
cRet := substr(acBuffer[nPort], anBufPos[nPort] +1, nLen)
if valtype(lHold) == "L" .and. lHold
   return cRet                         // don't remove
endif

anBufPos[nPort] := min(anBufPos[nPort] + nLen, anBufLen[nPort])

while lMustRead
   nSize  := max(nSize - nAvail, 0)
   nAvail := com_count(nPort)     // read into buffer if required

   lMustRead := .F.
   if nAvail > 0 .and. nSize > 0
      nLen := min(nSize, nAvail)
      cRet += substr(acBuffer[nPort], anBufPos[nPort] +1, nLen)
      lMustRead := nSize > nAvail
      anBufPos[nPort] := anBufPos[nPort] + nLen
   endif
enddo
return cRet

*****************************************************************************
* com_remote(nPort, [ncChar]) -> lAct
*
* CA3: Determines the clear char for the receiving buffer
* FS2: THIS FUNCTION AND MECHANISM IS NOT APPLICABLE ON UNIX
*      Therefore, default value .F. is returned here
*----------------------------------------------------------------------------
FUNCTION com_remote(nPort, ncChar)
fs2_devWarn(MSG_PORT + ".F.")
return .F.

*****************************************************************************
* com_ring(nPort) -> lAct
*
* CA3: Queries the ring line
* FS2: THIS FUNCTION AND MECHANISM IS NOT APPLICABLE ON UNIX
*      Therefore, default value .T. is returned here
*----------------------------------------------------------------------------
FUNCTION com_ring(nPort)
fs2_devWarn(MSG_PORT + ".T.")
return .T.

*****************************************************************************
* com_rts(nPort, [lnNew]) -> lActive
*
* CA3: Queries or sets the Request To Send (RTS)
* FS2: THIS FUNCTION AND MECHANISM IS NOT APPLICABLE ON UNIX
*         Therefore, default value .T. is returned here
*      In MS-Windows environment, this function is supported. If <lnNew>
*        is logical, DTR is set on/off, same as CA3.
*        FS2: if <lnNew> is numeric, 0=off, 1=on, 2 (or other num)=handshake
*----------------------------------------------------------------------------
FUNCTION com_rts(nPort, lnNew)
local lRet := .T.
#ifdef FS_WIN32
   if !com_isopen(nPort)
      return .F.
   endif
   #Cinline
   {
      DCB dcb;
      int iOldSet = 0, iNew;
      int iPort = VAR_INTNUM(VAR_NAME_LOCPAR(nport));

      if( GetCommState(aHANDLE[iPort], &dcb) ) {
         iOldSet = dcb.fRtsControl  != RTS_CONTROL_DISABLE;
         if(IS_VAR_LOG(VAR_NAME_LOCPAR(lnnew))) {
            dcb.fRtsControl = IS_VAR_TRUE(VAR_NAME_LOCPAR(lnnew)) ?
                                 RTS_CONTROL_ENABLE : RTS_CONTROL_DISABLE;
            SetCommState(aHANDLE[iPort], &dcb);
         } else if(IS_VAR_NUM(VAR_NAME_LOCPAR(lnnew))) {
            iNew = VAR_INTNUM(VAR_NAME_LOCPAR(lnnew));
            dcb.fRtsControl = (iNew == 0 ? RTS_CONTROL_DISABLE :
                                 (iNew == 1 ? RTS_CONTROL_ENABLE :
                                    RTS_CONTROL_HANDSHAKE));
            SetCommState(aHANDLE[iPort], &dcb);
         }
      }
      SET_VAR_LOG(VAR_NAME_LOCAL(lret), iOldSet);
   }
   #endCinline
#else
   fs2_devWarn(MSG_PORT + ".T.")
#endif
return lRet

*****************************************************************************
* com_scount(nPort) -> nChar
*
* Counts the number of chars in the internal buffer
*----------------------------------------------------------------------------
FUNCTION com_scount(nPort)
if !com_isopen(nPort)
   return -1
endif
return anBufLen[nPort] - anBufPos[nPort]

*****************************************************************************
* com_send(nPort, ncString) -> nFailed
*
* nPort   : corresponds to the <nPort> from com_open()
* ncString: a string or an ASCII value of one characted to be trensmitted
* nFailed : number of charecters not transmitted. The ok case is 0,
*          on parameter error or if the port is not open, -1 is returned.
*----------------------------------------------------------------------------
FUNCTION com_send(nPort, ncString)
local cBuff, ii
if !com_isopen(nPort)
   return -1
endif
if valtype(ncString) == "C"
   cBuff := ncString
elseif valtype(ncString) == "N" .and. ncString >= 0 .and. ncString <= 255
   cBuff := chr(ncString)
else
   return -1
endif
#ifdef FS_WIN32
  ii := _WinWrite(nPort, @cBuff, len(cBuff), @anLastEr[nPort])
#else
  ii := FWRITE(anFileHd[nPort], cBuff)  // == write (fd, buffer, len)
  anLastEr[nPort] := FERROR()
#endif
return len(cBuff) - ii

*****************************************************************************
* com_setio(nPort, [ncPortAdr]) -> lAct
*
* CA3: Changes the I/O address for a port
* FS2: THIS FUNCTION AND MECHANISM IS NOT APPLICABLE ON UNIX
*      Therefore, default value .F. is returned here
*----------------------------------------------------------------------------
FUNCTION com_setio(nPort, ncPortAdr)
fs2_devWarn(MSG_PORT + ".F.")
return .F.

*****************************************************************************
* com_setirq(nPort, [ncPortiIrq]) -> lAct
*
* CA3: Changes the IRQ for a port
* FS2: THIS FUNCTION AND MECHANISM IS NOT APPLICABLE ON UNIX
*      Therefore, default value .F. is returned here
*----------------------------------------------------------------------------
FUNCTION com_setirq(nPort, ncPortIrq)
fs2_devWarn(MSG_PORT + ".F.")
return .F.

*****************************************************************************
* com_sflush(nPort) -> lOk
*
* Flushes the internal buffer of com_count()
* However, it does not flush the rest of incomming chars on the serial line
*----------------------------------------------------------------------------
FUNCTION com_sflush(nPort)
return com_flush(nPort)

*****************************************************************************
* com_skey(nPort, [ncKey1], [ncKey2]) -> lOk
*
* CA3: Monitor the port using key traps
* FS2: THIS FUNCTION AND MECHANISM IS NOT APPLICABLE ON UNIX
*      Therefore, default value .F. is returned here
*      You may simply use cRead := com_read() ; KEYBOARD (cRead)
*----------------------------------------------------------------------------
FUNCTION com_skey(nPort, ncKey1, ncKey2)
fs2_devWarn(MSG_PORT0 + ";use cRead:=com_read() + KEYBOARD (cRead);Returning .F.")
return .F.

*****************************************************************************
* com_smode(nPort) -> lOk
*
* Determines the current transmission status
*
* lOk : 0 = port not open
*       1 = buffer is empty
*       2..8 = not applicable in FS2
*      -1 = buffer is available for com_read()
*----------------------------------------------------------------------------
FUNCTION com_smode(nPort)
if !com_isopen(nPort)
   return  0
endif
return if(anBufLen[nPort] - anBufPos[nPort] > 0, -1, 1)

*****************************************************************************
* com_soft(nPort, [lNew], [cXon], [cXoff]) -> lOk
*
* Queries or sets new XON/XOFF handshake or character
* FS2: setting of different [cXon], [cXoff] are not supported in Unix
*----------------------------------------------------------------------------
FUNCTION com_soft(nPort, lNew, cXon, cXoff)
if !com_isopen(nPort)
   return fs2_termio(6, nPort, lNew, cXon, cXoff)
endif
return .F.

*****************************************************************************
* com_soft_r(nPort, [lReset]) -> lXoff
*
* CA3: Tests if an XOFF character has been received, reset Xoff flag
* FS2: Checks if Ctrl-Q (XOFF) is available in receiving buffer
*----------------------------------------------------------------------------
FUNCTION com_soft_r(nPort, ldummy)
if !com_isopen(nPort) .or. anBufLen[nPort] <= 0 .or. ;
   anBufPos[nPort] >= anBufLen[nPort]
   return  .F.
endif
return chr(17) $ substr(acBuffer[nPort], anBufPos[nPort] +1)

*****************************************************************************
* com_soft_s(nPort) -> lXoff
*
* CA3: Tests if an XOFF character has been sent
* FS2: Not applicable. Therefore, default value .F. is returned here
*----------------------------------------------------------------------------
FUNCTION com_soft_s(nPort)
fs2_devWarn(MSG_PORT + ".F.")
return .F.

*****************************************************************************
* com_timeout(nPort, [nReadInterv], [nReadMulti], [nReadTime],
*                    [nWriteMulti], [nWriteTime ) -> aTimes
*
* CA3: n/a
* FS2: Sets/gets time-outs for open port. Applicable for MS-Windows only
*   nPort       = already open port number
*   nReadInterv = optional, specifies the maximum time, in milliseconds, allowed to
*                 elapse between the arrival of two characters on the communications
*                 line. During a com_read operation, the time period begins when the
*                 first character is received. If the interval between the arrival of
*                 any two characters exceeds this amount, the com_read operation is
*                 completed and any buffered data is returned. A value of zero
*                 indicates that interval time-outs are not used. A value of -1,
*                 combined with zero values for both the nReadMulti and nReadTime
*                 members, specifies that the read operation is to return immediately
*                 with the characters that have already been received, even if no
*                 characters have been received. Default: -1
*   nReadMulti  = optional, specifies the multiplier, in milliseconds, used to calculate
*                 the total time-out period for read operations. For each read operation,
*                 this value is multiplied by the requested number of bytes to be read.
*                 Default: 0
*   nReadTime   = optional, specifies the constant, in milliseconds, used to calculate
*                 the total time-out period for read operations. For each read operation,
*                 this value is added to the product of the nReadMulti member and the
*                 requested number of bytes. A value of zero for both the nReadMulti
*                 and nReadTime members indicates that total time-outs are not used for
*                 read operations. Default: 0
*   nWriteMult  = optional, specifies the multiplier, in milliseconds, used to calculate
*                 the total time-out period for write operations. For each write
*                 operation, this value is multiplied by the number of bytes to be
*                 written. Default: 0
*   nWriteTime  = optional, specifies the constant, in milliseconds, used to calculate
*                 the total time-out period for write operations. For each write
*                 operation, this value is added to the product of the nWriteMult
*                 member and the number of bytes to be written. A value of zero for both
*                 the nWriteMult and nWriteTime members indicates that total time-outs
*                 are not used for write operations. Default: 0
* If an application sets nReadInterv and nReadMulti to -1 and sets nReadTime to a value
* greater than zero and not -1, one of the following occurs when the com_read function
* is called:
* - If there are any characters in the input buffer, com_read returns immediately with
*   the characters in the buffer.
* - If there are no characters in the input buffer, com_read waits until a character
*   arrives and then returns immediately.
* - If no character arrives within the time specified by nReadTime, com_read times out.
* Note: above description is from Microsoft function SetCommTimeouts()
*
* Return: array of 5 elements corresponding nReadInterv ... nWriteTime (in that order)
*   at the time of calling this function (i.e. old values if data change requested)
*
* Default setting in com_open() -> _winopen():
*     [ aTimes := ] com_timeout(nPort, 100, 0, 100, 0, 200)
* Alternative setting without delays:
*     [ aTimes := ] com_timeout(nPort, -1, 0, 0, 0, 0)
* Get current data only:
*     aTimes := com_timeout(nPort)
*----------------------------------------------------------------------------
FUNCTION com_timeout(nPort, nReadInterv, nReadMulti, nReadTime, nWriteMult, nWriteTime)
/* C translation:
 *    FSvar *_bb_com_timeou (int parno, FSvar *parptr[])
 */

local aRet := {0,0,0,0,0}
#ifdef FS_WIN32
   if !com_isopen(nPort)
      return aRet
   endif

   #Cinline
   {
      COMMTIMEOUTS CommTimeouts;             // (Win32)
      int ok;
      int iPort = VAR_INTNUM(VAR_NAME_LOCPAR(nport));

      /* get current timeouts
       */
      ok = GetCommTimeouts(aHANDLE[iPort], &CommTimeouts);   // (Win32)
      if(ok) {
         SET_VAR_INT(VAR_ARRELEM(VAR_NAME_LOCAL(aret),1),
                     CommTimeouts.ReadIntervalTimeout == MAXDWORD ? -1 :
                        CommTimeouts.ReadIntervalTimeout);
         SET_VAR_INT(VAR_ARRELEM(VAR_NAME_LOCAL(aret),2),
                     CommTimeouts.ReadTotalTimeoutMultiplier == MAXDWORD ? -1 :
                        CommTimeouts.ReadTotalTimeoutMultiplier);
         SET_VAR_INT(VAR_ARRELEM(VAR_NAME_LOCAL(aret),3),
                     CommTimeouts.ReadTotalTimeoutConstant == MAXDWORD ? -1 :
                        CommTimeouts.ReadTotalTimeoutConstant);
         SET_VAR_INT(VAR_ARRELEM(VAR_NAME_LOCAL(aret),4),
                     CommTimeouts.WriteTotalTimeoutMultiplier);
         SET_VAR_INT(VAR_ARRELEM(VAR_NAME_LOCAL(aret),5),
                     CommTimeouts.WriteTotalTimeoutConstant);
      }
      if(ok && (parno > 1)) {
         /* set new timeouts
          */
         int iReadInterv, iReadMulti, iReadTime, iWriteMult = 0, iWriteTime = 0;
         DWORD dReadInterv = MAXDWORD, dReadMulti = 0, dReadTime = 0;
         switch(parno) {   /* note: falls thru, no break */
            case 7:
            case 6:
               if(IS_VAR_NUM(parptr[5]))
                  iWriteTime = VAR_INTNUM(parptr[5]);    // par6: nWriteTime
            case 5:
               if(IS_VAR_NUM(parptr[4]))
                  iWriteMult = VAR_INTNUM(parptr[4]);    // par5: nWriteMult
            case 4:
               if(IS_VAR_NUM(parptr[3])) {
                  iReadTime = VAR_INTNUM(parptr[3]);     // par4: nReadTime, can be -1
                  dReadTime = iReadTime == -1 ? MAXDWORD : iReadTime;
               }
            case 3:
               if(IS_VAR_NUM(parptr[2])) {
                  iReadMulti = VAR_INTNUM(parptr[2]);    // par3: nReadMulti, can be -1
                  dReadMulti = iReadMulti == -1 ? MAXDWORD : iReadMulti;
               }
            case 2:
               if(IS_VAR_NUM(parptr[1])) {
                   iReadInterv = VAR_INTNUM(parptr[1]);  // par2: nReadInterv, can be -1
                   dReadInterv = iReadInterv == -1 ? MAXDWORD : iReadInterv;
               }
         }
         CommTimeouts.ReadIntervalTimeout         = dReadInterv;     // stop if 100 msec elapses between two received bytes
         CommTimeouts.ReadTotalTimeoutMultiplier  = dReadMulti;      // no length sensitivity
         CommTimeouts.ReadTotalTimeoutConstant    = dReadTime;       // allow 100 (or 400) msec for a read (reset command can take a while)
         CommTimeouts.WriteTotalTimeoutMultiplier = iWriteMult;      // no length sensitivity
         CommTimeouts.WriteTotalTimeoutConstant   = iWriteTime;      // allow 200 msec for a write

         ok = SetCommTimeouts(aHANDLE[iPort], &CommTimeouts);  // set timeouts (Win32)
      }
   }
   #endCinline
#else
   // not applicable for Linux
#endif
return aRet

*****************************************************************************
* xmoblock(cString, nBlockNo, [lCRC], [nMode]) -> cXmodBlock
*
* Generates a block for XMODEM transmission
*
* cString : char string of 128 or 1024 length, depending on <nMode>.
*           Trailing chars are cut off, missing chars filled with binary 0.
* nBlockNo: The related XMODEM block number (BN). starting with 0
* lCRC    : .T. = the modem block contain checksum (CS)
*           .F. = the modem block contain 16-bit CRC (default)
* nMode   : 1 = the XMODEM block contains
*                 chr(1) + BM + ~BM + 128chars + (CS or CRC)
*           2 = the XMODEM block contains
*                 chr(2) + BM + ~BM + 1024chars + (CS or CRC)
* cXmodBlock = the returned XMODEM block
*----------------------------------------------------------------------------
FUNCTION xmoblock(cString, nBlockNo, lCRC, nMode)
local cBuff, iCheck, cCheck
local b1, b2 as intvar

lCRC  := if (valtype(lCRC) == "L", lCRC, .F.)
nMode := if (valtype(nMode) $ "NI", nMode, 1)

cBuff := padr(cString, if(nMode == 1, 128, 1024), chr(0))
b1 := nBlockNo % 256
b2 := numand (numnot(b1), 255)
if lCRC
   cCheck := chr(numlow(asciisum(cBuff)))
else
   iCheck:= crc16(cBuff, 0, 40961)
   cCheck := chr(numlow(iCheck)) + chr(numhigh(iCheck))
endif
return  if(nMode == 1, chr(1), chr(2)) + chr(b1) + chr(b2) + cBuff + cCheck

*****************************************************************************
* xmocheck(cString, [lCRC]) -> nBlockNo
*
* Tests a received XMODEM block
*
* cString : char string including a valid XMODEM block, see xmoblock()
* lCRC    : .T. = the modem block contain checksum (CS)
*           .F. = the modem block contain 16-bit CRC (default)
* nBlockNo: The received XMODEM block number (BN) (0..255)
*----------------------------------------------------------------------------
FUNCTION xmocheck(cString, lCRC)
local iCheck, xCheck, ok, ii, nMode, nSize
local b1, b2 as intvar

lCRC  := if (valtype(lCRC) == "L", lCRC, .F.)
nMode := strpeek(@cString, 1)
b1    := strpeek(@cString, 2)
b2    := strpeek(@cString, 3)

ok := nMode == 1 .or. nMode == 2
ok := ok .and. len(cString) >= if(nMode == 1, 131, 1027)
ok := ok .and. b2 == numand (numnot(b1), 255)
if !ok
   return -1
endif
nSize := if(nMode == 1, 128, 1024)
if lCRC
   xCheck := strpeek(@cString, nSize + 3)
   iCheck := numlow(asciisum(substr(cString,4,nSize)))
else
   if len(cString) < nSize + 4
      return -1
   endif
   ii := nSize +4
   xCheck := strpeek(@cString, ii--) * 256
   xCheck += strpeek(@cString, ii)
   iCheck := crc16(substr(cString,4,nSize), 0, 40961)
endif
if xCheck != iCheck
   return -1
endif
return b1


*****************************************************************************
* zeroinsert(cString) -> cNewStr
*
* Inserts a 0-bit after every fifth 1-bit to allow ISDN or X.25 transfer
*
* cString : char string
* cNewStr : <cString> with inserted 0-bits, the new length may be 20% larger
*----------------------------------------------------------------------------
FUNCTION zeroinsert(cString)
local cRet := ""
if valtype(cString) == "C" .and. len(cString) > 0
#Cinline
{
   int sizeIn, sizeOut, sizeInit, added = 0, move = 0, byteNo = 0;
   unsigned char *cIn, *cOut, *ptr, *buff;
   unsigned char cc, pp, dd, rr;

   sizeIn   = VAR_ISLEN(VAR_NAME_LOCPAR(cstring)) -1;
   sizeInit = sizeIn * 1.2 + 3;
   cIn  = VAR_CHR(VAR_NAME_LOCPAR(cstring));
   cOut = (unsigned char *) malloc((sizeInit +1) * sizeof(unsigned char));

   if (cOut) {
      memset(cOut,   0, sizeInit);
      memcpy(cOut, cIn, sizeIn  );
      pp = 0;                                        /* no previous byte yet */

      for (ptr=cOut; ptr <= (cOut + sizeIn + added / 8); ptr++, byteNo++) {
         move = 0;                         /* one or two shifts are possible */
         cc = *ptr;
         if (pp & 0x1F) {                            /* check previous byte? */
            if ((pp & 0x1F) == 0x1F) {       /* xxx11111,xxxxxxxx = 0xxxxxxx */
               cc = (cc >> 1) & 0x7F;
               move++;
            }
            else if (((pp & 0x0F) == 0x0F) &&
                     ((cc & 0x80) == 0x80)) { /* xxxx1111,1xxxxxxx = 10xxxxxx */
               cc = ((cc >> 1) & 0x3F) | 0x80;
               move++;
           }
            else if (((pp & 0x07) == 0x07) &&
                     ((cc & 0xC0) == 0xC0)) { /* xxxxx111,11xxxxxx = 110xxxxx */
               cc = ((cc >> 1) & 0x1F) | 0xC0;
               move++;
           }
            else if (((pp & 0x03) == 0x03) &&
                     ((cc & 0xE0) == 0xE0)) { /* xxxxxx11,111xxxxx = 1110xxxx */
               cc = ((cc >> 1) & 0x0F) | 0xE0;
               move++;
           }
            else if (((pp & 0x01) == 0x01) &&
                     ((cc & 0xF0) == 0xF0)) { /* xxxxxxx1,1111xxxx = 11110xxx */
               cc = ((cc >> 1) & 0x07) | 0xF0;
               move++;
            }
         }
         if (cc) {                                    /* check current byte? */
            if ((cc & 0xF8) == 0xF8) {                /* 11111xxx = 111110xx */
               cc = ((cc >> 1) & 0x03) | 0xF8;
               move++;
            }
            else if ((cc & 0x7C) == 0x7C) {           /* x11111xx = x111110x */
               dd = cc & 0xFC;
               cc = ((cc >> 1) & 0x01) | dd;
               move++;
            }
            else if ((cc & 0x3E) == 0x3E) {           /* xx11111x = xx111110 */
               cc = cc & 0xFE;
               move++;
            }
         }

         if (move) {
            rr = *ptr << (8 - move);                  /* remainder moved out */
            *ptr = cc;                                /* save current byte   */

            added += move;

            for (buff=ptr+1; buff <= (cOut + sizeIn + added / 8 +1); buff++) {
               dd = *buff << (8 - move);                    /* new remainder */
               *buff = (*buff >> move) | rr;   /* shift and insert remainder */
               rr = dd;
            }
         }

#ifdef TESTPRINT
         printbuff("zeroinsert()", cc, cOut, sizeInit);
#endif

         pp = cc;                                     /* next: previous byte */
      }

      sizeOut = sizeIn + added / 8;
      if ((added % 8) > 0)
         sizeOut++;
      SET_VAR_CHRLEN(VAR_NAME_LOCAL(cret), cOut, sizeOut);
      free(cOut);
   }
}
#endCinline
endif
return cRet

*****************************************************************************
* zeroremove(cString) -> cNewStr
*
* Removes a 0-bit after every fifth 1-bit - reverse of zeroinsert()
*
* cString : char string from zeroinsert()
* cNewStr : <cString> with removed 0-bits, the new length may be 20% shorter
*----------------------------------------------------------------------------
FUNCTION zeroremove(cString)
local cRet := ""
if valtype(cString) == "C" .and. len(cString) > 0
#Cinline
{
   int sizeIn, sizeOut, sizeInit, removed = 0, move = 0, byteNo = 0;
   unsigned char *cIn, *cOut, *ptr, *buff;
   unsigned char cc, pp, dd, rr;

   sizeIn   = VAR_ISLEN(VAR_NAME_LOCPAR(cstring)) -1;
   sizeInit = sizeIn +1;
   cIn  = VAR_CHR(VAR_NAME_LOCPAR(cstring));
   cOut = (unsigned char *) malloc((sizeInit +1) * sizeof(unsigned char));
   byteNo = sizeIn -1;

   if (cOut) {
      memset(cOut,   0, sizeInit);
      memcpy(cOut, cIn, sizeIn  );

      for (ptr=cOut + sizeIn -1; ptr >= cOut; ptr--, byteNo--) {
         move = 0;                         /* one or two shifts are possible */
         cc = *ptr;
         if (ptr > cOut)
            pp = *(ptr -1);
         else
            pp = 0;
         if (cc) {
            if ((cc & 0xFC) == 0xF8) {                /* 111110xx = 11111xx+ */
               cc = (cc << 1) | 0xF8;
               move++;
            }
            else if ((cc & 0x7E) == 0x7C) {           /* x111110x = x11111x+ */
               dd = cc & 0xFC;
               cc = ((cc & 1) << 1) | dd;
               move++;
            }
            else if ((cc & 0x3F) == 0x3E) {           /* xx111110 = xx11111+ */
               cc = cc & 0xFE;
               move++;
            }
         }
         if (move) {
            rr = *(ptr +1) >> (8 - move);            /* remainder moved in  */
            cc = cc | rr;                            /* save curr result    */
         }
         if (pp & 0x1F) {
            if (((pp & 0x1F) == 0x1F) &&
                ((cc & 0x80) == 0)) {        /* xxx11111,0xxxxxxx = xxxxxxx+ */
               cc = cc << 1;
               move++;
            }
            else if (((pp & 0x0F) == 0x0F) &&
                     ((cc & 0xC0) == 0x80)) { /* xxxx1111,10xxxxxx = 1xxxxxx+ */
               cc = (cc << 1) | 0x80;
               move++;
           }
            else if (((pp & 0x07) == 0x07) &&
                     ((cc & 0xE0) == 0xC0)) { /* xxxxx111,110xxxxx = 11xxxxx+ */
               cc = (cc << 1) | 0xC0;
               move++;
           }
            else if (((pp & 0x03) == 0x03) &&
                     ((cc & 0xF0) == 0xE0)) { /* xxxxxx11,1110xxxx = 111xxxx+ */
               cc = (cc << 1) | 0xE0;
               move++;
           }
            else if (((pp & 0x01) == 0x01) &&
                     ((cc & 0xF8) == 0xF0)) { /* xxxxxxx1,11110xxx = 1111xxx+ */
               cc = (cc << 1) | 0xF0;
               move++;
            }
         }

         if (move) {
            rr = *(ptr +1) >> (8 - move);            /* remainder moved in  */
            *ptr = cc | rr;                          /* save curr result    */

            removed += move;

            for (buff=ptr+1; buff <= (cOut + sizeIn - removed / 8); buff++) {
               rr = *(buff +1) >> (8 - move);
               *buff = (*buff << move) | rr;
            }
         }
      }

      sizeOut = sizeIn - removed / 8;
      if ((removed % 8) > 0)
         sizeOut--;
      SET_VAR_CHRLEN(VAR_NAME_LOCAL(cret), cOut, sizeOut);
      free(cOut);
   }
}
#endCinline
endif
return cRet

*****************************************************************************
* fs2_termio(nType, nPort, [value], [bit], [parity], [stop],
*            [lCtsRts], [lLocal], [nCanAss]) -> ok
*
* Helper for com_hard(), com_init(), com_init2()
* assumes the port is open. The parameters depends on the
* calling function and <nType>.
*
* nType = 0: init serial port with predefined values - for com_init()
*         1: return/set status of RTS/CTS (log)
*         2: value = baud (general, not port specific!) (num)
*         3: value = parity  (1/0)
*         4: value = dataLng (num 7/8)
*         5: value = stopBit (num 0/1/2)
*         6: return/set status of XON/XOFF (log)
*        -1: save port status    - for com_open() only
*        -2: restore port status - for com_close() only
* nPort = port (1..16)
*----------------------------------------------------------------------------
STATIC FUNCTION fs2_termio(nType, nPort, value, bit, parity, stop, ;
                           lCtsRts, lLocal, lCanon)
local RetVal := .T.
local_int iType, iFd, iPort

iType := nType
iFd   := anFileHd[nPort]
iPort := nPort

#Cinline
#ifdef FS_WIN32
/* ******************** MS-Windows part *****************************
 */
{
   int ok = 0, ii = 0, iVal = 0, iSpeed = 0;
   int iBit = 8, iParity = 0, iStop = 1;
   int iCtsRts = 1, iLocal = 1, iCanon = 0;
   int iSet, iDtr;
   FSvar *par3 = VAR_NAME_LOCPAR(value);
   FSvar *par4 = VAR_NAME_LOCPAR(bit);
   FSvar *par5 = VAR_NAME_LOCPAR(parity);
   FSvar *par6 = VAR_NAME_LOCPAR(stop);
   char cSet;
   DCB dcb;

   if(IS_VAR_NUM(VAR_NAME_LOCPAR(value)))
      iVal = VAR_INTNUM(VAR_NAME_LOCPAR(value));

   switch (itype) {
   case -1:
   case -2:
      break;

   case 0:     // init setup, called from com_init()
      if( ! GetCommState(aHANDLE[iport], &dcb) )
         return FALSE_VAR;

# ifdef TESTPRINT
      _winDisplayDcb("com_open() DCB stru Win defaults", &dcb);
# endif
      if(IS_VAR_LOG(VAR_NAME_LOCPAR(lctsrts)))
         iCtsRts = IS_VAR_TRUE(VAR_NAME_LOCPAR(lctsrts));

      dcb.BaudRate = _winCheckSpeed(iVal);   // 110..256000
      dcb.Parity   = _winCheckParity(par5);
      dcb.fParity  = 1;                      // check parity
      dcb.fBinary  = 1;                      // binary transfer
      dcb.ByteSize = IS_VAR_NUM(par4) ? VAR_INTNUM(par4) : iBit;
      dcb.StopBits = IS_VAR_NUM(par6) ? VAR_INTNUM(par6) : iStop;

      /* The RTS/CTS handshake is controlled
       * by input parameter
       */
      dcb.fOutxCtsFlow = iCtsRts;
      dcb.fRtsControl  = iCtsRts ? RTS_CONTROL_HANDSHAKE :
                                   RTS_CONTROL_DISABLE;

      /* The DTR (data-terminal-ready) flow control
       * is disabled per default (same as CA3),
       * use com_dtr() to enable
       */
      dcb.fDtrControl = DTR_CONTROL_DISABLE;

      /* The DSR (data-set-ready) signal is monitored for com_dsr()
       * for output flow control
       */
      dcb.fOutxDsrFlow = 1;

      /* The communications driver is sensitive to the state of
       * the DSR signal. The driver ignores any bytes received,
       * unless the DSR modem input line is high
       */
      dcb.fDsrSensitivity =   1;

      /* XON/XOFF handshaking is enabled when lCanon is ON.
       * The XON and XOFF chars are same as in CA3
       * You may activate the minimum number of bytes allowed in the
       * input buffer before flow control is activated to inhibit
       * the sender or to allow transmission by the sender.
       */
      if(IS_VAR_LOG(VAR_NAME_LOCPAR(lcanon)))
         iCanon = IS_VAR_TRUE(VAR_NAME_LOCPAR(lcanon));
      if(iCanon) {
         dcb.fInX  = 1;          // enable XON
         dcb.fOutX = 1;          // enable XOFF
         dcb.XoffChar = 19;      // Ctrl-S
         dcb.XonChar  = 17;      // Ctrl-Q
//       XonLim = 750;           // inhibit the sender after 750 chars
//       XoffLim = 750;          // allow transmission by sender after 750
      } else {
         dcb.fInX  = 0;          // disable XON
         dcb.fOutX = 0;          // disable XOFF
      }
      dcb.fNull          = 0;    // don't ignore 0 bytes
      dcb.fAbortOnError  = 1;    // abort on error

# ifdef TESTPRINT
      _winDisplayDcb("com_open() DCB stru Win set", &dcb);
# endif
      SetCommState(aHANDLE[iport], &dcb);
      break;

   case 1:     // set hanshake, called from com_hard()
      iSet = -1;
      if(IS_VAR_LOG(par3))
         iSet = IS_VAR_TRUE(par3);
      iDtr = IS_VAR_TRUE(par4);

      if( GetCommState(aHANDLE[iport], &dcb) ) {
          if(iDtr) {
            ok = dcb.fDtrControl;
            dcb.fDtrControl = iSet;
         } else {
            ok = dcb.fRtsControl;
            dcb.fRtsControl = iSet;
         }
         if(iSet >= 0)
            SetCommState(aHANDLE[iport], &dcb);
      }
      SET_VAR_LOG(VAR_NAME_LOCAL(retval), ok);
      break;

   case 2:     // set baud rate, called from com_init2()
      if( GetCommState(aHANDLE[iport], &dcb) ) {
         dcb.BaudRate = _winCheckSpeed(iVal);   // 110..256000
         SetCommState(aHANDLE[iport], &dcb);
         ok = dcb.BaudRate;
      }
      SET_VAR_INT(VAR_NAME_LOCAL(retval), ok);
      break;

   case 3:     // set parity, called from com_init2()
      if( GetCommState(aHANDLE[iport], &dcb) ) {
         dcb.Parity = _winCheckParity(par3);
         SetCommState(aHANDLE[iport], &dcb);
         ok = 1;
      }
      SET_VAR_LOG(VAR_NAME_LOCAL(retval), ok);
      break;

   case 4:     // set bit size, called from com_init2()
      if(iVal >= 5 && iVal <= 8 &&
         GetCommState(aHANDLE[iport], &dcb) )
      {
         dcb.ByteSize = iVal;
         SetCommState(aHANDLE[iport], &dcb);
         ok = 1;
      }
      SET_VAR_LOG(VAR_NAME_LOCAL(retval), ok);
      break;

   case 5:     // set stop bits, called from com_init2()
      if(iVal >= 0 && iVal <= 2 &&
         GetCommState(aHANDLE[iport], &dcb) )
      {
         dcb.StopBits = iVal;
         SetCommState(aHANDLE[iport], &dcb);
         ok = 1;
      }
      SET_VAR_LOG(VAR_NAME_LOCAL(retval), ok);
      break;

   case 6:     // set XON/XOFF and it chars, called from com_soft()
      iSet = -1;
      if( GetCommState(aHANDLE[iport], &dcb) ) {
         ok = dcb.fInX && dcb.fOutX;
         if(IS_VAR_LOG(par3))
            iSet = IS_VAR_TRUE(par3);
         if(iSet >= 0) {
            dcb.fInX  = iSet;                   // enable/disable XON
            dcb.fOutX = iSet;                   // enable/disable XOFF
            iSet = 0;
         }
         if(IS_VAR_CHR(par4)) {
            cSet = ((char *)VAR_CHR(par4))[0];
            dcb.XonChar = cSet;
            iSet = 0;
         }
         if(IS_VAR_CHR(par5)) {
            cSet = ((char *)VAR_CHR(par5))[0];
            dcb.XoffChar = cSet;
            iSet = 0;
         }
         if(iSet == 0)
            SetCommState(aHANDLE[iport], &dcb);
      }
      SET_VAR_LOG(VAR_NAME_LOCAL(retval), ok);
      break;
   }  /* end switch(itype) */
}
#else
/* ******************** Unix/Linux part *****************************
 */
{
# if defined(FGSLINUX) || defined(SUNsolaris)
#   include <termios.h>
#       define FGS_termio  termios
# else
#  if defined(NSCSinix)
#    include <termios.h>
#    define FGS_termio  termios
#  else
#    include <sys/ioctl.h>
#    if defined(FGSHPUX11)
#       include <sys/termio.h>
#    else
#       if defined(FGSHPUX) || defined(FGSDECAlpha)
#          include <termio.h>
#       endif
#    endif
#    define FGS_termio  termio
#  endif
# endif

# define FS2_TCSETATTR(fd,x)    tcsetattr(fd, TCSANOW, x)
# define FS2_TCGETATTR(fd,x)    tcgetattr(fd, x)
# define FS2_TCFLUSH(fd)        tcflush(fd, TCIFLUSH)

# if defined(FGSHPUX11)
#    define TCIFLUSH 0     /* from /usr/include/sys/termios.h */
#    define TCSANOW  0
#    define VSUSP    13    /* SUSP character */
#    define VSTART   14    /* START character */
#    define VSTOP    15    /* STOP character */
# endif

   struct FGS_termio term;
   struct FGS_termio *oldterm;

   int    ok, ii = 0, iVal = 0, iSpeed = 0;
   int    iBit = 8, cParity = 'N', iStop = 1;
   int    iCtsRts = 1, iLocal = 1, iCanon = 0;

   ok = FS2_TCGETATTR(ifd, &term);
   if (ok == 0) {
      iVal = 9600;   // default
      if(IS_VAR_NUM(VAR_NAME_LOCPAR(value)))
         iVal = VAR_INTNUM(VAR_NAME_LOCPAR(value));
      if(iVal >=    300) iSpeed = B300    ;
      if(iVal >=   1200) iSpeed = B1200   ;
      if(iVal >=   2400) iSpeed = B2400   ;
      if(iVal >=   4800) iSpeed = B4800   ;
      if(iVal >=   9600) iSpeed = B9600   ;
      if(iVal >=  19200) iSpeed = B19200  ;
      if(iVal >=  38400) iSpeed = B38400  ;
# ifdef B57600
      if(iVal >=  57600) iSpeed = B57600  ;
# endif
# ifdef B115200
      if(iVal >= 115200) iSpeed = B115200 ;
# endif
# ifdef B230400
      if(iVal >= 230400) iSpeed = B230400 ;
# endif
# ifdef B256000
      if(iVal >= 256000) iSpeed = B256000 ;
# endif

      switch (itype) {
      case -1:    // save curr.setting, called from com_open()
         oldterm = (struct FGS_termio *) malloc(sizeof(struct FGS_termio));
         if (oldterm) {
            memcpy(oldterm, &term, sizeof(struct FGS_termio));
            SET_VAR_SPECIAL(VAR_NAME_LOCPAR(value), oldterm);
         }
         break;
      case -2:    // restore curr.setting, called from com_close()
         oldterm = VAR_SPECIAL(VAR_NAME_LOCPAR(value));
         FS2_TCFLUSH(ifd);
         ok = FS2_TCSETATTR(ifd, oldterm);
         free(oldterm);
         break;

      case 0:     // init setup, called from com_init()
/*       bzero(&term, sizeof(term));   clear all flags */
         memset(&term, 0, sizeof(term));  /* clear all flags */

         term.c_cflag = iSpeed;  /* speed */

         if(IS_VAR_NUM(VAR_NAME_LOCPAR(bit)))
            iBit = VAR_INTNUM(VAR_NAME_LOCPAR(bit));
         switch (iBit) {
            case 5 : term.c_cflag |= CS5; break;
            case 6 : term.c_cflag |= CS6; break;
            case 7 : term.c_cflag |= CS7; break;
            case 8 :
            default: term.c_cflag |= CS8; break;
         }
         if(IS_VAR_CHR(VAR_NAME_LOCPAR(parity)))
            cParity = *(VAR_CHR(VAR_NAME_LOCPAR(parity)));
         switch (cParity) {
            case 'O':  term.c_cflag |= PARODD;  /* fall thru */
            case 'E':  term.c_cflag |= PARENB; break;
         }
         if(IS_VAR_NUM(VAR_NAME_LOCPAR(stop)))
            iStop = VAR_INTNUM(VAR_NAME_LOCPAR(stop));
         if (iStop == 2)
            term.c_cflag |= CSTOPB;

         if(IS_VAR_LOG(VAR_NAME_LOCPAR(lctsrts)))
            iCtsRts = IS_VAR_TRUE(VAR_NAME_LOCPAR(lctsrts));

# ifdef CRTSCTS
         if(iCtsRts)
            term.c_cflag |= CRTSCTS;
# endif

         if(IS_VAR_LOG(VAR_NAME_LOCPAR(llocal)))
            iLocal = IS_VAR_TRUE(VAR_NAME_LOCPAR(llocal));
         if(iLocal)
            term.c_cflag |= CLOCAL;

         term.c_cflag |= CREAD;    /* enable receiving chars */

         term.c_oflag = 0;                 /* raw output */

         if(IS_VAR_LOG(VAR_NAME_LOCPAR(lcanon)))
            iCanon = IS_VAR_TRUE(VAR_NAME_LOCPAR(lcanon));
         if(iCanon) {
            term.c_lflag = ICANON;            /* disable echo */
            term.c_iflag = IGNPAR | ICRNL;    /* ignore bytes with parity err, map CR to NL */
            term.c_cc[VINTR]    = 3;          /* Ctrl-c ENABLED */
            term.c_cc[VEOF]     = 4;          /* Ctrl-d ENABLED */
            term.c_cc[VMIN]     = 1;          /* blocking read until 1 character arrives */

            term.c_cc[VQUIT]    = 0;          /* Ctrl-\ */
            term.c_cc[VERASE]   = 0;          /* del */
            term.c_cc[VKILL]    = 0;          /* @ */
            term.c_cc[VTIME]    = 0;          /* inter-character timer unused */
# ifdef VSWTC
            term.c_cc[VSWTC]    = 0;          /* '\0' */
# endif
# ifdef VREPRINT
            term.c_cc[VREPRINT] = 0;          /* Ctrl-r */
# endif
# ifdef VDISCARD
            term.c_cc[VDISCARD] = 0;          /* Ctrl-u */
# endif
# ifdef VWERASE
            term.c_cc[VWERASE]  = 0;          /* Ctrl-w */
# endif
# ifdef VLNEXT
            term.c_cc[VLNEXT]   = 0;          /* Ctrl-v */
# endif
            term.c_cc[VSTART]   = 0;          /* Ctrl-q */
            term.c_cc[VSTOP]    = 0;          /* Ctrl-s */
            term.c_cc[VSUSP]    = 0;          /* Ctrl-z */
            term.c_cc[VEOL]     = 0;          /* '\0' */
            term.c_cc[VEOL2]    = 0;          /* '\0' */
         } else {
            term.c_lflag = 0;
            term.c_iflag = IGNPAR ;           /* ignore bytes with parity error */
            term.c_cc[VMIN]     = 1;          /* blocking read until 1 character arrives */
         }

         FS2_TCFLUSH(ifd);                    /* clean modem line */
         ok = FS2_TCSETATTR(ifd, &term);      /* set new mode */
         SET_VAR_LOG(VAR_NAME_LOCAL(retval), ok == 0 ? 1 : 0);
         break;

      case 1:     // set hanshake, called from com_hard()
         ok = 0;
# ifdef CRTSCTS
         ok = term.c_oflag & CRTSCTS;
         if(IS_VAR_LOG(VAR_NAME_LOCPAR(value))) {  /* set */
            ii = VAR_LOG(VAR_NAME_LOCPAR(value));
            if (ii) term.c_oflag |= CRTSCTS;
            else    term.c_oflag &= (~CRTSCTS);
            FS2_TCSETATTR(ifd, &term);
         }
# endif
         SET_VAR_LOG(VAR_NAME_LOCAL(retval), ok);
         break;

      case 2:     // set baud rate, called from com_init2()
         if (iSpeed) {
            ok  = cfsetispeed(&term, iSpeed);
            ok &= cfsetospeed(&term, iSpeed);
         } else
            ok = cfgetospeed(&term);
         SET_VAR_INT(VAR_NAME_LOCAL(retval),  ok);
         break;

      case 3:     // set parity, called from com_init2()
         iVal = VAR_INTNUM(VAR_NAME_LOCPAR(value));
         if (iVal) {
            term.c_iflag |= INPCK;
            term.c_oflag |= PARENB;
         } else {
            term.c_iflag |= INPCK;
            term.c_oflag &= (~PARENB);
         }
         ok = FS2_TCSETATTR(ifd, &term);
         SET_VAR_LOG(VAR_NAME_LOCAL(retval), !ok);
         break;

      case 4:     // set bit size, called from com_init2()
         iVal = VAR_INTNUM(VAR_NAME_LOCPAR(value));
         if (iVal == 7)
            ii = CS7;
         else if (iVal == 8 )
            ii = CS8;
         if (ii) {
            term.c_cflag &= (~CS7) & (~CS8);
            term.c_cflag |= ii;
            ok = FS2_TCSETATTR(ifd, &term);
         } else
           ok = -1;
         SET_VAR_LOG(VAR_NAME_LOCAL(retval), !ok);
         break;

      case 5:     // set stop bits, called from com_init2()
         iVal = VAR_INTNUM(VAR_NAME_LOCPAR(value));
         switch (iVal) {
            case 0:
            case 1:
               term.c_cflag &= (~CSTOPB);
               break;
            case 2:
               term.c_cflag |= CSTOPB;
               break;
         }
         ok = FS2_TCSETATTR(ifd, &term);
         SET_VAR_LOG(VAR_NAME_LOCAL(retval), !ok);
         break;

      case 6:     // set XON/XOFF and it chars, called from com_soft()
         ok = term.c_iflag & IXON;
         if(IS_VAR_LOG(VAR_NAME_LOCPAR(value))) {  /* set */
            ii = VAR_LOG(VAR_NAME_LOCPAR(value));
            if (ii) term.c_iflag |= IXON | IXOFF;
            else    term.c_iflag &= (~IXON) & (~IXOFF);
            FS2_TCSETATTR(ifd, &term);
         }
         SET_VAR_LOG(VAR_NAME_LOCAL(retval), ok);
         break;
      }  /* end switch(itype) */
   }  /* end if(...) */
}
#endif /* ifdef FS_WIN32 */
#endCinline
return RetVal

** eof fs2_com.prg

