* For FlagShip 6
* Copyright (c) 2004..06 by multisoft GmbH, Munich, Germany
*
* You may modify this source according to your needs and distribute the
* object code freely with your executable. You may distribute also the
* modified source code when your application is distributed in source,
* but may not remove the copyright notice.
*
* Release $Date: 2007/01/16 14:57:52 $
*
* Default Input handler for ListBox and ComboBox, called from
* ListBox:Show([nOpen],[nClose]) and ComboBox:Show([nOpen],[nClose])
*
* You may assign YOUR own handler UDF via code block :InputBlock, i.e.
* oLbox:InputBlock := {|obj,pos,nOpen,nClose| MyUdf(obj,pos,nOpen,nClose) }
* which is then called from :Show() automatically instead of this
* default handler.
*
* Compile     : FlagShip -c -na -m -io=a listboxhand.prg
* Link Linux  : FlagShip myapp*.prg listboxhand.o
* Link Windows: FlagShip myapp*.prg listboxhand.obj

#include "boxclass.fh"
#include "inkey.fh"
#include "box.fh"
#include "button.fh"
#include "set.fh"
#include "FSerrmsg.h"
#include "getexit.fh"

/* --------------------------------------------------------------------------
 * Default ListBox and ComboBox input handler
 * All parameters are mandatory and are not validated here, since
 * this function is invoked from oLbx:Show() or oLbx:Exec() only
 *
 * All parameters are passed from oBox:Show([naOpen],[naClose]):
 * obj    = current object
 * item   = current item#
 * naOpen = inkey value for "open"  ComboBox, or an array of num.values,
 *          passed to :Show(), default = 'Tab' or '#' = { 9, 35)
 * naClose= inkey value for "close" ComboBox, or an array of num.values
 *          passed to :Show(), default = 'shift-Tab' or '^' = { 271, 94}
 * returns: item# of selected item or 0 on abort by Escape
 */
function ListBoxHandler(obj, item, naOpen, naClose)
   local key, iLastKey, ii, iHeight, retVal := 0 as intvar
   local iLastPos := 0                           as intvar
   local iMouseRow := 0, iMouseCol := 0          as intvar
   local lFound, inGet, lHasFocus, lMouseTest    as logic
   local nCurr, iHitTest                         as usual

   if valtype(obj) != "O" .or. !IsObjClass(obj, "_ListBox")
      _Rt_Error(RTARGTYP, "ListBoxHandler(): wrong object class - " + ;
         if(valtype(obj) != "O", "not object", IsObjClass(obj)[1]) )
      return retVal
   endif

   obj:FilterKeys := .F.

   obj:SetFocus()
   obj:Display()

   inGet := valtype(obj:exitState) == "N" .and. obj:exitState >= GE_NOTSET

   while .T.

      if valtype(obj:Keypress) == "N" .and. obj:Keypress != 0
         key := obj:Keypress
         obj:Keypress := NIL
      else
         key := InkeyTrap(0, INKEY_ALL_BUT_MOVE) // same as Inkey(0) but considers SET KEY
      endif
      
      obj:Display()

      iHeight   := obj:ItemsVisible
      lHasFocus := obj:HasFocus

      if obj:DropDown
         /*
          * ComboBox
          */
         if valtype(naOpen) == "N" .and. key == naOpen
            obj:Open()
            loop
         elseif valtype(naClose) == "N" .and. key == naClose
            obj:Close()
            loop
         endif
         lFound := .F.
         if valtype(naOpen) == "A"
            for ii := 1 to len(naOpen)
               if valtype(naOpen[ii]) == "N" .and. key == naOpen[ii]
                  lFound := .T.
                  obj:Open()
                  exit
               endif
            next
         endif
         if !lFound .and. valtype(naClose) == "A"
            for ii := 1 to len(naClose)
               if valtype(naClose[ii]) == "N" .and. key == naClose[ii]
                  lFound := .T.
                  obj:Close()
                  exit
               endif
            next
         endif
         if lFound
            loop
         endif
      endif    // of ComboBox

      do case
      case key == K_ESC                         //--> ESC = exit
         retVal := 0
         obj:exitState := GE_ESCAPE             //    exit READ
         obj:KillFocus()
         exit

      case  key == K_ENTER       .or. ;         //--> Return
            key == K_CTRL_ENTER  .or. ;         //    Ctrl-Return
            (key == K_SPACE     .and. obj:SelectBySpace)  //--> space == Return
         retVal := obj:Buffer
         obj:exitState := GE_ENTER              //    next GET
         obj:KillFocus()
         exit

      case  key == K_HOME                       //--> Home
         obj:FirstItem()                        //    first item in list

      case  key == K_END  .or. ;                //--> End
            key == K_CTRL_DOWN .or. ;           //--> Ctrl-Cursor-Down
            key == K_CTRL_RIGHT                 //--> Ctrl-Cursor-Right
         if key != K_END .and. inGet
            obj:exitState := GE_DOWN            //    in GET: next GET item -->
            exit
         endif
         obj:Select(obj:ItemCount)              //    last item in list
         if !obj:GetItem(obj:Buffer)[3]         //    if not selectable,
            obj:PrevItem()                      //    skip to last selectable
         endif

      case  key == K_UP   .or. ;                //--> cursor up
            key == asc("-")
         nCurr := obj:Buffer
         obj:PrevItem()

      case  key == K_DOWN .or. ;                //--> cursor down:
            key == asc("+")
         nCurr := obj:Buffer
         obj:NextItem()

      case  key == K_CTRL_UP .or. ;             //--> Ctrl-Cursor-Up
            key == K_CTRL_LEFT                  //--> Ctrl-Cursor-Left
         if inGet
            obj:exitState := GE_UP              //    in GET: prev GET item -->
            exit
         endif
         obj:FirstItem()                        //    first enabled item, consid. SET WRAP

      case  key == K_LEFT .and. !inGet          //--> cursor left
         obj:ColumnLeft := max(1, obj:ColumnLeft - int((obj:Right - obj:Left) / 2) )
         obj:Display()

       case key == K_LEFT                       //--> cursor left
         if inGet
            obj:exitState := GE_UP              //    prev GET item -->
            exit
         endif
         obj:PrevItem()                         //    previous enabled item, consid. WRAP

      case  key == K_RIGHT .and. !inGet         //--> cursor right
         obj:ColumnLeft := obj:ColumnLeft + int((obj:Right - obj:Left) / 2)
         obj:Display()

      case key == K_RIGHT                       //--> cursor right
         if inGet
            obj:exitState := GE_DOWN            //    next GET item -->
            exit
         endif
         obj:NextItem()                         //    next enabled item, consid. WRAP

      case key == K_CTRL_HOME                   //--> Ctrl-Home -->
         if inGet
            obj:exitState := GE_TOP             //    first GET in list
            exit
         endif
         obj:FirstItem()                        //    first enabled item, consid. SET WRAP

      case key == K_CTRL_END                    //--> Ctrl-End -->
         if inGet
            obj:exitState := GE_BOTTOM          //    last GET in list
            exit
         end
         obj:Select(obj:ItemCount)              //    last item in list
         if !obj:GetItem(obj:Buffer)[3]         //    if not selectable,
            obj:PrevItem()                      //    skip to last selectable
         endif

      case key == K_PGUP .or. ;                 //--> PageUp
           (key == K_CTRL_PGUP .and. !inGet)
         nCurr := obj:Buffer
         if valtype(obj:VScroll) $ "CO"
            obj:Scroll(HTSCROLLBLOCKDEC)
         else
            obj:TopItem := max(1, obj:TopItem - iHeight)
            obj:Select(max(1, obj:Buffer - iHeight))
         endif
         if !obj:GetItem(obj:Buffer)[3]         //    if not selectable,
            obj:PrevItem()                      //    skip to last selectable
         endif
         if inGet .and. obj:Buffer == nCurr     //    already first?
            obj:exitState := GE_WRITE           //    exit READ
            exit
         endif

      case key == K_PGDN .or. ;                 //--> PageDown
           (key == K_CTRL_PGDN .and. !inGet)
         nCurr := obj:Buffer
         if valtype(obj:VScroll) $ "CO"
            obj:Scroll(HTSCROLLBLOCKINC)
         else
            obj:TopItem := min(obj:TopItem + iHeight, obj:ItemCount)
            obj:Select(min(obj:ItemCount, obj:Buffer + iHeight))
         endif
         if !obj:GetItem(obj:ItemCount)[3]      //    if not selectable,
            obj:NextItem()                      //    skip to last selectable
         endif
         if inGet .and. obj:Buffer == nCurr     //    already last?
            obj:exitState := GE_WRITE           //    exit READ
            exit
         endif

      case  key == K_CTRL_PGUP .or. ;           //--> Ctrl-PageUp
            key == K_CTRL_PGDN                  //--> Ctrl-PageDown
         if inGet
            obj:exitState := GE_WRITE           //    exit READ
            exit
         endif

      case  key == K_TAB                        //--> Tab -->
         obj:exitState := GE_DOWN               //    next GET item
         exit

      case  key == K_SH_TAB                     //--> Shift-Tab -->
         obj:exitState := GE_UP                 //    prev GET item
         exit

      CASE key = K_MOUSEWHEELUP .and. ;         // mouse wheel up, prev.record
           _aGlobSetting[GSET_L_ACCEPT_WHEEL] 
         nCurr := obj:Buffer
         obj:PrevItem()
                                   
      CASE key = K_MOUSEWHEELDN .and. ;         // mouse wheel down, next record
           _aGlobSetting[GSET_L_ACCEPT_WHEEL] 
         nCurr := obj:Buffer
         obj:NextItem()

      case key >= K_LBUTTONDOWN .and. ;         
            ascan(m->_aGlobSetting[GSET_G_A_LISTBOX_MOUSE], key) > 0
         iHitTest := obj:HitTest(MrowClick(.T.), McolClick(.T.), .T.)
         if iHitTest == 0                       // mouse click outside ListBox,
            if inGet
               obj:exitState := GE_MOUSE        // GetReader() will search 
            endif                               // for the corresp. GET field
            exit                          
         elseif iHitTest > 0  
            if key == K_LBUTTONUP 
               if _aGlobSetting[GSET_G_L_LISTBOX_SINGLE]
                  obj:exitState := GE_ENTER     // simulate Enter = next GET
                  exit
               endif
            else
               if obj:SelectBySingleClick       // click = exit ?
                  obj:exitState := GE_ENTER     // simulate Enter = next GET
                  exit
               endif     
            endif
         endif
                           
      case key >= K_MOUSEMOVE .and. ;
           key <= K_ALT_RBUTTONUP
           // ignore other mouse buttons
           loop

      case key > 0 .and. key <= 255             // check for key in list text
//       retVal := obj:FindText(chr(key), , .F.)            // ignore selectable
         retVal := obj:FindText(chr(key), , .F., , , ,.T.)  // only selectable
         if retVal > 0
            obj:Select(retVal)
         endif
      endcase
   enddo

   if !inGet  
      obj:exitState := GE_NOTUSED
   endif
   
   obj:KillFocus()                              // done, kill focus
   return retVal

/* eof */
