#ifdef COMMENT_ONLY
* For FlagShip 6
* Copyright (c) 2002..2008 by multisoft, Munich, Germany
*
* You may modify this source according to your needs and distribute the
* object code freely with your executable. You may distribute also the
* modified source code when your application is distributed in source,
* but may not remove the copyright notice.
*
* Release $Date: 2008/02/03 11:03:12 $
*
* Default Input handler for MemoEdit() for GUI mode called
* via MemoEdit(). You may replace this handler by your own,
* using the same function names MemoEditHandler(...) and
* _MemoEdHelp(...)
*
* Compile     : FlagShip -c -na -m -io=a memoedithand.prg
* Link Linux  : FlagShip  myapp*.prg  memoedithand.o   [your switches]
* Link Windows: FlagShip  myapp*.prg  memoedithand.obj [your switches]
#endif

#Cinline
#include "FSopenc.h"
#endCinline

// comment-out following #define statement to achive backward compatibility to pre-VFS 6.1.18 
#define NEW_SINCE_VFS6118   

#include "memoedit.fh"
#include "inkey.fh"
#include "dialog.fh"

#ifndef DEVWARNING
#define DEVWARNING   902      /* Develop warning */
// #define GENWARNING   901      /* General warning */
#endif

PROTO STATIC FUNCTION GuiMemoEdHelp(p1,p2,p3,p4)
PROTO STATIC FUNCTION TermMemoEdHelp(p1,p2,p3,p4)
PROTO static proc ShowScoreboard()
PROTO FUNCTION memoed_rc_set(p1) // fix row()/col() for GUI mode

/* --------------------------------------------------------------------------
 * MemoEdit() input handler (here for GUI mode only)
 * This function is trigerred at any key press in MemoEdit(). The editing
 * and movement is done internally, until ME_ABORT or ME_FINISH
 *
 * when pressing F1 key, the default HELP() function
 * calls the MemoEditHelp() defined below.
 *
 * cUdfName: name of user defined function or "" or NIL
 * iMode   : 0..3 = ME_IDLE, ME_UNKEY, ME_UNKEYX, ME_INIT
 * iRow    : current row (1..n)
 * iCol    : current column (0..n)
 * lReadWrite: parameter <expL6> of MemoEdit()
 * xWidget : variable holding the GUI widget ptr or NIL
 * return  :  ME_DEFAULT =   0 = continue
 *            ME_ABORT   = 201 = exit, return original string
 *            ME_FINISH  = 202 = exit, return edited string
 */
#ifdef NEW_SINCE_VFS6118
/* New version starting in VFS version 6.1.18, considering
 * the row() and col() in UDF
 * The old, backward compatible version (VFS 5.x to 6.1.17) is below
 */
FUNCTION MemoEditHandler(cUdfName, _iMode, _iRow, _iCol, lReadWrite, xWidget) as IntVar
   local_int ikey, iShift, iInsert, iCaps
   local_int iRow := _iRow, iCol := _iCol, iMode := _iMode
   local key                  as intvar
   local iUdfRet              as usual
   local retVal := ME_DEFAULT as IntVar
   local isUdf, lHelpSet      as logic
   local oBox                 as object
   local lFastRead, lProcessUdf as logic
   local cOldPrinterFile, cCurPrinterFile as char

   if AppIoMode() != "G"   // currently for GUI mode only
      return ME_DEFAULT
   endif

   isUdf := valtype(cUdfName) == "C" .and. !empty(cUdfName) .and. ;
            IsFunction(cUdfName)

   if iMode == ME_INIT
      ShowScoreboard()                                   // show scoreboard
   endif

   if _aGlobSetting[GSET_L_MEMOE_FAST] == NIL 
      lFastRead := _GlobSetting(GSET_L_READ_FAST, .F.)   // fast read from READ 
   else
      lFastRead := _GlobSetting(GSET_L_MEMOE_FAST, .F.)  // fast read MemoEdit 
   endif
      
   iUdfRet := 0
   lProcessUdf := .F.
   
   if isUdf .and. iMode == ME_INIT
      key := 0                // 1st call: bypass Inkey() to invoke UDF if any
   else
      /* Note that in GUI i/o mode, ISO inkey() input is required
       */
      key := InkeyTrap(0,,,,,.F.)  // = same as Inkey(0) but considers SET KEY

      iShift  := if(InkeyStatus(STATUS_SHIFT),  1, 0)       
      iCaps   := if(InkeyStatus(STATUS_CAPSLOCK),  1, 0)
      iInsert := if(Set(_SET_INSERT), 1, 0)
   endif

   do case
   case key == K_ESC .and. !isUdf            // ESC = exit
      if !lReadWrite
         retVal := ME_ABORT                  // abort, no warning on r/o
      else
         do case
         case set(_SET_MEMOEDITWARN) == 0
            retVal := ME_ABORT               // no warnings at all
         case set(_SET_MEMOEDITWARN) == 1 .and. iMode != ME_UNKEYX
            retVal := ME_ABORT               // warnings on change but no change
         otherwise
            // note: GUI Warning and InfoBox classes are used here
            if iMode == ME_UNKEYX
              oBox := _gWarningBox{m->oApplic, "Abort Edit ...", ;
                          "Do you really want to abort the editing;now and lose all changes ?"}
              oBox:DefButton := 1
           else
              oBox := _gInfoBox{m->oApplic, "Abort Edit ...", ;
                           "Do you really want to abort the editing now ?"}
              oBox:DefButton := 2
           endif
           oBox:Buttons := {"No", "Yes"}
           if oBox:Exec() == 2
              retVal := ME_ABORT             // abort
           endif
         endcase
      endif
   case key == K_CTRL_P .and. !isUdf         // Ctrl-P = print
      cOldPrinterFile := m->_oPrinter:InputFileName
      cCurPrinterFile := TempFileName(NIL,"memo")
      m->_oPrinter:InputFileName := cCurPrinterFile
      m->_oPrinter:Exec()                    // print
      m->_oPrinter:InputFileName := cOldPrinterFile
      retVal := ME_FINISH                    // exit

   case (key == K_CTRL_W .or. key == K_CTRL_END) .and. !isUdf
      retVal := ME_FINISH                    // finish, exit

   case key == K_INS // .and. !isUdf
      if InkeyStatus(STATUS_SHIFT) .or. InkeyStatus(STATUS_MOUSESHIFT)
         key := K_INS                        // Shift + Insert = paste
      else
         if !set(_SET_INSERT_INKEY)          // Insert = toggle insert mode
            Set( _SET_INSERT, !Set(_SET_INSERT) )
         endif
      endif
      retVal := ME_DEFAULT
      ShowScoreboard()                       // show scoreboard
      iInsert := if(Set(_SET_INSERT), 1, 0)

   case isUdf
      lProcessUdf := .T.
   endcase
   
   if (retVal != ME_ABORT .and. retVal != ME_FINISH) 
      iKey := key
      if key != 0
         /* process the key, get current row/col
          */
         #Cinline
         {
            extern int fgsUseMemoEditHandler;  // compatibility flag
            int    iUdfRet = VAR_INT(VAR_NAME_LOCAL(iudfret));
            void  *pWidget = IS_VAR_SPECIAL(VAR_NAME_LOCPAR(xwidget)) ?
                             VAR_SPECIAL(VAR_NAME_LOCPAR(xwidget)) : 0;

            if( pWidget ) {
               if(iinsert && ikey >= 32 && ikey <= 255 )
                  FsGuiMemoEdit_Insert(pWidget, ikey);  // insert key
               else {
                  if(fgsUseMemoEditHandler == 0) {
                     // FS614: compatibility mode to Terminal i/o
                     FsGuiMemoEdit_ProcessKey(pWidget, ikey, ishift, iinsert);
                  } else {
                     // FS613 backward compatibility
                     FsGuiMemoEdit_SetOverwrite(pWidget, !iinsert);
                  }
               }
               irow  = FsGuiMemoEdit_Row(pWidget);
               icol  = FsGuiMemoEdit_Col(pWidget);
               irow  = irow < 0 ? 0 : irow +1;                 /* range 1..n */
               icol  = icol < 0 ? 0 : icol;                    /* range 0..n */ 
               imode = FsGuiMemoEdit_Changed(pWidget, -1) ? 2 : 1;   
            }
         }
         #endCinline
      endif
      if isUdf
         /*
          * call user's UDF
          * in the UDF, the possible inkey() call
          * will not be echoed in the MemoEdit() GUI widget
          */
         memoed_rc_set(xWidget)                 // set row()/col() in GUI mode
         m->_oStdIo:Inkey2Widget(xWidget, .F.)  // don't pass inkey() to widget now
         iUdfRet := &cUdfName(iMode, iRow,iCol) // call user UDF
         m->_oStdIo:Inkey2Widget(xWidget, .T.)  // continue to pass inkey() to widget

         if valtype(iUdfRet) != "N"
             _rt_error(DEVWARNING, "MEMOEDIT(): UDF should return numeric;variable, not '" + ;
                        valtype(iUdfRet) + "' - assuming RETURN 0")
            iUdfRet := 0
         endif
      
         if iUdfRet == 0 .and. ;
            (key == K_ESC .or. key == K_CTRL_W .or. key == K_INS)
            iUdfRet := key
         endif

         do case
         case iUdfRet == K_ESC
            retVal := ME_ABORT
         case iUdfRet == K_CTRL_W .or. iUdfRet == K_CTRL_END
            retVal := ME_FINISH
         case iUdfRet == K_INS .and. iUdfRet != key
            if InkeyStatus(STATUS_SHIFT)
               iUdfRet := 0                       // Shift + Insert = paste
               key := K_INS
            else                                  // Insert = toggle insert mode
               if !set(_SET_INSERT_INKEY)
                  Set( _SET_INSERT, !Set(_SET_INSERT) )
               endif
            endif
            retVal := ME_DEFAULT
            ShowScoreboard()
         endcase
     
         if (retVal != ME_ABORT .and. retVal != ME_FINISH) .and. ;
            (iUdfRet > 2000 .and. iUdfRet <= 2255)

            /* insert the key, get current row/col
             */
            #Cinline
            {
               int    iUdfRet = VAR_INT(VAR_NAME_LOCAL(iudfret));
               void  *pWidget = IS_VAR_SPECIAL(VAR_NAME_LOCPAR(xwidget)) ?
                                VAR_SPECIAL(VAR_NAME_LOCPAR(xwidget)) : 0;

               if( pWidget ) {
                  FsGuiMemoEdit_Insert(pWidget, iUdfRet - 2000);  // insert key
                  irow  = FsGuiMemoEdit_Row(pWidget);
                  icol  = FsGuiMemoEdit_Col(pWidget);
                  irow  = irow < 0 ? 0 : irow +1;                 /* range 1..n */
                  icol  = icol < 0 ? 0 : icol;                    /* range 0..n */ 
                  imode = FsGuiMemoEdit_Changed(pWidget, -1) ? 2 : 1;   
               }
            }
            #endCinline      
            retVal := 0
         endif
      endif
   endif

   return retVal

#else    
/* ---------------------------------------------------------------------
 * Backward compatible version to VFS 6.1.1 to 6.1.17
 * To enable it, comment-out the #ifdef NEW_SINCE_VFS6118 
 * above, re-compile and add the object to your link list
 */
FUNCTION MemoEditHandler(cUdfName, iMode, iRow, iCol, lReadWrite, xWidget) as IntVar
   local_int ikey, iShift, iInsert, iCaps
   local key                  as intvar
   local iUdfRet              as usual
   local retVal := ME_DEFAULT as IntVar
   local isUdf, lHelpSet      as logic
   local oBox                 as object
   local lFastRead            as logic
   local cOldPrinterFile, cCurPrinterFile as char

   if AppIoMode() != "G"   // currently for GUI mode only
      return ME_DEFAULT
   endif

   isUdf := valtype(cUdfName) == "C" .and. !empty(cUdfName) .and. ;
            IsFunction(cUdfName)

   if iMode == ME_INIT
      ShowScoreboard()                                // show scoreboard
   endif

   if _aGlobSetting[GSET_L_MEMOE_FAST] == NIL 
      lFastRead := _GlobSetting(GSET_L_READ_FAST, .F.)   // fast read from READ 
   else
      lFastRead := _GlobSetting(GSET_L_MEMOE_FAST, .F.)  // fast read MemoEdit 
   endif
      
   iUdfRet := 0

   if isUdf .and. iMode == ME_INIT
      key := 0                // 1st call: bypass Inkey() to invoke UDF if any
   else
      /* Note that in GUI i/o mode, ISO inkey() input is required
       */
      key := InkeyTrap(0,,,,,.F.)  // = same as Inkey(0) but considers SET KEY

      iShift  := if(InkeyStatus(STATUS_SHIFT),  1, 0)       // FS6117: moved
      iCaps   := if(InkeyStatus(STATUS_CAPSLOCK),  1, 0)
      iInsert := if(Set(_SET_INSERT), 1, 0)
   endif

   do case
   case key == K_ESC .and. !isUdf            // ESC = exit
      if !lReadWrite
         retVal := ME_ABORT                  // abort, no warning on r/o
      else
         do case
         case set(_SET_MEMOEDITWARN) == 0
            retVal := ME_ABORT               // no warnings at all
         case set(_SET_MEMOEDITWARN) == 1 .and. iMode != ME_UNKEYX
            retVal := ME_ABORT               // warnings on change but no change
         otherwise
            // note: GUI Warning and InfoBox classes are used here
            if iMode == ME_UNKEYX
              oBox := _gWarningBox{m->oApplic, "Abort Edit ...", ;
                          "Do you really want to abort the editing;now and lose all changes ?"}
              oBox:DefButton := 1
           else
              oBox := _gInfoBox{m->oApplic, "Abort Edit ...", ;
                           "Do you really want to abort the editing now ?"}
              oBox:DefButton := 2
           endif
           oBox:Buttons := {"No", "Yes"}
           if oBox:Exec() == 2
              retVal := ME_ABORT             // abort
           endif
         endcase
      endif
   case key == K_CTRL_P .and. !isUdf         // Ctrl-P = print
      cOldPrinterFile := m->_oPrinter:InputFileName
      cCurPrinterFile := TempFileName(NIL,"memo")
      m->_oPrinter:InputFileName := cCurPrinterFile
      m->_oPrinter:Exec()                    // print
      m->_oPrinter:InputFileName := cOldPrinterFile
      retVal := ME_FINISH                    // exit

   case (key == K_CTRL_W .or. key == K_CTRL_END) .and. !isUdf
      retVal := ME_FINISH                    // finish, exit

   case key == K_INS .and. !isUdf
      if InkeyStatus(STATUS_SHIFT) .or. InkeyStatus(STATUS_MOUSESHIFT)
         key := K_INS                        // Shift + Insert = paste
      else
         if !set(_SET_INSERT_INKEY)          // Insert = toggle insert mode
            Set( _SET_INSERT, !Set(_SET_INSERT) )
         endif
      endif
      retVal := ME_DEFAULT
      ShowScoreboard()                       // show scoreboard

   case isUdf
      /*
       * call user's UDF
       * in the UDF, the possible inkey() call
       * will not be echoed in the MemoEdit() GUI widget
       */
      m->_oStdIo:Inkey2Widget(xWidget, .F.)  // don't pass inkey() to widget now
      iUdfRet := &cUdfName(iMode, iRow,iCol) // call user UDF
      m->_oStdIo:Inkey2Widget(xWidget, .T.)  // continue to pass inkey() to widget

      if valtype(iUdfRet) != "N"
          _rt_error(DEVWARNING, "MEMOEDIT(): UDF should return numeric;variable, not '" + ;
                     valtype(iUdfRet) + "' - assuming RETURN 0")
         iUdfRet := 0
      endif

      if iUdfRet == 0 .and. ;
         (key == K_ESC .or. key == K_CTRL_W .or. key == K_INS)
         iUdfRet := key
      endif

      do case
      case iUdfRet == K_ESC
         retVal := ME_ABORT
      case iUdfRet == K_CTRL_W .or. iUdfRet == K_CTRL_END
         retVal := ME_FINISH
      case iUdfRet == K_INS
         if InkeyStatus(STATUS_SHIFT)
            iUdfRet := 0                       // Shift + Insert = paste
            key := K_INS
         else                                  // Insert = toggle insert mode
            if !set(_SET_INSERT_INKEY)
               Set( _SET_INSERT, !Set(_SET_INSERT) )
            endif
         endif
         retVal := ME_DEFAULT
         ShowScoreboard()
      case iUdfRet > 2000      // insert character chr(iUdfRet - 2000)
         retVal := 0
      endcase
   endcase

   if (retVal != ME_ABORT .and. retVal != ME_FINISH) .and. ;
      (iUdfRet == 0 .and. key > 0) .or. (iUdfRet > 2000 .and. iUdfRet <= 2255)
      /*
       * Process key or the UDF return value.
       * The FsGuiMemoEdit_*() are internal GUI functions, handle with care.
       */
      iKey := key
      #Cinline
      {
         extern int fgsUseMemoEditHandler;  // compatibility flag
         int    iUdfRet = VAR_INT(VAR_NAME_LOCAL(iudfret));
         void  *pWidget = IS_VAR_SPECIAL(VAR_NAME_LOCPAR(xwidget)) ?
                          VAR_SPECIAL(VAR_NAME_LOCPAR(xwidget)) : 0;

//       fgsUseMemoEditHandler = 1; //enable for VFS 6.1.3 backward compatibility

         if( pWidget ) {
            if(iUdfRet > 2000)
               FsGuiMemoEdit_Insert(pWidget, iUdfRet - 2000);  // insert key
            else {
               if(fgsUseMemoEditHandler == 0) {
                  // FS614: compatibility mode to Terminal i/o
                  FsGuiMemoEdit_ProcessKey(pWidget, ikey, ishift, iinsert);
               } else {
                  // FS613 backward compatibility
                  FsGuiMemoEdit_SetOverwrite(pWidget, !iinsert);
               }
            }
         }
      }
      #endCinline
   endif

   return retVal
#endif

/**************************************************************************
 *  ShowScoreboard() displays the actual INSERT status, same as getsys.prg
 */
static proc ShowScoreboard()
   local nRow, nCol  AS Numer
   local cMsg        AS Char

#define SCORE_ROW    0
#define SCORE_COL    60

   if ( Set(_SET_SCOREBOARD) )                    // SET SCOREBOARD ON ?
      cMsg := _LANGNAMES (1)                      // get msg from FSsortab.xxx
      nRow := Row()                               // save actuall cursor
      nCol := Col()
      SetPos(SCORE_ROW, SCORE_COL)                // see define
      DispOut( if(Set(_SET_INSERT), cMsg, Space(Strlen2space(cMsg)) ) )
      SetPos(nRow, nCol)                          // restore cursor
   end
   return

/**************************************************************************
 * Glabaly available Help for MemoEdit
 * parameters (all optional) corresponds to HELP()
 */
FUNCTION _MemoEdHelp(p1,p2,p3,p4)
   if AppIoMode() == "G"
      GuiMemoEdHelp(p1,p2,p3,p4)
   elseif AppIoMode() == "T"
      TermMemoEdHelp(p1,p2,p3,p4)
   else
      ? 
      ? "** MemoEdit() for basic mode not available **" 
      ?
   endif
   return
   
/**************************************************************************
 * Help for MemoEdit in GUI mode, parameters not used
 */
STATIC FUNCTION GuiMemoEdHelp(p1,p2,p3,p4)
   local cHelp as char
   cHelp := ;
      '<HTML><B><Font color="#FF0000">Function keys</font> for this embedded editor</B><p><br>' + ;
      '<Font color="#0000FF">Exit/Abort</Font><br>' + ;
      "<table cellspacing=0 cellpadding=0>" + ;
      "<tr><td width=5%> </td><td width=30%><b>ESC</b></td><td width=65%>Abort editing withou save</td></tr>" + ;
      "<tr><td> </td><td><b>Ctrl-W</b></td><td>Save and exit</td></tr>" + ;
      "<tr><td> </td><td><b>Ctrl-End</b></td><td>Save and exit, same as Ctrl-W</td></tr></table>" + ;
      '<Font color="#0000FF">Print</Font><br>' + ;
      "<table cellspacing=0 cellpadding=0>" + ;
      "<tr><td width=5%0> </td><td width=30%><b>Ctrl-P</b></td><td width=65%>Print current content to default printer</td></tr></table>"
   cHelp += ;
      '<Font color="#0000FF">Movement</Font><br>' + ;
      "<table cellspacing=0 cellpadding=0>" + ;
      "<tr><td width=5%> </td><td width=30%><b>Cursor down,up</b></td><td width=65%>next/previous line</td></tr>" + ;
      "<tr><td> </td><td><b>Cursor left,right</b></td><td>previous/next character</td></tr>" + ;
      "<tr><td> </td><td><b>Ctrl-Cursor left,right</b></td><td>previous/next word</td></tr>" + ;
      "<tr><td> </td><td><b>Home,End</b></td><td>first/last character in current line</td></tr>" + ;
      "<tr><td> </td><td><b>Ctrl-Home</b></td><td>begin of current page</td></tr>" + ;
      "<tr><td> </td><td><b>Ctrl-L</b></td><td>end of current page</td></tr>" + ;
      "<tr><td> </td><td><b>PgUp,PgDn</b></td><td>previous/next window page</td></tr>" + ;
      "<tr><td> </td><td><b>Ctrl-PgUp,-PgDn</b></td><td>begin/end of text</td></tr></table>"
   cHelp += ;
      '<Font color="#0000FF">Editing</Font><br>' + ;
      "<table cellspacing=0 cellpadding=0>" + ;
      "<tr><td width=5%> </td><td width=30%><b>Enter, Return</b></td><td width=65%>new line/paragraph</td></tr>" + ;
      "<tr><td> </td><td><b>Insert</b></td><td>toggle insert/overwrite mode</td></tr>" + ;
      "<tr><td> </td><td><b>Delete</b></td><td>delete character under cursor</td></tr>" + ;
      "<tr><td> </td><td><b>Backspace</b></td><td>delete character left of cursor</td></tr>" + ;
      "<tr><td> </td><td><b>Ctrl-Y</b></td><td>delete current line</td></tr>" + ;
      "<tr><td> </td><td><b>Ctrl-T</b></td><td>delete word right</td></tr></table>"

   _gInfoBox{NIL, "MemoEdit Help", cHelp}:Show()
   return NIL


/**************************************************************************
 * Help for MemoEdit in Zerminal i/o mode, parameters not used
 * colors are controlled by _aGlobSetting[GSET_T_C_INFBOXCOLOR]
 */
STATIC FUNCTION TermMemoEdHelp(p1,p2,p3,p4)
   local cHelp as char
   local iMaxRow := MaxRow() as intvar
   
   cHelp := ;
      "Function keys for this embedded editor; ;" + ;
      "<<!Exit/Abort;" + ;
      "<<!  ESC                   Abort editing withou save;" + ;
      "<<!  Ctrl-W                Save and exit;" + ;
      "<<!  Ctrl-End              Save and exit, same as Ctrl-W;" + ;
      "<<!Print;" + ;
      "<<!  Ctrl-P                Print curr. content to def. printer;"
   
   if iMaxRow < 21
      InfoBox(cHelp, "MemoEdit Help")
      cHelp :=  "Function keys for this embedded editor; ;" 
   endif
   
   cHelp += ;
      "<<!Movement;" + ;
      "<<!  Cursor down,up        next/previous line;" + ;
      "<<!  Cursor left,right     previous/next character;" + ;
      "<<!  Ctrl-Curs left,right  previous/next word;" + ;
      "<<!  Home,End              first/last character in current line;" + ;
      "<<!  Ctrl-Home             begin of current page;" + ;
      "<<!  Ctrl-L                end of current page;" + ;
      "<<!  PgUp,PgDn             previous/next window page;" + ;
      "<<!  Ctrl-PgUp,-PgDn       begin/end of text;"
      
   if iMaxRow < 28
      InfoBox(cHelp, "MemoEdit Help")
      cHelp :=  "Function keys for this embedded editor; ;" 
   endif
   
   cHelp += ;
      "<<!Editing;" + ;
      "<<!  Enter, Return         new line/paragraph;" + ;
      "<<!  Insert                toggle insert/overwrite mode;" + ;
      "<<!  Delete                delete character under cursor;" + ;
      "<<!  Backspace             delete character left of cursor;" + ;
      "<<!  Ctrl-Y                delete current line;" + ;
      "<<!  Ctrl-T                delete word right"

   InfoBox(cHelp, "MemoEdit Help")
   return NIL

* eof *
