/**************************************************************************
 * FlagShip 5.x : the GUI Printer class, reimplemented default _Printer class
 * Prototypes   : printerclass.fh 
 *
 * Compile with -na -c -io=g switch.
 *
 * For FlagShip 5. Copyright (c) 2002 by multisoft GmbH, Munich, Germany
 * You may modify this source according to your needs and distribute the
 * object code freely with your executable. You may distribute also the 
 * modified source code when your application is distributed in source, 
 * but may not remove the copyright notice.
 * 
 * Note: this source is intended mainly for your information. It may be 
 * changed without further notification in later FS release.
 *
 * Release $Date: 2005/09/27 17:33:03 $
 */
 
#Cinline
static char *fileID="@@#IOGUI# _gPrinter{} $Revision: 1.7 $";
# include "FSopenc.h"
# include "FsGuiPrinter.h"  /* Note: you may disable this line */
# ifndef __FSGUIPRINTER_H__
typedef void QPrinter;
typedef void QFont;
# endif
#endCinline

PROTO STATIC FUNCTION CheckMacro(p1,p2) as char

/**************************** _gPrinter *************************************/
#include "printerclass.fh"

CLASS _gPrinter INHERIT _Printer ALIAS Printer

/* --- instances are in printerclass.fh
 * --- basic _Printer methods are in o_printer.prg
 */

/*---------------------------------------------------------------------------
 *  :Axit() -> NIL
 *    destructor
 */
METHOD Axit()                             CLASS _gPrinter
   #Cinline
   {
      FSvar *vIntern = OBJ_INSTANCE_DIRECT_SELF(_gprinterInternalProtect); 
      
      if(IS_VAR_SPECIAL(vIntern)) 
         FsGuiPrinter_Delete(VAR_SPECIAL(vIntern));
      SET_VAR_COPY(vIntern, NIL_VAR);
   }
   #endCinline
   return NIL

/*---------------------------------------------------------------------------
 * :Driver?margin -> iPixel
 *   Get the horizontal/vertical margin of the printer driver
 *   If the printer is not set yet, use default printer
 */
ACCESS DriverHmargin                      CLASS _gPrinter
   local_int retVar := 0
   #Cinline
   {
      FSvar *vIntern = OBJ_INSTANCE_DIRECT_SELF(_gprinterInternalProtect); 
      QPrinter *pSetUp;
      if(!IS_VAR_SPECIAL(vIntern)) {
         pSetUp = FsGuiPrinter_Create();
         if(pSetUp) 
            SET_VAR_SPECIAL(vIntern, pSetUp);
      }
      if(IS_VAR_SPECIAL(vIntern) && VAR_SPECIAL(vIntern) ) 
         retvar = FsGuiPrinter_GetHmargin( VAR_SPECIAL(vIntern) );
   }
   #endCinline
   self:iDriverHmargin := retVar
   return retVar
      
ACCESS DriverVmargin                      CLASS _gPrinter
   local_int retVar := 0
   #Cinline
   {
      FSvar *vIntern = OBJ_INSTANCE_DIRECT_SELF(_gprinterInternalProtect); 
      QPrinter *pSetUp;
      if(!IS_VAR_SPECIAL(vIntern)) {
         pSetUp = FsGuiPrinter_Create();
         if(pSetUp) 
            SET_VAR_SPECIAL(vIntern, pSetUp);
      }
      if(IS_VAR_SPECIAL(vIntern) && VAR_SPECIAL(vIntern) ) 
         retvar = FsGuiPrinter_GetVmargin( VAR_SPECIAL(vIntern) );
   }
   #endCinline
   self:iDriverVmargin := retVar
   return retVar
   
/*---------------------------------------------------------------------------
 * :Driver?ldpi -> iLogPixel_per_inch
 *   Get the horizontal/vertical "logical" driver resolution in DPI (dots
 *   per inch). It is not the real printer resolution, but the unit
 *   used for margins. For laser printers,the returned value is usually 
 *   72 x 72 dpi = 2.83 x 2.83 dots per mm.
 -> reimplemented in GUI
 *   If the printer is not set yet, use default printer
 */
ACCESS DriverHldpi                        CLASS _gPrinter
   local_int retVar := 0
   #Cinline
   {
      FSvar *vIntern = OBJ_INSTANCE_DIRECT_SELF(_gprinterInternalProtect); 
      QPrinter *pSetUp;
      if(!IS_VAR_SPECIAL(vIntern)) {
         pSetUp = FsGuiPrinter_Create();
         if(pSetUp) 
            SET_VAR_SPECIAL(vIntern, pSetUp);
      }
      if(IS_VAR_SPECIAL(vIntern) && VAR_SPECIAL(vIntern) ) 
         retvar = FsGuiPrinter_GetHldpi( VAR_SPECIAL(vIntern) );
   }
   #endCinline
   self:iDriverHmargin := retVar
   return retVar

ACCESS DriverVldpi                        CLASS _gPrinter
   local_int retVar := 0
   #Cinline
   {
      FSvar *vIntern = OBJ_INSTANCE_DIRECT_SELF(_gprinterInternalProtect); 
      QPrinter *pSetUp;
      if(!IS_VAR_SPECIAL(vIntern)) {
         pSetUp = FsGuiPrinter_Create();
         if(pSetUp) 
            SET_VAR_SPECIAL(vIntern, pSetUp);
      }
      if(IS_VAR_SPECIAL(vIntern) && VAR_SPECIAL(vIntern) ) 
         retvar = FsGuiPrinter_GetVldpi( VAR_SPECIAL(vIntern) );
   }
   #endCinline
   self:iDriverHmargin := retVar
   return retVar
   
/*---------------------------------------------------------------------------
 * :ExecFormatted([fileName], [oFont]) -> lOk
 *    reads the <fileName>, reformats according to font, headers and margins,
 *    invokes the printer-spooler according to :SetUp()
 -> Reimplemented in Terminal and GUI i/o
 *  In GUI, following font attributes within the text are supported:
 *    <b>...</b> - Bold font style. 
 *    <i>...</i> - Italic font style. 
 *    <u>...</u> - Underlined font style.         
 */
METHOD ExecFormatted([p1],[p2])            CLASS _gPrinter
   local_int iOk := 0
   local retVal := .F. as logic
   local cLeft, cMid, cRight as char

   cLeft  := CheckMacro(self:cHeadLeft, p1) 
   cMid   := CheckMacro(self:cHeadMid,  p1)
   cRight := CheckMacro(self:cHeadRight,p1)
   
   if valtype(p1) != "C" .or. empty(p1) .or. !file(p1)
      if !empty(self:InputFileName)
         p1 := self:InputFileName
      else
         return .F.
      endif
   endif
   
   #Cinline
   {
      QPrinter *pSetUp;
      FSvar *vIntern   = OBJ_INSTANCE_DIRECT_SELF(_gprinterInternalProtect); 
      FSvar *vaTabStop = OBJ_INSTANCE_DIRECT_SELF(_gprinterAtabsProtect); 
      FSvar *ovFont, *vFontPtr;
      QFont *pFont = 0;
      int   *aiTabs = 0;
      int    ii, iCopies;
      
      /* create 0-terminated int C-array of tabs, 
       * when ::TabStop array is specified
       */
      if(VAR_ISDIM(vaTabStop) > 0) {
         aiTabs = (int *) malloc( (VAR_ISDIM(vaTabStop) +1) * sizeof(int));
         for(ii=1; ii <= VAR_ISDIM(vaTabStop); ii++) 
            aiTabs[ii-1] = VAR_INTNUM(VAR_ARRELEM(vaTabStop,ii));
         aiTabs[ii-1] = 0;
      }
      
      /* set font if given as param2
       */
      ovFont = VAR_NAME_LOCPAR(p2);
      if(IS_VAR_OBJ(ovFont)) {
         vFontPtr = (FSvar *) OBJ_ACCEXEC(ovFont, "fontptr", -1);
         pFont    = (QFont *) VAR_SPECIAL(vFontPtr);
      }
            
      /* instantiate QPrinter if not already done
       */
      if(!IS_VAR_SPECIAL(vIntern)) {
         pSetUp = FsGuiPrinter_Create();
         if(pSetUp) 
            SET_VAR_SPECIAL(vIntern, pSetUp);
      }
      
      /* if printer available: print formatted
       * using the given font or "Courier",10 otherwise
       */
      if(IS_VAR_SPECIAL(vIntern)) {
         pSetUp = VAR_SPECIAL(vIntern);

         iCopies = VAR_INT(OBJ_INSTANCE_DIRECT_SELF(_gprinterInumcopiesProtect));
         if(iCopies < 1)
            iCopies = 1;            
         iok = 1;
         
         for(ii = 1; ii <= iCopies && iok; ii++)
            iok = FsGuiPrinter_PrintFormated(pSetUp, 
                     VAR_CHR(VAR_NAME_LOCPAR(p1)), 
                     VAR_INT(OBJ_INSTANCE_DIRECT_SELF(_gprinterIpageleftmProtect)),
                     VAR_INT(OBJ_INSTANCE_DIRECT_SELF(_gprinterIpagetopmaProtect)),
                     VAR_INT(OBJ_INSTANCE_DIRECT_SELF(_gprinterIpagerightProtect)),
                     VAR_INT(OBJ_INSTANCE_DIRECT_SELF(_gprinterIpagebotmaProtect)),
                     VAR_CHR(VAR_NAME_LOCAL(cleft)),
                     VAR_CHR(VAR_NAME_LOCAL(cmid)),
                     VAR_CHR(VAR_NAME_LOCAL(cright)),
                     aiTabs, pFont );
      } 
      SET_VAR_LOG(VAR_NAME_LOCAL(retval), iok); 
      if(aiTabs)
         free(aiTabs);
   }
   #endCinline
   return retVal

/*---------------------------------------------------------------------------
 * :ExecPrintScreen([lAdapt]) -> lOk
 *    prints the whole user-screen to default spooler according to :SetUp()
 * The _oprinter:PrintScreen() is also executed from main menu, see 
 *    initiomenu.prg either via STATIC FUNCTION InitIoScrPrint(obj, menuID) 
 *    or user redefined
 -> Reimplemented in Terminal and GUI i/o
 */
METHOD ExecPrintScreen([p1])                  CLASS _gPrinter
   local_int iOk := 0
   local_int iTopMar   := self:iPageTopMarg
   local_int iBotMar   := self:iPageBotMarg
   local_int iLeftMar  := self:iPageLeftMarg
   local_int iRightMar := self:iPageRightMarg
   local_int iAdapt := if(valtype(p1) == "L" .and. p1, 1, 0)
   local retVal := .F. as logic
   
   #Cinline
   {
      QPrinter *pSetUp;
      FSvar *vIntern = OBJ_INSTANCE_DIRECT_SELF(_gprinterInternalProtect); 

      /* instantiate QPrinter if not already done
       */
      if(!IS_VAR_SPECIAL(vIntern)) {
         pSetUp = FsGuiPrinter_Create();
         if(pSetUp) 
            SET_VAR_SPECIAL(vIntern, pSetUp);
      }
      if(IS_VAR_SPECIAL(vIntern)) {
         pSetUp = VAR_SPECIAL(vIntern);
    
         iok = FsGuiPrinter_PrintScreen(pSetUp, 
                                        ileftmar, itopmar, irightmar, ibotmar,
                                        iadapt );
      } 
      SET_VAR_LOG(VAR_NAME_LOCAL(retval), iok); 
   }
   #endCinline
   return retVal

/*---------------------------------------------------------------------------
 * Dialog to select priner properties.
 * The _oprinter:Setup() is also executed from main menu, see initiomenu.prg
 *    either via STATIC FUNCTION InitIoPrSet(obj, menuID) or user redefined
 * Reimplemented for GUI i/o
 */
METHOD Setup()                            CLASS _gPrinter   
   local_int iSetupOk := -1
   #Cinline
   {
      QPrinter *pSetUp;
      FSvar *vIntern = OBJ_INSTANCE_DIRECT_SELF(_gprinterInternalProtect); 
      char  *cOutputFile = 0;
      
      /* instantiate QPrinter if not already done
       */
      if(!IS_VAR_SPECIAL(vIntern)) {
         pSetUp = FsGuiPrinter_Create();
         if(pSetUp) 
            SET_VAR_SPECIAL(vIntern, pSetUp);
      }
      if(IS_VAR_SPECIAL(vIntern)) {
         pSetUp = VAR_SPECIAL(vIntern);
         
         /* fill/update class instances
          */
         if(!IS_VAR_EMPTY(OBJ_INSTANCE_DIRECT_SELF(_gprinterCoutputfilProtect)))
            cOutputFile = VAR_CHR(OBJ_INSTANCE_DIRECT_SELF(_gprinterCoutputfilProtect));

         FsGuiPrinter_SetValues(pSetUp, 
                 cOutputFile,
                 VAR_LOG(OBJ_INSTANCE_DIRECT_SELF(_gprinterLlandscapeProtect)),
                 VAR_INT(OBJ_INSTANCE_DIRECT_SELF(_gprinterInumcopiesProtect)),
                 VAR_LOG(OBJ_INSTANCE_DIRECT_SELF(_gprinterLlastfirstProtect)),
                 VAR_INT(OBJ_INSTANCE_DIRECT_SELF(_gprinterIpagefirstProtect)),
                 VAR_INT(OBJ_INSTANCE_DIRECT_SELF(_gprinterIpagelastProtect )),
                 VAR_CHR(OBJ_INSTANCE_DIRECT_SELF(_gprinterCpagesizeProtect )),
                 VAR_LOG(OBJ_INSTANCE_DIRECT_SELF(_gprinterLcolorProtect    )) );
                 
         /* user-dialog
          */
         isetupok = FsGuiPrinter_Setup(pSetUp);  // 1=ok, 0=failure,abort
         
         /* store new settings
          */
         if( isetupok ) {
            SET_VAR_CHR(OBJ_INSTANCE_DIRECT_SELF(_gprinterCoutputfilProtect), 
                        (char *)FsGuiPrinter_GetOutputFile(pSetUp)); 
            SET_VAR_LOG(OBJ_INSTANCE_DIRECT_SELF(_gprinterLlandscapeProtect),
                        FsGuiPrinter_GetLandscape(pSetUp));
            SET_VAR_INT(OBJ_INSTANCE_DIRECT_SELF(_gprinterInumcopiesProtect),
                        FsGuiPrinter_GetNumCopies(pSetUp));
            SET_VAR_LOG(OBJ_INSTANCE_DIRECT_SELF(_gprinterLlastfirstProtect),
                        FsGuiPrinter_GetLastFirst(pSetUp));
            SET_VAR_INT(OBJ_INSTANCE_DIRECT_SELF(_gprinterIpagefirstProtect),
                        FsGuiPrinter_GetPageFirst(pSetUp));
            SET_VAR_INT(OBJ_INSTANCE_DIRECT_SELF(_gprinterIpagelastProtect),
                        FsGuiPrinter_GetPageLast(pSetUp));
            SET_VAR_CHR(OBJ_INSTANCE_DIRECT_SELF(_gprinterCpagesizeProtect), 
                        (char *)FsGuiPrinter_GetPageSize(pSetUp)); 
            SET_VAR_CHR(OBJ_INSTANCE_DIRECT_SELF(_gprinterCprintdrivProtect), 
                        (char *)FsGuiPrinter_GetDriver(pSetUp)); 
            SET_VAR_CHR(OBJ_INSTANCE_DIRECT_SELF(_gprinterCprintexenProtect), 
                        (char *)FsGuiPrinter_GetExe(pSetUp)); 
            SET_VAR_LOG(OBJ_INSTANCE_DIRECT_SELF(_gprinterLcolorProtect),
                        FsGuiPrinter_GetColor(pSetUp));
         }
      }
   }
   #endCinline
   if iSetupOk != -1    // change only if FsGuiPrinter_Setup() was invoked
      self:lAborted := (iSetupOk == 0)
   endif
   return self


/****************************************************************************
 * helper
 */
STATIC FUNCTION CheckMacro(p1, p2) as char
   local retVal as char
   
   retVal := strtran(p1,     "<date>", dtoc(date()) )
   retVal := strtran(retVal, "<time>", time() )
   retVal := strtran(retVal, "<file>", p2 )
   return retVal
    
/* eof */
