* For FlagShip 6 
* Copyright (c) 2002..06 by multisoft GmbH, Munich, Germany 
*
* You may modify this source according to your needs and distribute the
* object code freely with your executable. You may distribute also the 
* modified source code when your application is distributed in source, 
* but may not remove the copyright notice.
* 
* Release $Date: 2006/07/29 07:58:33 $
*
* Default Input handler for one RadioGroup, called from RadioGroup:Show() 
*
* You may assign YOUR own handler UDF via code block :Handler, i.e.
* oRadGrp:Handler := {|obj| MyUdf(obj) }
* which is then called from oRadGrp:Show() automatically instead 
* of this default handler.
*
* Compile     : FlagShip -c -na -m -io=a radgrouphand.prg
* Link Linux  : FlagShip myapp*.prg radgrouphand.o
* Link Windows: FlagShip myapp*.prg radgrouphand.obj

#include "boxclass.fh"
#include "inkey.fh"
#include "box.fh"
#include "button.fh"
#include "set.fh"
#include "FSerrmsg.h"
#include "getexit.fh"

/* --------------------------------------------------------------------------
 * Default RadioGroup input handler for a GROUP of Radio Buttons
 * This function is called from oRadGrp:Show() or oRadGrp:Exec() only
 *
 * Parameter passed from oRadGrp:Show() via oRadGrp:Handler block,
 * i.e. here oRadGrp:Handler := {|obj| RadGroupHandler(obj) }
 *    obj = current object
 * Returns current obj:Buffer status or NIL on error
 */
function RadGroupHandler(obj)
   local key, iPos, iHitTest  as intvar
   local iMouseRow, iMouseCol as intvar
   local cAppType             as char
   local inGet                as logic
   local nCurr                as usual
   local lRigthButt := .F., lMidButt := .F.  as logic
   
   if valtype(obj) != "O" .or. !IsObjClass(obj, "_BoxGroup") 
      _Rt_Error(RTARGTYP, "RadGroupHandler(): wrong object class - " + ;
         if(valtype(obj) != "O", "not object", IsObjClass(obj)[1]) )
      return NIL 
   endif

   obj:SetFocus()                // set focus to group
   obj:Display()

   inGet := valtype(obj:exitState) == "N" .and. obj:exitState >= GE_NOTSET

   if obj:Buffer == 0
      obj:FirstItem()
   else
      obj:GetItem(obj:Buffer)
      obj:Select(.T.)
   endif

   /* remove all pending keys in buffer
    */
   clear typeahead

   while .T.
      if valtype(obj:Keypress) == "N" .and. obj:Keypress != 0
         key := obj:Keypress
         obj:Keypress := NIL
      else
         key := InkeyTrap(0, INKEY_ALL_BUT_MOVE)  // same as Inkey(0) but considers SET KEY
      endif

      iMouseCol := MCol(.T.)
      iMouseRow := MRow(.T.)

      do case
      case key == K_HOME                  //--> Home
         obj:FirstItem()                  //    first enabled item, consid. SET WRAP

      case key ==  K_END                  //--> End
         obj:LastItem()                   //    last enabled item, consid. SET WRAP

      case key == K_DOWN                  //--> cursor down:
         nCurr := obj:Buffer
         obj:NextItem()                   //    next enabled item, consid. SET WRAP
         if inGet .and. obj:Buffer == nCurr
            obj:exitState := GE_DOWN      //    if already last: next GET item 
            exit
         endif

      case key == K_UP                    //--> cursor up:
         nCurr := obj:Buffer
         obj:PrevItem()                   //    previous enabled item, consid. WRAP
         if inGet .and. obj:Buffer == nCurr
            obj:exitState := GE_UP        //    if already first: prev GET item 
            exit
         endif

      case  key == K_CTRL_UP .or. ;       //--> Ctrl-Cursor-Up
            key == K_CTRL_LEFT            //--> Ctrl-Cursor-Left
         if inGet 
            obj:exitState := GE_UP        //    in GET: prev GET item -->
            exit
         endif
         obj:FirstItem()                  //    first enabled item, consid. SET WRAP

      case  key == K_CTRL_DOWN .or. ;     //--> Ctrl-Cursor-Down
            key == K_CTRL_RIGHT           //--> Ctrl-Cursor-Right
         if inGet 
            obj:exitState := GE_DOWN      //    in GET: next GET item -->
            exit
         endif
         obj:LastItem()                   //    last enabled item, consid. SET WRAP

       case key == K_LEFT                 //--> cursor left 
         if inGet 
            obj:exitState := GE_UP        //    prev GET item -->
            exit
         endif
         obj:PrevItem()                   //    previous enabled item, consid. WRAP
         
      case key == K_RIGHT                 //--> cursor right 
         if inGet 
            obj:exitState := GE_DOWN      //    next GET item -->
            exit
         endif
         obj:NextItem()                   //    next enabled item, consid. WRAP

      case key == K_CTRL_HOME             //--> Ctrl-Home -->
         if inGet
            obj:exitState := GE_TOP       //    first GET in list
            exit
         endif
         obj:FirstItem()                  //    first enabled item, consid. SET WRAP
         
      case key == K_CTRL_END              //--> Ctrl-End -->
         if inGet
            obj:exitState := GE_BOTTOM    //    last GET in list 
            exit
         endif
         obj:LastItem()                   //    last enabled item, consid. SET WRAP

      case  key == K_PGUP  .or. ;         //--> PageUp -->
            key == K_PGDN  .or. ;         //--> PageDown (in GET only) -->
            key == K_CTRL_PGUP .or. ;     //--> Ctrl-PageUp 
            key == K_CTRL_PGDN            //--> Ctrl-PageDown 
         obj:exitState := GE_WRITE        //    exit READ
         exit

      case  key == K_ESC                  //--> Esc -->   
         obj:exitState := GE_ESCAPE  
         exit                             //    exit READ
         
      case  key == K_ENTER                //--> Return -->
         obj:Select(NIL, .T.)             //    select ON & exit
         obj:exitState := GE_ENTER        //    next GET item 
         exit

      case  key == K_TAB                  //--> Tab -->
         obj:exitState := GE_DOWN         //    next GET item
         exit
         
      case  key == K_SH_TAB               //--> Shift-Tab -->
         obj:exitState := GE_UP           //    prev GET item
         exit

      /* MidMouseUp and RightMouseUp (if allowed) will toggle 
       * item in non-exclusive RadioGroup
       */
      case key == K_MBUTTONUP .and. ; 
            ascan(m->_aGlobSetting[GSET_G_A_RADGRP_MOUSE], key) > 0
         if !obj:Exclusive
            inkey(0.2)
            obj:Select(NIL, NIL)		   // toggle on/off -->
         else
            lMidButt := .T.                  // considered in LeftMouseUp below
            obj:Keypress := K_LBUTTONUP
         endif
         
      case key == K_RBUTTONUP .and. ; 
            ascan(m->_aGlobSetting[GSET_G_A_RADGRP_MOUSE], key) > 0
         lRigthButt := .T.                // considered in LeftMouseUp below
         obj:Keypress := K_LBUTTONUP

      /* LeftMouseUp (and Mid/RightMouseUp if allowed) will select item.
       * Note, tricky: the default GUI behavior is only Mouse-Left-Button,
       * but when _aGlobSetting[GSET_G_A_RADGRP_MOUSE] allows another
       * buttons, the driver simulates/sends Mouse-Left-Button event/key
       */
      case key >= K_LBUTTONDOWN .and. ;         
            ascan(m->_aGlobSetting[GSET_G_A_RADGRP_MOUSE], key) > 0
         iHitTest := obj:HitTest(MrowClick(.T.), McolClick(.T.), .T.)
         while inkey() != 0
         enddo
         if iHitTest == 0                 //  - mouse click outside RadioGroup,
            if inGet
               obj:exitState := GE_MOUSE  //    GetReader() will search 
            endif                         //    for the corresp. GET field
            exit                          
         endif    
         if !obj:Exclusive                      
            obj:Select(NIL, NIL)          // toggle current item on/off
         else
            obj:Select(NIL, .T.)          // set current item on 
         endif
         if !(lMidButt .or. lRigthButt) .and. ;
               _aGlobSetting[GSET_G_L_RADGRP_SINGLE]  
            obj:exitState := GE_ENTER  
            exit                          // LMB: exit RadioGroup to next GET
         endif               
         lMidButt := lRigthButt := .F.    // MMB, RMB: stay in RadioGroup

      case key > 0 .and. key < 255 .and. chr(key) $ " xX+yYtT-nNfF"
         if obj:Exclusive                 // exclusive: set current on
            obj:Select(NIL, .T.)
            obj:exitState := GE_ENTER     //    next GET item 
            exit
         else                             // non-exclusive mode:
            do case              
            case chr(key) $ " xX"         // space or X
               obj:Select(NIL, NIL)		   // toggle on/off -->
               obj:exitState := GE_ENTER  // next GET item 
               exit
            case chr(key) $ "+yYtT"       // + or Y or T
               obj:Select(NIL, .T.)       // select ON 
            case chr(key) $ "-nNfF"       // - or N or F
               obj:Select(NIL, .F.)       // select OFF 
            endcase
         endif
         
      otherwise
         iPos := obj:GetAccel(key)
         if iPos > 0
            if !obj:GetItem(iPos):Enabled    // go to accel.key item
               obj:NextItem()                // or the next enabled
            endif
         endif
      endcase       
   enddo

   if !inGet  
      obj:exitState := GE_NOTUSED
   endif

   obj:KillFocus()               // done, kill focus

   clear typeahead
   inkey()                       // clear
   return obj:Buffer
   
/* eof */
