# FSansi2oem.pl = Polish tables used by Ansi2Oem() and Oem2Ansi()
#   CP-852    <-->   ISO-8559-2 
#    134    c-acute       230
#    136    l-stroke      179
#    141    Z-acute       172
#    143    C-acute       198
#    151    S-acute       166
#    152    s-acute       182
#    157    L-stroke      163
#    164    A-cedilla     161
#    165    a-cedilla     177
#    168    E-cedilla     202
#    169    e-cedilla     234
#    171    z-acute       188
#    189    Z-dot above   175
#    190    z-dot above   191
#    224    O-acute       211
#    227    N-acute       209
#    228    n-acute       241
#  
# Assign by FS_SET("ansi2oem","FSansi2oem.pl")
#
# Both ":pc8_to_iso" and ":iso_to_pc8" table identifiers are required
# Character-Value replacement syntax: 
#  chr: val
#  chr-chr: val
# where <chr> = decimal equivalence of the character within the table
#           * = the same resulting value, i.e. <chr> = <chr>
#       <val> = replacement value (0..255) 
#      -<val> = close-to but not exact char available, result = abs(<val>) 
#               or "?", depending on param2 of Ansi2oem() and Oem2Ansi()
#        -999 = replacement value not available, return "?" or the
#               character passed in param3 of Ansi2oem() and Oem2Ansi()
#
# comments: rest of line starting with the '#' character is ignored
#           empty lines are ignored

#--------------------------------------------------------------------
# (left:) ASCII / IBM-PC-8 / OEM (CP-852)
# (right) Ansi / ISO-8559-2 / Latin1 equivalence

:pc8_to_iso
   0-20:    *           #  (chars are equivalent)
   21:      167         # paragraph
   22-127:  *           #  (chars are equivalent)
   128:     199         # C-cedilla    
   129:     252         # u-umlaut     
   130:     233         # e-acute      
   131:     226         # a-circumflex 
   132:     228         # a-umlaut     
   133:     224         # a-grave      
   134:     229         # a-circle     
   135:     231         # c-cedilla    
   136:     234         # e-circumflex 
   137:     235         # e-dieresis   
   138:     232         # e-grave      
   139:     239         # i-dieresis   
   140:     238         # i-circumflex 
   141:     236         # i-grave      
   142:     196         # A-umlaut     
   143:     197         # A-circle     
   144:     201         # E-acute      
   145:     230         # ae
   146:     198         # AE
   147:     244         # o-circumflex 
   148:     246         # o-umlaut     
   149:     242         # o-grave      
   150:     251         # u-circumflex 
   151:     249         # u-grave      
   152:     255         # y-dieresis   
   153:     214         # O-umlaut     
   154:     220         # U-umlaut     
   155:     162         # c-cedilla
   156:     163         # pound
   157:     165         # yen
   158:     -80         # Pt                      (-> P)
   159:     -102        # f                       (-> f)
   160:     225         # a-acute
   161:     237         # i-acute
   162:     243         # o-acute
   163:     250         # u-acute
   164:     241         # n-tilde
   165:     209         # N-tilde
   166:     170         # a-subscript
   167:     186         # o-subscript
   168:     191         # ?-reverse
   169:     -999        # graphical symbol IBM-PC (-> n/a)
   170:     172         # graphical symbol IBM-PC
   171:     189         # 1/2
   172:     188         # 1/4
   173:     161         # !-reverse
   174:     171         # <<
   175:     187         # >>
#note: chars 176-223 are drawn at output when SET GUITRANSL TEXT+ASCII are ON
   176-223: -999        # graphical symbol IBM-PC (-> n/a)
#some of above can redefined now if not drawn graphically, e.g.:
   179:     166         # graphical symbol IBM-PC (vertical line   -> |)
   191:     -43         # graphical symbol IBM-PC (right top edge  -> +)
   192:     -43         # graphical symbol IBM-PC (left bott edge  -> +)
   196:     -45         # graphical symbol IBM-PC (horiz line      -> -)
   217:     -43         # graphical symbol IBM-PC (right bott edge -> +)
   218:     -43         # graphical symbol IBM-PC (left  top  edge -> +)
#end of redefinition
   224:     -999        # alpha                   (-> n/a)
   225:     223         # beta, ss, sz
   226-229: -999        # greek chars             (-> n/a)
   230:     181         # greek my
   231-232: -999        # greek chars             (-> n/a)
   233:     -999        # greek theta             (-> n/a)
   234:     -999        # greek sigma             (-> n/a)
   235:     -999        # greek delta             (-> n/a)
   236:     -999        # endless symbol          (-> n/a)          
   237:     248         # greek phi
   238-240: -999        # greek chars             (-> n/a)
   241:     177         # +-
   242-247: -999        # math symbols            (-> n/a)
   248:     176         # o-subscript
   249:     -999        # dot                     (-> n/a)
   250:     183         # dot
   251:     -999        # math symbol             (-> n/a)
   252:     -110        # n-subscript             (-> n)
   253:     178         # 2-subscript
   254:     -999        # graphical symbol IBM-PC (-> n/a)
   255:     -32         # none                    (-> space)

#redefinitions of above for CP-852 --> ISO-8559-2
   134:     230         # c-acute
   136:     179         # l-stroke
   141:     172         # Z-acute
   143:     198         # C-acute
   151:     166         # S-acute
   152:     182         # s-acute
   157:     163         # L-stroke
   164:     161         # A-cedilla
   165:     177         # a-cedilla
   168:     202         # E-cedilla
   169:     234         # e-cedilla
   171:     188         # z-acute
   189:     175         # Z-dot above
   190:     191         # z-dot above
   224:     211         # O-acute
   227:     209         # N-acute
   228:     241         # n-acute

#--------------------------------------------------------------------
# (left:) Ansi / ISO-8559-1 / Latin1 
# (right) ASCII / IBM-PC-8 / OEM (CP-437 or CP-850) equivalence

:iso_to_pc8  
   0-127:     *         #  (chars are equivalent)
   128-160: -999        #  (chars not defined in ISO-8559-1)
   161:     173         # !-reverse
   162:     155         # c-cedilla
   163:     156         # pound
   164:     -15         # graphical symbol
   165:     157         # yen
   166:     179         # graphical symbol vertical line 
   167:      21         # paragraph
   168:     -34         # dieresis
   169:     -99         # copyright                (-> c)
   170:     166         # a-subscript
   171:     174         # <<
   172:     170         # graphical symbol  
   173:      45         # dash
   174:     -82         # registered               (-> R)
   175:     -45         # dash                     (-> -)
   176:     248         # o-subscript
   177:     241         # +-
   178:     253         # 2-subscript
   179:     -51         # 3-subscript              (-> 3)
   180:      39         # acute accent
   181:     230         # greek my  
   182:      20         # end-of-paragraph
   183:     250         # dot   
   184:      44         # comma
   185:     -49         # 1-subscript              (-> 1)
   186:     167         # o-subscript
   187:     175         # >>
   188:     172         # 1/4
   189:     171         # 1/2   
   190:     -999        # 3/4                      (-> n/a)   
   191:     168         # ?-reverse
   192-195: -65         # A-diacrtitical mark      (-> A)
   196:     142         # A-umlaut
   197:     143         # A-circle
   198:     146         # AE
   199:     128         # C-cedilla
   200:     -69         # E-grave                  (-> E)
   201:     144         # E-acute
   202-203: -69         # E-diacrtitical mark      (-> E)   
   204-207: -73         # I-diacrtitical mark      (-> I)
   208:     -68         # D-diacrtitical mark      (-> D)
   209:     165         # N-tilde
   210-213: -79         # O-diacrtitical mark      (-> O)
   214:     153         # O-umlaut
   215:    -120         # cross                    (-> x)
   216:     -79         # O-diacrtitical mark      (-> O)
   217-219: -85         # U-diacrtitical mark      (-> U)
   220:     154         # U-umlaut
   221:     -89         # Y-diacrtitical mark      (-> Y)
   222:    -112         # P-diacrtitical mark      (-> p)
   223:     225         # beta, ss, sz
   224:     133         # a-grave
   225:     160         # a-acute
   226:     131         # a-circumflex
   227:     -97         # a-tilde                  (-> a)
   228:     132         # a-umlaut
   229:     134         # a-circle
   230:     145         # ae
   231:     135         # c-cedilla
   232:     138         # e-grave
   233:     130         # e-acute
   234:     136         # e-circumflex
   235:     137         # e-dieresis
   236:     141         # i-grave
   237:     161         # i-acute
   238:     140         # i-circumflex
   239:     139         # i-dieresis
   240:    -235         # d-diacrtitical mark      (-> delta)
   241:     164         # n-tilde
   242:     149         # o-grave
   243:     162         # o-acute
   244:     147         # o-circumflex
   245:    -111         # o-tilde                  (-> o)
   246:     148         # o-umlaut
   247:     246         # -:-
   248:     237         # greek phi
   249:     151         # u-grave
   250:     163         # u-acute
   251:     150         # u-circumflex
   252:     129         # u-umlaut
   253:    -121         # y-acute                  (-> y)
   254:    -112         # p-diacrtitical mark      (-> p)
   255:     152         # y-dieresis
   
#redefinitions of above for ISO-8559-2 --> CP-852 
   161:     164         # A-cedilla
   163:     157         # L-stroke
   166:     151         # S-acute
   172:     141         # Z-acute
   175:     189         # Z-dot above
   177:     165         # a-cedilla
   179:     136         # l-stroke
   182:     152         # s-acute
   188:     171         # z-acute
   191:     190         # z-dot above
   198:     143         # C-acute
   202:     168         # E-cedilla
   209:     227         # N-acute
   211:     224         # O-acute
   230:     134         # c-acute
   234:     169         # e-cedilla
   241:     228         # n-acute

#--------------------------------------------------------------------
# (left:) ASCII / IBM-PC-8 / OEM (CP-852)
# (right) >0: draw as graphic char (179..218) if SET GUITRANS TEXT is ON
#         you may redefine e.g. double-line to single-line by 205:196 etc.

:pc8_draw_text
   0-178:   0           # don't draw
   179-218: *           # draw as graphic
   219-255: 0           # don't draw
   189-190: 0           # CP-852 exception: don't draw #189,190
   
#eof
