/* --------------------------------------------------------
 * Stand-alone demo for the windowing system of FS2-Toolbox
 * you will need FS2 Toolbox installed to be able to use it
 *
 * For FlagShip 5 and 6
 * Copyright (c) 2003..07 by multisoft GmbH, Munich, Germany 
 *
 * You may freely modify this source according to your needs and/or
 * re-use any parts of it in your source code.
 * Release $Date: 2007/06/27 07:31:31 $
 *
 * Copy the source to your working directory, and
 *    Compile Linux: FlagShip windemo.prg -o windemo
 *    or MS-Windows: FlagShip windemo.prg
 *    Run Linux    : "./windemo" -or- "newfswin ./windemo"
 *    Run Windows  : windemo -or- windemo -io=t
 */
#include "fspreset.fh"        // changes frame of ErrorBox
SET GUICOLOR OFF
_aGlobSetting[GSET_T_C_ERRBOX]  := B_PLAIN // reset default frame
_aGlobSetting[GSET_T_C_TEXTBOX] := B_PLAIN // reset default frame

set font "courier"            // apply for GUI mode
setcolor("R+/B")              // apply for Terminal i/o mode

private aWinID := {}          // array holding Window-ID numbers

/* First lets create some Windows   
 */
setclearb("")                 // set clear character within window
for ii = 2 to 16 step 2       
   top    = ii / 2 
   left   = ii 
   bottom = min(top + max(ii, 2), max_row())
   right  = left + ii * 3 
   iWinID = wopen(top, left, bottom, right)  // create window
   if iWinID < 0
      select(0)
      alert("could not open subwindow #" + ltrim(ii))
      quit
   endif
   aadd(aWinID, iWinID)       // add window-ID into array
   wbox(0)                    // draw frame around window using setclearb() char
next

wformat(2,2,2,2)              // reduce usable window area (terminal i/o only)
row := 0
col := 0
if AppIoMode() == "G"         // since wformat() does not reduce GUI area,
   row := 2                   // we need to add these values to coordinates
   col := 2
   set color to "B"
// set guicolors on           // accept SET COLOR in GUI mode
   cls
endif
                              // display some lines in the current window:
@ row++,9+col say "*********  *********    *********  "
@ row++,8+col say "*********  *********    ********** "
@ row++,7+col say "**         **                   ** "
@ row++,6+col say "**         **                   ** "
@ row++,5+col say "******     *********     ********  "
@ row++,4+col say "**                **    **         "
@ row++,3+col say "**                **    **         "
@ row++,2+col say "**        **********    ********** "
@ row++,1+col say "**        **********    ********** "
setclearb(" ")                // reset clear character
?
set guicolors off

/* infoboxes within sub-window, 
 * header left justified
 */
cText := "<<!Even in sub-window,;<<!you may invoke pop-ups;by Alert(), InfoBox() etc"
InfoBox(cText)

cText := "<<!However, if the sub-window is;<<!too small to display your text," + ;
         ";<<!separate pop-up window is used; "
for ii := 5 to 20
   cText += ";Line " + ltrim(ii)   
next 
InfoBox(cText, "Pop-up") 
for ii := 21 to 60
   cText += ";Line " + ltrim(ii)  // too many lines for screen
next 
TextBox(cText, "TextBox()")       // oversized lines are truncated

/* still in sub-window with "FS2" text:
 */
row := row()+1                // wait will start with NewLine
wait timeout 5 "Press any key or wait 5 seconds..." color "RG+/B" 
@ row,0 clear                 // clear previous message
if lastkey() == K_ESC 
   WaClose()        
   return                     // exit when ESC was pressed
endif

/* create new status sub-window (small at top) 
 */
iStatID = wopen(0,0, 2,60)    // open sub-window
SET COLOR TO "W+/B"           // window-colors 
// SET GUICOLORS OFF          // disable SET COLOR for GUI mode
cls                           // clear current window
wbox()                        // draw frame around window

/* now lets open a lot of windows   
 */
iStart := len(aWinID) +1
iMax   := if(AppIoMode() == "G", 20, 100)

jj = 0
firstwin = 0
maxwin = 0
iStatID =  wopen(0,0, 2, 60)                 
setcolor("W+/B")
wbox()
do while .T.
   height = min(max(4, int(rand() * 10)), 10)   // using random
	width  = min(max(7, int(rand() * 20)), 20)
	top    = max(0, rand() * max_row() - height)
	left   = max(0, rand() * max_col() - width)
	right  = left + width 
	bottom = top + height
	xx= wopen(top, left, bottom, right)
	if firstwin == 0
		firstwin = xx
   endif
	if xx > maxwin
		maxwin = xx
   endif
   if (xx == -1)
	  exit
   endif
   do colors with jj
	jj ++          // counter
	bb = jj % 8
	if bb >= 4
	   bb = bb + 4
   endif
	wbox(bb)       
	@ 0,0 say "W:" + str(xx, 3)
	@ 1,0 say "At" + str(wrow(), 3)
	@ 2,0 say " /" + str(wcol(), 3)
   wselect(iStatID)
   
 	@ 0, 1 say "Window No.:" + str(xx, 3)  + " "
	inkey()
	if jj > iMax .or. lastkey() = 27
exit
	endif
enddo

/* select a view windows
 */
for ii = firstwin to maxwin
   wselect (ii)         // popup selected sub-window on top
   wselect(iStatID)     // write into status sub-window
	@ 0, 0 say "Select Window No.:" + str(ii) + " "
	inkey()
   if (lastkey() = 27)
exit
   endif
next
setlastkey(1)

/* and close them all again 
 */
for ii = maxwin to firstwin step -1
   wselect(ii)          // popup selected sub-window on top
   wclose()             // and close it
   wselect(iStatID)     // write into status sub-window
   @ 0, 0 say "Close Window No.:" + str(ii) + " "
	inkey()
   if (lastkey() = 27)
exit
   endif
next
inkey(2)

/* now you can move current window 
 * handled manually via wmove*()
 */
wselect(iStatID)
@ 0, 0 say "move with CURSOR or scroll by 's','b', ESCAPE to stop"

wopen (10,15, max_row(.t.) -10, max_col(.t.) - 15)
setcolor("GR+/B, GR+/B*")
wbox()
clear screen
formboard = 0
l_maxcol := maxcol() - 28
do while .t.
   if nextkey() != 0
      iKey = inkey()
      do case 
      case iKey = 4
         wmoveright()
      case iKey = 5
         wmoveup()
      case iKey = 19
         wmoveleft()
      case iKey = 24
         wmovedown()
      case iKey = 115    // 's' = form/scroll right 
         if (2 * formboard + 2) < l_maxcol
            formboard++
            clear screen
            wformat(0,1,0,1)	
         endif
      case iKey = 98     // 'b' = form/scroll left
         if formboard > 0
            formboard--
            clear screen
            wformat(0,-1,0,-1)	
         endif
      case iKey = 27
         exit
      endcase
   endif
   ? "This is Outputline: ", str(jj,5)
   jj = jj + 1 
enddo

/* Move Window by cursor keys, handled automatically
 */
WmoveUser(.T.)    // allow subwindow-movement
xColor := setcolor("W+/B,N/W,,N/W")
dummy = "Now edit or move with cursor keys "
@ 1,0 get dummy
read
WmoveUser(.F.)
/*
 * Now move by Alt + cursor keys
 */
WmoveUser(.T., {K_ALT_LEFT, K_ALT_RIGHT, K_ALT_UP, K_ALT_DOWN} ) 
dummy = "Now edit or move with Alt-cursor keys "
@ 1,0 get dummy
read
WmoveUser(.F.)
setcolor(xColor)

/* Thats all Folks 
 */
waclose()
wselect(0)
clear screen
wait timeout 5 "Done, bye bye ..."

return

/**************************************************************************
 * procedure colors         
 *    helper, sets colors according to counter (permuted) 
 * param: numeric counter                                             
 */
procedure colors
parameters val
col1 = val % 14
col2 = val / 14
col2 = col2 % 8
col1 = col1 + 1
col2 = col2 + 1

col1 = int(col1)
col2 = int(col2)

do case
case col1 = 1 .or. col1 = 8
	x1 = "B"
case col1 = 2 .or. col1 = 9
	x1 = "G"
case col1 = 3 .or. col1 = 10
	x1 = "BG"
case col1 = 4 .or. col1 = 11
	x1 = "R"
case col1 = 5 .or. col1 = 12
	x1 = "RB"
case col1 = 6 .or. col1 = 13
	x1 = "GR"
case col1 = 7 .or. col1 = 14
	x1 = "W"
endcase
if col1 > 7
	x1 = x1 + "+"
endif
do case
case col2 = 1 
	x2 = "B"
case col2 = 2
	x2 = "G"
case col2 = 3
	x2 = "BG"
case col2 = 4
	x2 = "R"
case col2 = 5
	x2 = "RB"
case col2 = 6
	x2 = "GR"
case col2 = 7
	x2 = "N"
case col2 = 8
	x2 = "W"
endcase
if (x1 == x2) 
  x1 = "W+"
endif
setcolor(x1 + "/" + x2)
return
