/* FlagShip calable function FREADSTDIN(...)
 * (c) 1999 by multisoft GmbH
 * will be also available in the library of FlagShip rel. 4.5 or later
 */

#include "FSopenc.h"

#ifndef EOF
# define EOF -1
#endif

/* ------------------------------------------------------------------------
 * FReadStdin (@cBuff, nBytes, [nFilter]) -> nStored
 *
 * Reads data from stdin into character buffer. 
 * The function is similar to FRead(), but this one does
 * not require the file name and performs additional
 * filtering on request.
 * 
 * - When reading data from console, the system transmittes it
 *   when the RETURN key or ^D was pressed regardless the buffer
 *   size - but only the required char number is stored, the
 *   rest is ignored.
 * - When reading data from a file (by an input redirection on
 *   command-line), the data are read until EOF or the 
 *   required buffer size whichever comes first.  
 *
 * Note: Since cBuff may include \0, use strlen() instead of len()
 *       and left(cBuff,nStored) instead of the cBuff string in
 *       the application.
 * Note: additional replacement/filter by cBuff=strtran(cBuff,x,y)
 *       in the application
 *
 * cBuff  = by caller specified buffer (passed by reference) of at
 *          least <nBytes> chars, otherwise nBytes=len(cBuff).
 * nBytes = max. number of bytes to store. If the length
 *          of <cBuff> is smaller, nBytes=len(cBuff)
 * nFilter= optional, flag combination is possible by its addition
 *           0: default, no filter
 *           1: ignore all CR and LF at the end of input (or buffer)
 *           2: ignore all CR ( 0x0D )
 *           4: ignore all LF ( 0x0A )
 *           8: ignore all ^Z ( 0x1A )
 *          16: ignore chars < " " ( 0x20)
 *          32: replace TAB ( 0x07) by space
 *          64: replace all chars < " " ( 0x20) by space
 *
 * nStored= number of bytes stored in cBuff (w/o trailing \0)
 *          -1: less than 2 arguments passed
 *          -2: wrong type of argument 1 or 2
 *          -3: wrong type of argument 3 if given
 *          -4: par1 not passed by reference
 *          -5: len(cBuff) < 1 or nBytes < 1
 * ----------------------------------------------------------------------- */
UDF_DECL(freadstdin)
{
   int nStored = 0; 
   int nBytes  = 0;
   int nFilter = 0;
   FSvar *varBuff;
   unsigned char *buff;
   int chr, ii;

   if(argc < 2) {
      nStored = -1;
      goto done;
   }

   if( (!IS_VAR_CHR(argv[0])) || (!IS_VAR_NUM(argv[1])) ) {
      nStored = -2;
      goto done;
   }

   if(argc > 2) {
      if (!IS_VAR_NUM(argv[2])) {
         nStored = -3;
         goto done;
      }
      nFilter = VAR_INTNUM(argv[2]); 
      if (nFilter < 1 || nFilter > 127) nFilter = 0;
   }

   varBuff = argv[0];

   if(! IS_VAR_BYREF(varBuff)) {               /* pass by refence required */
      nStored = -4;
      goto done;
   }

   nBytes = VAR_INTNUM(argv[1]);
   if(VAR_ISLEN(varBuff) <= nBytes)
      nBytes = VAR_ISLEN(varBuff) -1;

   if(nBytes <= 0) {
      nStored = -5;
      goto done;
   }

   buff = VAR_CHR(varBuff);

   while (nStored < nBytes) {
      chr = fgetc(stdin);
      if(chr == EOF) 
         break;
      if(nFilter) {
         if ((nFilter &  2) && (chr == '\r')) continue;
         if ((nFilter &  4) && (chr == '\n')) continue;
         if ((nFilter &  8) && (chr == 0x1a)) continue;
         if ((nFilter & 16) && (chr  < ' ' )) continue;
         if ((nFilter & 32) && (chr == '\t')) chr = ' ';
         if ((nFilter & 64) && !(nFilter & 1) && (chr  < ' ' )) chr = ' ';
      }
      *buff = chr;
      buff++;
      nStored++;
   }
   *buff = 0;

   if(nFilter & 1) {                          /* remove trailing CR and LF */
      buff--;
      while ((nStored > 0) &&  
             ((*buff == '\r') || (*buff == '\n')) ) {
         nStored--;
         *buff = 0;
         buff--;
      }
      if(nFilter & 64) {           /* transform remaining special to Space */
         buff = VAR_CHR(varBuff);
         for (ii = 0; ii < nStored; ii++, buff++) {
            if (*buff < ' ') *buff = ' ';
         }
      }
   }

done:
   return set_Iv(nStored);
}

