* webutil.prg  (see also webmail.c, freadstd.c, bbsleep.c)
*
* Some commonly used functions for Web based FlagShip applications
* (c) 1999..2007 by multisoft GmbH
* This source may be freely used or modified by the FlagShip Licensee
*
* compile: cd /usr/FSsrc/webtools ; make -or- make test
*          FlagShip mywebapplic.prg /usr/FSsrc/webtools/*.o ....
*
* if compiled with -DTESTING (or make test), no e-mail on error is sent
* and additional data are passed to the html output

#Cinline
static char *fileID="@@# $Name:  $ $Revision: 1.2 $ $Date: 2002/04/10 10:32:32 $";
#endCinline

#ifndef LF_CHAR
# define LF_CHAR chr(10)
#endif
#ifndef BRK
# define BRK "<BR>"
#endif

******************************************************************************
* WebHtmlBegin([cbHeader], [cBackColor])  -> NIL
*
* print HTML preface
*
* cbHeader: optional string or code block with the output of additional user 
*           text, e.g. {|| qout(text1,text2), qout(text3)}
* cBackColor: optional background color string, either 6-char "RRGGBB" or
*             7-char "#RRGGBB"  ->  <BODY BGCOLOR="#RRGGBB">
*             white = "FFFFFF"
* global : cWebHeaderTitle (if not set, default = "HTML Request")
*          lWebHeaderOut   (header already printed ?) --> .T.
******************************************************************************
FUNCTION WebHtmlBegin ([cbHeader], [cBackColor])
memvar lWebHeaderOut, cWebHeaderTitl

if type("lWebHeaderOut") != "L"
   public lWebHeaderOut := .F.
endif
if type("cWebHeaderTitle") != "C"
   public cWebHeaderTitle := "HTML Request"
endif

if !lWebHeaderOut
   ?? "Content-Type: text/html"
   ?
   ? "<HTML>"
   ? "<HEAD><TITLE>" + cWebHeaderTitle + "</TITLE></HEAD>"
   if valtype(cBackColor) == "C" .and. len(cBackColor) >= 6
      if len(cBackColor) == 6
         cBackColor :=  "#" + cBackColor
      endif
      ? '<BODY BGCOLOR="' + cBackColor + '">'
   else
      ? "<BODY>"
   endif
   if valtype(cbHeader) == "B"
      eval(cbHeader)
   elseif valtype(cbHeader) == "C"
      ? cbHeader
   endif
endif
lWebHeaderOut := .T.
return NIL

******************************************************************************
* WebHtmlEnd([cbFooter])  -> NIL
*
* print HTML footer
*
* bFooter: optional string or codeblock for user defined footer, e.g. {||qout(text)}
* global : lWebHeaderOut (if still .F. or undefined, print header) --> .T.
******************************************************************************
FUNCTION WebHtmlEnd ([cbFooter])
memvar lWebHeaderOut

if type("lWebHeaderOut") != "L"
   public lWebHeaderOut := .F.
endif
if !lWebHeaderOut
   WebHtmlBegin()
endif
if valtype(cbFooter) == "B"
   eval(cbFooter)
elseif valtype(cbFooter) == "C"
   ? cbFooter
endif

? "</BODY>"
? "</HTML>"
?
return NIL

******************************************************************************
* WebGetFormData([errUDF], [lUpper]) -> aFields
*
* Detects the used HTML Form Request Method, receives the data and
* decodes the fields into array elements. 
*
* errUDF : optional UDF (name or code block) for handling Method error,
*          it receives the REQUEST_METHOD value in the 1st parameter
*          and the text message in the 2nd parameter
* lUpper : translate 1st char of each field to oppercase
*
* aFields: array {{"NAME", "content"}, {"NAME", "content"}, ...} 
******************************************************************************
FUNCTION WebGetFormData ([cbErrUDF], [lUpper])
local  isGETform := .T.                            as logic
local  cQuery := "", cMethod := "", nLength := -1  as usual
local  pos, pos2, sep, ii, cField, token, cTxt     as usual
local  aFields := {}                               as array
static hexa := "0123456789ABCDEF"                  as char
memvar cFormString

#define MAX_INPUT_LNG 30000

cMethod := alltrim(upper(getenv("REQUEST_METHOD")))
lUpper  := if (valtype(lUpper) != "L", .F., lUpper)

#ifdef TESTING
   WebHtmlBegin()
   ? "---WebGetFormData()"
   ? 'REQUEST_METHOD="' + cMethod + '"' + BRK
#endif

do case
case cMethod == "GET"
   cQuery  := getenv("QUERY_STRING")               // get data via envir var
   nLength := len(cQuery)
case cMethod == "POST"
   isGETform := .F.
   nLength := val(alltrim(getenv("CONTENT_LENGTH")))
   if nLength > 0 .and. nLength < MAX_INPUT_LNG
      cQuery  := space(nLength +2)
      nLength := FreadStdin(@cQuery, len(cQuery), 1)  // read data via stdin
   else
      nLength := -1
   endif
   if nLength < 0
      cTxt := "FReadStdin('"+cQuery+"',"+ltrim(len(cQuery))+",1) retured: nLength=" + ltrim(nLength)
      if valtype(cbErrUDF) == "C" .and. isFunction(trim(cbErrUDF))
         do &cbErrUDF with cMethod, cTxt 
      elseif valtype(cbErrUDF) == "B"
         eval(cbErrUDF, cMethod, cTxt)
      endif
   endif
#ifdef TESTING
   ? 'CONTENT_LENGTH=',  val(alltrim(getenv("CONTENT_LENGTH"))), BRK
   ? 'cQuery="' + cQuery + '"' + BRK
   ? 'nLength=' + ltrim(nLength) + BRK
#endif
otherwise
   cTxt := "Unknown/unsupported REQUEST_METHOD=" + cMethod
   if valtype(cbErrUDF) == "C" .and. isFunction(trim(cbErrUDF))
      do &cbErrUDF with cMethod, cTxt 
   elseif valtype(cbErrUDF) == "B"
      eval(cbErrUDF, cMethod, cTxt)
   endif
endcase

while nLength > 0 .and. strpeek(@cQuery, nLength) == 10
   nLength--                                       // remove trailing LF
enddo

if nLength <= 0
   return aFields
endif

//****
// WebHtmlBegin()
// ? "----- " + procname() + "/" + ltrim(procline()) + BRK
// ? "cQuery=[" + cQuery + "]" + BRK
//***

if type("cFormString") == "U"
   public cFormString
endif
cFormString := cQuery                              // store to public var

// the fields are separated by "&"
pos := 1
while pos <= nLength
   if strpeek(@cQuery, pos) == asc("&")
      pos++                                        // skip leading separators
      loop
   endif
   sep := at("&", cQuery, pos)
   if sep == 0
      sep := nLength +1
   endif
   token := substr(cQuery, pos, sep - pos)         // field token
//****
//   ? "-pos=" + ltrim(pos) + " sep=" + ltrim(sep) + BRK
//   ? " token=[" + token + "]" + BRK
//****
   pos := sep +1                                   // next token pos
 
   ii := strlen(token) 
   while ii > 0 .and. strpeek(@cQuery, ii) == 10
      ii--                                         // remove trailing LF
   enddo
   if ii == 0
      loop
   endif
   if ii < strlen(token)
      token := left(token, ii)
   endif

   
// the field name and content are separated by "="
   cField := ""
   ii := at("=", token)
   if ii > 0
      cField := left(token, ii-1)
      token  := substr(token, ii+1) 
   endif

// translate all special characters in token
   token := strtran(token, "+", " ")
   pos2 := at("%", token)
   while pos2 > 0
      cTxt := upper(substr(token, pos2 +1, 2))
      if len(cTxt) < 2
         exit
      endif
      ii := at(substr(cTxt,1,1), hexa)
      if ii > 0
         sep := (ii -1) * 16
         ii := at(substr(cTxt,2,1), hexa)
         if ii > 0
            sep += (ii -1)
            token := strtran (token, substr(token, pos2,3), chr(sep))
         endif
      endif 
      pos2 := at("%", token, pos2 +1)
   enddo

// avoid intentional macro evaluation in text
   pos2 := at("&", token)
   while  pos2 > 0
      if substr(token, pos2+1,1) != " "
         token := left(token,pos2) + " " + substr(token,pos2 +1)
      endif
      pos2 :=  at("&", token, pos2+1)
   enddo

// convert 1st char to upper case if requested
   if lUpper .and. len(token) > 0
      token := upper(left(token,1)) + substr(token,2) 
   endif

// add name and token into array
   aadd(aFields, {alltrim(upper(cField)), token})  // NAME, content
enddo
return aFields
 
******************************************************************************
* WebOutData(aFields) -> NIL
*
* Prints the input data in the output html document
******************************************************************************
FUNCTION WebOutData([aFields])
if valtype(aFields) != "A" .or. empty(aFields)
   aFields := WebGetFormData()
endif
? "---Input data/fields:" + BRK
?? "<TT>"
aeval (aFields, {|x| qout(x[1] + '=[' + x[2] + ']' + BRK) })
?? "</TT>"
return NIL

******************************************************************************
* WebGetEnvir()
*
* fills the commonly used environment vars into an array
*   { {"envir_name", "content"}, {"envir_name", ""}, ...}
*
* See the description of these variables in webgetenvir.html
* and http://www.ncsa.uiuc.edu/demoweb/url-primer.html
******************************************************************************
FUNCTION WebGetEnvir()
local aEnvir := {}   as array
local ii             as intvar
static aVars := {"SERVER_SOFTWARE", "SERVER_NAME", "GATEWAY_INTERFACE", ;
   "SERVER_PROTOCOL", "SERVER_PORT", "REQUEST_METHOD", "PATH_INFO", ;
   "PATH_TRANSLATED", "SCRIPT_NAME", "QUERY_STRING", "REMOTE_HOST", ;
   "REMOTE_ADDR", "REMOTE_USER", "REMOTE_IDENT", "AUTH_TYPE", ;
   "CONTENT_TYPE", "CONTENT_LENGTH", "HTTP_ACCEPT", "HTTP_USER_AGENT"}  as array

for ii := 1 to len(aVars)
   aadd(aEnvir, {aVars[ii], getenv(aVars[ii])} )
next
return aEnvir

******************************************************************************
* WebMailDomain (cEmail) -> cDomain
*
* Extracts e-mail address "xxx@yyy.zz" from e.g. "aaa bbb" <xxx@yyy.zz> qqq"
* on error, " " is returned
******************************************************************************
FUNCTION WebMailDomain (email)
local ii, jj   as intvar
local cDomain  as usual                         

cDomain := email                               // e.g.: ["aaa bbb" <xxx@yyy.zz> qqq]
for ii := 1 to strlen(email)             
   if strpeek(@cDomain, ii) < 32               // replace spec.char
       cDomain := strpoke(@cDomain, ii, 32)    // by space
   endif
next
cDomain := strtran(cDomain, "<", " ")
cDomain := strtran(cDomain, ">", " ")          // ["aaa bbb"  xxx@yyy.zz  qqq]

ii := at("@", cDomain)                         // here = 15
if ii == 0
   return " "
endif
jj := rat(" ", left(cDomain, ii))              // 11 in ["aaa bbb"  xxx@]
if jj > 0
   cDomain := substr(cDomain, jj+1)            // now: [xxx@yyy.zz  qqq]
endi
ii := at(" ", cDomain)                         // here = 11
if ii > 0
   cDomain := left(cDomain, ii-1)              // now: [xxx@yyy.zz]
endif
cDomain := alltrim(cDomain)
return cDomain

******************************************************************************
* WebDate ([dDatum], [cTime]) -> cDatum  e.g. "01-Jan-2001 [13:56:05]"
*
* dDatum: opt, date value. If not specified, the current date() is used
* cTime : opt, time value. If empty string or .T., the current time() is used, 
*         or ignored if <cTime> is NIL or .F.
******************************************************************************
FUNCTION WebDate ([dDatum], [clTime])
local cRet := ""  as char
static aMon := {"000", "Jan","Feb","Mar","Apr","May","Jun", ;
                       "Jul","Aug","Sep","Oct","Nov","Dec" }   as array
if valtype(dDatum) != "D"
   dDatum := date()
endif
cRet := strzero(day(dDatum),2) + ;
        "-" + aMon[month(dDatum) +1] + ;
        "-" + strzero(year(dDatum),4)
do case
case valtype(clTime) == "C"
    cRet += " " + if(empty(clTime), time(), clTime)
case valtype(clTime) == "L" .and. clTime
    cRet += " " + time()
endcase
return cRet

******************************************************************************
* WebLogErr(cText, cFile, [cEmail], [aMailHead]) -> NIL
*
* Store message or error text in a ascii file <cFile> 
* On open error, send e-mail to <cEmail> if required
*
* cText     = the message stored in <cFile>
* cFile     = error-log file name including path 
* cEmail    = optional, e-mail address where error message should be send
* aMailHead = optional, e-mail header (see WebSendMail() in webmail.c)
******************************************************************************
FUNCTION WebLogErr (cText, cFile, [cEmail], [aMailHead])
local fd := 0     as usual
local count, pos  as usual

if empty(cText) .or. empty(cFile)
   return NIL
endif

if !file(cFile)
   fd := Fcreate(cFile, 0)
else
   for count := 1 to 5                        // retry for 5 seconds
      fd := fopen(cFile, 1)
      if fd > 0
         exit
      endif
      Sleep(1)                                // see bbsleep.c
   next
endif
if fd < 1 
   cText := strtran(cText, ";", LF_CHAR)
#ifdef TESTING
   cText := strtran(cText, LF_CHAR, BRK)
   WebHtmlBegin()
   ? "** canot open error file [" + cFile + "]" + BRK
   ? "   text:" + cText + BRK
#else
   if valtype(cEmail) == "C" .and. !empty(cEmail) .and. "@" $ cEmail
      WebSendMail (if(valtype(aMailHead)=="A", aMailHead, cEmail), ;
                   "ERROR in " + execname(),  ;
                   "Cannot open " + cFile + " - check permissions", ;
                   cEmail)
   endif
#endif      
else
   pos := fseek(fd, 0, 2)                     // go to EOF
   for count := 1 to 5                     
      if flockf(fd, len(cText) + 25, 2) != 0  // lock the new text area
         exit                                 // retry for 5 seconds,
      endif                                   // then ignore lock
      Sleep(1)
   next
   fwrite(fd, WebDate(NIL,.T.) + ": " + cText + LF_CHAR)
   fseek (fd, 0, 0)
   flockf(fd, 0, 0)                           // unlock
   fclose(fd)
endif
return NIL

******************************************************************************
* WebErrorHandler(oErr, cFile, [cEmail],[aMailHead],[bAddOut]) -> BREAK or NIL
*
* Store on-line errors in a file <cFile> and e-mail errors to <cEmail>.
* Issue a BREAK oErr i.e. jump into RECOVER USING ... section of BEGIN
* SEQUENCE .. RECOVER .. END in your application if such is specified, or
* return NIL otherwise
*
* oErr      = the error object passed from RTE
* cFile     = error-log file name including path 
* cEmail    = optional, e-mail address where error message should be send
* aMailHead = optional, e-mail header (see WebSendMail() in webmail.c)
* bAddOut   = optional, code block returning char string for additional 
*             data display
*
* usage: bOldErrBlock := ErrorBlock({|obj| WebErrorHandler(obj, ;
*                         "/cgi-bin/errlog.txt", "webmaster@my.domain.com") })
******************************************************************************
FUNCTION WebErrorHandler (oErr, cFile, [cEmail], [aMailHead], [bAddOut])
local cTxt, cTmp        as usual
local lEmailOut := .T.  as logic
static lRecursion := .F. as logic
static nCount := 0      as intvar

#ifndef TESTING
   lEmailOut := valtype(cEmail) == "C" .and. !empty(cEmail) .and. "@" $ cEmail
#endif

nCount++
if nCount > 5
   WebHtmlBegin()
   ? " --- more than 5 run-time errors, aborted ---" + BRK
   WebHtmlEnd()
   quit
endif
   
if !lRecursion       // ignore errors occured within this exception handlers
   lRecursion := .T. 

   cTxt := ";********* Run-time error in " + execname() + "*********;"
   cTxt += "FATAL ERROR " + ltrim(oErr:GenCode) + ";" 
   if !empty(oErr:FILENAME)
      cTxt += "file " + oErr:FILENAME
   else
      cTxt += "in " + PROCNAME(3) + "(" + LTRIM(STR(PROCLINE(3),5)) + ")"
   endif
   cTxt += " " + oErr:Description
   cTxt += if(!empty(oErr:SUBSYSTEM) .and. (oErr:SUBSYSTEM != "BASE"), oErr:SUBSYSTEM + "/", "")
   cTxt += if(!empty(oErr:SUBCODE), ltrim(oErr:SUBCODE) + ";", "")
   cTxt += if(!empty(oErr:Operation), oErr:Operation , "")
 
   WebLogErr(cTxt, cFile, cEmail, aMailHead)   // see above

   if lEmailOut
      cTxt += ";********* end of Run-time error message"
      cTxt := strtran(cTxt, ";", LF_CHAR)
      cTxt += LF_CHAR + "Environment:" + LF_CHAR
      cTxt := strtran(cTxt, ";", LF_CHAR)
      // add the current environment
      aeval(WebGetEnvir(), ;
          {|x| cTxt += if(empty(x[2]),"", x[1] + '="' +x[2] + '"' +LF_CHAR) })
      // additional data
      if valtype(bAddOut) == "B" 
          cTmp := eval(bAddOut)
          if valtype(cTmp) == "C"
             cTxt += LF_CHAR + cTmp
          endif
      endif
      cTxt += LF_CHAR
      // see webmail.c
#ifdef TESTING
      WebHtmlBegin()
      cTxt := strtran(cTxt, LF_CHAR, BRK)
      ? cTxt
#else
      WebSendMail (if(valtype(aMailHead)=="A", aMailHead, cEmail), ;
                   "ERROR in " + execname(),  ;
                   WebDate(NIL,.T.) + ": " + cTxt, ;
                   cEmail)
#endif
   endif
   lRecursion := .F.
   if IsBegSeq()           // BEGIN SEQUENCE active ?
      BREAK oErr
   endif
endif
return NIL

* eof *
